#--------------------------------------------------------------------------------
# before compiling for the first time, use "make makedir" to create all the
# necessary subdirectories
#--------------------------------------------------------------------------------

# uncomment build english version of NMAME X II
BUILD_ENGLISH = -DBUILD_ENGLISH

# Set the version here
VERSION = -DVERSION=13

# uncomment out the BETA_VERSION = line to build a beta version of MAME
# BETA_VERSION = -DBETA_VERSION=0

#--------------------------------------------------------------------------------

# Define this to enable the fittest Pentium Pro for compiling NMAME X II
COMPILE_PPRO = -DCOMPILE_PPRO

# uncomment out the MAME_DEBUG = line to build a version of MAME for debugging
# games 
# MAME_DEBUG = -DMAME_DEBUG

# if MAME_MMX is defined, MMX will be compiled in
MAME_MMX = -DMAME_MMX

# if KAILLERA is defined, Kaillera support will be compiled in
KAILLERA = -DKAILLERA

# whether to use fastcall or stdcall
USE_FASTCALL = 1

# uncomment to build without MIDAS
# NOMIDAS = 1

# uncomment to build Helpfiles
# HELPFILE = Mame32.hlp

# uncomment out the DEBUG = line to build a debugging version of mame32.
# DEBUG = 1

# uncomment enable MMX Render
MMXRENDER = -DENABLE_MMX_RENDER

# REMOVE_TESTDRIVER = -DREMOVE_TESTDRIVER

# MEM_DEBUG = -DMEM_DEBUG

#--------------------------------------------------------------------------------

CC      = @cl.exe
LD      = @link.exe
RSC     = @rc.exe
ASM     = @nasmw.exe

!ifdef BUILD_ENGLISH
OBJ     = obj_eng
EXEFILE = nmamex2e.exe
RESFILE = mame32e.res
!else
OBJ     = obj_jap
EXEFILE = nmamex2.exe
RESFILE = mame32.res
!endif

ASMFLAGS = -f win32
ASMDEFS  =

!if "$(PROCESSOR_ARCHITECTURE)" == ""
PROCESSOR_ARCHITECTURE=x86
!endif

HAS_CPUS  = -DHAS_Z80=1 -DHAS_M68000=1
HAS_SOUND = -DHAS_YM2610=1 -DCUT_RESAMPLING=1

DEFS   = -DLSB_FIRST=1 -DWIN32 -DPI=3.1415926535 \
         -DINLINE="static __inline" -Dinline=__inline -D__inline__=__inline \
         -DPNG_SAVE_SUPPORT -D_WINDOWS -DZLIB_DLL

RCFLAGS = -l 0x411 -DNDEBUG -I./Win32 $(MAME_MMX)

#--------------------------------------------------------------------------------
# MIDAS setting
#--------------------------------------------------------------------------------

!ifndef NOMIDAS
AUDIOFLAGS = -IMIDAS
AUDIOOBJS  = $(OBJ)/Win32/MIDASSound.o
#  DLLgȂƂ͂
# AUDIOLIBS  = MIDAS/MIDAS.lib
AUDIOLIBS  = MIDAS/MIDASDLL.lib
!else
AUDIOFLAGS = -DNOMIDAS
AUDIOOBJS  =
AUDIOLIBS  =
!endif

#--------------------------------------------------------------------------------
# net play settings
#--------------------------------------------------------------------------------

!ifdef KAILLERA
KAILLERA_LIBS = kailleraclient.lib ws2_32.lib imm32.lib
!endif

#--------------------------------------------------------------------------------
# compiler options
#--------------------------------------------------------------------------------

CFLAGSGLOBAL = -Gr -I. -Isrc -Isrc/Win32 \
               -I$(OBJ)/cpu/m68000 -Isrc/cpu/m68000 \
               -IZLIB $(AUDIOFLAGS) -W3 -nologo -MT \
               $(MAME_DEBUG) $(BETA_VERSION) $(VERSION) \
               $(MAME_MMX) $(HAS_CPUS) $(HAS_SOUND) \
               $(JPBIOS) $(NMAMECOMBO) $(MMXRENDER) $(BUILD_ENGLISH) \
               $(SORT_DI_JOYSTICK) $(REMOVE_TESTDRIVER) $(MEM_DEBUG) $(KAILLERA)

CFLAGSDEBUG = -Zi -Od

!ifdef COMPILE_PPRO
CFLAGSOPTIMIZED = -DNDEBUG -Ox -G6 -Ob2
!else
CFLAGSOPTIMIZED = -DNDEBUG -Ox -GB -Ob2
!endif

!ifdef DEBUG
CFLAGS = $(CFLAGSGLOBAL) $(CFLAGSDEBUG)
!else
CFLAGS = $(CFLAGSGLOBAL) $(CFLAGSOPTIMIZED)
!endif

!ifdef USE_FASTCALL
CFLAGS = $(CFLAGS) -DCLIB_DECL=__cdecl -DDECL_SPEC=__cdecl
!endif

#--------------------------------------------------------------------------------
# linker options
#--------------------------------------------------------------------------------

LDFLAGSGLOBAL = -subsystem:windows -machine:$(PROCESSOR_ARCHITECTURE) -nologo

LDFLAGSDEBUG = -debug:full
LDFLAGSOPTIMIZED = -release -incremental:no -map

!ifdef DEBUG
LDFLAGS = $(LDFLAGSGLOBAL) $(LDFLAGSDEBUG)
!else
LDFLAGS = $(LDFLAGSGLOBAL) $(LDFLAGSOPTIMIZED)
!endif


#--------------------------------------------------------------------------------
# objects setting
#--------------------------------------------------------------------------------

!ifdef MEM_DEBUG
MEM_OBJ = $(OBJ)/memmngr.o
!else
MEM_OBJ =
!endif

WIN32_OBJS = \
         $(OBJ)/Win32/osdepend.o \
         $(OBJ)/Win32/MAME32.o $(OBJ)/Win32/M32Util.o \
         $(OBJ)/Win32/DirectInput.o $(OBJ)/Win32/DIKeyboard.o $(OBJ)/Win32/DIJoystick.o \
         $(OBJ)/Win32/uclock.o \
         $(OBJ)/Win32/DirectDraw.o $(OBJ)/Win32/DDrawDisplay.o \
         $(OBJ)/Win32/led.o $(OBJ)/Win32/status.o \
         $(AUDIOOBJS) $(OBJ)/Win32/DirectSound.o $(OBJ)/Win32/NullSound.o \
         $(OBJ)/Win32/Keyboard.o $(OBJ)/Win32/Joystick.o $(OBJ)/Win32/Trak.o \
         $(OBJ)/Win32/file.o $(OBJ)/Win32/Directories.o $(OBJ)/Win32/mzip.o \
         $(OBJ)/Win32/audit32.o $(OBJ)/Win32/Win32text.o \
         $(OBJ)/Win32/Win32ui.o $(OBJ)/Win32/Properties.o $(OBJ)/Win32/ColumnEdit.o \
         $(OBJ)/Win32/Screenshot.o $(OBJ)/Win32/TreeView.o $(OBJ)/Win32/Splitters.o \
         $(OBJ)/Win32/options.o $(OBJ)/Win32/Bitmask.o $(OBJ)/Win32/DataMap.o \
         $(OBJ)/Win32/Avi.o $(OBJ)/Win32/dxdecode.o \
		 $(OBJ)/Win32/KailleraChat.o

CPUOBJS = \
         $(OBJ)/cpu/m68000/asmintf.o $(OBJ)/cpu/m68000/68kem.oa \
         $(OBJ)/cpu/m68000/m68kops.og $(OBJ)/cpu/m68000/m68kopac.og \
         $(OBJ)/cpu/m68000/m68kopdm.og $(OBJ)/cpu/m68000/m68kopnz.og \
         $(OBJ)/cpu/m68000/m68kcpu.o $(OBJ)/cpu/m68000/m68kmame.o \
         $(OBJ)/cpu/z80/z80.o

!ifdef MAME_DEBUG
DBGOBJS = \
         $(OBJ)/cpu/m68000/m68kdasm.o \
         $(OBJ)/cpu/z80/z80dasm.o \
         $(OBJ)/mamedbg.o \
         $(OBJ)/window.o \
         $(OBJ)/Win32/debug.o
!else
DBGOBJS =
!endif

SNDOBJS = \
         $(OBJ)/sound/fm.o \
         $(OBJ)/sound/2610intf.o \
         $(OBJ)/sound/ymdeltat.o \
         $(OBJ)/sound/ay8910.o \

COREOBJS = \
         $(OBJ)/version.o $(OBJ)/driver.o $(OBJ)/mame.o \
         $(OBJ)/drawgfx.o $(OBJ)/common.o $(OBJ)/usrintrf.o \
         $(OBJ)/cpuintrf.o $(OBJ)/memory.o $(OBJ)/timer.o $(OBJ)/palette.o \
         $(OBJ)/input.o $(OBJ)/inptport.o $(OBJ)/unzip.o $(OBJ)/ui_text.o \
         $(OBJ)/audit.o $(OBJ)/info.o $(OBJ)/png.o $(OBJ)/cheat.o \
         $(OBJ)/datafile.o $(OBJ)/ui_font.o \
         $(OBJ)/wav.o \
         $(CPUOBJS) \
         $(OBJ)/sndintrf.o \
         $(OBJ)/sound/streams.o $(OBJ)/sound/mixer.o \
         $(SNDOBJS) \
         $(OBJ)/profiler.o \
         $(DBGOBJS) \
         $(MEM_OBJ)

NEO_OBJS = \
         $(OBJ)/machine/neogeo.o $(OBJ)/machine/pd4990a.o \
         $(OBJ)/vidhrdw/neogeo.o $(OBJ)/drivers/neogeo.o \
		 $(OBJ)/machine/neocrypt.o \
         $(OBJ)/patch/ffevpat.o $(OBJ)/patch/garoupat.o


OBJS = $(COREOBJS) $(WIN32_OBJS) $(NEO_OBJS)

RES = $(OBJ)/Win32/$(RESFILE)

LIBS   = kernel32.lib user32.lib gdi32.lib comctl32.lib comdlg32.lib \
         advapi32.lib winmm.lib shell32.lib dinput.lib dxguid.lib vfw32.lib \
         ZLIB\zlib.lib $(AUDIOLIBS) $(KAILLERA_LIBS)


!ifdef HELPFILE
all: $(EXEFILE) $(HELPFILE)
!else
all: $(EXEFILE)
!endif

# workaround for vc4.2 compiler optimization bug:
!ifndef DEBUG
$(OBJ)/vidhrdw/timeplt.o: src/vidhrdw/timeplt.c
	@echo Compiling $<...
	$(CC) $(DEFS) $(CFLAGSGLOBAL) -Oi -Ot -Oy -Ob1 -Gs -G5 -Gr -Fo$@ -c src/vidhrdw/timeplt.c
!endif

#--------------------------------------------------------------------------------

$(EXEFILE): $(OBJS) $(RES)
        @echo Linking $(EXEFILE)...
        $(LD) @<<
        $(LDFLAGS) -out:$(EXEFILE) $(OBJS) $(RES) $(LIBS)
<<

!ifdef HELPFILE
$(HELPFILE): src\Win32\Hlp\Mame32.cnt src\Win32\Hlp\Mame32.hlp
        @echo Compiling help...
        @Makehelp.bat
!endif

#--------------------------------------------------------------------------------

.asm.oa:
	@echo Assembling $<...
	$(ASM) -o $@ $(ASMFLAGS) $(ASMDEFS) $<

{src}.c{$(OBJ)}.o:
	@echo Compiling $<...
	$(CC) $(DEFS) $(CFLAGS) -Fo$@ -c $<

{src/cpu/z80}.c{$(OBJ)/cpu/z80}.o:
	@echo Compiling $<...
	$(CC) $(DEFS) $(CFLAGS) -Fo$@ -c $<

{src/cpu/m68000}.c{$(OBJ)/cpu/m68000}.o:
	@echo Compiling $<...
	$(CC) $(DEFS) $(CFLAGS) -Fo$@ -c $<

{src/vidhrdw}.c{$(OBJ)/vidhrdw}.o:
	@echo Compiling $<...
	$(CC) $(DEFS) $(CFLAGS) -Fo$@ -c $<

{src/sndhrdw}.c{$(OBJ)/sndhrdw}.o:
	@echo Compiling $<...
	$(CC) $(DEFS) $(CFLAGS) -Fo$@ -c $<

{src/machine}.c{$(OBJ)/machine}.o:
	@echo Compiling $<...
	$(CC) $(DEFS) $(CFLAGS) -Fo$@ -c $<

{src/drivers}.c{$(OBJ)/drivers}.o:
	@echo Compiling $<...
	$(CC) $(DEFS) $(CFLAGS) -Fo$@ -c $<

{src/sound}.c{$(OBJ)/sound}.o:
	@echo Compiling $<...
	$(CC) $(DEFS) $(CFLAGS) -Fo$@ -c $<

{src/patch}.c{$(OBJ)/patch}.o:
	@echo Compiling $<...
	$(CC) $(DEFS) $(CFLAGS) -Fo$@ -c $<

{src/Win32}.c{$(OBJ)/Win32}.o:
	@echo Compiling $<...
	$(CC) $(DEFS) $(CFLAGS) -Fo$@ -c $<

{src/Win32}.rc{$(OBJ)/Win32}.res:
	@echo Compiling Resource...
	$(RSC) $(RCFLAGS) -Fo$@ $<

CFLAGS_MAKE_68K = $(CFLAGS) -DWIN32
!ifdef USE_FASTCALL
CFLAGS_MAKE_68K = $(CFLAGS_MAKE_68K) -DFASTCALL
!endif

# generate C source files for the 68000 emulator
$(OBJ)/cpu/m68000/m68kops.h \
$(OBJ)/cpu/m68000/m68kops.c \
$(OBJ)/cpu/m68000/m68kopac.c \
$(OBJ)/cpu/m68000/m68kopdm.c \
$(OBJ)/cpu/m68000/m68kopnz.c : src/cpu/m68000/m68kmake.c src/cpu/m68000/m68k_in.c
	$(CC) $(CFLAGS_MAKE_68K) /Fe$(OBJ)/cpu/m68000/m68kmake.exe /Fo$(OBJ)/cpu/m68000/m68kmake.$(OBJ) \
	src/cpu/m68000/m68kmake.c
	@echo Generating MC68K(C)...
	@$(OBJ)\\cpu\\m68000\\m68kmake $(OBJ)/cpu/m68000 src/cpu/m68000/m68k_in.c

# generated C files for the 68000 emulator
{$(OBJ)/cpu/m68000}.c{$(OBJ)/cpu/m68000}.og:
	@echo Compiling $<...
	$(CC) $(DEFS) $(CFLAGS) -Fo$@ -c $<

# generate asm source files for the 68000 emulator
$(OBJ)/cpu/m68000/68kem.asm:  src/cpu/m68000/make68k.c
	$(CC) $(CFLAGS_MAKE_68K) /Fe$(OBJ)/cpu/m68000/make68k.exe /Fo$(OBJ)/cpu/m68000/make68k.$(OBJ) \
	src/cpu/m68000/make68k.c
	@echo Generating MC68K(ASM)...
	@$(OBJ)\\cpu\\m68000\\make68k $@ $(OBJ)/cpu/m68000/comptab.asm

# dependencies
$(OBJ)/cpu/z80/z80.o:          src/cpu/z80/z80.c src/cpu/z80/z80.h src/cpu/z80/z80daa.h
$(OBJ)/cpu/m68000/m68kcpu.o:   $(OBJ)/cpu/m68000/m68kops.c src/cpu/m68000/m68kmake.c src/cpu/m68000/m68k_in.c

.IGNORE:

#--------------------------------------------------------------------------------

maketree:
	@echo Making object tree...
	@md $(OBJ)
	@md $(OBJ)\cpu
	@md $(OBJ)\cpu\z80
	@md $(OBJ)\cpu\m68000
	@md $(OBJ)\sound
	@md $(OBJ)\drivers
	@md $(OBJ)\machine
	@md $(OBJ)\vidhrdw
	@md $(OBJ)\sndhrdw
	@md $(OBJ)\patch
	@md $(OBJ)\Win32

clean:
	@echo Clear object files...
	@del $(OBJ)\*.o
	@del $(OBJ)\cpu\z80\*.o
	@del $(OBJ)\cpu\m68000\*.o
	@del $(OBJ)\cpu\m68000\*.oa
	@del $(OBJ)\cpu\m68000\*.og
	@del $(OBJ)\cpu\m68000\*.c
	@del $(OBJ)\cpu\m68000\*.h
	@del $(OBJ)\cpu\m68000\*.asm
	@del $(OBJ)\cpu\m68000\*.exe
	@del $(OBJ)\sound\*.o
	@del $(OBJ)\drivers\*.o
	@del $(OBJ)\machine\*.o
	@del $(OBJ)\vidhrdw\*.o
	@del $(OBJ)\sndhrdw\*.o
	@del $(OBJ)\patch\*.o
	@del $(OBJ)\Win32\*.o
	@del src\Win32\*.aps
	@del *.map
	@del $(EXEFILE)
!ifdef HELPFILE
	del mame32.hlp
	del mame32.cnt
!endif

cleandebug:
	@echo Clear object files...
	@del $(OBJ)\*.o
	@del $(OBJ)\cpu\z80\*.o
	@del $(OBJ)\cpu\m68000\*.o
	@del $(OBJ)\cpu\m68000\*.oa
	@del $(OBJ)\cpu\m68000\*.og
	@del $(OBJ)\cpu\m68000\*.c
	@del $(OBJ)\cpu\m68000\*.h
	@del $(OBJ)\cpu\m68000\*.asm
	@del $(OBJ)\cpu\m68000\*.exe
	@del $(EXEFILE)

cleantiny:
	@echo Clear object files...
	@del $(OBJ)\driver.o
	@del $(OBJ)\usrintrf.o
	@del $(OBJ)\cheat.o
	@del $(OBJ)\Win32\*.o
