/***************************************************************************

    M.A.M.E.32  -  Multiple Arcade Machine Emulator for Win32
    Win32 Portions Copyright (C) 1997-99 Michael Soderstrom and Chris Kirmse

    This file is part of MAME32, and may only be used, modified and
    distributed under the terms of the MAME license, in "readme.txt".
    By continuing to use, modify or distribute this file you indicate
    that you have read the license and understand and accept it fully.

***************************************************************************/

/***************************************************************************

  Avi.c

  AVI capture code contributed by John Zissopoulos
                <zissop@theseas.softlab.ece.ntua.gr>

***************************************************************************/

#include "driver.h"

#include <windows.h>
#include <shellapi.h>
#include <windowsx.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <commctrl.h>
#include <commdlg.h>
#include <strings.h>
#include <sys/stat.h>
#include <wingdi.h>
#include <tchar.h>
/* to shutup the warnings about functions were NOT calling in vfw.h */
#ifdef CDECL
#undef CDECL
#define CDECL CLIB_DECL
#endif
#include <vfw.h>

#include "Mame32.h"
#include "DDrawDisplay.h"
#include "Avi.h"
#include "win32text.h"


/***************************************************************************
    function prototypes
 ***************************************************************************/

static BOOL AVI_Check_Version(void);
static BOOL AVI_Init(char* filename, int min_x, int min_y, int max_x, int max_y);
static void AVI_Close(char* filename);
static void AVI_Convert_Bitmap(struct osd_bitmap* tBitmap);
static void AVI_Message_Error(void);

/***************************************************************************
    External variables
 ***************************************************************************/

/***************************************************************************
    Internal variables
 ***************************************************************************/

static PAVIFILE     pfile = NULL;
static PAVISTREAM   psCompressed = NULL;
static PAVISTREAM   ps = NULL;
static BOOL         bAviError = FALSE;
static int          nAviFrames = 0;
static BOOL         bAviCapture = FALSE;

static DWORD        nAviFPS = 0;
static char         aviFilename[MAX_PATH * 2];

static tRect        rectAvi;
static char         *pAviBitmap = NULL;

/***************************************************************************
    External variables
 ***************************************************************************/

/* Set frames per second */
void SetAviFPS(int fps)
{
    nAviFPS = fps;
}

/* Is AVI capturing initialized? */
BOOL GetAviCapture(void)
{
    return bAviCapture;
}

/* Start capture, initialize */
BOOL AviStartCapture(char* filename, int min_x, int min_y, int max_x, int max_y)
{
    if (bAviCapture || (!AVI_Check_Version()))
        return FALSE;

    AVI_Init(filename, min_x, min_y, max_x, max_y);
    bAviCapture = TRUE;
    strcpy(aviFilename, filename);

    return TRUE;
}

/* End capture, close file */
void AviEndCapture(void)
{
    if (pAviBitmap)
    {
        MyFree(pAviBitmap);
        pAviBitmap = NULL;
    }

    if (!bAviCapture)
        return;

    AVI_Close(aviFilename);
    bAviCapture = FALSE;
    strcpy(aviFilename, "");
}

/* Add a frame to an open AVI */
void AviAddBitmap(struct osd_bitmap *tBitmap)
{
    HRESULT hr;
    LPBITMAPINFOHEADER lpbit;

    if (! bAviCapture)
        return;

    AVI_Convert_Bitmap(tBitmap);
    lpbit = (LPBITMAPINFOHEADER)pAviBitmap;

    if (!nAviFrames)
    {
        hr = AVIStreamSetFormat(psCompressed, 0,
                            lpbit,              // stream format
                            lpbit->biSize);     // format size

        if (hr != AVIERR_OK)
        {
            bAviError = TRUE;
        }
    }

    if (!bAviError)
    {
        hr = AVIStreamWrite(psCompressed,       // stream pointer
                        nAviFrames,             // time of this frame
                        1,                      // number to write
                        (LPBYTE)lpbit +         // pointer to data
                        lpbit->biSize,
                        lpbit->biSizeImage,     // size of this frame
                        AVIIF_KEYFRAME,         // flags....
                        NULL,
                        NULL);
    }

    nAviFrames++;
}

/***************************************************************************
    External variables
 ***************************************************************************/

static BOOL AVI_Init(char* filename, int min_x, int min_y, int max_x, int max_y)
{
    HRESULT             hr;
    AVISTREAMINFO       strhdr;
    DWORD               dwSize;
    DWORD               biSizeImage;
    AVICOMPRESSOPTIONS  opts;
    AVICOMPRESSOPTIONS* aopts[1];
    BITMAPINFOHEADER    bmiHeader;

    aopts[0] = &opts;

    AVIFileInit();

    hr = AVIFileOpen(&pfile,                    // returned file pointer
                   filename,                    // file name
                   OF_WRITE | OF_CREATE,        // mode to open file with
                   NULL);                       // use handler determined
                                                // from file extension....
    if (hr != AVIERR_OK)
    {
        AVI_Message_Error();
        return FALSE;
    }

    rectAvi.m_Left   = min_x;
    rectAvi.m_Top    = min_y;
    rectAvi.m_Width  = (max_x - min_x) + 1;
    rectAvi.m_Height = (max_y - min_y) + 1;

    dwSize = sizeof(BITMAPINFOHEADER) + 2 * (DWORD)rectAvi.m_Width * (DWORD)rectAvi.m_Height;

    biSizeImage = dwSize - sizeof(BITMAPINFOHEADER);

    // Fill in the header for the video stream....
    // The video stream will run in 60ths of a second....To be changed

    _fmemset(&strhdr, 0, sizeof(strhdr));
    strhdr.fccType                = streamtypeVIDEO;// stream type
    strhdr.fccHandler             = 0;
    strhdr.dwScale                = 1;
    strhdr.dwRate                 = 60;         // 60 fps
    strhdr.dwSuggestedBufferSize  = biSizeImage;

    SetRect(&strhdr.rcFrame, 0, 0,          // rectangle for stream
            rectAvi.m_Width,
            rectAvi.m_Height);


    // And create the stream;
    hr = AVIFileCreateStream(pfile,         // file pointer
                             &ps,           // returned stream pointer
                             &strhdr);      // stream header
    if (hr != AVIERR_OK)
    {
        AVI_Message_Error();
        return FALSE;
    }

    _fmemset(&opts, 0, sizeof(opts));
    if (!AVISaveOptions(NULL, 0, 1, &ps, aopts))
    {
        AVI_Message_Error();
        return FALSE;
    }

    hr = AVIMakeCompressedStream(&psCompressed, ps, &opts, NULL);
    if (hr != AVIERR_OK)
    {
        AVI_Message_Error();
        return FALSE;
    }

    pAviBitmap = MyMalloc(dwSize);

    /* This is the BITMAPINFOHEADER structure */
    bmiHeader.biSize          = sizeof(BITMAPINFOHEADER);
    bmiHeader.biWidth         = rectAvi.m_Width;
    bmiHeader.biHeight        = rectAvi.m_Height;
    bmiHeader.biPlanes        = 1;
    bmiHeader.biBitCount      = 16;
    bmiHeader.biCompression   = BI_RGB;
    bmiHeader.biSizeImage     = dwSize - sizeof(BITMAPINFOHEADER);
    bmiHeader.biXPelsPerMeter = 0;
    bmiHeader.biYPelsPerMeter = 0;
    bmiHeader.biClrUsed       = 0;
    bmiHeader.biClrImportant  = 0;

    memcpy(pAviBitmap, (void *)&bmiHeader, sizeof(BITMAPINFOHEADER));

    return TRUE;
}

static void AVI_Close(char* filename)
{
    if (ps)
        AVIStreamClose(ps);

    if (psCompressed)
        AVIStreamClose(psCompressed);

    if (pfile)
        AVIFileClose(pfile);

    AVIFileExit();

    if (nAviFrames)
    {
        if (MessageBox(GetActiveWindow(),
			win32_getstring(WIN32_avi_reflectionFPS),
			win32_getstring(WIN32_avi_reflectionFPStitle),
			MB_YESNO | MB_ICONINFORMATION) == IDYES)
        {
            FILE* fp;

            /* Directly edit the frame rate information */
            if ((fp = fopen(filename, "rb+")) != NULL)
            {
                fseek(fp, 132, SEEK_SET);
                fwrite(&nAviFPS, 1, sizeof(DWORD), fp);
                fclose(fp);
            }
        }
    }
    else
    {
        remove(filename);
    }

    pfile = NULL;
    ps = NULL;
    psCompressed = NULL;
    nAviFrames = 0;
    bAviError = FALSE;
}


static BOOL AVI_Check_Version(void)
{
    WORD wVer;

    /* first let's make sure we are running on 1.1 */
    wVer = HIWORD(VideoForWindowsVersion());
    if (wVer < 0x010a)
    {
        return FALSE;
    }
    return TRUE;
}

static void AVI_Message_Error(void)
{
    MessageBox(NULL, "AVI File Error!", "Error", MB_OK);
}

static void AVI_Convert_Bitmap(struct osd_bitmap* tBitmap)
{
    int x, y;
    WORD *src, *dest;
    extern unsigned short *game_pen;

    dest = (WORD *)(pAviBitmap + sizeof(BITMAPINFOHEADER));
    y    = rectAvi.m_Height;

    while (y--)
    {
        src = tBitmap->line[rectAvi.m_Top + y] + rectAvi.m_Left;
        x   = rectAvi.m_Width;

        while (x--)
            *dest++ = game_pen[*src++];
    }
}
