/***************************************************************************

    M.A.M.E.32  -  Multiple Arcade Machine Emulator for Win32
    Win32 Portions Copyright (C) 1997-98 Michael Soderstrom and Chris Kirmse

    This file is part of MAME32, and may only be used, modified and
    distributed under the terms of the MAME license, in "readme.txt".
    By continuing to use, modify or distribute this file you indicate
    that you have read the license and understand and accept it fully.

 ***************************************************************************/

/***************************************************************************

  DDrawDisplay.c

 ***************************************************************************/

#include "driver.h"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h>
#include <windowsx.h>
#include <stdio.h>
#define DIRECTDRAW_VERSION 0x0700
#include <ddraw.h>
#include <assert.h>
#include <math.h>
#include "MAME32.h"
#include "resource.h"
#include "DDrawDisplay.h"
#include "M32Util.h"
#include "driver.h"
#include "uclock.h"
#include "status.h"
#include "DirectDraw.h"
#include "led.h"
#include "avi.h"
#include "dxdecode.h"
#include "win32text.h"
#ifdef KAILLERA
#include "KailleraChat.h"
#endif

#ifdef MAME_DEBUG
#include "debug.h"
#endif



/***************************************************************************
    Macros
 ***************************************************************************/

#define FRAMESKIP_LEVELS 12
#define INTENSITY_LEVELS 256
#define FRAMES_TO_SKIP   20

#define RENDERARGS   BYTE *pSrc, BYTE *pDst, int nSrcLen, int nDstLen, int nWidth, int nHeight

#define MAKECOL(r, g, b) (((r & 0xFF) >> (8 - Red.m_nSize))   << Red.m_nShift   | \
                          ((g & 0xFF) >> (8 - Green.m_nSize)) << Green.m_nShift | \
                          ((b & 0xFF) >> (8 - Blue.m_nSize))  << Blue.m_nShift)

#define GETC(col, pi)    (((col & (pi)->m_dwMask) >> (pi)->m_nShift)) << (8 - (pi)->m_nSize);

 /***************************************************************************
    Internal structures
 ***************************************************************************/

struct tPixelInfo
{
    int   m_nShift;
    int   m_nSize;
    DWORD m_dwMask;
};

typedef void (*RenderMethod)(RENDERARGS);

/* Gamma/Brightness */
static unsigned char       pColorMap[INTENSITY_LEVELS];
static double              dGamma;
static int                 nBrightness;

/* Frames per second (FPS) info */
static BOOL                bShowFPS;
static int                 nShowFPSTemp;

/* Speed throttling / frame skipping */
static BOOL                bThrottled;
static BOOL                bAutoFrameskip;
static int                 nFrameSkip;
static int                 nFrameskipCounter;
static int                 nFrameskipAdjust;
static int                 nFPS;
static int                 nSpeed;
static uclock_t            PrevMeasure;
static uclock_t            ThisFrameBase;
static uclock_t            Prev;

/* average frame rate data */
static int                 frames_displayed;
static uclock_t            start_time;
static uclock_t            end_time;        /* to calculate fps average on exit */

static tRect               GameRect;        /* hCỏʗ̈                       */
static tRect               ClientRect;      /* EBhẺʗ̈                     */
static tRect               WindowRect;      /* EBhË̗                           */
static RECT                rectPrimary;     /* tgobt@̓]̈                 */
static RECT                rectBack;        /* obNobt@̓]̈                   */

static BOOL                bWindowMode;     /* EChE[h                         */

#ifdef ENABLE_MMX_RENDER
static BOOL                bDisableMMX;     /* MMX𖳌ɂ邩                          */
#endif
static BOOL                bSleep;          /* Sleep()邩                              */

/* Options/Parameters */
static UINT                nScreenWidth;    /* ʂ̕                                   */
static UINT                nScreenHeight;   /* ʂ̍                                 */
static UINT                nScale;          /* ʂ̊g嗦(݂͓{A2{̂)          */
static BOOL                bHScanLines;     /* XLC\邩                   */
static BOOL                bDouble;         /* ʂg`悷邩                       */
static BOOL                bBltDouble;      /* gɃn[hEFAblit()gp邩 */
static BOOL                bStretch;        /* Xgb`gp邩                     */
static BOOL                bFastDouble;     /* Ȋg`gp邩                 */
static BOOL                bTripleBuffer;   /* gvobt@gp邩               */
static BOOL                bVsync;          /* Vsync҂                              */
static BOOL                bSystemMemory;   /* obNobt@VXeɂƂ邩     */
static BOOL                bUseBackBuffer;  /* obNobt@gp邩                 */

/* AVI Capture params */
static BOOL                bAviCapture;      /* Capture AVI mode */
static BOOL                bAviRun;          /* Capturing frames */
static int                 nAviShowMessage;  /* Show status if !0 */

/* Palette/Color */
static BOOL                bUpdatePalette;
static BOOL                bUpdateBackground;
static PALETTEENTRY*       pPalEntries;
static PALETTEENTRY*       pAdjustedPalette;
static UINT32*             pPalLookup;
static UINT32              nTotalColors;
static UINT                nBytesPerPixel;
static struct tPixelInfo   Red;
static struct tPixelInfo   Green;
static struct tPixelInfo   Blue;

/* DirectDraw */
static IDirectDraw7*        pDDraw;
static IDirectDrawSurface7* pDDSPrimary;
static IDirectDrawSurface7* pDDSBack;
static IDirectDrawSurface7* pDDSFlip;
static IDirectDrawClipper*  pDDClipper;

static UINT                nDisplayIndex;
static BOOL                bBestFit;
static BOOL                bInitialized;
static BOOL                bPaused;
static BOOL                bMinimized;

static BYTE*               pBitmapOffset;
static int                 nBitmapPitch;
static BYTE*               pSurfaceBits;
static int                 nSurfacePitch;

static RenderMethod        Render;

static HMENU               hMenu;
static HWND                hWndAbout;

static int                 already_synced;


/***************************************************************************
    Function prototypes
 ***************************************************************************/

static int                DDraw_init(options_type *options);
static int                DDraw_init2(void);
static void               DDraw_exit(void);
static struct osd_bitmap* DDraw_alloc_bitmap(int width, int height);
static void               DDraw_free_bitmap(struct osd_bitmap* bitmap);
static int                DDraw_create_display(int width, int height, float fps);
static void               DDraw_close_display(void);
static void               DDraw_set_visible_area(int min_x, int max_x, int min_y, int max_y);
#ifdef MAME_DEBUG
static void               DDraw_set_debugger_focus(int debugger_has_focus);
static int                DDraw_allocate_colors(const UINT8 *debug_palette, UINT16 *debug_pens);
#else
static int                DDraw_allocate_colors(void);
#endif
static void               DDraw_modify_pen(int pen, unsigned char red, unsigned char green, unsigned char blue);
static void               DDraw_get_pen(int pen, unsigned char* pRed, unsigned char* pGreen, unsigned char* pBlue);
static int                DDraw_skip_this_frame(void);
#ifdef MAME_DEBUG
static void               DDraw_update_display(struct osd_bitmap *game_bitmap, struct osd_bitmap *debugger_bitmap);
#else
static void               DDraw_update_display(struct osd_bitmap *game_bitmap);
#endif
static void               DDraw_led_w(int leds_status);
static void               DDraw_set_gamma(float gamma);
static float              DDraw_get_gamma(void);
static void               DDraw_set_brightness(int brightness);
static int                DDraw_get_brightness(void);
static void               DDraw_save_snapshot(struct osd_bitmap *bitmap);
static BOOL               DDraw_OnMessage(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam, LRESULT* pResult);
static void               DDraw_Refresh(void);
static void               DDraw_UpdateFPS(BOOL bShow, BOOL bAuto, int nSpeed, int nFPS, int nMachineFPS, int nFrameskip);

static void               UpdateDisplay(void);

static void               SelectRenderMethod(void);

static void               Render2BytesPerPixel(RENDERARGS);
static void               Render3BytesPerPixel(RENDERARGS);
static void               Render4BytesPerPixel(RENDERARGS);
static void               Render2BytesPerPixel_BltHScanLines(RENDERARGS);
static void               Render3BytesPerPixel_BltHScanLines(RENDERARGS);
static void               Render4BytesPerPixel_BltHScanLines(RENDERARGS);
static void               Render2BytesPerPixel_HScanLines(RENDERARGS);
static void               Render3BytesPerPixel_HScanLines(RENDERARGS);
static void               Render4BytesPerPixel_HScanLines(RENDERARGS);
static void               Render2BytesPerPixel_Double(RENDERARGS);
static void               Render3BytesPerPixel_Double(RENDERARGS);
static void               Render4BytesPerPixel_Double(RENDERARGS);

#ifdef ENABLE_MMX_RENDER
static void               MMXRender2BytesPerPixel(RENDERARGS);
static void               MMXRender3BytesPerPixel(RENDERARGS);
static void               MMXRender4BytesPerPixel(RENDERARGS);
static void               MMXRender2BytesPerPixel_BltHScanLines(RENDERARGS);
static void               MMXRender3BytesPerPixel_BltHScanLines(RENDERARGS);
static void               MMXRender4BytesPerPixel_BltHScanLines(RENDERARGS);
static void               MMXRender2BytesPerPixel_HScanLines(RENDERARGS);
static void               MMXRender3BytesPerPixel_HScanLines(RENDERARGS);
static void               MMXRender4BytesPerPixel_HScanLines(RENDERARGS);
static void               MMXRender2BytesPerPixel_Double(RENDERARGS);
static void               MMXRender3BytesPerPixel_Double(RENDERARGS);
static void               MMXRender4BytesPerPixel_Double(RENDERARGS);
#endif

static BOOL               CreateDDrawObjects(int width, int height, float fps);
static BOOL               CreatePalettes(void);
static void               ReleaseDDrawObjects(void);
static void               ReleasePalettes(void);

static void               MapColor(unsigned char *pRed, unsigned char *pGreen, unsigned char *pBlue);
static void               SetPen(int pen, unsigned char red, unsigned char green, unsigned char blue);
static void               SetStaticPen(int pen, unsigned char red, unsigned char green, unsigned char blue);
static void               SetPaletteColors(void);
static void               AdjustPalette(void);
static void               AdjustColorMap(void);

static void               AdjustVisibleRect(int min_x, int min_y, int max_x, int max_y);
static void               GetPixelInfo(DWORD dwMask, struct tPixelInfo* pPixelInfo);

static int                CheckScreenDepth(void);

static void               SelectDisplayMode(int width, int height);
static BOOL               FindDisplayMode(DWORD dwWidth, DWORD dwHeight);
static BOOL               FindBestDisplayMode(DWORD dwWidthIn, DWORD  dwHeightIn, DWORD *pdwWidthOut, DWORD *pdwHeightOut);

static void               DrawSurface(void);
static void               FlipSurface(void);
static void               DoubleSurface(void);
static void               ClearSurface(IDirectDrawSurface7 *pddSurface);
static void               FillSurface(void);
static BOOL               LockSurface(IDirectDrawSurface7 *pddSurface);
static BOOL               UnlockSurface(IDirectDrawSurface7 *pddSurface);
static BOOL               RestoreSurface(void);
static BOOL               SurfaceLockable(IDirectDrawSurface7 *pddSurface);

static void               OnActivateApp(HWND hWnd, BOOL fActivate, DWORD dwThreadId);
static void               OnGetMinMaxInfo(HWND hWnd, MINMAXINFO* pMinMaxInfo);
static void               OnPaint(HWND hWnd);
static void               OnSize(HWND hwnd,UINT state,int cx,int cy);
static void               OnDrawItem(HWND hwnd, const DRAWITEMSTRUCT *lpDrawItem);
static void               OnCommand(HWND hWnd, int CmdID, HWND hWndCtl, UINT codeNotify);
static BOOL               OnSetCursor(HWND hWnd, HWND hWndCursor, UINT codeHitTest, UINT msg);

static INT_PTR CALLBACK   AboutDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam);

static void               ForceWindowMode(void);


/***************************************************************************
    External variables
 ***************************************************************************/

struct OSDDisplay DDrawDisplay =
{
    { DDraw_init },                 /* init              */
    { DDraw_exit },                 /* exit              */
    { DDraw_alloc_bitmap },         /* alloc_bitmap      */
    { DDraw_free_bitmap },          /* free_bitmap       */
    { DDraw_create_display },       /* create_display    */
    { DDraw_close_display },        /* close_display     */
    { DDraw_set_visible_area },     /* set_visible_area  */
#ifdef MAME_DEBUG
    { DDraw_set_debugger_focus },   /* set_debugger_focus*/
#endif
    { DDraw_allocate_colors },      /* allocate_colors   */
    { DDraw_modify_pen },           /* modify_pen        */
    { DDraw_get_pen },              /* get_pen           */
    { DDraw_skip_this_frame },      /* skip_this_frame   */
    { DDraw_update_display },       /* update_display    */
    { DDraw_led_w },                /* led_w             */
    { DDraw_set_gamma },            /* set_gamma         */
    { DDraw_get_gamma },            /* get_gamma         */
    { DDraw_set_brightness },       /* set_brightness    */
    { DDraw_get_brightness },       /* get_brightness    */
    { DDraw_save_snapshot },        /* save_snapshot     */

    { DDraw_OnMessage },            /* OnMessage         */
    { DDraw_Refresh },              /* Refresh           */
};

#ifdef KAILLERA
extern int kPlay;
#endif

/***************************************************************************
    Internal variables
 ***************************************************************************/

static const int safety = 8;
static options_type       tOptions;

/***************************************************************************
    External OSD function definitions
 ***************************************************************************/

/*
    put here anything you need to do when the program is started. Return 0 if
    initialization was successful, nonzero otherwise.
*/

UINT32 makecolor(unsigned char r, unsigned char g, unsigned char b)
{
	return MAKECOL(r, g, b);
}

UINT32 makecolor2(unsigned char r, unsigned char g, unsigned char b)
{
	unsigned char r1,g1,b1;

	if (options.ui_transparency)
	{
		r1 = (unsigned char)(r * options.ui_transparency / 100);
		g1 = (unsigned char)(g * options.ui_transparency / 100);
		b1 = (unsigned char)(b * options.ui_transparency / 100);
	}
	else
	{
		r1 = g1 = b1 = 0;
	}

	return MAKECOL(r1, g1, b1);
}

static int DDraw_init(options_type *options)
{
    memcpy(&tOptions, options, sizeof(options_type));

    dGamma            = max(0, tOptions.gamma);
    nBrightness       = max(0, tOptions.brightness);
    bShowFPS          = FALSE;
    nShowFPSTemp      = 0;
    bThrottled        = TRUE;
#ifdef ENABLE_MMX_RENDER
    bDisableMMX       = options->disable_mmx;
#endif

#ifdef KAILLERA
	if (kPlay)	
	{
        bAutoFrameskip = TRUE;
        nFrameSkip     = 0;
	}
	else
	{
	    if (tOptions.frame_skip == 0)
		{
		    bAutoFrameskip = TRUE;
		    nFrameSkip     = 0;
		}
	    else
	    {
	        bAutoFrameskip = FALSE;
	        nFrameSkip     = tOptions.frame_skip - 1;
	    }
	}
#else
	if (tOptions.frame_skip == 0)
	{
	    bAutoFrameskip = TRUE;
	    nFrameSkip     = 0;
	}
    else
    {
        bAutoFrameskip = FALSE;
        nFrameSkip     = tOptions.frame_skip - 1;
	}
#endif
    nFrameskipCounter = 0;
    nFrameskipAdjust  = 0;
    nFPS              = 60;
    nSpeed            = 100;
    PrevMeasure       = uclock();
    ThisFrameBase     = uclock();
    Prev              = uclock();
    frames_displayed  = 0;
    start_time        = uclock();
    end_time          = uclock();

    if (bAutoFrameskip == TRUE)
        nFrameSkip = 0;

    /* Initialize map. */
    AdjustColorMap();

    DDraw_init2();

#ifdef MAME_DEBUG
    MAME32Debug_init(options);
#endif

    return 0;
}


static int DDraw_init2(void)
{
    int nRenderMode;

    bWindowMode          = (tOptions.is_window) ? TRUE : FALSE;

    GameRect.m_Top       = 0;
    GameRect.m_Left      = 0;
    GameRect.m_Width     = 0;
    GameRect.m_Height    = 0;

    ClientRect.m_Left    = 0;
    ClientRect.m_Top     = 0;
    ClientRect.m_Height  = 0;
    ClientRect.m_Width   = 0;

    WindowRect.m_Left    = 0;
    WindowRect.m_Top     = 0;
    WindowRect.m_Height  = 0;
    WindowRect.m_Width   = 0;

    rectPrimary.left     = 0;
    rectPrimary.top      = 0;
    rectPrimary.right    = 0;
    rectPrimary.bottom   = 0;

    rectBack.left        = 0;
    rectBack.top         = 0;
    rectBack.right       = 0;
    rectBack.bottom      = 0;

    nDisplayIndex        = tOptions.display_monitor;
    bBestFit             = tOptions.display_best_fit;
    nScreenWidth         = tOptions.width;
    nScreenHeight        = tOptions.height;
    bSystemMemory        = tOptions.systemmemory;

    bAviCapture          = GetAviCapture();
    bAviRun              = FALSE;
    nAviShowMessage      = (bAviCapture) ? 10: 0;

    bUpdatePalette       = FALSE;
    bUpdateBackground    = FALSE;
    pPalEntries          = NULL;
    pAdjustedPalette     = NULL;
    pPalLookup           = NULL;
    nTotalColors         = 0;
    nBytesPerPixel       = 2;

    bInitialized         = FALSE;
    bPaused              = FALSE;
    bMinimized           = FALSE;

    pDDraw               = NULL;
    pDDSPrimary          = NULL;
    pDDSBack             = NULL;
    pDDSFlip             = NULL;
    pDDClipper           = NULL;

    Render               = NULL;

    pBitmapOffset        = NULL;
    nBitmapPitch         = 0;
    pSurfaceBits         = NULL;
    nSurfacePitch        = 0;

    hMenu                = NULL;
    hWndAbout            = NULL;

    memset(&Red,   0, sizeof(struct tPixelInfo));
    memset(&Green, 0, sizeof(struct tPixelInfo));
    memset(&Blue,  0, sizeof(struct tPixelInfo));

    if (tOptions.sound == SOUND_DIRECT)
        already_synced = 1;
    else
        already_synced = 0;

    /* Check for inconsistent parameters. */
    if (bWindowMode == TRUE)
    {
        nRenderMode   = tOptions.render_window;
        bVsync        = tOptions.vsync_window;
        bTripleBuffer = FALSE;
        bSleep        = (tOptions.dontsleep) ? FALSE : TRUE;
    }
    else
    {
        nRenderMode   = tOptions.render_fullscreen;
        bVsync        = tOptions.vsync_fullscreen;
        bTripleBuffer = tOptions.triple_buffer;
        bSleep        = FALSE;

        if (bTripleBuffer)
            bVsync = FALSE;
    }

    bUseBackBuffer = TRUE;

    switch (nRenderMode)
    {
        case 0: /* Normal */
            bHScanLines = FALSE;
            bDouble     = FALSE;
            bBltDouble  = FALSE;
            bFastDouble = FALSE;
            bStretch    = FALSE;
            nScale      = 1;
            if (bTripleBuffer)
                bUseBackBuffer = FALSE;
            break;

        case 1: /* Double */
            bHScanLines = FALSE;
            bDouble     = TRUE;
            bBltDouble  = FALSE;
            bFastDouble = FALSE;
            bStretch    = FALSE;
            nScale      = 2;
            if (bTripleBuffer)
                bUseBackBuffer = FALSE;
            break;

        case 2: /* Blit Fast Double */
            bHScanLines = FALSE;
            bDouble     = TRUE;
            bBltDouble  = FALSE;
            bFastDouble = TRUE;
            bStretch    = FALSE;
            nScale      = 2;
            break;

        case 3: /* Blit Double */
        case 4: /* Blit Triple */
        case 5: /* Blit Quad */
            bHScanLines = FALSE;
            bDouble     = TRUE;
            bBltDouble  = TRUE;
            bFastDouble = FALSE;
            bStretch    = FALSE;
            nScale      = nRenderMode - 1;
            break;

        case 6: /* Horizontal Scan Lines */
            bHScanLines = TRUE;
            bDouble     = TRUE;
            bBltDouble  = FALSE;
            bFastDouble = FALSE;
            bStretch    = FALSE;
            nScale      = 2;
            if (bTripleBuffer)
                bUseBackBuffer = FALSE;
            break;

        case 7: /* Blit Horizontal Scan Lines */
            bHScanLines = TRUE;
            bDouble     = TRUE;
            bBltDouble  = TRUE;
            bFastDouble = FALSE;
            bStretch    = FALSE;
            nScale      = 2;
            break;

        case 8: /* Stretch (full screen only) */
            bHScanLines = FALSE;
            bDouble     = FALSE;
            bBltDouble  = TRUE;
            bFastDouble = FALSE;
            bStretch    = TRUE;
            nScale      = 1;
            break;
    }

#if 0
    ErrorMsg("bDouble       = %s\n"
             "bHScanLines   = %s\n"
             "bFastDouble   = %s\n"
             "bBltDouble    = %s\n"
             "bStretch      = %s\n"
             "bTripleBuffer = %s\n",
             bDouble ? "TRUE" : "FALSE",
             bHScanLines ? "TRUE" : "FALSE",
             bFastDouble ? "TRUE" : "FALSE",
             bBltDouble ? "TRUE" : "FALSE",
             bStretch ? "TRUE" : "FALSE",
             bTripleBuffer ? "TRUE" : "FALSE");
#endif

    LED_init();

    return 0;
}


static void DDraw_exit(void)
{
    if (frames_displayed > FRAMES_TO_SKIP)
        printf("Average FPS: %f\n",
               (double)UCLOCKS_PER_SEC / (end_time - start_time) *
               (frames_displayed - FRAMES_TO_SKIP));

    ReleaseDDrawObjects();
    ReleasePalettes();
    LED_exit();

#ifdef MAME_DEBUG
    MAME32Debug_exit();
#endif
}


static struct osd_bitmap* DDraw_alloc_bitmap(int width, int height)
{
    struct osd_bitmap *pBitmap;

    pBitmap = (struct osd_bitmap *)MyMalloc(sizeof(struct osd_bitmap));

    if (pBitmap != NULL)
    {
        unsigned char *bm;
        int     i;
        int     rdwidth;
        int     rowlen;

        pBitmap->width  = width;
        pBitmap->height = height;
        pBitmap->depth  = 16;

        rdwidth = (width + 7) & ~7;     /* round width to a quadword */

        rowlen = 2 * (rdwidth + 2 * safety) * sizeof(unsigned char);

        bm = (unsigned char *)MyMalloc((height + 2 * safety) * rowlen);
        if (bm == NULL)
        {
            MyFree(pBitmap);
            return NULL;
        }

        /* clear ALL bitmap, including safety area, to avoid garbage on right */
        /* side of screen is width is not a multiple of 4 */
        memset(bm, 0, (height + 2 * safety) * rowlen);

        if ((pBitmap->line = MyMalloc((height + 2 * safety) * sizeof(unsigned char *))) == 0)
        {
            MyFree(bm);
            MyFree(pBitmap);
            return 0;
        }

        for (i = 0; i < height + 2 * safety; i++)
            pBitmap->line[i] = (unsigned short *)&bm[i * rowlen + safety * 2];

        pBitmap->line += safety;

        pBitmap->modulo = pBitmap->line[1] - pBitmap->line[0];

        pBitmap->_private = bm;

        if (pBitmap == Machine->scrbitmap)
            schedule_full_refresh();
    }

    return pBitmap;
}


static void DDraw_free_bitmap(struct osd_bitmap *pBitmap)
{
    if (pBitmap != NULL)
    {
        pBitmap->line -= safety;
        MyFree(pBitmap->line);
        MyFree(pBitmap->_private);
        MyFree(pBitmap);
    }
}


static int DDraw_create_display(int width, int height, float fps)
{
    char TitleBuf[256];

    AdjustVisibleRect(0, 0, width - 1, height - 1);

    ShowWindow(MAME32App.m_hWnd, SW_HIDE);

    sprintf(TitleBuf, "%s - %s", Machine->gamedrv->description, MAME32App.m_Name);
    SetWindowText(MAME32App.m_hWnd, TitleBuf);

    if (bWindowMode)
    {
        BOOL b_Vsync = tOptions.is_window ? tOptions.vsync_window : tOptions.vsync_fullscreen;

        hMenu = LoadMenu(GetModuleHandle(NULL), MAKEINTRESOURCE(IDR_MAIN_MENU));
        SetMenu(MAME32App.m_hWnd, hMenu);
        CheckMenuItem(hMenu, ID_CHANGE_MODE1 + tOptions.render_window, MF_CHECKED);
        CheckMenuItem(hMenu, ID_MENU_VSYNC, b_Vsync ? MF_CHECKED : MF_UNCHECKED);
        CheckMenuItem(hMenu, ID_MENU_FULLSCREEN, tOptions.is_window ? MF_UNCHECKED : MF_CHECKED);

        SetWindowLong(MAME32App.m_hWnd, GWL_STYLE, WS_OVERLAPPEDWINDOW & ~WS_THICKFRAME | WS_BORDER);

        StatusCreate();

        WindowRect.m_Width  = (GameRect.m_Width * nScale) + (2 * GetSystemMetrics(SM_CXFIXEDFRAME));
        WindowRect.m_Height = (GameRect.m_Height * nScale) + (2 * GetSystemMetrics(SM_CYFIXEDFRAME)) + GetSystemMetrics(SM_CYCAPTION) + GetStatusHeight();
        WindowRect.m_Height += (hMenu != NULL) ? GetSystemMetrics(SM_CYMENU) : 0;

#ifdef MAME_DEBUG
        WindowRect.m_Left   = 0;
        WindowRect.m_Top    = 0;
#else
        WindowRect.m_Left   = (GetSystemMetrics(SM_CXSCREEN) - WindowRect.m_Width)  / 2;
        WindowRect.m_Top    = (GetSystemMetrics(SM_CYSCREEN) - WindowRect.m_Height) / 2;
#endif

        SetWindowPos(MAME32App.m_hWnd,
                     HWND_NOTOPMOST,
                     WindowRect.m_Left,
                     WindowRect.m_Top,
                     WindowRect.m_Width + 1,
                     WindowRect.m_Height,
                     SWP_NOZORDER);

        hWndAbout = CreateDialog(GetModuleHandle(NULL),
                                 MAKEINTRESOURCE(IDD_ABOUT),
                                 MAME32App.m_hWnd,
                                 AboutDialogProc);
    }
    else
    {
        SetWindowLong(MAME32App.m_hWnd, GWL_STYLE, WS_POPUP);

        WindowRect.m_Left   = 0;
        WindowRect.m_Top    = 0;
        WindowRect.m_Width  = nScreenWidth;
        WindowRect.m_Height = nScreenHeight;

        SetWindowPos(MAME32App.m_hWnd,
                     HWND_TOPMOST,
                     0,
                     0,
                     GetSystemMetrics(SM_CXSCREEN) / 4,
                     GetSystemMetrics(SM_CYSCREEN) / 4,
                     0);
    }

    if (CreateDDrawObjects(width, height, fps) == FALSE)
    {
        DDraw_close_display();
        return 1;
    }

    ShowWindow(MAME32App.m_hWnd, SW_SHOW);

#ifdef MAME_DEBUG
    if (bWindowMode)
    {
        MAME32Debug_create_display(options.debug_width,
                                   options.debug_height,
                                   16, fps, 0, 0);
    }
#endif

    SetAbsoluteForegroundWindow(MAME32App.m_hWnd);

    return 0;
}


static int DDraw_change_display(void)
{
    int i;
    unsigned char palbak[4096][3];

    osd_stop_audio_stream();

    for (i = 0; i < 4096; i++)
        osd_get_pen(i, &palbak[i][0], &palbak[i][1], &palbak[i][2]);

    DDraw_close_display();

    DDraw_init2();

    if (DDraw_create_display(
            Machine->drv->default_visible_area.max_x - Machine->drv->default_visible_area.min_x + 1,
            Machine->drv->default_visible_area.max_y - Machine->drv->default_visible_area.min_y + 1,
            Machine->drv->frames_per_second))
        return 1;

    set_visible_area(
            Machine->drv->default_visible_area.min_x,
            Machine->drv->default_visible_area.max_x,
            Machine->drv->default_visible_area.min_y,
            Machine->drv->default_visible_area.max_y);

    if (palette_init())
        return 1;

    for (i = 0; i < 4096; i++)
        palette_change_color(i, palbak[i][0], palbak[i][1], palbak[i][2]);

    DDraw_Refresh();

    for (i = 0; i < 10; i++)
    {
        UpdateDisplay();
        Sleep(1);
    }

    osd_start_audio_stream(tOptions.stereo);

    return 0;
}


static void DDraw_close_display(void)
{
    extern int usres;

    bInitialized = FALSE;

    ShowWindow(MAME32App.m_hWnd, SW_HIDE);

    if (bWindowMode)
    {
        if (hMenu)
        {
            DestroyMenu(hMenu);
            hMenu = NULL;
        }

        if (IsWindow(hWndAbout))
        {
            EndDialog(hWndAbout, 0);
            hWndAbout = NULL;
        }

        StatusDelete();
    }
    else if (pDDraw != NULL)
    {
        IDirectDraw7_RestoreDisplayMode(pDDraw);
        IDirectDraw7_SetCooperativeLevel(pDDraw, MAME32App.m_hWnd, DDSCL_NORMAL);
        ShowWindow(GetMainWindow(), SW_HIDE);
    }

    ReleaseDDrawObjects();

    ReleasePalettes();

#ifdef MAME_DEBUG
    MAME32Debug_close_display();
#endif
}


static void DDraw_set_visible_area(int min_x, int max_x, int min_y, int max_y)
{
    AdjustVisibleRect(min_x, min_y, max_x, max_y);

    set_ui_visarea(GameRect.m_Left,
                   GameRect.m_Top,
                   GameRect.m_Left + GameRect.m_Width  - 1,
                   GameRect.m_Top  + GameRect.m_Height - 1);
}


#ifdef MAME_DEBUG
static void DDraw_set_debugger_focus(int debugger_has_focus)
{
    if (!debugger_has_focus)
    {
        SetForegroundWindow(MAME32App.m_hWnd);
    }

    MAME32Debug_set_debugger_focus(debugger_has_focus);
}
#endif

#ifdef MAME_DEBUG
static int DDraw_allocate_colors(const UINT8 *debug_palette, UINT16 *debug_pens)
#else
static int DDraw_allocate_colors(void)
#endif
{
    unsigned int i;

    if (CreatePalettes() == FALSE)
        return 1;

    for (i = 0; i < nTotalColors; i++)
        SetPen(i, 0, 0, 0);

    /* reserve colors for the user interface text */
    Machine->uifont->colortable[0] = 0;
    Machine->uifont->colortable[1] = 8192;
    Machine->uifont->colortable[2] = 8192;
    Machine->uifont->colortable[3] = 0;

    SetStaticPen(0x2000, 0xff, 0xff, 0xff);	//FONT_COLOR_NORMAL
    SetStaticPen(0x2001, 0xff, 0x00, 0x00);	//BUTTON_COLOR_RED
    SetStaticPen(0x2002, 0xff, 0xee, 0x00);	//BUTTON_COLOR_YELLOW
    SetStaticPen(0x2003, 0x00, 0xff, 0x33);	//BUTTON_COLOR_GREEN
    SetStaticPen(0x2004, 0x00, 0xaa, 0xff);	//BUTTON_COLOR_BLUE
	SetStaticPen(0x2005, 0x99, 0x00, 0x00); //VOLUME_DARKRED
	SetStaticPen(0x2006, 0x99, 0x88, 0x00); //VOLUME_DARKYELLOW
    SetStaticPen(0x2007, 0x00, 0x88, 0x33);	//VOLUME_DARKGREEN
	SetStaticPen(0x2008, 0xc0, 0xc0, 0xc0);	//SYSTEM_COLOR_FRAMEMIDIUM
	SetStaticPen(0x2009, 0xe0, 0xe0, 0xe0);	//SYSTEM_COLOR_FRAMELIGHT
	SetStaticPen(0x200a, 0x80, 0x80, 0x80);	//SYSTEM_COLOR_FRAMEDARK
	SetStaticPen(0x200b, 0x3c, 0x78, 0xf0);	//CURSOR_COLOR

    bUpdatePalette = TRUE;

    SelectRenderMethod();

    /* Clear buffers. */
    ClearSurface(pDDSPrimary);
    ClearSurface(pDDSBack);
    UpdateWindow(MAME32App.m_hWnd);

    DDraw_Refresh();

    bInitialized = TRUE;

#ifdef MAME_DEBUG
    MAME32Debug_allocate_colors(FALSE, debug_palette, debug_pens);
#endif

    return 0;
}


static void DDraw_modify_pen(int pen, unsigned char red, unsigned char green, unsigned char blue)
{
    assert(0 <= pen && pen < nTotalColors);

#if 0
    if (pPalEntries[pen].peRed   == red
    &&  pPalEntries[pen].peGreen == green
    &&  pPalEntries[pen].peBlue  == blue)
        return;
#endif

    SetPen(pen, red, green, blue);

    bUpdatePalette = TRUE;
}


static void DDraw_get_pen(int pen, unsigned char* pRed, unsigned char* pGreen, unsigned char* pBlue)
{
    assert(pRed   != NULL);
    assert(pGreen != NULL);
    assert(pBlue  != NULL);
    assert(0 <= pen && pen < nTotalColors);

    if (nTotalColors <= pen)
        pen = 0;

    *pRed   = pPalEntries[pen].peRed;
    *pGreen = pPalEntries[pen].peGreen;
    *pBlue  = pPalEntries[pen].peBlue;
}


static int DDraw_skip_this_frame(void)
{
    static const int skiptable[FRAMESKIP_LEVELS][FRAMESKIP_LEVELS] =
    {
        { 0,0,0,0,0,0,0,0,0,0,0,0 },
        { 0,0,0,0,0,0,0,0,0,0,0,1 },
        { 0,0,0,0,0,1,0,0,0,0,0,1 },
        { 0,0,0,1,0,0,0,1,0,0,0,1 },
        { 0,0,1,0,0,1,0,0,1,0,0,1 },
        { 0,1,0,0,1,0,1,0,0,1,0,1 },
        { 0,1,0,1,0,1,0,1,0,1,0,1 },
        { 0,1,0,1,1,0,1,0,1,1,0,1 },
        { 0,1,1,0,1,1,0,1,1,0,1,1 },
        { 0,1,1,1,0,1,1,1,0,1,1,1 },
        { 0,1,1,1,1,1,0,1,1,1,1,1 },
        { 0,1,1,1,1,1,1,1,1,1,1,1 }
    };

    assert(0 <= nFrameSkip && nFrameSkip < FRAMESKIP_LEVELS);
    assert(0 <= nFrameskipCounter && nFrameskipCounter < FRAMESKIP_LEVELS);

    return skiptable[nFrameSkip][nFrameskipCounter];
}


#ifdef MAME_DEBUG
static void DDraw_update_display(struct osd_bitmap *game_bitmap, struct osd_bitmap *debug_bitmap)
#else
static void DDraw_update_display(struct osd_bitmap *game_bitmap)
#endif
{
    uclock_t curr;

    if (bInitialized == FALSE)
    {
        PrevMeasure = uclock() - FRAMESKIP_LEVELS * UCLOCKS_PER_SEC / Machine->drv->frames_per_second;
        return;
    }

    if (nFrameskipCounter == 0)
        ThisFrameBase = PrevMeasure + FRAMESKIP_LEVELS * UCLOCKS_PER_SEC / Machine->drv->frames_per_second;

    curr = uclock();

    if (osd_skip_this_frame() == 0)
    {
        /* obNobt@ɕ` */
        profiler_mark(PROFILER_RENDER);
        UpdateDisplay();
        profiler_mark(PROFILER_RENDER);

        if (nShowFPSTemp)
        {
            nShowFPSTemp--;

            if (bShowFPS == 0 && nShowFPSTemp == 0)
                DDraw_UpdateFPS(FALSE, 0, 0, 0, 0, 0);
        }

        if (input_ui_pressed(IPT_UI_SHOW_FPS))
        {
            if (nShowFPSTemp)
            {
                nShowFPSTemp = 0;
                DDraw_UpdateFPS(FALSE, 0, 0, 0, 0, 0);
            }
            else
            {
                bShowFPS = !bShowFPS;

                if (bShowFPS == 0)
                    DDraw_UpdateFPS(FALSE, 0, 0, 0, 0, 0);
            }
        }

        curr = uclock();

        if (bThrottled) // && !already_synced)
        {
            uclock_t target;

            profiler_mark(PROFILER_IDLE);

            /* wait until enough time has passed since last frame... */
            target = ThisFrameBase + nFrameskipCounter * UCLOCKS_PER_SEC / (Machine->drv->frames_per_second + 1);

            if (bSleep)
            {
                if (target - curr > 0)
                {
                    uclock_t sleeptime = (target - curr) / (UCLOCKS_PER_SEC/1000) - 1;

                    if (sleeptime < 0)
                        sleeptime = 0;

                    Sleep(sleeptime);
                }
            }

            do
            {
                curr = uclock();
            }
            while (curr - target < 0);

            profiler_mark(PROFILER_END);
        }

        if (++frames_displayed == FRAMES_TO_SKIP)
            start_time = curr;
        else
            end_time = curr;

        if (nFrameskipCounter == 0)
        {
            int divdr;

            divdr = Machine->drv->frames_per_second * (curr - PrevMeasure) / (100 * FRAMESKIP_LEVELS);
            nSpeed = (UCLOCKS_PER_SEC + divdr / 2) / divdr;
            PrevMeasure = curr;
        }

        Prev = curr;

        if (bShowFPS || nShowFPSTemp || bAviCapture)
        {
            int divdr;

            divdr = 100 * FRAMESKIP_LEVELS;
            nFPS = (Machine->drv->frames_per_second * (FRAMESKIP_LEVELS - nFrameSkip) * nSpeed + (divdr / 2)) / divdr;
        }

        /* ۂ̉ʂɓ] */
        profiler_mark(PROFILER_BLIT);
        if (bTripleBuffer)
            FlipSurface();
        else
            DrawSurface();
        profiler_mark(PROFILER_END);

#ifdef MAME_DEBUG
        MAME32Debug_update_display(debug_bitmap);
#endif

        if (bWindowMode)
            StatusUpdate();

        MAME32App.ProcessMessages();

        if (bThrottled && bAutoFrameskip && nFrameskipCounter == 0)
        {
            if (nSpeed >= 100)
            {
                nFrameskipAdjust++;
                if (nFrameskipAdjust >= 3)
                {
                    nFrameskipAdjust = 0;
                    if (nFrameSkip > 0) nFrameSkip--;
                }
            }
            else
            {
                if (nSpeed < 80)
                {
                    nFrameskipAdjust -= (90 - nSpeed) / 5;
                }
                else
                {
                    /* don't push frameskip too far if we are close to 100% speed */
                    if (nFrameSkip < 8)
                        nFrameskipAdjust--;
                }

                while (nFrameskipAdjust <= -2)
                {
                    nFrameskipAdjust += 2;
                    if (nFrameSkip < FRAMESKIP_LEVELS - 1)
                        nFrameSkip++;
                }
            }
        }
    }

#ifdef KAILLERA
	if (!kPlay)
#endif
    if (input_ui_pressed(IPT_UI_FRAMESKIP_INC))
    {
        if (bAutoFrameskip == TRUE)
        {
            bAutoFrameskip = FALSE;
            nFrameSkip     = 0;
        }
        else
        {
            if (nFrameSkip == FRAMESKIP_LEVELS - 1)
            {
                nFrameSkip     = 0;
                bAutoFrameskip = TRUE;
            }
            else
                nFrameSkip++;
        }

        if (bShowFPS == FALSE)
            nShowFPSTemp = 2 * Machine->drv->frames_per_second;

        /* reset the frame counter every time the frameskip key is pressed, so */
        /* we'll measure the average FPS on a consistent status. */
        frames_displayed = 0;
    }

#ifdef KAILLERA
	if (!kPlay)
#endif
    if (input_ui_pressed(IPT_UI_FRAMESKIP_DEC))
    {
        if (bAutoFrameskip)
        {
            bAutoFrameskip = FALSE;
            nFrameSkip     = FRAMESKIP_LEVELS - 1;
        }
        else
        {
            if (nFrameSkip == 0)
                bAutoFrameskip = TRUE;
            else
                nFrameSkip--;
        }

        if (bShowFPS == FALSE)
            nShowFPSTemp = 2 * Machine->drv->frames_per_second;

        /* reset the frame counter every time the frameskip key is pressed, so */
        /* we'll measure the average FPS on a consistent status. */
        frames_displayed = 0;
    }

#ifdef KAILLERA
	if (!kPlay)
#endif
    if (input_ui_pressed(IPT_UI_THROTTLE))
    {
        bThrottled = !bThrottled;

        /* reset the frame counter every time the throttle key is pressed, so */
        /* we'll measure the average FPS on a consistent status. */
        frames_displayed = 0;
    }

    nFrameskipCounter = (nFrameskipCounter + 1) % FRAMESKIP_LEVELS;

#ifdef KAILLERA
	if (!kPlay)
#endif
    if (input_ui_pressed(IPT_UI_TOGGLE_DISPLAY))
    {
        tOptions.is_window = !tOptions.is_window;
        DDraw_change_display();
    }
}


/* control keyboard leds or other indicators */
static void DDraw_led_w(int leds_status)
{
    if (bWindowMode)
        StatusWrite(leds_status);
    else
        LED_StatusWrite(leds_status);
}


static void DDraw_set_gamma(float gamma)
{
    if (gamma < 0.5) gamma = 0.5;
    if (gamma > 2.0) gamma = 2.0;

    dGamma = gamma;
    AdjustColorMap();
    AdjustPalette();
}


static float DDraw_get_gamma(void)
{
    return (float)dGamma;
}


static void DDraw_set_brightness(int brightness)
{
    nBrightness = brightness;
    AdjustColorMap();
    AdjustPalette();
}


static int DDraw_get_brightness(void)
{
    return nBrightness;
}


static void DDraw_save_snapshot(struct osd_bitmap *bitmap)
{
    if (bAviCapture)
    {
        nAviShowMessage = 120;       /* show message for 120 frames */
        bAviRun = !bAviRun;   /* toggle capture on/off */
    }
    else
        save_screen_snapshot(bitmap);
}


static BOOL DDraw_OnMessage(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam, LRESULT* pResult)
{
    if (bWindowMode)
    {
        switch (Msg)
        {
            PEEK_MESSAGE(hWnd, WM_ACTIVATEAPP,     OnActivateApp);
            HANDLE_MESSAGE(hWnd, WM_GETMINMAXINFO, OnGetMinMaxInfo);
            HANDLE_MESSAGE(hWnd, WM_COMMAND,       OnCommand);
            HANDLE_MESSAGE(hWnd, WM_PAINT,         OnPaint);
            HANDLE_MESSAGE(hWnd, WM_SIZE,          OnSize);
            HANDLE_MESSAGE(hWnd, WM_DRAWITEM,      OnDrawItem);
        }
    }
    else
    {
        switch (Msg)
        {
            HANDLE_MESSAGE(hWnd, WM_SETCURSOR,     OnSetCursor);
        }
    }

    return FALSE;
}


static void DDraw_Refresh(void)
{
    if (bWindowMode)
    {
        InvalidateRect(MAME32App.m_hWnd, NULL, FALSE);
    }
    else
    {
        bUpdateBackground = TRUE;
#ifdef MAME_DEBUG
        DDraw_update_display(Machine->scrbitmap, Machine->debug_bitmap);
#else
        DDraw_update_display(Machine->scrbitmap);
#endif
    }
}


static void DDraw_UpdateFPS(BOOL bShow, BOOL bAuto, int nSpeed, int nFPS, int nMachineFPS, int nFrameskip)
{
    char buf[64];

    if (bWindowMode)
    {
        StatusUpdateFPS(bShow, bAuto, nSpeed, nFPS, nMachineFPS, nFrameskip);
    }
    else if (bShow)
    {
        sprintf(buf, "%s%2d%4d%%%4d/%d fps", bAuto ? "auto" : "fskp", nFrameskip, nSpeed, nFPS, nMachineFPS);
        ui_text(Machine->scrbitmap, buf, Machine->uiwidth - strlen(buf) * Machine->uifontwidth, 0);
    }
}


BOOL Display_Throttled(void)
{
    return bThrottled;
}


/***************************************************************************
    Internal functions
 ***************************************************************************/

/*===========================================================================
    ʍXV
 ==========================================================================*/

static void UpdateDisplay(void)
{
#if 0
    if (pDDSBack    == NULL
    ||  pDDSPrimary == NULL)
        return;
#endif

    if (bUpdatePalette == TRUE)
    {
        SetPaletteColors();
        bUpdatePalette = FALSE;
    }

    if (bAviRun)
        AviAddBitmap(Machine->scrbitmap);

    if (bInitialized == TRUE && bMinimized == FALSE)
    {
        if (nAviShowMessage > 0)
        {
            char buf[32];

            nAviShowMessage--;

            sprintf(buf, "AVI Capture %s", (bAviRun) ? "ON" : "OFF");
            if (bWindowMode)
                StatusSetString(buf);
            else
                ui_text(Machine->scrbitmap, buf, Machine->uiwidth - strlen(buf) * Machine->uifontwidth, 0);
        }

        if (bAviCapture)
            SetAviFPS(nFPS);

        if (bShowFPS || nShowFPSTemp)
            DDraw_UpdateFPS(TRUE, bAutoFrameskip, nSpeed, nFPS, Machine->drv->frames_per_second, nFrameSkip);

        if (bUpdateBackground == TRUE || bTripleBuffer == TRUE)
        {
            ClearSurface(pDDSPrimary);
            bUpdateBackground = FALSE;
        }

        if (pDDSBack)
        {
            if (bFastDouble)
                FillSurface();

            if (LockSurface(pDDSBack))
            {
                Render(pBitmapOffset, pSurfaceBits, nBitmapPitch, nSurfacePitch, GameRect.m_Width, GameRect.m_Height);
                UnlockSurface(pDDSBack);
            }

            if (bFastDouble)
                DoubleSurface();
        }

        if (bTripleBuffer && bUseBackBuffer)
            DrawSurface();
    }
}


static void SelectRenderMethod(void)
{
#ifdef ENABLE_MMX_RENDER
    if (bDisableMMX)
    {
#endif
        switch (nBytesPerPixel)
        {
            case 2:
                if (bHScanLines == TRUE && bBltDouble == TRUE)
                {
                    Render = Render2BytesPerPixel_BltHScanLines;
                }
                else if (bHScanLines == TRUE || bFastDouble == TRUE)
                {
                    Render = Render2BytesPerPixel_HScanLines;
                }
                else if (bDouble == TRUE && bBltDouble == FALSE)
                {
                    Render = Render2BytesPerPixel_Double;
                }
                else
                {
                    Render = Render2BytesPerPixel;
                }
                break;

            case 3:
                if (bHScanLines == TRUE && bBltDouble == TRUE)
                {
                    Render = Render3BytesPerPixel_BltHScanLines;
                }
                else if (bHScanLines == TRUE || bFastDouble == TRUE)
                {
                    Render = Render3BytesPerPixel_HScanLines;
                }
                else if (bDouble == TRUE && bBltDouble == FALSE)
                {
                    Render = Render3BytesPerPixel_Double;
                }
                else
                {
                    Render = Render3BytesPerPixel;
                }
                break;

            case 4:
                if (bHScanLines == TRUE && bBltDouble == TRUE)
                {
                    Render = Render4BytesPerPixel_BltHScanLines;
                }
                else if (bHScanLines == TRUE || bFastDouble == TRUE)
                {
                    Render = Render4BytesPerPixel_HScanLines;
                }
                else if (bDouble == TRUE && bBltDouble == FALSE)
                {
                    Render = Render4BytesPerPixel_Double;
                }
                else
                {
                    Render = Render4BytesPerPixel;
                }
                break;

            default:
                Render = NULL;
                break;
        }
#ifdef ENABLE_MMX_RENDER
    }
    else
    {
        switch (nBytesPerPixel)
        {
            case 2:
                if (bHScanLines == TRUE && bBltDouble == TRUE)
                {
                    Render = MMXRender2BytesPerPixel_BltHScanLines;
                }
                else if (bHScanLines == TRUE || bFastDouble == TRUE)
                {
                    Render = MMXRender2BytesPerPixel_HScanLines;
                }
                else if (bDouble == TRUE && bBltDouble == FALSE)
                {
                    Render = MMXRender2BytesPerPixel_Double;
                }
                else
                {
                    Render = MMXRender2BytesPerPixel;
                }
                break;

            case 3:
                if (bHScanLines == TRUE && bBltDouble == TRUE)
                {
                    Render = MMXRender3BytesPerPixel_BltHScanLines;
                }
                else if (bHScanLines == TRUE || bFastDouble == TRUE)
                {
                    Render = MMXRender3BytesPerPixel_HScanLines;
                }
                else if (bDouble == TRUE && bBltDouble == FALSE)
                {
                    Render = MMXRender3BytesPerPixel_Double;
                }
                else
                {
                    Render = MMXRender3BytesPerPixel;
                }
                break;

            case 4:
                if (bHScanLines == TRUE && bBltDouble == TRUE)
                {
                    Render = MMXRender4BytesPerPixel_BltHScanLines;
                }
                else if (bHScanLines == TRUE || bFastDouble == TRUE)
                {
                    Render = MMXRender4BytesPerPixel_HScanLines;
                }
                else if (bDouble == TRUE && bBltDouble == FALSE)
                {
                    Render = MMXRender4BytesPerPixel_Double;
                }
                else
                {
                    Render = MMXRender4BytesPerPixel;
                }
                break;

            default:
                Render = NULL;
                break;
        }
    }
#endif

    if (bFastDouble)
    {
        DDCOLORKEY ddck;

        ddck.dwColorSpaceLowValue  = 0x1234;
        ddck.dwColorSpaceHighValue = 0x1234;
        IDirectDrawSurface7_SetColorKey(pDDSBack, DDCKEY_SRCBLT, &ddck);
    }
}


/*===========================================================================
    _[ (MAMEzʂDirectDrawT[tFXւ̕`)
 ==========================================================================*/

static void Render2BytesPerPixel(RENDERARGS)
{
    __asm {
        shr       nWidth, 1
        mov       ebx, pPalLookup
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        xor       eax, eax
        mov       ax,  [esi]
        mov       edx, [ebx + eax * 4]
        and       edx, 0000ffffh
        mov       ax,  [esi + 2]
        mov       eax, [ebx + eax * 4]
        and       eax, 0ffff0000h
        or        eax, edx
        mov       [edi], eax
        lea       esi, [esi + 4]
        lea       edi, [edi + 4]
        dec       ecx
        jnz       L2

        mov       eax, nSrcLen
        mov       edx, nDstLen
        add       pSrc, eax
        add       pDst, edx
        dec       nHeight
        jnz       L1
    }
}

static void Render3BytesPerPixel(RENDERARGS)
{
    __asm {
        mov       ebx, pPalLookup
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        xor       eax, eax
        mov       ax,  [esi]
        mov       eax, [ebx + eax * 4]
        mov       [edi], eax
        lea       esi, [esi + 2]
        lea       edi, [edi + 3]
        dec       ecx
        jnz       L2

        mov       eax, nSrcLen
        mov       edx, nDstLen
        add       pSrc, eax
        add       pDst, edx
        dec       nHeight
        jnz       L1
    }
}

static void Render4BytesPerPixel(RENDERARGS)
{
    __asm {
        mov       ebx, pPalLookup
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        xor       eax, eax
        mov       ax,  [esi]
        mov       eax, [ebx + eax * 4]
        mov       [edi], eax
        lea       esi, [esi + 2]
        lea       edi, [edi + 4]
        dec       ecx
        jnz       L2

        mov       eax, nSrcLen
        mov       edx, nDstLen
        add       pSrc, eax
        add       pDst, edx
        dec       nHeight
        jnz       L1
    }
}

static void Render2BytesPerPixel_BltHScanLines(RENDERARGS)
{
    __asm {
        shl       nDstLen, 1
        shr       nWidth, 1
        mov       ebx, pPalLookup
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        xor       eax, eax
        mov       ax,  [esi]
        mov       edx, [ebx + eax * 4]
        and       edx, 0000ffffh
        mov       ax,  [esi + 2]
        mov       eax, [ebx + eax * 4]
        and       eax, 0ffff0000h
        or        eax, edx
        mov       [edi], eax
        lea       esi, [esi + 4]
        lea       edi, [edi + 4]
        dec       ecx
        jnz       L2

        mov       eax, nSrcLen
        mov       edx, nDstLen
        add       pSrc, eax
        add       pDst, edx
        dec       nHeight
        jnz       L1
    }
}

static void Render3BytesPerPixel_BltHScanLines(RENDERARGS)
{
    __asm {
        shl       nDstLen, 1
        mov       ebx, pPalLookup
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        xor       eax, eax
        mov       ax,  [esi]
        mov       eax, [ebx + eax * 4]
        mov       [edi], eax
        lea       esi, [esi + 2]
        lea       edi, [edi + 3]
        dec       ecx
        jnz       L2

        mov       eax, nSrcLen
        mov       edx, nDstLen
        add       pSrc, eax
        add       pDst, edx
        dec       nHeight
        jnz       L1
    }
}

static void Render4BytesPerPixel_BltHScanLines(RENDERARGS)
{
    __asm {
        shl       nDstLen, 1
        mov       ebx, pPalLookup
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        xor       eax, eax
        mov       ax,  [esi]
        mov       eax, [ebx + eax * 4]
        mov       [edi], eax
        lea       esi, [esi + 2]
        lea       edi, [edi + 4]
        dec       ecx
        jnz       L2

        mov       eax, nSrcLen
        mov       edx, nDstLen
        add       pSrc, eax
        add       pDst, edx
        dec       nHeight
        jnz       L1
    }
}

/*----- doublen ------*/

static void Render2BytesPerPixel_HScanLines(RENDERARGS)
{
    __asm {
        shl       nDstLen, 1
        mov       ebx, pPalLookup
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        xor       eax, eax
        mov       ax,  [esi]
        mov       eax, [ebx + eax * 4]
        mov       [edi], eax
        lea       esi, [esi + 2]
        lea       edi, [edi + 4]
        dec       ecx
        jnz       L2

        mov       eax, nSrcLen
        mov       edx, nDstLen
        add       pSrc, eax
        add       pDst, edx
        dec       nHeight
        jnz       L1
    }
}

static void Render3BytesPerPixel_HScanLines(RENDERARGS)
{
    __asm {
        shl       nDstLen, 1
        mov       ebx, pPalLookup
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        xor       eax, eax
        mov       ax,  [esi]
        mov       eax, [ebx + eax * 4]
        mov       [edi], eax
        mov       [edi + 3], eax
        lea       esi, [esi + 2]
        add       edi, [edi + 6]
        dec       ecx
        jnz       L2

        mov       eax, nSrcLen
        mov       edx, nDstLen
        add       pSrc, eax
        add       pDst, edx
        dec       nHeight
        jnz       L1
    }
}

static void Render4BytesPerPixel_HScanLines(RENDERARGS)
{
    __asm {
        shl       nDstLen, 1
        mov       ebx, pPalLookup
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        xor       eax, eax
        mov       ax,  [esi]
        mov       eax, [ebx + eax * 4]
        mov       [edi], eax
        mov       [edi + 4], eax
        lea       esi, [esi + 2]
        lea       edi, [edi + 8]
        dec       ecx
        jnz       L2

        mov       eax, nSrcLen
        mov       edx, nDstLen
        add       pSrc, eax
        add       pDst, edx
        dec       nHeight
        jnz       L1
    }
}

static void Render2BytesPerPixel_Double(RENDERARGS)
{
    __asm {
        mov       ebx, pPalLookup
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        xor       eax, eax
        mov       ax,  [esi]
        mov       eax, [ebx + eax * 4]
        mov       [edi], eax
        lea       esi, [esi + 2]
        lea       edi, [edi + 4]
        dec       ecx
        jnz       L2

        mov       eax, nDstLen
        add       pDst, eax
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L3:
        xor       eax, eax
        mov       ax,  [esi]
        mov       eax, [ebx + eax * 4]
        mov       [edi], eax
        lea       esi, [esi + 2]
        lea       edi, [edi + 4]
        dec       ecx
        jnz       L3

        mov       eax, nSrcLen
        mov       edx, nDstLen
        add       pSrc, eax
        add       pDst, edx
        dec       nHeight
        jnz       L1
    }
}

static void Render3BytesPerPixel_Double(RENDERARGS)
{
    __asm {
        shl       nWidth, 1
        mov       ebx, pPalLookup
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        xor       eax, eax
        mov       ax,  [esi]
        mov       eax, [ebx + eax * 4]
        mov       [edi], eax
        mov       [edi + 3], eax
        lea       esi, [esi + 2]
        lea       edi, [edi + 6]
        dec       ecx
        jnz       L2

        mov       eax, nDstLen
        add       pDst, eax
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L3:
        xor       eax, eax
        mov       ax,  [esi]
        mov       eax, [ebx + eax * 4]
        mov       [edi], eax
        mov       [edi + 3], eax
        lea       esi, [esi + 2]
        lea       edi, [edi + 6]
        dec       ecx
        jnz       L3

        mov       eax, nSrcLen
        mov       edx, nDstLen
        add       pSrc, eax
        add       pDst, edx
        dec       nHeight
        jnz       L1
    }
}

static void Render4BytesPerPixel_Double(RENDERARGS)
{
    __asm {
        shl       nWidth, 1
        mov       ebx, pPalLookup
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        xor       eax, eax
        mov       ax,  [esi]
        mov       eax, [ebx + eax * 4]
        mov       [edi], eax
        mov       [edi + 4], eax
        lea       esi, [esi + 2]
        lea       edi, [edi + 8]
        dec       ecx
        jnz       L2

        mov       eax, nDstLen
        add       pDst, eax
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L3:
        xor       eax, eax
        mov       ax,  [esi]
        mov       eax, [ebx + eax * 4]
        mov       [edi], eax
        mov       [edi + 4], eax
        lea       esi, [esi + 2]
        lea       edi, [edi + 8]
        dec       ecx
        jnz       L3

        mov       eax, nSrcLen
        mov       edx, nDstLen
        add       pSrc, eax
        add       pDst, edx
        dec       nHeight
        jnz       L1
    }
}


/*===========================================================================
    _[ (MAMEzʂDirectDrawT[tFXւ̕` / MMXgp)

  24bit(3 byte per pixcel)̓`FbNłȂ̂ŁAǂ܂
 ==========================================================================*/

#ifdef ENABLE_MMX_RENDER
static void MMXRender2BytesPerPixel(RENDERARGS)
{
    __asm {
        shr       nWidth, 2
        mov       ebx, pPalLookup
        xor       eax, eax
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
        xor       edx, edx
L2:
        mov       ax, [esi]
        mov       dx, [esi + 2]
        movd      mm0, [ebx + eax * 4]
        movd      mm1, [ebx + edx * 4]
        mov       ax, [esi + 4]
        mov       dx, [esi + 6]
        movd      mm2, [ebx + eax * 4]
        movd      mm3, [ebx + edx * 4]
        punpcklwd mm0, mm1
        punpcklwd mm2, mm3
        punpckldq mm0, mm2
        movq      [edi], mm0
        lea       esi, [esi + 8]
        lea       edi, [edi + 8]
        dec       ecx
        jnz       L2

        mov       edx, nSrcLen
        add       pSrc, edx
        mov       edx, nDstLen
        add       pDst, edx
        dec       nHeight
        jnz       L1

        emms
    }
}

static void MMXRender3BytesPerPixel(RENDERARGS)
{
    __asm {
        shr       nWidth, 3
        mov       ebx, pPalLookup
        xor       eax, eax
L1:
        xor       edx, edx
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        mov       ax, [esi]
        mov       dx, [esi + 2]
        movd      mm0, [ebx + eax * 4]
        movd      mm1, [ebx + edx * 4]
        mov       ax, [esi + 4]
        mov       dx, [esi + 6]
        movd      mm2, [ebx + eax * 4]
        movd      mm3, [ebx + edx * 4]
        mov       ax, [esi + 8]
        mov       dx, [esi + 10]
        movd      mm4, [ebx + eax * 4]
        movd      mm5, [ebx + edx * 4]
        mov       ax, [esi + 12]
        mov       dx, [esi + 14]
        movd      mm6, [ebx + eax * 4]
        movd      mm7, [ebx + edx * 4]

        psll      mm1, 24
        por       mm0, mm1
        movq      mm1, mm2
        psll      mm1, 48
        por       mm0, mm1

        movq      mm1, mm5
        psll      mm1, 56
        psrl      mm2, 16
        por       mm1, mm2
        psll      mm3, 8
        por       mm1, mm3
        psll      mm4, 32
        por       mm1, mm4

        psrl      mm5, 8
        psll      mm6, 16
        por       mm5, mm6
        psll      mm7, 40
        por       mm5, mm7

        movq      [edi], mm0
        movq      [edi + 8], mm1
        movq      [edi + 16], mm5

        lea       esi, [esi + 16]
        lea       edi, [edi + 24]
        dec       ecx
        jnz       L2

        mov       edx, nSrcLen
        add       pSrc, edx
        mov       edx, nDstLen
        add       pDst, edx
        dec       nHeight
        jnz       L1

        emms
    }
}

static void MMXRender4BytesPerPixel(RENDERARGS)
{
    __asm {
        shr       nWidth, 1
        mov       ebx, pPalLookup
        xor       eax, eax
L1:
        xor       edx, edx
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        mov       ax, [esi]
        mov       dx, [esi + 2]
        movd      mm0, [ebx + eax * 4]
        movd      mm1, [ebx + edx * 4]
        punpckldq mm0, mm1
        movq      [edi], mm0
        lea       esi, [esi + 4]
        lea       edi, [edi + 8]
        dec       ecx
        jnz       L2

        mov       edx, nSrcLen
        add       pSrc, edx
        mov       edx, nDstLen
        add       pDst, edx
        dec       nHeight
        jnz       L1

        emms
    }
}

static void MMXRender2BytesPerPixel_BltHScanLines(RENDERARGS)
{
    __asm {
        shl       nDstLen, 1
        shr       nWidth, 2
        mov       ebx, pPalLookup
        xor       eax, eax
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
        xor       edx, edx
L2:
        mov       ax, [esi]
        mov       dx, [esi + 2]
        movd      mm0, [ebx + eax * 4]
        movd      mm1, [ebx + edx * 4]
        mov       ax, [esi + 4]
        mov       dx, [esi + 6]
        movd      mm2, [ebx + eax * 4]
        movd      mm3, [ebx + edx * 4]
        punpcklwd mm0, mm1
        punpcklwd mm2, mm3
        punpckldq mm0, mm2
        movq      [edi], mm0
        lea       esi, [esi + 8]
        lea       edi, [edi + 8]
        dec       ecx
        jnz       L2

        mov       edx, nSrcLen
        add       pSrc, edx
        mov       edx, nDstLen
        add       pDst, edx
        dec       nHeight
        jnz       L1

        emms
    }
}

static void MMXRender3BytesPerPixel_BltHScanLines(RENDERARGS)
{
    __asm {
        shl       nDstLen, 1
        shr       nWidth, 3
        mov       ebx, pPalLookup
        xor       eax, eax
L1:
        xor       edx, edx
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        mov       ax, [esi]
        mov       dx, [esi + 2]
        movd      mm0, [ebx + eax * 4]
        movd      mm1, [ebx + edx * 4]
        mov       ax, [esi + 4]
        mov       dx, [esi + 6]
        movd      mm2, [ebx + eax * 4]
        movd      mm3, [ebx + edx * 4]
        mov       ax, [esi + 8]
        mov       dx, [esi + 10]
        movd      mm4, [ebx + eax * 4]
        movd      mm5, [ebx + edx * 4]
        mov       ax, [esi + 12]
        mov       dx, [esi + 14]
        movd      mm6, [ebx + eax * 4]
        movd      mm7, [ebx + edx * 4]

        psll      mm1, 24
        por       mm0, mm1
        movq      mm1, mm2
        psll      mm1, 48
        por       mm0, mm1

        movq      mm1, mm5
        psll      mm1, 56
        psrl      mm2, 16
        por       mm1, mm2
        psll      mm3, 8
        por       mm1, mm3
        psll      mm4, 32
        por       mm1, mm4

        psrl      mm5, 8
        psll      mm6, 16
        por       mm5, mm6
        psll      mm7, 40
        por       mm5, mm7

        movq      [edi], mm0
        movq      [edi + 8], mm1
        movq      [edi + 16], mm5

        lea       esi, [esi + 16]
        lea       edi, [edi + 24]
        dec       ecx
        jnz       L2

        mov       edx, nSrcLen
        add       pSrc, edx
        mov       edx, nDstLen
        add       pDst, edx
        dec       nHeight
        jnz       L1

        emms
    }
}

static void MMXRender4BytesPerPixel_BltHScanLines(RENDERARGS)
{
    __asm {
        shl       nDstLen, 1
        shr       nWidth, 1
        mov       ebx, pPalLookup
        xor       eax, eax
L1:
        xor       edx, edx
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        mov       ax, [esi]
        mov       dx, [esi + 2]
        movd      mm0, [ebx + eax * 4]
        movd      mm1, [ebx + edx * 4]
        punpckldq mm0, mm1
        movq      [edi], mm0
        lea       esi, [esi + 4]
        lea       edi, [edi + 8]
        dec       ecx
        jnz       L2

        mov       edx, nSrcLen
        add       pSrc, edx
        mov       edx, nDstLen
        add       pDst, edx
        dec       nHeight
        jnz       L1

        emms
    }
}

/*----- doublen ------*/

static void MMXRender2BytesPerPixel_HScanLines(RENDERARGS)
{
    __asm {
        shr       nWidth, 1
        shl       nDstLen, 1
        mov       ebx, pPalLookup
        xor       eax, eax
L1:
        xor       edx, edx
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        mov       ax,  [esi]
        mov       dx,  [esi + 2]
        movd      mm0, [ebx + eax * 4]
        movd      mm1, [ebx + edx * 4]
        punpckldq mm0, mm1
        movq      [edi], mm0
        lea       esi, [esi + 4]
        lea       edi, [edi + 8]
        dec       ecx
        jnz       L2

        mov       edx, nSrcLen
        add       pSrc, edx
        mov       edx, nDstLen
        add       pDst, edx
        dec       nHeight
        jnz       L1

        emms
    }
}

static void MMXRender3BytesPerPixel_HScanLines(RENDERARGS)
{
    __asm {
        shr       nWidth, 2
        shl       nDstLen, 1
        mov       ebx, pPalLookup
        xor       eax, eax
L1:
        xor       edx, edx
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        mov       ax, [esi]
        mov       dx, [esi + 2]
        movd      mm0, [ebx + eax * 4]
        movd      mm1, [ebx + edx * 4]
        mov       ax, [esi + 4]
        mov       dx, [esi + 6]
        movd      mm2, [ebx + eax * 4]
        movd      mm3, [ebx + edx * 4]

        movq      mm4, mm1
        movq      mm5, mm1
        movq      mm6, mm2

        psll      mm4, 48
        por       mm4, mm0
        psll      mm0, 24
        por       mm0, mm4

        psrl      mm1, 16
        psll      mm5, 8
        por       mm1, mm5
        psll      mm2, 32
        por       mm1, mm2
        psll      mm2, 24
        por       mm1, mm2

        psrl      mm6, 8
        psll      mm3, 16
        por       mm6, mm3
        psll      mm3, 24
        por       mm6, mm3

        movq      [edi], mm0
        movq      [edi + 8], mm1
        movq      [edi + 16], mm6

        lea       esi, [esi + 8]
        lea       edi, [edi + 24]
        dec       ecx
        jnz       L2

        mov       edx, nSrcLen
        add       pSrc, edx
        mov       edx, nDstLen
        add       pDst, edx
        dec       nHeight
        jnz       L1

        emms
    }
}

static void MMXRender4BytesPerPixel_HScanLines(RENDERARGS)
{
    __asm {
        shl       nDstLen, 1
        mov       ebx, pPalLookup
        xor       eax, eax
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        mov       ax,  [esi]
        movd      mm0, [ebx + eax * 4]
        punpckldq mm0, mm0
        movq      [edi], mm0
        lea       esi, [esi + 2]
        lea       edi, [edi + 8]
        dec       ecx
        jnz       L2

        mov       edx, nSrcLen
        add       pSrc, edx
        mov       edx, nDstLen
        add       pDst, edx
        dec       nHeight
        jnz       L1

        emms
    }
}

static void MMXRender2BytesPerPixel_Double(RENDERARGS)
{
    __asm {
        shr       nWidth, 1
        mov       ebx, pPalLookup
        xor       eax, eax
L1:
        xor       edx, edx
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        mov       ax, [esi]
        mov       dx, [esi + 2]
        movd      mm0, [ebx + eax * 4]
        movd      mm1, [ebx + edx * 4]
        punpckldq mm0, mm1
        movq      [edi], mm0
        lea       esi, [esi + 4]
        lea       edi, [edi + 8]
        dec       ecx
        jnz       L2

        mov       edx, nDstLen
        add       pDst, edx

        xor       edx, edx
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L3:
        mov       ax, [esi]
        mov       dx, [esi + 2]
        movd      mm0, [ebx + eax * 4]
        movd      mm1, [ebx + edx * 4]
        punpckldq mm0, mm1
        movq      [edi], mm0
        lea       esi, [esi + 4]
        lea       edi, [edi + 8]
        dec       ecx
        jnz       L3

        mov       edx, nSrcLen
        add       pSrc, edx
        mov       edx, nDstLen
        add       pDst, edx
        dec       nHeight
        jnz       L1

        emms
    }
}

static void MMXRender3BytesPerPixel_Double(RENDERARGS)
{
    __asm {
        shr       nWidth, 2
        mov       ebx, pPalLookup
        xor       eax, eax
L1:
        xor       edx, edx
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        mov       ax, [esi]
        mov       dx, [esi + 2]
        movd      mm0, [ebx + eax * 4]
        movd      mm1, [ebx + edx * 4]
        mov       ax, [esi + 4]
        mov       dx, [esi + 6]
        movd      mm2, [ebx + eax * 4]
        movd      mm3, [ebx + edx * 4]

        movq      mm4, mm1
        movq      mm5, mm1
        movq      mm6, mm2

        psll      mm4, 48
        por       mm4, mm0
        psll      mm0, 24
        por       mm0, mm4

        psrl      mm1, 16
        psll      mm5, 8
        por       mm1, mm5
        psll      mm2, 32
        por       mm1, mm2
        psll      mm2, 24
        por       mm1, mm2

        psrl      mm6, 8
        psll      mm3, 16
        por       mm6, mm3
        psll      mm3, 24
        por       mm6, mm3

        movq      [edi], mm0
        movq      [edi + 8], mm1
        movq      [edi + 16], mm6

        lea       esi, [esi + 8]
        lea       edi, [edi + 24]
        dec       ecx
        jnz       L2

        mov       edx, nDstLen
        add       pDst, edx

        xor       edx, edx
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L3:
        mov       ax, [esi]
        mov       dx, [esi + 2]
        movd      mm0, [ebx + eax * 4]
        movd      mm1, [ebx + edx * 4]
        mov       ax, [esi + 4]
        mov       dx, [esi + 6]
        movd      mm2, [ebx + eax * 4]
        movd      mm3, [ebx + edx * 4]

        movq      mm4, mm1
        movq      mm5, mm1
        movq      mm6, mm2

        psll      mm4, 48
        por       mm4, mm0
        psll      mm0, 24
        por       mm0, mm4

        psrl      mm1, 16
        psll      mm5, 8
        por       mm1, mm5
        psll      mm2, 32
        por       mm1, mm2
        psll      mm2, 24
        por       mm1, mm2

        psrl      mm6, 8
        psll      mm3, 16
        por       mm6, mm3
        psll      mm3, 24
        por       mm6, mm3

        movq      [edi], mm0
        movq      [edi + 8], mm1
        movq      [edi + 16], mm6

        lea       esi, [esi + 8]
        lea       edi, [edi + 24]
        dec       ecx
        jnz       L3

        mov       edx, nSrcLen
        add       pSrc, edx
        mov       edx, nDstLen
        add       pDst, edx
        dec       nHeight
        jnz       L1

        emms
    }
}

static void MMXRender4BytesPerPixel_Double(RENDERARGS)
{
    __asm {
        mov       ebx, pPalLookup
        xor       eax, eax
L1:
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L2:
        mov       ax,  [esi]
        movd      mm0, [ebx + eax * 4]
        punpckldq mm0, mm0
        movq      [edi], mm0
        lea       esi, [esi + 2]
        lea       edi, [edi + 8]
        dec       ecx
        jnz       L2

        mov       edx, nDstLen
        add       pDst, edx
        mov       esi, pSrc
        mov       edi, pDst
        mov       ecx, nWidth
L3:
        mov       ax,  [esi]
        movd      mm0, [ebx + eax * 4]
        punpckldq mm0, mm0
        movq      [edi], mm0
        lea       esi, [esi + 2]
        lea       edi, [edi + 8]
        dec       ecx
        jnz       L3

        mov       edx, nSrcLen
        add       pSrc, edx
        mov       edx, nDstLen
        add       pDst, edx
        dec       nHeight
        jnz       L1

        emms
    }
}
#endif


/*===========================================================================
    DirectDrawIuWFNgƃpbg̍쐬 / 
 ==========================================================================*/

static BOOL CreateDDrawObjects(int width, int height, float fps)
{
    DDSURFACEDESC2 ddSurfaceDesc;
    HRESULT        hResult = DD_OK;
    UINT           nCurrentDepth = 2;

    if (!bWindowMode)
        DirectDraw_CreateByIndex(nDisplayIndex);

    pDDraw = dd;
    if (pDDraw == NULL)
    {
        ErrorMsg(win32_getstring(WIN32_ddraw_notinitialized));
        return FALSE;
    }

    if (bWindowMode)
    {
        hResult = IDirectDraw7_SetCooperativeLevel(pDDraw,
                                                   MAME32App.m_hWnd,
                                                   DDSCL_NORMAL);
    }
    else
    {
        if (bBestFit == TRUE)
        {
            SelectDisplayMode(width, height);
        }
        else
        {
            BOOL bFound = FALSE;
            /*
                Check to see if display mode exists since user
                can specify any arbitrary size.
            */
            bFound = FindDisplayMode(nScreenWidth, nScreenHeight);
            if (bFound == FALSE)
            {
                ErrorMsg(win32_getstring(WIN32_ddraw_not16bppsupport),
                         nScreenWidth,
                         nScreenHeight);
                return FALSE;
            }
        }

        hResult = IDirectDraw7_SetCooperativeLevel(pDDraw,
                                                   MAME32App.m_hWnd,
                                                   DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN);
    }

    if (FAILED(hResult))
    {
        ErrorMsg("IDirectDraw7.SetCooperativeLevel %s%s",
				 win32_getstring(WIN32_errorcode),
                 DirectXDecodeError(hResult));
        return FALSE;
    }

    if (bWindowMode)
    {
        /* F`FbN */
        nCurrentDepth = CheckScreenDepth();
        if (nCurrentDepth == 1)
            return FALSE;
    }
    else
    {
        int frame_rate = (int)fps;

        // most monitors can't go this low
        if (frame_rate < 50)
            frame_rate *= 2;

        if (bTripleBuffer == FALSE)
            frame_rate = 0; // old behavior--monitor default when not triple buffering

try_set_display_mode:
        hResult = IDirectDraw7_SetDisplayMode(pDDraw,
                                              nScreenWidth,
                                              nScreenHeight,
                                              16,
                                              frame_rate,
                                              0);

        if (FAILED(hResult))
        {
            if (frame_rate == 0)
            {
                ErrorMsg("IDirectDraw7.SetDisplayMode %s%s",
						 win32_getstring(WIN32_errorcode),
                         DirectXDecodeError(hResult));
                return FALSE;
            }

            frame_rate = 0; // tried to get the correct rate, now try anything

            goto try_set_display_mode;
        }
    }

    nBytesPerPixel = nCurrentDepth;

    /* tgobt@̍쐬 */
    if (bTripleBuffer == TRUE)
    {
        /* tbvp̘AꂽT[tFX쐬 */
        memset(&ddSurfaceDesc, 0, sizeof(ddSurfaceDesc));

        ddSurfaceDesc.dwSize            = sizeof(ddSurfaceDesc);
        ddSurfaceDesc.dwFlags           = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
        ddSurfaceDesc.ddsCaps.dwCaps    = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP |
                                          DDSCAPS_COMPLEX | DDSCAPS_VIDEOMEMORY;
        ddSurfaceDesc.dwBackBufferCount = 2;

        /* tgobt@AeՂɂ邽߂pDDSFlipɍ쐬 */
        hResult = IDirectDraw7_CreateSurface(pDDraw,
                                            &ddSurfaceDesc,
                                            &pDDSFlip,
                                            NULL);

        if (FAILED(hResult))
        {
            ErrorMsg("IDirectDraw7.CreateSurface(Flip) %s%s",
					 win32_getstring(WIN32_errorcode),
                     DirectXDecodeError(hResult));
            return FALSE;
        }

        /* T[tFXbNł邩`FbN */
        if (SurfaceLockable(pDDSFlip) == TRUE)
        {
            DDSCAPS2 ddscaps;

            /* ̂ŁAobNobt@A^b` */

            memset(&ddscaps, 0, sizeof(ddscaps));
            ddscaps.dwCaps = DDSCAPS_BACKBUFFER;

            /* obNobt@AeՂɂ邽߂pDDSPrimaryɃA^b` */
            hResult = IDirectDrawSurface7_GetAttachedSurface(pDDSFlip,
                                                            &ddscaps,
                                                            &pDDSPrimary);
            if (FAILED(hResult))
            {
                ErrorMsg("IDirectDrawSurface7.GetAttachedSurface %s%s",
						 win32_getstring(WIN32_errorcode),
                         DirectXDecodeError(hResult));
                return FALSE;
            }
        }
       else
        {
            /* ŝŁAtbv~߂ */
            IDirectDraw7_Release(pDDSFlip);
            pDDSFlip = NULL;
            bTripleBuffer  = FALSE;
            bUseBackBuffer = TRUE;
        }
    }

    /* gvobt@pT[tFX̍쐬ɎsAgvobt@̏ꍇ */
    if (pDDSFlip == NULL)
    {
        /* Create the primary surface with no back buffers. */
        memset(&ddSurfaceDesc, 0, sizeof(ddSurfaceDesc));
        ddSurfaceDesc.dwSize         = sizeof(ddSurfaceDesc);
        ddSurfaceDesc.dwFlags        = DDSD_CAPS;
        ddSurfaceDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;

        hResult = IDirectDraw7_CreateSurface(pDDraw,
                                            &ddSurfaceDesc,
                                            &pDDSPrimary,
                                            NULL);

        if (FAILED(hResult))
        {
            ErrorMsg("IDirectDraw7.CreateSurface(Primary Surface) %s%s",
					 win32_getstring(WIN32_errorcode),
                     DirectXDecodeError(hResult));
            return FALSE;
        }
    }

    if (bUseBackBuffer)
    {
        /* obNobt@iȗ̂ߏɎgpj̍쐬 */
        memset(&ddSurfaceDesc, 0, sizeof(ddSurfaceDesc));
        ddSurfaceDesc.dwSize         = sizeof(ddSurfaceDesc);
        ddSurfaceDesc.dwFlags        = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
        ddSurfaceDesc.dwWidth        = ClientRect.m_Width * nScale;
        ddSurfaceDesc.dwHeight       = ClientRect.m_Height * nScale;
        if (!bSystemMemory)
            ddSurfaceDesc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_VIDEOMEMORY;
        else
            ddSurfaceDesc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;

        hResult = IDirectDraw7_CreateSurface(pDDraw, &ddSurfaceDesc, &pDDSBack, NULL);
        if (FAILED(hResult))
        {
            if (bSystemMemory == FALSE)
            {
                /* rfIɊmۂłȂ̂ŁAVXeōĒ */
                ddSurfaceDesc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;

                hResult = IDirectDraw7_CreateSurface(pDDraw, &ddSurfaceDesc, &pDDSBack, NULL);
            }

            if (FAILED(hResult))
            {
                ErrorMsg("IDirectDraw7.CreateSurface(Back Buffer) %s%s",
						 win32_getstring(WIN32_errorcode),
                         DirectXDecodeError(hResult));
                return FALSE;
            }
        }
    }
    else
    {
        pDDSBack = pDDSPrimary;
    }

    /* Nbp쐬 */
    if (bWindowMode)
    {
        hResult = IDirectDraw7_CreateClipper(pDDraw, 0, &pDDClipper, NULL);
        if (FAILED(hResult))
        {
            ErrorMsg("IDirectDraw7.CreateClipper %s%s",
					 win32_getstring(WIN32_errorcode),
                     DirectXDecodeError(hResult));
            return FALSE;
        }

        hResult = IDirectDrawClipper_SetHWnd(pDDClipper, 0, MAME32App.m_hWnd);
        if (FAILED(hResult))
        {
            ErrorMsg("IDirectDrawClipper.SetHWnd %s%s",
					 win32_getstring(WIN32_errorcode),
                     DirectXDecodeError(hResult));
            return FALSE;
        }

        hResult = IDirectDrawSurface7_SetClipper(pDDSPrimary, pDDClipper);
        if (FAILED(hResult))
        {
            ErrorMsg("IDirectDrawSurface7.SetClipper %s%s",
					 win32_getstring(WIN32_errorcode),
                     DirectXDecodeError(hResult));
            return FALSE;
        }
    }

    if (pDDSFlip)    ClearSurface(pDDSFlip);
    if (pDDSPrimary) ClearSurface(pDDSPrimary);
    if (pDDSBack)    ClearSurface(pDDSBack);

    return TRUE;
}


static BOOL CreatePalettes(void)
{
    BOOL          bResult = TRUE;
    HRESULT       hResult = DD_OK;
    DDPIXELFORMAT ddPixelFormat;

    nTotalColors = TRANS_COLOR;

    pPalLookup       = (UINT32 *)      MyMalloc((nTotalColors * 2 + 16) * sizeof(UINT32));
    pPalEntries      = (PALETTEENTRY *)MyMalloc((nTotalColors * 2) * sizeof(PALETTEENTRY));
    pAdjustedPalette = (PALETTEENTRY *)MyMalloc((nTotalColors * 2) * sizeof(PALETTEENTRY));
	
    if (pPalLookup == NULL || pPalEntries == NULL || pAdjustedPalette == NULL)
        return FALSE;

//	memset(pPalLookup, 0, sizeof(UINT32));
    memset(&ddPixelFormat, 0, sizeof(DDPIXELFORMAT));
    ddPixelFormat.dwSize  = sizeof(DDPIXELFORMAT);
    ddPixelFormat.dwFlags = DDPF_RGB;

    hResult = IDirectDrawSurface7_GetPixelFormat(pDDSPrimary, &ddPixelFormat);
    if (FAILED(hResult))
    {
        ErrorMsg("IDirectDrawSurface7.getPixelFormat %s%s",
				 win32_getstring(WIN32_errorcode),
                 DirectXDecodeError(hResult));
        return FALSE;
    }

    GetPixelInfo(ddPixelFormat.dwRBitMask, &Red);
    GetPixelInfo(ddPixelFormat.dwGBitMask, &Green);
    GetPixelInfo(ddPixelFormat.dwBBitMask, &Blue);

    return TRUE;
}


static void ReleaseDDrawObjects(void)
{
    if (pDDraw != NULL)
    {
        if (pDDClipper != NULL)
        {
            IDirectDrawClipper_Release(pDDClipper);
            pDDClipper = NULL;
        }

        if (pDDSBack != NULL)
        {
            if (bUseBackBuffer)
                IDirectDrawSurface7_Release(pDDSBack);
            pDDSBack = NULL;
        }

        if (pDDSFlip != NULL)
        {
            IDirectDrawSurface7_Release(pDDSFlip);
            pDDSFlip    = NULL;
            pDDSPrimary = NULL;
        }

        if (pDDSPrimary != NULL)
        {
            IDirectDrawSurface7_Release(pDDSPrimary);
            pDDSPrimary = NULL;
        }

        pDDraw = NULL;
    }
}


static void ReleasePalettes(void)
{
    if (pPalEntries != NULL)
    {
        MyFree(pPalEntries);
        pPalEntries = NULL;
    }

    if (pAdjustedPalette != NULL)
    {
        MyFree(pAdjustedPalette);
        pAdjustedPalette = NULL;
    }

    if (pPalLookup != NULL)
    {
        MyFree(pPalLookup);
        pPalLookup = NULL;
    }
}


/*===========================================================================
    pbg̑
 ==========================================================================*/

static void MapColor(unsigned char *pRed, unsigned char *pGreen, unsigned char *pBlue)
{
    *pRed   = pColorMap[*pRed];
    *pGreen = pColorMap[*pGreen];
    *pBlue  = pColorMap[*pBlue];
}


static void SetPen(int pen, unsigned char red, unsigned char green, unsigned char blue)
{
    assert(0 <= pen && pen < nTotalColors);
    assert(pPalEntries      != NULL);
    assert(pAdjustedPalette != NULL);

    pPalEntries[pen].peRed    = red;
    pPalEntries[pen].peGreen  = green;
    pPalEntries[pen].peBlue   = blue;
    pPalEntries[pen].peFlags  = PC_NOCOLLAPSE;

    memcpy(&pAdjustedPalette[pen], &pPalEntries[pen], sizeof(PALETTEENTRY));
}


void SetStaticPen(int pen, unsigned char red, unsigned char green, unsigned char blue)
{
//  assert(pen >= nTotalColors && pen < nTotalColors + 16);

    pPalLookup[pen] = makecolor(red, green, blue);

    if (nBytesPerPixel == 2)
        pPalLookup[pen] *= 0x10001;
}


static void SetPaletteColors(void)
{
    UINT i;
	AdjustPalette();

    assert(pPalEntries      != NULL);
    assert(pAdjustedPalette != NULL);


    switch (nBytesPerPixel)
    {
        case 2:
            for (i = 0; i < nTotalColors; i++)
            {
                pPalLookup[i] = makecolor(pAdjustedPalette[i].peRed,
                                        pAdjustedPalette[i].peGreen,
                                        pAdjustedPalette[i].peBlue) * 0x10001;
                pPalLookup[i + TRANS_COLOR] = makecolor2(pAdjustedPalette[i].peRed,
                                        pAdjustedPalette[i].peGreen,
                                        pAdjustedPalette[i].peBlue) * 0x10001;
            }
            break;

        default:
            for (i = 0; i < nTotalColors; i++)
            {
                pPalLookup[i] = makecolor(pAdjustedPalette[i].peRed,
                                        pAdjustedPalette[i].peGreen,
                                        pAdjustedPalette[i].peBlue);
                pPalLookup[i + TRANS_COLOR] = makecolor2(pAdjustedPalette[i].peRed,
                                        pAdjustedPalette[i].peGreen,
                                        pAdjustedPalette[i].peBlue);
            }
            break;
    }
}


static void AdjustPalette(void)
{
    UINT i;

    memcpy(pAdjustedPalette, pPalEntries, nTotalColors * sizeof(PALETTEENTRY));

    for (i = 0; i < nTotalColors; i++)
    {
        MapColor(&pAdjustedPalette[i].peRed,
                 &pAdjustedPalette[i].peGreen,
                 &pAdjustedPalette[i].peBlue);
    }

    bUpdatePalette = TRUE;
}


static void AdjustColorMap(void)
{
    int i;

    for (i = 0; i < INTENSITY_LEVELS; i++)
    {
        pColorMap[i] = (unsigned char)((INTENSITY_LEVELS - 1.0)
                       * (nBrightness / 100.0)
                       * pow(((double)i / (INTENSITY_LEVELS - 1.0)), 1.0 / dGamma));
    }
}


/*===========================================================================
    fBXvC̐ݒ
 ==========================================================================*/

static void AdjustVisibleRect(int min_x, int min_y, int max_x, int max_y)
{
    /* MAMẼrbg}bv(hCo)̗̈ݒ */
    GameRect.m_Left   = min_x;
    GameRect.m_Top    = min_y;
    GameRect.m_Width  = (max_x - min_x) + 1;
    GameRect.m_Height = (max_y - min_y) + 1;

    if (bWindowMode)
    {
        /* XN[TCY̐ݒ */
        nScreenWidth  = GameRect.m_Width * nScale;
        nScreenHeight = GameRect.m_Height * nScale;

        /* XN[̈̐ݒ */
        ClientRect.m_Left   = 0;
        ClientRect.m_Top    = 0;
        ClientRect.m_Width  = nScreenWidth;
        ClientRect.m_Height = nScreenHeight;
    }
    else
    {
        /* tXN[̓EBhËʒuŒȂ̂ŁAɐݒ肵Ă */
        if (bStretch)
        {
            /* Xgb`̓n[hEFAɔĈŁAʃTCŶ܂܎gp */
            ClientRect.m_Left   = 0;
            ClientRect.m_Top    = 0;
            ClientRect.m_Width  = nScreenWidth;
            ClientRect.m_Height = nScreenHeight;
        }
        else
        {
            int gfx_xoffset, gfx_yoffset;

            gfx_xoffset = (nScreenWidth - GameRect.m_Width * nScale) / 2;
            if (bDouble == TRUE)
            {
                if (GameRect.m_Width > nScreenWidth / 2)
                    GameRect.m_Width = nScreenWidth / 2;
            }
            else
            {
                if (GameRect.m_Width > nScreenWidth)
                    GameRect.m_Width = nScreenWidth;
            }

            gfx_yoffset = (nScreenHeight - GameRect.m_Height * nScale) / 2;
            if (bDouble == TRUE)
            {
                if (GameRect.m_Height > nScreenHeight / 2)
                    GameRect.m_Height = nScreenHeight / 2;
            }
            else
            {
                if (GameRect.m_Height > nScreenHeight)
                    GameRect.m_Height = nScreenHeight;
            }

            ClientRect.m_Left   = gfx_xoffset;
            ClientRect.m_Top    = gfx_yoffset;
            ClientRect.m_Width  = GameRect.m_Width * nScale;
            ClientRect.m_Height = GameRect.m_Height * nScale;
        }

        /* tgobt@̗̈ݒ */
        rectPrimary.left   = ClientRect.m_Left;
        rectPrimary.top    = ClientRect.m_Top;
        rectPrimary.right  = ClientRect.m_Left + ClientRect.m_Width;
        rectPrimary.bottom = ClientRect.m_Top  + ClientRect.m_Height;
    }

    /* ŌɃobNobt@̗̈ݒ */
    rectBack.left   = 0;
    rectBack.top    = 0;
    rectBack.right  = rectBack.left + GameRect.m_Width;
    rectBack.bottom = rectBack.top  + GameRect.m_Height;

#if 0
            ErrorMsg("Screen : width = %d, height = %d\n"
                     "Client : (%4d, %4d)-(%4d,%4d)\n"
                     "Driver : (%4d, %4d)-(%4d,%4d)\n"
                     "Primary: (%4d, %4d)-(%4d,%4d)\n"
                     "Back   : (%4d, %4d)-(%4d,%4d)\n",
                          nScreenWidth,
                          nScreenHeight,
                          ClientRect.m_Left,
                          ClientRect.m_Top,
                          ClientRect.m_Width,
                          ClientRect.m_Height,
                          GameRect.m_Left,
                          GameRect.m_Top,
                          GameRect.m_Width,
                          GameRect.m_Height,
                          rectPrimary.left,
                          rectPrimary.top,
                          rectPrimary.right,
                          rectPrimary.bottom,
                          rectBack.left,
                          rectBack.top,
                          rectBack.right,
                          rectBack.bottom);
#endif

    /* blitgpȂ2{Ɋg傷鎞́A]̈2{ */
    if (bDouble)
    {
        if (bHScanLines && bBltDouble)
        {
            rectBack.bottom *= nScale;
        }
        else if (!bBltDouble)
        {
            rectBack.right  *= nScale;
            rectBack.bottom *= nScale;
        }
    }

    pBitmapOffset = (BYTE *)(Machine->scrbitmap->line[GameRect.m_Top] + GameRect.m_Left);
    nBitmapPitch  = Machine->scrbitmap->modulo * 2;
}


static void GetPixelInfo(DWORD dwMask, struct tPixelInfo* pPixelInfo)
{
    int nShift = 0;
    int nCount = 0;
    int i;

    pPixelInfo->m_dwMask = dwMask;

    while ((dwMask & 1) == 0 && nShift < sizeof(DWORD) * 8)
    {
        nShift++;
        dwMask >>= 1;
    }
    pPixelInfo->m_nShift = nShift;

    for (i = 0; i < sizeof(DWORD) * 8; i++)
    {
        if (dwMask & (1 << i))
            nCount++;
    }
    pPixelInfo->m_nSize = nCount;
}


/*===========================================================================
    EBhE[hp֐
 ==========================================================================*/

static int CheckScreenDepth(void)
{
    DDSURFACEDESC2 ddSurfaceDesc;
    HRESULT        hResult = DD_OK;

    ddSurfaceDesc.dwSize = sizeof(ddSurfaceDesc);
    hResult = IDirectDraw7_GetDisplayMode(pDDraw, &ddSurfaceDesc);
    if (hResult != DD_OK)
    {
        ErrorMsg("IDirectDraw7.GetDisplayMode %s%s",
				 win32_getstring(WIN32_errorcode),
                 DirectXDecodeError(hResult));
        return 1;
    }
    if ((ddSurfaceDesc.ddpfPixelFormat.dwFlags & DDPF_PALETTEINDEXED8) ||
        (ddSurfaceDesc.ddpfPixelFormat.dwRBitMask == 0 &&
         ddSurfaceDesc.ddpfPixelFormat.dwGBitMask == 0 &&
         ddSurfaceDesc.ddpfPixelFormat.dwBBitMask == 0))
    {
        ErrorMsg(win32_getstring(WIN32_ddraw_notwork8bit));
        return 1;
    }

    return ddSurfaceDesc.ddpfPixelFormat.dwRGBBitCount / 8;
}


/*===========================================================================
    tXN[[hp֐
 ==========================================================================*/

static BOOL FindBestDisplayMode(DWORD  dwWidthIn, DWORD  dwHeightIn,
                                DWORD* pdwWidthOut, DWORD* pdwHeightOut)
{
    struct tDisplayModes*   pDisplayModes;
    int     i;
    BOOL    bFound = FALSE;
    DWORD   dwBestWidth  = 10000;
    DWORD   dwBestHeight = 10000;
    DWORD   dwBiggestWidth  = 0;
    DWORD   dwBiggestHeight = 0;

    pDisplayModes = DirectDraw_GetDisplayModes();

    assert(0 < pDisplayModes->m_nNumModes);

    for (i = 0; i < pDisplayModes->m_nNumModes; i++)
    {
        if (pDisplayModes->m_Modes[i].m_dwBPP != 16)
            continue;

        /* Find a mode big enough for input size. */
        if (dwWidthIn  <= pDisplayModes->m_Modes[i].m_dwWidth
        &&  dwHeightIn <= pDisplayModes->m_Modes[i].m_dwHeight)
        {
            /* Get the smallest display size. */
            if (pDisplayModes->m_Modes[i].m_dwWidth  < dwBestWidth
            ||  pDisplayModes->m_Modes[i].m_dwHeight < dwBestHeight)
            {
                dwBestWidth  = pDisplayModes->m_Modes[i].m_dwWidth;
                dwBestHeight = pDisplayModes->m_Modes[i].m_dwHeight;
                bFound = TRUE;
            }
        }

        /*
            Keep track of the biggest size in case
            we can't find a mode that works.
        */
        if (dwBiggestWidth  < pDisplayModes->m_Modes[i].m_dwWidth
        ||  dwBiggestHeight < pDisplayModes->m_Modes[i].m_dwHeight)
        {
            dwBiggestWidth  = pDisplayModes->m_Modes[i].m_dwWidth;
            dwBiggestHeight = pDisplayModes->m_Modes[i].m_dwHeight;
        }
    }

    assert(pdwWidthOut  != NULL);
    assert(pdwHeightOut != NULL);

    if (bFound == TRUE)
    {
        *pdwWidthOut  = dwBestWidth;
        *pdwHeightOut = dwBestHeight;
    }
    else
    {
        *pdwWidthOut  = dwBiggestWidth;
        *pdwHeightOut = dwBiggestHeight;
    }

    return bFound;
}


static BOOL FindDisplayMode(DWORD dwWidth, DWORD dwHeight)
{
    int    i;
    BOOL   bModeFound = FALSE;
    struct tDisplayModes* pDisplayModes;

    pDisplayModes = DirectDraw_GetDisplayModes();

    for (i = 0; i < pDisplayModes->m_nNumModes; i++)
    {
        if (pDisplayModes->m_Modes[i].m_dwWidth  == dwWidth
        &&  pDisplayModes->m_Modes[i].m_dwHeight == dwHeight
        &&  pDisplayModes->m_Modes[i].m_dwBPP    == 16)
        {
            bModeFound = TRUE;
            break;
        }
    }

    return bModeFound;
}


static void SelectDisplayMode(int width, int height)
{
    if (bBestFit == TRUE)
    {
        BOOL bResult;

        if (bDouble == TRUE && bStretch == FALSE)
        {
            width  *= nScale;
            height *= nScale;

            bResult = FindBestDisplayMode(width, height,
                                          &nScreenWidth,
                                          &nScreenHeight);

            if (bResult == FALSE)
            {
                /* no pixel doubled modes fit, revert to not doubled */
                bDouble = FALSE;
                width  /= nScale;
                height /= nScale;
            }
        }
        else
        {
            FindBestDisplayMode(width, height,
                                &nScreenWidth,
                                &nScreenHeight);
        }
    }
}


/*===========================================================================
    DirectDrawT[tFX̑
 ==========================================================================*/

static void DrawSurface(void)
{
    HRESULT hResult;
    extern int in_jukebox;

    if (!pDDSPrimary || !pDDSBack)
        return;

    if (bWindowMode)
    {
        POINT p;

        p.x = 0;
        p.y = 0;
        ClientToScreen(MAME32App.m_hWnd, &p);

        if (rectPrimary.left != p.x ||
            rectPrimary.top  != p.y)
        {
            rectPrimary.left   = p.x;
            rectPrimary.top    = p.y;
            rectPrimary.right  = rectPrimary.left + nScreenWidth;
            rectPrimary.bottom = rectPrimary.top  + nScreenHeight;
        }
    }

    if (bVsync && bThrottled && !in_jukebox)
    {
        BOOL bResult = FALSE;

        while (bResult == FALSE)
            IDirectDraw7_GetVerticalBlankStatus(pDDraw, &bResult);
    }

    hResult = IDirectDrawSurface7_Blt(pDDSPrimary,
                                      &rectPrimary,
                                      pDDSBack,
                                      &rectBack,
                                      DDBLT_WAIT,
                                      NULL);

    if (hResult == DDERR_SURFACELOST)
        IDirectDrawSurface7_Restore(pDDSPrimary);
}


static void FlipSurface(void)
{
    HRESULT hResult;

    if (!pDDSPrimary || !pDDSBack)
        return;

    hResult = IDirectDrawSurface7_Flip(pDDSFlip, NULL, DDFLIP_WAIT);

    if (hResult == DDERR_SURFACELOST)
        IDirectDrawSurface7_Restore(pDDSFlip);
}


static void DoubleSurface(void)
{
    HRESULT hResult;
    RECT rectSrc, rectDst;

    if (!pDDSBack)
        return;

    rectSrc = rectDst = rectBack;
    rectDst.top    = rectBack.top + 1;
    rectSrc.bottom = rectBack.bottom - 1;

    hResult = IDirectDrawSurface7_Blt(pDDSBack,
                                      &rectDst,
                                      pDDSBack,
                                      &rectSrc,
                                      DDBLT_WAIT | DDBLT_KEYSRC,
                                      NULL);

    if (hResult == DDERR_SURFACELOST)
        IDirectDrawSurface7_Restore(pDDSBack);
}


static void ClearSurface(IDirectDrawSurface7 *pddSurface)
{
    HRESULT hResult;
    DDBLTFX ddBltFX;

    if (!pddSurface)
        return;

    memset(&ddBltFX, 0, sizeof(DDBLTFX));
    ddBltFX.dwSize = sizeof(DDBLTFX);

    hResult = IDirectDrawSurface7_Blt(pddSurface,
                                      NULL,
                                      NULL,
                                      NULL,
                                      DDBLT_WAIT | DDBLT_COLORFILL,
                                      &ddBltFX);

    if (hResult == DDERR_SURFACELOST)
        IDirectDrawSurface7_Restore(pddSurface);

    return;
}


static void FillSurface(void)
{
    HRESULT hResult;
    DDBLTFX ddBltFX;

    if (!pDDSBack)
        return;

    memset(&ddBltFX, 0, sizeof(DDBLTFX));
    ddBltFX.dwSize = sizeof(DDBLTFX);
    ddBltFX.dwFillColor = 0x1234;

    hResult = IDirectDrawSurface7_Blt(pDDSBack,
                                      NULL,
                                      NULL,
                                      NULL,
                                      DDBLT_WAIT | DDBLT_COLORFILL,
                                      &ddBltFX);

    if (hResult == DDERR_SURFACELOST)
        IDirectDrawSurface7_Restore(pDDSBack);

    return;
}


static BOOL LockSurface(IDirectDrawSurface7 *pddSurface)
{
    DDSURFACEDESC2 ddSurfaceDesc;
    HRESULT        hResult;

    if (!pddSurface)
        return FALSE;

    ddSurfaceDesc.dwSize = sizeof(ddSurfaceDesc);

    hResult = IDirectDrawSurface7_Lock(pddSurface, NULL, &ddSurfaceDesc, DDLOCK_WAIT, NULL);

    if (hResult == DDERR_SURFACELOST)
    {
        IDirectDrawSurface7_Restore(pddSurface);
        return FALSE;
    }

    pSurfaceBits  = ddSurfaceDesc.lpSurface;
    nSurfacePitch = ddSurfaceDesc.lPitch;

    if (!bUseBackBuffer)
        pSurfaceBits  += nSurfacePitch * rectPrimary.top + 2 * rectPrimary.left;

    return TRUE;
}


static BOOL UnlockSurface(IDirectDrawSurface7 *pddSurface)
{
    HRESULT hResult;

    if (!pddSurface)
        return FALSE;

    hResult = IDirectDrawSurface7_Unlock(pddSurface, NULL);
    if (FAILED(hResult))
        return FALSE;

    return TRUE;
}


static BOOL RestoreSurface(void)
{
    HRESULT hResult;

    if (!pDDSPrimary || !pDDSBack)
        return TRUE;

    hResult = IDirectDrawSurface7_IsLost(pDDSPrimary);
    if (hResult == DDERR_SURFACELOST)
    {
        hResult = IDirectDrawSurface7_Restore(pDDSPrimary);
        if (hResult != DD_OK)
            return FALSE;
    }

    hResult = IDirectDrawSurface7_IsLost(pDDSBack);
    if (hResult == DDERR_SURFACELOST)
    {
        hResult = IDirectDrawSurface7_Restore(pDDSBack);
        if (hResult != DD_OK)
            return FALSE;
    }

    return TRUE;
}


static BOOL SurfaceLockable(IDirectDrawSurface7 *pddSurface)
{
    DDSURFACEDESC2 ddSurfaceDesc;
    HRESULT        hResult;

    if (!pddSurface)
        return FALSE;

    memset(&ddSurfaceDesc, 0, sizeof(ddSurfaceDesc));
    ddSurfaceDesc.dwSize = sizeof(ddSurfaceDesc);

    while (1)
    {
        hResult = IDirectDrawSurface7_Lock(pddSurface, NULL, &ddSurfaceDesc, DDLOCK_WAIT, NULL);

        if (hResult == DD_OK)
        {
            hResult = IDirectDrawSurface7_Unlock(pddSurface, NULL);
            return TRUE;
        }

        if (hResult == DDERR_CANTLOCKSURFACE)
        {
            return FALSE;
        }

        if (hResult == DDERR_SURFACELOST)
        {
            hResult = IDirectDrawSurface7_Restore(pddSurface);
            if (FAILED(hResult))
                break;
        }
        else if (hResult != DDERR_WASSTILLDRAWING)
            return FALSE;
    }

    ErrorMsg("IDirectDrawSurface7.Lock %s%s",
			 win32_getstring(WIN32_errorcode),
             DirectXDecodeError(hResult));

    return FALSE;
}


/*===========================================================================
    WindowsbZ[W
 ==========================================================================*/

static void OnActivateApp(HWND hWnd, BOOL fActivate, DWORD dwThreadId)
{
    if (bInitialized == FALSE)
        return;

    if (fActivate && MAME32App.m_hWnd == hWnd)
    {
        UINT nCurrentDepth;

        nCurrentDepth = CheckScreenDepth();

        if (nCurrentDepth != nBytesPerPixel)
            DDraw_change_display();
        else
            bPaused = FALSE;
    }
    else if (!fActivate)
    {
        bPaused = TRUE;
    }
}


static void OnGetMinMaxInfo(HWND hWnd, MINMAXINFO* pMinMaxInfo)
{
    pMinMaxInfo->ptMaxSize.x      = WindowRect.m_Width;
    pMinMaxInfo->ptMaxSize.y      = WindowRect.m_Height;
    pMinMaxInfo->ptMaxTrackSize.x = WindowRect.m_Width;
    pMinMaxInfo->ptMaxTrackSize.y = WindowRect.m_Height;
    pMinMaxInfo->ptMinTrackSize.x = WindowRect.m_Width;
    pMinMaxInfo->ptMinTrackSize.y = WindowRect.m_Height;
}


static void OnPaint(HWND hWnd)
{
    PAINTSTRUCT ps;
    HDC         hdc;

    hdc = BeginPaint(hWnd, &ps);

    if (bInitialized == TRUE && bMinimized == FALSE)
        DrawSurface();

    EndPaint(hWnd, &ps);
}


static void OnSize(HWND hwnd, UINT state, int cx, int cy)
{
    if (bWindowMode)
        StatusWindowSize(state, cx, cy);

    if (bInitialized == FALSE)
        return;

    if (bMinimized == TRUE && state == SIZE_RESTORED)
    {
        HRESULT hResult1 = DD_OK;
        HRESULT hResult2 = DD_OK;
        HRESULT hResult3 = DD_OK;

        hResult1 = IDirectDrawSurface7_IsLost(pDDSPrimary);
        if (hResult1 == DDERR_SURFACELOST)
            hResult1 = IDirectDrawSurface7_Restore(pDDSPrimary);

        hResult2 = IDirectDrawSurface7_IsLost(pDDSBack);
        if (hResult2 == DDERR_SURFACELOST)
            hResult2 = IDirectDrawSurface7_Restore(pDDSBack);

        if (pDDSFlip)
        {
            hResult3 = IDirectDrawSurface7_IsLost(pDDSFlip);
            if (hResult3 == DDERR_SURFACELOST)
                hResult3 = IDirectDrawSurface7_Restore(pDDSFlip);
        }

        if (hResult1 == DD_OK && hResult2 == DD_OK && hResult3 == DD_OK)
        {
            bMinimized = FALSE;
        }
        else
        {
            ErrorMsg(win32_getstring(WIN32_ddraw_restoreerrorcode),
                     DirectXDecodeError(hResult1),
                     DirectXDecodeError(hResult2),
                     DirectXDecodeError(hResult3));
            ForceWindowMode();
        }
    }
    else if (bMinimized == FALSE && state == SIZE_MINIMIZED)
    {
        bMinimized = TRUE;
    }
}


static void OnDrawItem(HWND hwnd, const DRAWITEMSTRUCT *lpDrawItem)
{
    StatusDrawItem(lpDrawItem);
}


static void OnCommand(HWND hWnd, int CmdID, HWND hWndCtl, UINT codeNotify)
{
    switch (CmdID)
    {
        case ID_FILE_EXIT:
            /* same as double-clicking on main window close box */
            SendMessage(hWnd, WM_CLOSE, 0, 0);
            break;

        case ID_MENU_FULLSCREEN:
            tOptions.is_window = !tOptions.is_window;
            CheckMenuItem(hMenu, ID_MENU_FULLSCREEN, tOptions.is_window ? MF_UNCHECKED : MF_CHECKED);
            DDraw_change_display();
            break;

        case ID_MENU_VSYNC:
            if (tOptions.is_window)
            {
                bVsync = tOptions.vsync_window = !tOptions.vsync_window;
                CheckMenuItem(hMenu, ID_MENU_VSYNC, tOptions.vsync_window ? MF_CHECKED : MF_UNCHECKED);
                if (bVsync && tOptions.triple_buffer)
                {
                    tOptions.triple_buffer = FALSE;
                    bTripleBuffer = FALSE;
                    CheckMenuItem(hMenu, ID_MENU_TRIPLEBUFFER, MF_UNCHECKED);
                    DDraw_change_display();
                }
            }
            else
            {
                bVsync = tOptions.vsync_fullscreen = !tOptions.vsync_fullscreen;
                CheckMenuItem(hMenu, ID_MENU_VSYNC, tOptions.vsync_fullscreen ? MF_CHECKED : MF_UNCHECKED);
            }
            break;

        case ID_CHANGE_MODE1:
        case ID_CHANGE_MODE2:
        case ID_CHANGE_MODE3:
        case ID_CHANGE_MODE4:
        case ID_CHANGE_MODE5:
        case ID_CHANGE_MODE6:
        case ID_CHANGE_MODE7:
        case ID_CHANGE_MODE8:
        case ID_CHANGE_MODE9:
            {
                CheckMenuItem(hMenu, ID_CHANGE_MODE1 + tOptions.render_window, MF_UNCHECKED);
                tOptions.render_window = (CmdID - ID_CHANGE_MODE1);
                CheckMenuItem(hMenu, CmdID, MF_CHECKED);
                DDraw_change_display();
            }
            break;

        case ID_ABOUT:
            ShowWindow(hWndAbout, SW_SHOW);
            break;
    }
}


static BOOL OnSetCursor(HWND hWnd, HWND hWndCursor, UINT codeHitTest, UINT msg)
{
    /* Remove the cursor only after display is setup. */
    if (Machine->scrbitmap != NULL)
        SetCursor(NULL);
    return TRUE;
}


/*===========================================================================
    R[obN֐
 ==========================================================================*/

static INT_PTR CALLBACK AboutDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    switch (Msg)
    {
        case WM_INITDIALOG:
            {
				char stuff[4096];

				connect_strings(stuff, WIN32_about01, WIN32_about54);

                Edit_SetText(GetDlgItem(hDlg, IDC_EDIT_ABOUT), stuff);
            }
            return 1;

        case WM_COMMAND:
	        if (LOWORD(wParam) == IDC_NMX_WEB_PAGE)
				ShellExecute(MAME32App.m_hwndUI/*hMain*/,NULL,"http://go.to/nmamex/",NULL,NULL,SW_SHOWNORMAL);

			if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL || LOWORD(wParam) == IDC_NMX_WEB_PAGE)
            {
                ShowWindow(hDlg, SW_HIDE);
                return 1;
            }
    }
    return 0;
}


static void ForceWindowMode(void)
{
    tOptions.is_window = TRUE;
    DDraw_change_display();
}

