/***************************************************************************

    M.A.M.E.32  -  Multiple Arcade Machine Emulator for Win32
    Win32 Portions Copyright (C) 1997-98 Michael Soderstrom and Chris Kirmse
    
    This file is part of MAME32, and may only be used, modified and
    distributed under the terms of the MAME license, in "readme.txt".
    By continuing to use, modify or distribute this file you indicate
    that you have read the license and understand and accept it fully.

 ***************************************************************************/

#ifndef __DISPLAY_H__
#define __DISPLAY_H__

#include "osdepend.h"

typedef struct
{
    int m_Top;
    int m_Left;
    int m_Width;
    int m_Height;
} tRect;

struct OSDDisplay
{
    int                 (*init)(options_type *options);
    void                (*exit)(void);
    struct osd_bitmap*  (*alloc_bitmap)(int width, int height);
    void                (*free_bitmap)(struct osd_bitmap* bitmap);
    int                 (*create_display)(int width, int height, float fps);
    void                (*close_display)(void);
    void                (*set_visible_area)(int min_x, int max_x, int min_y, int max_y);
#ifdef MAME_DEBUG
    void                (*set_debugger_focus)(int debugger_has_focus);
    int                 (*allocate_colors)(const UINT8 *debug_palette, UINT16 *debug_pens);
#else
    int                 (*allocate_colors)(void);
#endif
    void                (*modify_pen)(int pen, unsigned char red, unsigned char green, unsigned char blue);
    void                (*get_pen)(int pen, unsigned char* pRed, unsigned char* pGreen, unsigned char* pBlue);
    int                 (*skip_this_frame)(void);
#ifdef MAME_DEBUG
    void                (*update_display)(struct osd_bitmap *game_bitmap, struct osd_bitmap *debug_bitmap);
#else
    void                (*update_display)(struct osd_bitmap *bitmap);
#endif
    void                (*led_w)(int leds_status);
    void                (*set_gamma)(float gamma);
    float               (*get_gamma)(void);
    void                (*set_brightness)(int brightness);
    int                 (*get_brightness)(void);
    void                (*save_snapshot)(struct osd_bitmap *bitmap);

    BOOL                (*OnMessage)(HWND hWnd, UINT Msg, WPARAM wParam, LPARAM lParam, LRESULT* pResult);
    void                (*Refresh)(void);
};

extern BOOL  Display_Throttled(void);

extern struct OSDDisplay DDrawDisplay;

#endif
