/***************************************************************************

    M.A.M.E.32  -  Multiple Arcade Machine Emulator for Win32
    Win32 Portions Copyright (C) 1997-98 Michael Soderstrom and Chris Kirmse

    This file is part of MAME32, and may only be used, modified and
    distributed under the terms of the MAME license, in "readme.txt".
    By continuing to use, modify or distribute this file you indicate
    that you have read the license and understand and accept it fully.

 ***************************************************************************/

 /***************************************************************************

  win32ui.c

  Win32 GUI code.

  Created 8/12/97 by Christopher Kirmse (ckirmse@ricochet.net)
  Additional code November 1997 by Jeff Miller (miller@aa.net)
  More July 1998 by Mike Haaland (mhaaland@hypertech.com)
  Added Spitters/Property Sheets/Removed Tabs/Added Tree Control in
  Nov/Dec 1998 - Mike Haaland

***************************************************************************/

#include "driver.h"
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shellapi.h>
#include <windowsx.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <commctrl.h>
#include <commdlg.h>
#include <strings.h>
#include <sys/stat.h>
#include <zmouse.h>         /* Intellimouse */
#include <wingdi.h>
#include <tchar.h>
#include <time.h>

#include "resource.h"
#include "resource.hm"
#include "mame32.h"
#include "M32Util.h"
#include "DirectDraw.h"
#include "DirectInput.h"
#include "DDrawDisplay.h"   /* For display modes. */
#include "Joystick.h"       /* For joystick avalibility. */
#include "DIJoystick.h"     /* For DIJoystick avalibility. */
#include "DIKeyboard.h"     /* For DIKeyboard avalibility. */
#include "osdepend.h"
#include "mzip.h"
#include "file.h"
#include "audit32.h"
#include "Directories.h"
#include "Screenshot.h"
#include "Properties.h"
#include "ColumnEdit.h"
#include "TreeView.h"
#include "Splitters.h"
#include "Avi.h"
#include "unzip.h"
#include "datafile.h"
#include "win32text.h"
#ifdef KAILLERA
#include "KailleraChat.h"
#endif

#ifdef KAILLERA
#include "kailleraclient.h"
int kPlay = 0;
#endif /* KAILLERA */

#include "drivers/neogeo.h"
#ifndef BUILD_ENGLISH
#include <Winnls32.h>       /* IME on/off */
#endif

// Uncomment this to use internal background bitmaps
// #define INTERNAL_BKBITMAP


#define MM_PLAY_GAME (WM_APP + 15000)

/* Max size of a sub-menu */
#define DBU_MIN_WIDTH  292
#define DBU_MIN_HEIGHT 190

int MIN_WIDTH  = DBU_MIN_WIDTH;
int MIN_HEIGHT = DBU_MIN_HEIGHT;

typedef BOOL (WINAPI *common_file_dialog_proc)(LPOPENFILENAME lpofn);

/***************************************************************************
    function prototypes
 ***************************************************************************/

static BOOL             Win32UI_init(HINSTANCE hInstance, LPSTR lpCmdLine, int nCmdShow);
static void             Win32UI_exit(void);

static BOOL             PumpMessage(void);
static BOOL             PumpAndReturnMessage(MSG *pmsg);
static void             OnIdle(void);
static void             OnSize(HWND hwnd, UINT state, int width, int height);
static long WINAPI      MameWindowProc(HWND hwnd,UINT message,UINT wParam,LONG lParam);

static void             SetView(int menu_id,int listview_style);
static void             ResetListView(void);
static void             UpdateGameList(void);
static void             ReloadIcons(HWND hWnd);
static void             PollGUIJoystick(void);
static void             PressKey(HWND hwnd,UINT vk);
static BOOL             MameCommand(HWND hwnd,int id, HWND hwndCtl, UINT codeNotify);

static void             UpdateStatusBar(void);
static BOOL             PickerHitTest(HWND hWnd);
static BOOL             MamePickerNotify(NMHDR *nm);
static BOOL             TreeViewNotify(NMHDR *nm);
static char*            ConvertAmpersandString(const char *s);

static void             LoadListBitmap(void);
static void             PaintControl(HWND hWnd, BOOL useBitmap);

static void             InitPicker(void);
static int CALLBACK     ListCompareFunc(LPARAM index1, LPARAM index2, int sort_subitem);

static void             DisableSelection(void);
static void             EnableSelection(int nGame);

static int              GetSelectedPick(void);
static int              GetSelectedPickItem(void);
static void             SetSelectedPick(int new_index);
static void             SetSelectedPickItem(int val);
static void             SetRandomPickItem();

static BOOL             ParseCommandLine(char *command_line);
static INT_PTR CALLBACK AboutDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam);
static INT_PTR CALLBACK DirectXDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam);
static INT_PTR CALLBACK MameHelpDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
static INT_PTR CALLBACK VersionDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
static void             MamePlayRecordGame(void);
static void             MamePlayBackGame(void);
static BOOL             CommonFileDialog(common_file_dialog_proc cfd,char *filename, BOOL bZip);
static void             MamePlayRecordAVI(void);
static BOOL             CommonFileDialogAVI(common_file_dialog_proc cfd,char *filename);
static void             MamePlayNetGame(void);
static void             MamePlayGame(void);
static void             MamePlayGameWithOptions(void);
static INT_PTR CALLBACK LoadProgressDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam);
static void             ColumnSort(int column, BOOL bColumn);
static BOOL             GameCheck(void);

static void             ToggleScreenShot(void);
static void             AdjustMetrics(void);
static void             EnablePlayOptions(int nIndex, options_type *o);

/* for header */
static LRESULT CALLBACK HeaderWndProc(HWND, UINT, WPARAM, LPARAM);
static void             Header_SetSortInfo(HWND hwndList, int nCol, BOOL bAsc);
static void             Header_Initialize(HWND hwndList);
static BOOL             ListCtrlOnErase(HWND hWnd, HDC hDC);
static BOOL             ListCtrlOnPaint(HWND hWnd, UINT uMsg);

/* Icon routines */
static DWORD            GetShellLargeIconSize(void);
static BOOL             CreateIcons(HWND hWnd);
static int              WhichIcon(int nItem);
static void             AddIcon(int index);

/* Context Menu handlers */
static void             UpdateMenu(HWND hwndList, HMENU hMenu);
static BOOL             HandleContextMenu( HWND hWnd, WPARAM wParam, LPARAM lParam);
static BOOL             HeaderOnContextMenu(HWND hWnd, WPARAM wParam, LPARAM lParam);
static BOOL             HandleTreeContextMenu( HWND hWnd, WPARAM wParam, LPARAM lParam);

/* Re/initialize the ListView header columns */
static void             ResetColumnDisplay(BOOL firstime);

/* Custom Draw item */
static void             DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct);
static int              GetNumColumns(HWND hWnd);

static void             CopyToolTipText (LPTOOLTIPTEXT lpttt);

static void             ProgressBarShow(void);
static void             ProgressBarHide(void);
static void             ResizeProgressBar(void);
static void             ProgressBarStep(void);

static HWND             InitProgressBar(HWND hParent);
static HWND             InitToolbar(HWND hParent);
static HWND             InitStatusBar(HWND hParent);

static LRESULT          Statusbar_MenuSelect (HWND hwnd, WPARAM wParam, LPARAM lParam);

static void             UpdateHistory(int gameNum);

static INT_PTR CALLBACK LoadErrorDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam);
static char *           GetRomList(int iGame);

/***************************************************************************
    External variables
 ***************************************************************************/

int win32_debug = 0;

/***************************************************************************
    Internal structures
 ***************************************************************************/

/*
 * These next two structs represent how the icon information
 * is stored in an ICO file.
 */
typedef struct
{
    BYTE    bWidth;               /* Width of the image */
    BYTE    bHeight;              /* Height of the image (times 2) */
    BYTE    bColorCount;          /* Number of colors in image (0 if >=8bpp) */
    BYTE    bReserved;            /* Reserved */
    WORD    wPlanes;              /* Color Planes */
    WORD    wBitCount;            /* Bits per pixel */
    DWORD   dwBytesInRes;         /* how many bytes in this resource? */
    DWORD   dwImageOffset;        /* where in the file is this image */
} ICONDIRENTRY, *LPICONDIRENTRY;

typedef struct
{
    UINT            Width, Height, Colors; /* Width, Height and bpp */
    LPBYTE          lpBits;                /* ptr to DIB bits */
    DWORD           dwNumBytes;            /* how many bytes? */
    LPBITMAPINFO    lpbi;                  /* ptr to header */
    LPBYTE          lpXOR;                 /* ptr to XOR image bits */
    LPBYTE          lpAND;                 /* ptr to AND image bits */
} ICONIMAGE, *LPICONIMAGE;

/* Which edges of a control are anchored to the corresponding side of the parent window */
#define RA_LEFT     0x01
#define RA_RIGHT    0x02
#define RA_TOP      0x04
#define RA_BOTTOM   0x08
#define RA_ALL      0x0F

#define RA_END  0
#define RA_ID   1
#define RA_HWND 2

typedef struct
{
    int         type;       /* Either RA_ID or RA_HWND, to indicate which member of u is used; or RA_END
                               to signify last entry */
    union {                 /* Can identify a child window by control id or by handle */
        int     id;         /* Window control id */
        HWND    hwnd;       /* Window handle */
    } u;
    int         action;     /* What to do when control is resized */
    void        *subwindow; /* Points to a Resize structure for this subwindow; NULL if none */
} ResizeItem;

typedef struct {
    RECT rect;          /* Client rect of window; must be initialized before first resize */
    ResizeItem *items;  /* Array of subitems to be resized */
} Resize;

static void             ResizeWindow(HWND hParent, Resize *r);

/* List view Icon defines */
#define LG_ICONMAP_WIDTH    GetSystemMetrics(SM_CXICON)
#define LG_ICONMAP_HEIGHT   GetSystemMetrics(SM_CYICON)
#define ICONMAP_WIDTH       GetSystemMetrics(SM_CXSMICON)
#define ICONMAP_HEIGHT      GetSystemMetrics(SM_CYSMICON)

typedef struct tagPOPUPSTRING
{
    HMENU hMenu;
    UINT uiString;
} POPUPSTRING;

#define MAX_MENUS 3

#define VERSIONTEXT_LEN 256

/***************************************************************************
    Internal variables
 ***************************************************************************/

static HWND hMain     = NULL;
static HACCEL hAccel  = NULL;

static HWND hPicker   = NULL;
static HWND hwndList  = NULL;
static HWND hTreeView = NULL;
static HWND hProgWnd  = NULL;
static HWND hHelp     = NULL;

static HWND hLoad     = NULL;
static HWND hAbort    = NULL;
static BOOL abort_loading;

static HINSTANCE hInst = NULL;

static HFONT hFont = NULL;     /* Font for list view */

static int              page_index;
static game_data_type*  game_data;
static int              game_count;

static int  last_sort = 0;

/* global data--know where to send messages */
#ifndef MEM_DEBUG
static BOOL in_emulation;
#else
BOOL in_emulation;
#endif

/* quit after game */
static BOOL quit = 0;

/* idle work at startup */
static BOOL idle_work;

static int  game_index = 0;
static int  progBarStep;

static BOOL bDoGameCheck = FALSE;

/* current menu check for listview */
static int current_view_id;

/* Tree control variables */
static BOOL bShowTree = 1;
static BOOL bShowToolBar = 1;
static BOOL bShowStatusBar = 1;
static BOOL bProgressShown = FALSE;
static BOOL bListReady = FALSE;
static BOOL bUseJapanese  = FALSE;
static BOOL bUseModifyThe = TRUE;
#ifdef KAILLERA
static BOOL bUseImeInChat = FALSE;
static BOOL bUseFixedForm = FALSE;
#endif

/* use a joystick subsystem in the gui? */
static struct OSDJoystick *joygui;

/* Intellimouse available? (0 = no) */
static UINT uiMsh_MsgMouseWheel;

/* sort columns in reverse order */
static BOOL reverse_sort = FALSE;
static UINT lastColumnClick = 0;
static UINT m_uHeaderSortCol = 0;
static BOOL m_fHeaderSortAsc = TRUE;
static WNDPROC g_lpHeaderWndProc = NULL;

static POPUPSTRING popstr[MAX_MENUS];

/* Tool and Status bar variables */
static HWND hStatusBar = 0;
static HWND hToolBar = 0;

/* Column Order as Displayed */
static BOOL oldControl = FALSE;
static int  realColumn[COLUMN_MAX];

/* Used to recalculate the main window layout */
static int  bottomMargin;
static int  topMargin;
static int  have_history = FALSE;

static BOOL bScreenShotAvailable = FALSE;
static BOOL nPictType = PICT_SCREENSHOT;
static BOOL have_selection = FALSE;

/* Icon variables */
static HIMAGELIST   hLarge;
static HIMAGELIST   hSmall;
static int          *icon_index = NULL;

/* Kaillera Version Name */
static char version_name[VERSIONTEXT_LEN];

/* Icon names we will load and use */
static char *icon_names[] = {
    "noroms",
    "roms",
    "unknown",
    "warning"
};

#define NUM_ICONS           (sizeof(icon_names) / sizeof(icon_names[0]))


#ifdef BUILD_ENGLISH

#ifdef KAILLERA
#define NUM_TOOLBUTTONS 14
#define NUM_TOOLTIPS 8
#else
#define NUM_TOOLBUTTONS 12
#define NUM_TOOLTIPS 7
#endif

static TBBUTTON tbb[NUM_TOOLBUTTONS] = {
    { 0, 0,                  TBSTATE_ENABLED, TBSTYLE_SEP,        0, 0},
    { 0, ID_VIEW_FOLDERS,    TBSTATE_ENABLED, TBSTYLE_CHECK,      0, 0},
    { 1, ID_VIEW_SCREEN_SHOT,TBSTATE_ENABLED, TBSTYLE_CHECK,      0, 1},
    { 0, 0,                  TBSTATE_ENABLED, TBSTYLE_SEP,        0, 0},
    { 2, ID_VIEW_ICON,       TBSTATE_ENABLED, TBSTYLE_CHECKGROUP, 0, 2},
    { 3, ID_VIEW_SMALL_ICON, TBSTATE_ENABLED, TBSTYLE_CHECKGROUP, 0, 3},
    { 4, ID_VIEW_LIST_MENU,  TBSTATE_ENABLED, TBSTYLE_CHECKGROUP, 0, 4},
    { 5, ID_VIEW_DETAIL,     TBSTATE_ENABLED, TBSTYLE_CHECKGROUP, 0, 5},
    { 0, 0,                  TBSTATE_ENABLED, TBSTYLE_SEP,        0, 0},
    { 7, ID_MODIFY_THE,      TBSTATE_ENABLED, TBSTYLE_CHECK,      0, 7},
#ifdef KAILLERA
    { 0, 0,                  TBSTATE_ENABLED, TBSTYLE_SEP,        0, 0},
    { 9, ID_USE_FIXED_FORM,  TBSTATE_ENABLED, TBSTYLE_CHECK,      0, 7},
#endif
    { 0, 0,                  TBSTATE_ENABLED, TBSTYLE_SEP,        0, 0},
    {10, ID_HELP_ABOUT,      TBSTATE_ENABLED, TBSTYLE_BUTTON,     0, 8}
};
#else

#ifdef KAILLERA
#define NUM_TOOLBUTTONS 16
#define NUM_TOOLTIPS 10
#else
#define NUM_TOOLBUTTONS 13
#define NUM_TOOLTIPS 8
#endif

static TBBUTTON tbb[NUM_TOOLBUTTONS] = {
    { 0, 0,                  TBSTATE_ENABLED, TBSTYLE_SEP,        0, 0},
    { 0, ID_VIEW_FOLDERS,    TBSTATE_ENABLED, TBSTYLE_CHECK,      0, 0},
    { 1, ID_VIEW_SCREEN_SHOT,TBSTATE_ENABLED, TBSTYLE_CHECK,      0, 1},
    { 0, 0,                  TBSTATE_ENABLED, TBSTYLE_SEP,        0, 0},
    { 2, ID_VIEW_ICON,       TBSTATE_ENABLED, TBSTYLE_CHECKGROUP, 0, 2},
    { 3, ID_VIEW_SMALL_ICON, TBSTATE_ENABLED, TBSTYLE_CHECKGROUP, 0, 3},
    { 4, ID_VIEW_LIST_MENU,  TBSTATE_ENABLED, TBSTYLE_CHECKGROUP, 0, 4},
    { 5, ID_VIEW_DETAIL,     TBSTATE_ENABLED, TBSTYLE_CHECKGROUP, 0, 5},
    { 0, 0,                  TBSTATE_ENABLED, TBSTYLE_SEP,        0, 0},
    { 6, ID_USE_JLIST,       TBSTATE_ENABLED, TBSTYLE_CHECK,      0, 6},
    { 7, ID_MODIFY_THE,      TBSTATE_ENABLED, TBSTYLE_CHECK,      0, 7},
#ifdef KAILLERA
    { 0, 0,                  TBSTATE_ENABLED, TBSTYLE_SEP,        0, 0},
    { 8, ID_USE_IME_IN_CHAT, TBSTATE_ENABLED, TBSTYLE_CHECK,      0, 7},
    { 9, ID_USE_FIXED_FORM,  TBSTATE_ENABLED, TBSTYLE_CHECK,      0, 8},
#endif
    { 0, 0,                  TBSTATE_ENABLED, TBSTYLE_SEP,        0, 0},
    {10, ID_HELP_ABOUT,      TBSTATE_ENABLED, TBSTYLE_BUTTON,     0, 9}
};
#endif

static const char *szTbStrings[NUM_TOOLTIPS + 1];

static int CommandToString[] = {
    ID_VIEW_FOLDERS,
    ID_VIEW_SCREEN_SHOT,
    ID_VIEW_ICON,
    ID_VIEW_SMALL_ICON,
    ID_VIEW_LIST_MENU,
    ID_VIEW_DETAIL,
#ifndef BUILD_ENGLISH
	ID_USE_JLIST,
#endif
	ID_MODIFY_THE,
#ifdef KAILLERA
#ifndef BUILD_ENGLISH
    ID_USE_IME_IN_CHAT,
#endif
	ID_USE_FIXED_FORM,
#endif
    ID_HELP_ABOUT,
    -1
};

/* How to resize main window */
static ResizeItem main_resize_items[] = {
    { RA_HWND, 0,            RA_LEFT  | RA_RIGHT  | RA_TOP,     NULL },
    { RA_HWND, 0,            RA_LEFT  | RA_RIGHT  | RA_BOTTOM,  NULL },
    { RA_ID,   IDC_DIVIDER,  RA_LEFT  | RA_RIGHT  | RA_TOP,     NULL },
    { RA_ID,   IDC_TREE,     RA_LEFT  | RA_BOTTOM | RA_TOP,     NULL },
    { RA_ID,   IDC_LIST,     RA_ALL,                            NULL },
    { RA_ID,   IDC_SPLITTER, RA_LEFT  | RA_BOTTOM | RA_TOP,     NULL },
    { RA_ID,   IDC_SPLITTER2,RA_RIGHT | RA_BOTTOM | RA_TOP,     NULL },
    { RA_ID,   IDC_SSFRAME,  RA_RIGHT | RA_BOTTOM | RA_TOP,     NULL },
    { RA_ID,   IDC_SSPICTURE,RA_RIGHT | RA_BOTTOM | RA_TOP,     NULL },
    { RA_ID,   IDC_HISTORY,  RA_RIGHT | RA_BOTTOM | RA_TOP,     NULL },
    { RA_ID,   IDC_SSDEFPIC, RA_RIGHT | RA_TOP,                 NULL },
    { RA_END,  0,            0,                                 NULL }
};

static Resize main_resize = { {0, 0, 0, 0}, main_resize_items };

/* our dialog/configured options */
static options_type playing_game_options;

/* last directory for common file dialogs */
static char last_directory[MAX_PATH];

/* system-wide window message sent out with an ATOM of the current game name
   each time it changes */
static UINT mame32_message;
static BOOL bDoBroadcast;

static int mame_debug = 0;


/***************************************************************************
    Global variables
 ***************************************************************************/

/* Background Image handles also accessed from TreeView.c */
HPALETTE         hPALbg   = 0;
HBITMAP          hBitmap  = 0;
MYBITMAPINFO     bmDesc;

/* List view Column text */
char *column_names[COLUMN_MAX];


/***************************************************************************
    Message Macros
 ***************************************************************************/

#ifndef StatusBar_GetItemRect
#define StatusBar_GetItemRect(hWnd, iPart, lpRect) \
    SendMessage(hWnd, SB_GETRECT, (WPARAM) iPart, (LPARAM) (LPRECT) lpRect)
#endif

#ifndef ToolBar_CheckButton
#define ToolBar_CheckButton(hWnd, idButton, fCheck) \
    SendMessage(hWnd, TB_CHECKBUTTON, (WPARAM)idButton, (LPARAM)MAKELONG(fCheck, 0))
#endif

/***************************************************************************
    External functions
 ***************************************************************************/

int WINAPI WinMain(HINSTANCE    hInstance,
                   HINSTANCE    hPrevInstance,
                   LPSTR        lpCmdLine,
                   int          nCmdShow)
{
    MSG     msg;

#ifdef KAILLERA
    kailleraInit();
#endif /* KAILLERA */

    if (!Win32UI_init(hInstance, lpCmdLine, nCmdShow))
        return 1;

    /*
        Simplified MFC Run() alg. See mfc/src/thrdcore.cpp.
    */
    for (;;)
    {
        /* phase1: check to see if we can do idle work */
        while (idle_work && !PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
        {
            /* call OnIdle while idle_work */
            OnIdle();
        }

        /* phase2: pump messages while available */
        do
        {
            /* pump message, but quit on WM_QUIT */
            if (!PumpMessage())
            {
                Win32UI_exit();
                return msg.wParam;
            }

        } while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE));
    }

#ifdef KAILLERA
    kailleraShutdown();
#endif /* KAILLERA */

    return 0;
}


HWND GetMainWindow(void)
{
    return hMain;
}

int GetNumColumns(HWND hWnd)
{
    int  nColumn = 0;
    int  i;
    HWND hwndHeader;
    int  shown[COLUMN_MAX];

    GetColumnShown(shown);
    hwndHeader = GetDlgItem(hWnd, 0);

    if (oldControl || (nColumn = Header_GetItemCount(hwndHeader)) < 1)
    {
        nColumn = 0;
        for (i = 0; i < COLUMN_MAX ; i++ )
        {
            if (shown[i])
                nColumn++;
        }
    }
    return nColumn;
}

void GetRealColumnOrder(int order[])
{
    int tmpOrder[COLUMN_MAX];
    int nColumnMax;
    int i;

    nColumnMax = GetNumColumns(hwndList);

    /* Get the Column Order and save it */
    if (!oldControl)
    {
        ListView_GetColumnOrderArray(hwndList, nColumnMax, tmpOrder);

        for (i = 0; i < nColumnMax; i++)
        {
            order[i] = realColumn[tmpOrder[i]];
        }
    }
}

int UpdateLoadProgress(const char* name, int current, int total)
{
    MSG Msg;

    if (hLoad == NULL)
        return 0;

    while (PeekMessage(&Msg, NULL, 0, 0, PM_REMOVE))
    {
        if (!IsDialogMessage(hLoad, &Msg))
        {
            TranslateMessage(&Msg);
            DispatchMessage(&Msg);
        }
    }

    if (abort_loading)
        return 1;

    SendDlgItemMessage(hLoad, IDC_LOAD_PROGRESS, PBM_SETRANGE, 0, MAKELPARAM(0, total));
    SendDlgItemMessage(hLoad, IDC_LOAD_PROGRESS, PBM_SETPOS, current, 0);

    SetWindowText(GetDlgItem(hLoad, IDC_LOAD_ROMNAME), name);

    if (current == total)
    {
        ShowWindow(hMain, SW_HIDE);
        ShowWindow(hLoad, SW_HIDE);
        EndDialog(hLoad, 0);
        hLoad = NULL;
    }

    return 0;
}

BOOL GameUsesTrackball(int game)
{
    int port;

    /* new trackball support */
    if (drivers[game]->input_ports != 0)
    {
        port = 0;
        while (drivers[game]->input_ports[port].type != IPT_END)
        {
            int type = drivers[game]->input_ports[port].type & ~IPF_MASK;
            if (type == IPT_TRACKBALL_X || type == IPT_TRACKBALL_Y)
            {
                return TRUE;
            }
            port++;
        }
        return FALSE;
    }
    return FALSE;
}

BOOL GameUsesRasterDriver(int game)
{
    if (drivers[game]->flags & GAME_RASTER_DRIVER)
        return TRUE;
    return FALSE;
}

BOOL GameUsesRomPatch(int game)
{
    if (!strcmp(drivers[game]->name, "garoup")    ||
        !strcmp(drivers[game]->name, "fightfev"))
        return TRUE;
    return FALSE;
}

BOOL GameUsesOverClock(int game)
{
	if (!strcmp(drivers[game]->name, "s1945p")  ||
		!strcmp(drivers[game]->name, "s1945pn") ||
		!strcmp(drivers[game]->name, "aof3")    ||
		!strcmp(drivers[game]->name, "zedblade"))
		return TRUE;
    return FALSE;
}

BOOL GameOnlyJukeBox(int game)
{
    if (!strcmp(drivers[game]->name, "mslug3")   ||
		!strcmp(drivers[game]->name, "kof2000")  ||
		!strcmp(drivers[game]->name, "nitd")     ||
		!strcmp(drivers[game]->name, "sengoku3") ||
		!strcmp(drivers[game]->name, "zupapa"))
        return TRUE;
    return FALSE;
}

/*
 * PURPOSE: Format raw data read from an ICO file to an HICON
 * PARAMS:  PBYTE ptrBuffer  - Raw data from an ICO file
 *          UINT nBufferSize - Size of buffer ptrBuffer
 * RETURNS: HICON - handle to the icon, NULL for failure
 * History: July '95 - Created
 *          March '00- Seriously butchered from MSDN for mine own
 *          purposes, sayeth H0ek.
 */
HICON FormatICOInMemoryToHICON( PBYTE ptrBuffer, UINT nBufferSize )
{
    ICONIMAGE           IconImage;
    LPICONDIRENTRY      lpIDE = NULL;
    UINT                nNumImages;
    UINT                nBufferIndex = 0;
    HICON               hIcon = NULL;

    /* Is there a WORD? */
    if( nBufferSize < sizeof( WORD ) )
    {
        return NULL;
    }

    /* Was it 'reserved' ?   (ie 0) */
    if( (WORD)(ptrBuffer[nBufferIndex]) != 0 )
    {
        return NULL;
    }

    nBufferIndex += sizeof( WORD );

    /* Is there a WORD? */
    if( nBufferSize - nBufferIndex < sizeof( WORD ) )
    {
        return NULL;
    }

    /* Was it type 1? */
    if( (WORD)(ptrBuffer[nBufferIndex]) != 1 )
    {
        return NULL;
    }

    nBufferIndex += sizeof( WORD );

    /* Is there a WORD? */
    if( nBufferSize - nBufferIndex < sizeof( WORD ) )
    {
        return NULL;
    }

    /* Then that's the number of images in the ICO file */
    nNumImages = (WORD)(ptrBuffer[nBufferIndex]);

    /* Is there at least one icon in the file? */
    if( nNumImages < 1 )
    {
        return NULL;
    }

    nBufferIndex += sizeof( WORD );

    /* Is there enough space for the icon directory entries? */
    if( ( nBufferIndex + nNumImages * sizeof( ICONDIRENTRY ) ) > nBufferSize )
    {
        return NULL;
    }

    /* Assign icon directory entries from buffer */
    lpIDE = ( LPICONDIRENTRY )( &ptrBuffer[nBufferIndex] );
    nBufferIndex += nNumImages * sizeof( ICONDIRENTRY );

    IconImage.dwNumBytes = lpIDE->dwBytesInRes;

    /* Seek to beginning of this image */
    if( lpIDE->dwImageOffset > nBufferSize )
    {
        return NULL;
    }

    nBufferIndex = lpIDE->dwImageOffset;

    /* Read it in */
    if( ( nBufferIndex + lpIDE->dwBytesInRes ) > nBufferSize )
    {
        return NULL;
    }

    IconImage.lpBits = &ptrBuffer[nBufferIndex];
    nBufferIndex += lpIDE->dwBytesInRes;

    hIcon = CreateIconFromResourceEx( IconImage.lpBits, IconImage.dwNumBytes, TRUE, 0x00030000,
            (*(LPBITMAPINFOHEADER)(IconImage.lpBits)).biWidth, (*(LPBITMAPINFOHEADER)(IconImage.lpBits)).biHeight/2, 0 );

    /* It failed, odds are good we're on NT so try the non-Ex way */
    if( hIcon == NULL )
    {
        /* We would break on NT if we try with a 16bpp image */
        if(((LPBITMAPINFO)IconImage.lpBits)->bmiHeader.biBitCount != 16)
        {
            hIcon = CreateIconFromResource( IconImage.lpBits, IconImage.dwNumBytes, TRUE, 0x00030000 );
        }
    }
    return hIcon;
}

HICON LoadIconFromFile(char *iconname)
{
    HICON       hIcon = 0;
    struct stat file_stat;
    char        tmpStr[MAX_PATH];
    char        tmpIcoName[MAX_PATH];
    const char* sDirName = GetIconDir();
    PBYTE       bufferPtr;
    UINT        bufferLen;

    sprintf(tmpStr, "%s/%s.ico", sDirName, iconname);
    if (stat(tmpStr, &file_stat) != 0
    || (hIcon = ExtractIcon(hInst, tmpStr, 0)) == 0)
    {
        sprintf(tmpStr, "%s/icons.zip", sDirName);
        sprintf(tmpIcoName, "%s.ico", iconname);
        if (stat(tmpStr, &file_stat) == 0)
        {
            if (load_zipped_file(tmpStr, tmpIcoName, &bufferPtr, &bufferLen) == 0)
            {
                hIcon = FormatICOInMemoryToHICON(bufferPtr, bufferLen);
                MyFree(bufferPtr);
            }
        }
    }
    return hIcon;
}

/* Return the number of games currently displayed */
int GetNumGames(void)
{
   return game_count;
}

game_data_type * GetGameData(void)
{
    return game_data;
}

/* Adjust the list view and screenshot button based on GetShowScreenShot() */
void UpdateScreenShot(void)
{
    HWND hList = GetDlgItem(hPicker, IDC_LIST);
    HWND hParent = GetParent(hList);
    RECT rect;
    int  nWidth, show;
    int  nTreeWidth = 0;

    /* Size the List Control in the Picker */
    GetClientRect(hParent, &rect);

    if (bShowStatusBar)
        rect.bottom -= bottomMargin;
    if (bShowToolBar)
        rect.top += topMargin;

    if (GetShowScreenShot())
    {
        nWidth = nSplitterOffset[SPLITTER_MAX-1];
        show   = SW_SHOW;
        CheckMenuItem(GetMenu(hMain),ID_VIEW_SCREEN_SHOT, MF_CHECKED);
        ToolBar_CheckButton(hToolBar, ID_VIEW_SCREEN_SHOT, MF_CHECKED);
    }
    else
    {
        nWidth = rect.right;
        show   = SW_HIDE;
        CheckMenuItem(GetMenu(hMain),ID_VIEW_SCREEN_SHOT, MF_UNCHECKED);
        ToolBar_CheckButton(hToolBar, ID_VIEW_SCREEN_SHOT, MF_UNCHECKED);
    }
    if (show == SW_HIDE)
    {
        ShowWindow(GetDlgItem(hParent, IDC_SSPICTURE), SW_HIDE);
        ShowWindow(GetDlgItem(hParent, IDC_SSFRAME),   SW_HIDE);
        ShowWindow(GetDlgItem(hParent, IDC_SSDEFPIC),  SW_HIDE);
    }

    ShowWindow(GetDlgItem(hParent, IDC_HISTORY),   SW_HIDE);

    /* Tree control */
    if (bShowTree)
    {
        nTreeWidth = nSplitterOffset[0];
        if (nTreeWidth >= nWidth)
        {
            nTreeWidth = nWidth - 10;
        }

        MoveWindow(GetDlgItem(hParent, IDC_TREE), 0, rect.top + 2, nTreeWidth - 2,
            (rect.bottom - rect.top) - 4 , TRUE);

        MoveWindow(GetDlgItem(hParent, IDC_SPLITTER), nTreeWidth, rect.top + 2,
            4, (rect.bottom - rect.top) - 4, TRUE);

        ShowWindow(GetDlgItem(hParent, IDC_TREE), SW_SHOW);
    }
    else
    {
        ShowWindow(GetDlgItem(hParent, IDC_TREE), SW_HIDE);
    }

    /* Splitter window #2 */
    MoveWindow(GetDlgItem(hParent, IDC_SPLITTER2), nSplitterOffset[1],
        rect.top + 2, 4, (rect.bottom - rect.top) - 2, FALSE);

    /* List control */
    MoveWindow(hList, 2 + nTreeWidth, rect.top + 2,
        (nWidth - nTreeWidth) - 2, (rect.bottom - rect.top) - 4 , TRUE);

    if (GetShowScreenShot())
        ShowWindow(GetDlgItem(hParent, IDC_SSFRAME), SW_SHOW);

    UpdateHistory(GetSelectedPickItem());

    if (GetShowScreenShot() && !nPictType && have_selection && have_history)
        ShowWindow(GetDlgItem(hParent, IDC_HISTORY), SW_SHOW);
}

void ResizePickerControls(HWND hWnd)
{
    RECT picRect;
    RECT frameRect;
    RECT rect, sRect;
    int  nListWidth, nScreenShotWidth, picX, picY, nTreeWidth;
    static BOOL firstTime = TRUE;
    int  doSSControls = TRUE;

//    printf("ResizePickerControls() called\n");

    /* Size the List Control in the Picker */
    GetClientRect(hWnd, &rect);

    /* Calc 1/2 of the display for the screenshot and 1/4 for the treecontrol */
    if (firstTime)
    {
        RECT rWindow;

        nListWidth = rect.right / 2;
        nSplitterOffset[1] = nListWidth;
        nSplitterOffset[0] = nListWidth / 2;
        GetWindowRect(hStatusBar, &rWindow);
        bottomMargin = rWindow.bottom - rWindow.top;
        GetWindowRect(hToolBar, &rWindow);
        topMargin = rWindow.bottom - rWindow.top;
        //buttonMargin = (sRect.bottom + 4);

        firstTime = FALSE;
    }
    else
    {
        doSSControls = GetShowScreenShot();
        nListWidth = nSplitterOffset[1];
    }

    if (bShowStatusBar)
        rect.bottom -= bottomMargin;

    if (bShowToolBar)
        rect.top += topMargin;

    MoveWindow(GetDlgItem(hWnd, IDC_DIVIDER), rect.left, rect.top - 4, rect.right, 2, TRUE);

    nScreenShotWidth = (rect.right - nListWidth) - 4;

    /* Tree Control */
    nTreeWidth = nSplitterOffset[0];
    MoveWindow(GetDlgItem(hWnd, IDC_TREE), 4, rect.top + 2,
        nTreeWidth - 4, (rect.bottom - rect.top) - 4, TRUE);

    /* Splitter window #1 */
    MoveWindow(GetDlgItem(hWnd, IDC_SPLITTER), nTreeWidth, rect.top + 2,
        4, (rect.bottom - rect.top) - 4, TRUE);

    /* List control */
    MoveWindow(GetDlgItem(hWnd, IDC_LIST), 4 + nTreeWidth, rect.top + 2,
        (nListWidth - nTreeWidth) - 4, (rect.bottom - rect.top) - 4, TRUE);

    /* Splitter window #2 */
    MoveWindow(GetDlgItem(hWnd, IDC_SPLITTER2), nListWidth, rect.top + 2,
        4, (rect.bottom - rect.top) - 2, doSSControls);

    /* resize the Screen shot frame */
    MoveWindow(GetDlgItem(hWnd, IDC_SSFRAME), nListWidth + 4, rect.top + 2,
        nScreenShotWidth - 2, (rect.bottom - rect.top) - 4, doSSControls);

    /* The screen shot controls */
    GetClientRect(GetDlgItem(hWnd, IDC_SSFRAME), &frameRect);
    GetClientRect(GetDlgItem(hWnd, IDC_SSDEFPIC), &picRect);

    picX = (frameRect.right - (picRect.right - picRect.left)) / 2;
    picY = (frameRect.bottom - (picRect.bottom - picRect.top)) / 2;

    /* Mame logo */
    MoveWindow(GetDlgItem(hWnd, IDC_SSDEFPIC), nListWidth + picX + 4,
        rect.top + 24, picRect.right + 2, picRect.bottom + 2, doSSControls);

    /* Screen shot sunken frame */
    MoveWindow(GetDlgItem(hWnd, IDC_SSPICTURE), nListWidth + picX + 4,
        rect.top + picY , picRect.right, picRect.bottom, doSSControls);

    /* Text control - game history */
    sRect.left = nListWidth + 14;
    sRect.right = sRect.left + (nScreenShotWidth - 22);
    sRect.top = rect.top + 264;
    sRect.bottom = (rect.bottom - rect.top) - 278;

    MoveWindow(GetDlgItem(hWnd, IDC_HISTORY),
        sRect.left, sRect.top,
        sRect.right - sRect.left, sRect.bottom, doSSControls);
}

char *ModifyThe(const char *str)
{
    if (UseModifyThe())
    {
	    static int bufno = 0;
	    static char buffer[4][255];

	    if (strncmp(str, "The ", 4) == 0)
	    {
	        char *s, *p;
	        char temp[255];

	        strcpy(temp, &str[4]);

	        bufno = (bufno + 1) % 4;

	        s = buffer[bufno];

	        /* Check for version notes in parens */
	        p = strchr(temp, '(');
	        if (p)
	        {
	            p[-1] = '\0';
	        }

	        strcpy(s, temp);
	        strcat(s, ", The");

	        if (p)
	        {
		        strcat(s, " ");
		        strcat(s, p);
	        }

	        return s;
	    }
	}
    return (char *)str;
}

/***************************************************************************
    Internal functions
 ***************************************************************************/

static BOOL WINAPI HandlerRoutine( DWORD dwCtrlType )
{
    if (MAME32App.m_bIsInitialized)
    {
        MAME32App.Quit();
    }
    return TRUE;
}

static BOOL Win32UI_init(HINSTANCE hInstance, LPSTR lpCmdLine, int nCmdShow)
{
    WNDCLASS    wndclass;
    RECT        rect;
    int         i;
    struct GameDriver * drvr;

    memory_manager_init();

    mame32_message = RegisterWindowMessage("NMAME X II");
    bDoBroadcast = FALSE;

    srand((unsigned)time(NULL));

    joygui = NULL;

    game_count = 0;
    while (drivers[game_count] != 0)
        game_count++;

    game_data = (game_data_type *)MyMalloc(game_count * sizeof(game_data_type));

    wndclass.style           = CS_HREDRAW | CS_VREDRAW;
    wndclass.lpfnWndProc     = MameWindowProc;
    wndclass.cbClsExtra      = 0;
    wndclass.cbWndExtra      = DLGWINDOWEXTRA;
    wndclass.hInstance       = hInstance;
    wndclass.hIcon           = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_MAME32_ICON));
    wndclass.hCursor         = NULL;
    wndclass.hbrBackground   = (HBRUSH)(COLOR_3DFACE + 1);
    wndclass.lpszMenuName    = MAKEINTRESOURCE(IDR_UI_MENU);
    wndclass.lpszClassName   = "NMAME X II";

    RegisterClass(&wndclass);

    InitCommonControls();

#if 0
    if (!MAME32App.DetectMMX())
    {
        char *buf = MAME32NAME " MMX MMXpłB\n\n"
                    "Ȃ̃p\RCPÚAMMX𓋍ڂĂ܂B\n"
                    "m[}łNMAME X IIgpB";

        MessageBox(0, buf, MAME32NAME " - MMXoł܂", MB_OK | MB_ICONWARNING);

        return FALSE;
    }
#endif

    // Are we using an Old comctl32.dll?
    oldControl = (GetDllVersion() == FALSE);
    if (oldControl)
    {
		char buf[512];

		connect_strings(buf, WIN32_oldcontrol01, WIN32_oldcontrol07);

        if (IDNO == MessageBox(0, buf, win32_getstring(WIN32_oldcontrol_title), MB_YESNO | MB_ICONWARNING))
            return FALSE;
    }

    OptionsInit(game_count);

    for (i = 0; i < game_count; i++)
    {
        int j;
        BOOL neogeo_clone = FALSE;
        BOOL neogeo       = FALSE;
        BOOL found        = FALSE;

        drvr = (struct GameDriver *)drivers[i]->clone_of;

        if (drvr == DRIVER_NEOGEO)
            neogeo = TRUE;

        if (drvr != DRIVER_ROOT)
        {
            for (j = 0; j < game_count; j++)
            {
                if (drvr == drivers[j])
                {
                    found = TRUE;
                    if (drvr->clone_of == DRIVER_NEOGEO)
                    {
                        neogeo_clone = TRUE;
                        neogeo       = TRUE;
                    }
                    break;
                }
            }
        }
        if (!found
        &&  drivers[i]->clone_of != DRIVER_NEOGEO
        &&  drivers[i]->clone_of != DRIVER_ROOT)
            found = TRUE;

        game_data[i].in_list      = found;
        game_data[i].neogeo_clone = neogeo_clone;
        game_data[i].neogeo       = neogeo;
    }

    // init files after OptionsInit to init paths
    File.init();
    strcpy(last_directory, GetInpDir());

    hMain = MAME32App.m_hwndUI = CreateDialog(hInstance,
        MAKEINTRESOURCE(IDD_MAIN), 0, NULL);

    hPicker = hMain;

    /* Stash hInstance for later use */
    hInst = hInstance;

    hToolBar   = InitToolbar(hMain);
    hStatusBar = InitStatusBar(hMain);
    hProgWnd   = InitProgressBar(hStatusBar);

    main_resize_items[0].u.hwnd = hToolBar;
    main_resize_items[1].u.hwnd = hStatusBar;

    /* In order to handle 'Large Fonts' as the Windows
     * default setting, we need to make the dialogs small
     * enough to fit in our smallest window size with
     * large fonts, then resize the picker, tab and button
     * controls to fill the window, no matter which font
     * is currently set.  This will still look like bad
     * if the user uses a bigger default font than 125%
     * (Large Fonts) on the Windows display setting tab.
     *
     * NOTE: This has to do with Windows default font size
     * settings, NOT our picker font size.
     */

    GetClientRect(hMain, &rect);

    hTreeView       = GetDlgItem(hPicker, IDC_TREE);
    hwndList        = GetDlgItem(hPicker, IDC_LIST);

    InitSplitters();


    AddSplitter(GetDlgItem(hPicker, IDC_SPLITTER), hTreeView, hwndList,
                AdjustSplitter1Rect);
    AddSplitter(GetDlgItem(hPicker, IDC_SPLITTER2), hwndList,
                GetDlgItem(hPicker,IDC_SSFRAME),AdjustSplitter2Rect);

    /* Initial adjustment of controls on the Picker window */
    ResizePickerControls(hPicker);

    /* Reset the font */
    {
        LOGFONT     logfont;

        GetListFont(&logfont);
        hFont = CreateFontIndirect(&logfont);
        if (hFont != NULL)
        {
            if (hwndList != NULL)
                SetWindowFont(hwndList, hFont, FALSE);
            if (hTreeView != NULL)
                SetWindowFont(hTreeView, hFont, FALSE);
            SetWindowFont(GetDlgItem(hPicker, IDC_HISTORY), hFont, FALSE);
        }
    }

    nPictType = GetShowPictType();

    bDoGameCheck = GetGameCheck();
    idle_work = TRUE;
    game_index = 0;

    bShowTree = GetShowFolderList();
    bShowToolBar = GetShowToolBar();
    bShowStatusBar = GetShowStatusBar();
#ifndef BUILD_ENGLISH
    bUseJapanese   = UseJapaneseList();
#endif
    bUseModifyThe  = UseModifyThe();
#ifdef KAILLERA
    bUseImeInChat  = GetUseImeInChat();
    bUseFixedForm  = GetUseFixedForm();
#endif

    CheckMenuItem(GetMenu(hMain), ID_VIEW_FOLDERS, (bShowTree) ? MF_CHECKED : MF_UNCHECKED);
    ToolBar_CheckButton(hToolBar, ID_VIEW_FOLDERS, (bShowTree) ? MF_CHECKED : MF_UNCHECKED);
    CheckMenuItem(GetMenu(hMain), ID_VIEW_TOOLBARS, (bShowToolBar) ? MF_CHECKED : MF_UNCHECKED);
    ShowWindow(hToolBar, (bShowToolBar) ? SW_SHOW : SW_HIDE);
    CheckMenuItem(GetMenu(hMain), ID_VIEW_STATUS, (bShowStatusBar) ? MF_CHECKED : MF_UNCHECKED);
    ShowWindow(hStatusBar, (bShowStatusBar) ? SW_SHOW : SW_HIDE);
#ifndef BUILD_ENGLISH
    ToolBar_CheckButton(hToolBar, ID_USE_JLIST, (bUseJapanese) ? MF_CHECKED : MF_UNCHECKED);
#endif
    ToolBar_CheckButton(hToolBar, ID_MODIFY_THE, (bUseModifyThe) ? MF_CHECKED : MF_UNCHECKED);
#ifdef KAILLERA
#ifndef BUILD_ENGLISH
    ToolBar_CheckButton(hToolBar, ID_USE_IME_IN_CHAT, (bUseImeInChat) ? MF_CHECKED : MF_UNCHECKED);
#endif
    ToolBar_CheckButton(hToolBar, ID_USE_FIXED_FORM, (bUseFixedForm) ? MF_CHECKED : MF_UNCHECKED);
#endif

	for (i = 0 ; i < COLUMN_MAX ; i++)
		column_names[i] = (char *)win32_getstring(i + WIN32_columntext_game);

    if (oldControl)
    {
        EnableMenuItem(GetMenu(hMain), ID_CUSTOMIZE_FIELDS, MF_GRAYED);
        EnableMenuItem(GetMenu(hMain), ID_GAME_PROPERTIES,  MF_GRAYED);
        EnableMenuItem(GetMenu(hMain), ID_OPTIONS_DEFAULTS, MF_GRAYED);
    }

    // Init DirectDraw
    if (!DirectDraw_Initialize())
    {
       DialogBox(GetModuleHandle(NULL),MAKEINTRESOURCE(IDD_DIRECTX), NULL, DirectXDialogProc);
       return FALSE;
    }

    LoadListBitmap();

    InitTree(hTreeView, game_count);

    /* Subclass the tree control for painting */
    Tree_Initialize(hTreeView);

    SetCurrentFolder(GetFolderByID(GetSavedFolderID()));

    switch(GetViewMode())
    {
    case VIEW_LARGE_ICONS:
        SetView(ID_VIEW_ICON,LVS_ICON);
        break;
    case VIEW_SMALL_ICONS:
        SetView(ID_VIEW_SMALL_ICON,LVS_SMALLICON);
        break;
    case VIEW_INLIST:
        SetView(ID_VIEW_LIST_MENU,LVS_LIST);
        break;
    case VIEW_REPORT:
    default:
        SetView(ID_VIEW_DETAIL,LVS_REPORT);
        break;
    }

    // Initialize listview columns
    InitPicker();
    SetFocus(hwndList);

    // Init Intellimouse
    {
        UINT uiMsh_Msg3DSupport;
        UINT uiMsh_MsgScrollLines;
        BOOL f3DSupport;
        INT  iScrollLines;
        HWND hwndMsWheel;

        hwndMsWheel = HwndMSWheel(
            &uiMsh_MsgMouseWheel, &uiMsh_Msg3DSupport,
            &uiMsh_MsgScrollLines, &f3DSupport, &iScrollLines);
    }

    // Init DirectInput
    if (!DirectInputInitialize())
    {
       DialogBox(GetModuleHandle(NULL),MAKEINTRESOURCE(IDD_DIRECTX), NULL, DirectXDialogProc);
       return FALSE;
    }

    if (ParseCommandLine(lpCmdLine))
    {
        PostMessage(hMain, MM_PLAY_GAME, 0, 0);
    }

    AdjustMetrics();

    UpdateScreenShot();

#ifdef MAME_DEBUG
    if (win32_debug)
    {
        int     cFile;
        FILE*   pFILE;

        AllocConsole();

        SetConsoleCtrlHandler(HandlerRoutine,1);
        SetConsoleMode(GetStdHandle(STD_INPUT_HANDLE),0);

        cFile = _open_osfhandle((long)GetStdHandle(STD_OUTPUT_HANDLE), _O_TEXT);
        pFILE = _fdopen(cFile, "w");
        *stdout = *pFILE;

        cFile = _open_osfhandle((long)GetStdHandle(STD_ERROR_HANDLE), _O_TEXT);
        pFILE = _fdopen(cFile, "w");
        *stderr = *pFILE;
#if 0
        cFile = _open_osfhandle((long)GetStdHandle(STD_INPUT_HANDLE), _O_TEXT);
        pFILE = _fdopen(cFile, "r");
        *stdin = *pFILE;
#endif
        setvbuf(stdout, NULL, _IONBF, 0);
        setvbuf(stderr, NULL, _IONBF, 0);
    }
#endif

    hAccel = LoadAccelerators(hInstance, MAKEINTRESOURCE(IDA_TAB_KEYS));

    InitZip(hMain);

    {
        options_type o;

        memcpy(&o, GetDefaultOptions(), sizeof(options_type));
        o.use_joystick = TRUE;
        DIJoystick.init(&o);
    }

    if (joygui != NULL)
    {
        /* hack to get joystick to start */
        options_type o;

        memcpy(&o, GetDefaultOptions(), sizeof(options_type));
        o.use_joystick = TRUE;
        if (joygui->init(&o) != 0)
            joygui = NULL;
        else
            SetTimer(hMain,0,50,NULL);
    }

    if (!quit)
        ShowWindow(hMain,nCmdShow);

    return TRUE;
}

static void Win32UI_exit(void)
{
    unzip_cache_clear();
    free_datafile();

    ExitZip();

    if (joygui != NULL)
        joygui->exit();

    DIJoystick.exit();

    DestroyAcceleratorTable(hAccel);

    if (game_data != NULL)
        MyFree(game_data);
    if (icon_index != NULL)
        MyFree(icon_index);

    DirectInputClose();
    DirectDraw_Close();

    if (!quit)
    {
        SetSavedFolderID(GetCurrentFolderID());
        FreeFolders();
        SetViewMode(current_view_id - ID_VIEW_ICON);
    }

    if (hBitmap)
        DeleteObject(hBitmap);

    if (hPALbg)
        DeleteObject(hPALbg);

    FreeScreenShot();

    if (!quit)
        OptionsExit();

    File.exit();
}

static long WINAPI MameWindowProc(HWND hWnd,UINT message,UINT wParam,LONG lParam)
{
    static int  nTab;
    MINMAXINFO  *mminfo;
    int         i;

    switch (message)
    {
    case WM_INITDIALOG:
        /* Initialize info for resizing subitems */
        GetClientRect(hWnd, &main_resize.rect);
        return TRUE;

    case WM_SETFOCUS:
        SetFocus(hwndList);
        break;

    case WM_SETTINGCHANGE:
        AdjustMetrics();
        return 0;

    case WM_SIZE:
        OnSize(hWnd, wParam, LOWORD(lParam), HIWORD(wParam));
        return TRUE;

    case WM_MENUSELECT:
        return Statusbar_MenuSelect(hWnd, wParam, lParam);

    case MM_PLAY_GAME:
        MamePlayGame();
        return TRUE;

    case WM_INITMENUPOPUP:
        UpdateMenu(GetDlgItem(hWnd, IDC_LIST), GetMenu(hWnd));
        break;

    case WM_CONTEXTMENU:
        if (HandleTreeContextMenu( hWnd, wParam, lParam) ||
            HandleContextMenu(hWnd, wParam, lParam))
            return FALSE;
        break;

    case WM_COMMAND:
        return MameCommand(hWnd,(int)(LOWORD(wParam)),(HWND)(lParam),(UINT)HIWORD(wParam));

    case WM_GETMINMAXINFO:
        /* Don't let the window get too small; it can break resizing */
        mminfo = (MINMAXINFO *) lParam;
        mminfo->ptMinTrackSize.x = MIN_WIDTH;
        mminfo->ptMinTrackSize.y = MIN_HEIGHT;
        return 0;

    case WM_TIMER:
        PollGUIJoystick();
        return TRUE;

    case WM_CLOSE:
        {
            /* save current item */
            RECT rect;
            AREA area;
            int widths[COLUMN_MAX];
            int order[COLUMN_MAX];
            int shown[COLUMN_MAX];
            int tmpOrder[COLUMN_MAX];
            int nItem;
            int nColumnMax = 0;

            /* Restore the window before we attempt to save parameters,
             * This fixed the lost window on startup problem.
             */
            ShowWindow(hWnd, SW_RESTORE);

            GetColumnOrder(order);
            GetColumnShown(shown);
            GetColumnWidths(widths);

            nColumnMax = GetNumColumns(hwndList);

            if (oldControl)
            {
                for (i = 0; i < nColumnMax; i++)
                {
                    widths[realColumn[i]] = ListView_GetColumnWidth(hwndList, i);
                }
            }
            else
            {
                /* Get the Column Order and save it */
                ListView_GetColumnOrderArray(hwndList, nColumnMax, tmpOrder);

                for (i = 0; i < nColumnMax; i++)
                {
                    widths[realColumn[i]] = ListView_GetColumnWidth(hwndList, i);
                    order[i] = realColumn[tmpOrder[i]];
                }
            }

            SetColumnWidths(widths);
            SetColumnOrder(order);

            for (i = 0; i < SPLITTER_MAX; i++)
                SetSplitterPos(i, nSplitterOffset[i]);

            GetWindowRect(hWnd, &rect);
            area.x = rect.left;
            area.y = rect.top;
            area.width  = rect.right - rect.left;
            area.height = rect.bottom - rect.top;
            SetWindowArea(&area);

            /* Save the users current game options and default game */
            nItem = GetSelectedPickItem();
#ifdef BUILD_ENGLISH
            SetDefaultGame(ModifyThe(drivers[nItem]->description));
#else
            if (UseJapaneseList())
                SetDefaultGame(drivers[nItem]->description);
            else
	            SetDefaultGame(ModifyThe(drivers[nItem]->descriptione));
#endif

            /* hide window to prevent orphan empty rectangles on the taskbar */
            ShowWindow(hWnd,SW_HIDE);

            PostQuitMessage(0);
            break;
        }

    case WM_LBUTTONDOWN:
        OnLButtonDown(hWnd, (UINT)wParam, MAKEPOINTS(lParam));
        break;

    case WM_MOUSEMOVE:
        OnMouseMove(hWnd, (UINT)wParam, MAKEPOINTS(lParam));
        break;

    case WM_LBUTTONUP:
        OnLButtonUp(hWnd, (UINT)wParam, MAKEPOINTS(lParam));
        break;

    case WM_NOTIFY:
        // Where is this message intended to go
        {
            LPNMHDR lpNmHdr = (LPNMHDR)lParam;

            // Fetch tooltip text
            if (lpNmHdr->code == TTN_NEEDTEXT)
            {
                LPTOOLTIPTEXT lpttt = (LPTOOLTIPTEXT)lParam;
                CopyToolTipText( lpttt );
            }
            if (lpNmHdr->hwndFrom == hwndList)
                return MamePickerNotify( lpNmHdr );
            if (lpNmHdr->hwndFrom == hTreeView)
                return TreeViewNotify( lpNmHdr );
        }
        break;

    case WM_DRAWITEM:
        {
            LPDRAWITEMSTRUCT lpDis = (LPDRAWITEMSTRUCT)lParam;
            switch(lpDis->CtlID)
            {
            case IDC_LIST:
                DrawItem((LPDRAWITEMSTRUCT)lParam);
                break;
            }
        }
        break;

    case WM_PAINT:
        if (GetShowScreenShot())
        {
            PAINTSTRUCT ps;
            RECT        rect;
            RECT        fRect;
            POINT       p = {0, 0};

            BeginPaint(hWnd, &ps);

            if (!have_history)
                ShowWindow(GetDlgItem(hWnd, IDC_HISTORY), SW_HIDE);

            ClientToScreen(hWnd, &p);
            GetWindowRect(GetDlgItem(hWnd, IDC_SSFRAME), &fRect);
            OffsetRect(&fRect, -p.x, -p.y);

            if (GetScreenShotRect(GetDlgItem(hWnd, IDC_SSFRAME), &rect, have_history && !nPictType))
            {
                DWORD dwStyle;
                DWORD dwStyleEx;

                dwStyle   = GetWindowLong(GetDlgItem(hWnd, IDC_SSPICTURE), GWL_STYLE);
                dwStyleEx = GetWindowLong(GetDlgItem(hWnd, IDC_SSPICTURE), GWL_EXSTYLE);

                AdjustWindowRectEx(&rect, dwStyle, FALSE, dwStyleEx);
                MoveWindow(GetDlgItem(hWnd, IDC_SSPICTURE),
                    fRect.left  + rect.left,
                    fRect.top   + rect.top,
                    rect.right  - rect.left,
                    rect.bottom - rect.top,
                    TRUE);
            }

            if (ScreenShotLoaded() && have_selection)
            {
                ShowWindow(GetDlgItem(hWnd, IDC_SSDEFPIC),  SW_HIDE);
                ShowWindow(GetDlgItem(hWnd, IDC_SSPICTURE), SW_HIDE);
                ShowWindow(GetDlgItem(hWnd, IDC_SSPICTURE), SW_SHOW);
            }
            else
            {
                ShowWindow(GetDlgItem(hWnd, IDC_SSDEFPIC),  SW_HIDE);
                ShowWindow(GetDlgItem(hWnd, IDC_SSPICTURE), SW_HIDE);
                ShowWindow(GetDlgItem(hWnd, IDC_SSDEFPIC),  SW_SHOW);
            }

            if (!nPictType && have_selection && have_history)
            {
                ShowWindow(GetDlgItem(hWnd, IDC_HISTORY), SW_HIDE);
                ShowWindow(GetDlgItem(hWnd, IDC_HISTORY), SW_SHOW);
            }

            /* Paint Screenshot Frame rect with the background image */
            PaintControl(GetDlgItem(hWnd, IDC_SSFRAME), TRUE);

            DrawScreenShot(GetDlgItem(hWnd, IDC_SSPICTURE));

            EndPaint(hWnd, &ps);

        }
        break;

    case WM_DESTROY:
        if (hFont != NULL)
        {
            DeleteObject(hFont);
            hFont = NULL;
        }
        PostQuitMessage(0);
        return 0;

    case WM_MOUSEWHEEL:
        {
            // WM_MOUSEWHEEL will be in Win98
            short zDelta = (short)HIWORD(wParam);

            if (zDelta < 0)
                SetSelectedPickItem(GetSelectedPick() + 1);
            else
                SetSelectedPickItem(GetSelectedPick() - 1);
            return 0;
        }
        break;

    default:
        if (message == uiMsh_MsgMouseWheel)
        {
            int zDelta = (int)wParam; /* wheel rotation */
            // WORD xPos = LOWORD(lParam); /* horizontal position of pointer */
            // WORD yPos = HIWORD(lParam); /* vertical position of pointer */

            if (zDelta < 0)
                SetSelectedPick(GetSelectedPick() + 1);
            else
                SetSelectedPick(GetSelectedPick() - 1);
            return 0;
        }
        break;
    }

    return DefWindowProc(hWnd, message, wParam, lParam);
}

static BOOL PumpMessage(void)
{
    MSG msg;

    return PumpAndReturnMessage(&msg);
}

static BOOL PumpAndReturnMessage(MSG *pmsg)
{
    if (!(GetMessage(pmsg, NULL, 0, 0)))
        return FALSE;

    if (!TranslateAccelerator(hMain, hAccel, pmsg))
    {
        if (!IsDialogMessage(hMain,pmsg))
        {
            TranslateMessage(pmsg);
            DispatchMessage(pmsg);
        }
    }
    return TRUE;
}

static void EmptyQueue(void)
{
    MSG msg;

    while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
    {
        if (!PumpMessage())
            ExitProcess(0);
    }
}

static void ColumnSort(int column, BOOL bColumn)
{
    LV_FINDINFO lvfi;
    int         save_column = 0;
    int         use_column = column;

    if (bColumn) // Was a column enum used??
    {
        int i;

        for (i = 0; i < COLUMN_MAX; i++)
        {
            if(realColumn[i] == column)
            {
                use_column = i;
                break;
            }
        }
    }

    /* Sort 'em */
    save_column = realColumn[use_column] + 1;

    reverse_sort = (column == last_sort && !reverse_sort) ? TRUE : FALSE;
    ListView_SortItems(hwndList, ListCompareFunc, realColumn[use_column]);
    last_sort = use_column;

    if (reverse_sort)
        save_column = -(save_column);

    SetSortColumn(save_column);

    lvfi.flags = LVFI_PARAM;
    lvfi.lParam = GetSelectedPickItem();
    ListView_EnsureVisible(hwndList, ListView_FindItem(hwndList, -1, &lvfi), FALSE);
    Header_SetSortInfo(hwndList, use_column, !reverse_sort);
}

static BOOL IsGameRomless(int iGame)
{
    const struct RomModule *region, *rom;

    /* loop over regions, then over files */
    for (region = rom_first_region(drivers[iGame]); region; region = rom_next_region(region))
        for (rom = rom_first_file(region); rom; rom = rom_next_file(rom))
            return FALSE;

    return TRUE;
}

static BOOL GameCheck(void)
{

    LV_FINDINFO lvfi;
    int         i;
    BOOL        success;
    BOOL        changed = FALSE;

    if (game_index == 0)
        ProgressBarShow();

    if (game_count <= game_index)
    {
        bDoGameCheck = FALSE;
        ProgressBarHide();
        return FALSE;
    }

    if (GetHasRoms(game_index) == UNKNOWN)
    {
        if (IsGameRomless(game_index))
        {
            success = TRUE;
        }
        else
        {
            /* Only check for the ZIP file once */
            success = File_ExistZip(drivers[game_index]->name, OSD_FILETYPE_ROM);
        }

        if (!success)
            success = FindRomSet(game_index);

        SetHasRoms(game_index, success);
        changed = TRUE;
    }

    lvfi.flags = LVFI_PARAM;
    lvfi.lParam = game_index;

    if (changed && (i = ListView_FindItem(hwndList, -1, &lvfi)) != -1)
        ListView_RedrawItems(hwndList, i, i);
    if ((game_index % progBarStep) == 0)
        ProgressBarStep();
    game_index++;

    return changed;
}

static void OnIdle(void)
{
    LV_FINDINFO lvfi;
    int         i;
    static int  bFirstTime = TRUE;
    static int  bResetList = TRUE;

    if (bFirstTime)
    {
        bResetList = FALSE;
        bFirstTime = FALSE;
    }
    if (bDoGameCheck)
    {
        bResetList |= GameCheck();
        return;
    }

    lvfi.flags = LVFI_STRING;
    lvfi.psz   = GetDefaultGame();
    i = ListView_FindItem(hwndList, -1, &lvfi);

    SetSelectedPick((i != -1) ? i : 0);
    i = GetSelectedPickItem();
#ifdef BUILD_ENGLISH
    SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 0, (LPARAM) ModifyThe(drivers[i]->description));
#else
    if (UseJapaneseList())
        SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 0, (LPARAM) drivers[i]->description);
    else
	    SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 0, (LPARAM) ModifyThe(drivers[i]->descriptione));
#endif
    if (bResetList || (GetViewMode() == VIEW_LARGE_ICONS))
    {
        InitGames(game_count);
        ResetListView();
    }
    idle_work = FALSE;
    UpdateStatusBar();
    bFirstTime = TRUE;
}


static void OnSize(HWND hWnd, UINT nState, int nWidth, int nHeight)
{
    static BOOL firstTime = TRUE;

    if (nState != SIZE_MAXIMIZED && nState != SIZE_RESTORED)
        return;

    ResizeWindow(hWnd, &main_resize);
    ResizeProgressBar();
    if (!firstTime)
        OnSizeSplitter(hMain);
    else
        firstTime = FALSE;
    /* Update the splitters structures as appropriate */
    RecalcSplitters();
}

static void ResizeWindow(HWND hParent, Resize *r)
{
    int index = 0, dx, dy;
    HWND hControl;
    RECT parent_rect, rect;
    ResizeItem *ri;
    POINT p = {0, 0};

    if (hParent == NULL)
        return;

    /* Calculate change in width and height of parent window */
    GetClientRect(hParent, &parent_rect);
    dx = parent_rect.right - r->rect.right;
    dy = parent_rect.bottom - r->rect.bottom;
    ClientToScreen(hParent, &p);

    while (r->items[index].type != RA_END)
    {
        ri = &r->items[index];
        if (ri->type == RA_ID)
            hControl = GetDlgItem(hParent, ri->u.id);
        else
            hControl = ri->u.hwnd;

        if (hControl == NULL)
        {
            index++;
            continue;
        }

        /* Get control's rectangle relative to parent */
        GetWindowRect(hControl, &rect);
        OffsetRect(&rect, -p.x, -p.y);

        if (!(ri->action & RA_LEFT))
            rect.left += dx;

        if (!(ri->action & RA_TOP))
            rect.top += dy;

        if (ri->action & RA_RIGHT)
            rect.right += dx;

        if (ri->action & RA_BOTTOM)
            rect.bottom += dy;

        MoveWindow(hControl, rect.left, rect.top,
            (rect.right - rect.left), (rect.bottom - rect.top), TRUE);

        /* Take care of subcontrols, if appropriate */
        if (ri->subwindow != NULL)
            ResizeWindow(hControl, ri->subwindow);

        index++;
    }

    /* Record parent window's new location */
    memcpy(&r->rect, &parent_rect, sizeof(RECT));
}

static void ProgressBarShow(void)
{
    RECT rect;
    int  widths[2] = {150, -1};

    if (game_count < 100)
        progBarStep = 100 / game_count;
    else
        progBarStep = (game_count / 100);

    SendMessage(hStatusBar, SB_SETPARTS, (WPARAM) 2, (LPARAM) (LPINT) widths);
    SendMessage(hProgWnd, PBM_SETRANGE, 0, (LPARAM) MAKELONG(0, game_count));
    SendMessage(hProgWnd, PBM_SETSTEP, (WPARAM)progBarStep, 0);
    SendMessage(hProgWnd, PBM_SETPOS, 0, 0);

    StatusBar_GetItemRect(hStatusBar, 1, &rect);

    MoveWindow(hProgWnd, rect.left, rect.top,
        rect.right - rect.left, rect.bottom - rect.top, TRUE);

    bProgressShown = TRUE;
}

static void ProgressBarHide(void)
{
    RECT rect;
    int  widths[4];
    HDC  hDC = GetDC(hProgWnd);
    SIZE size;
    int  numParts = 4;

    ShowWindow(hProgWnd, SW_HIDE);

    GetTextExtentPoint32(hDC, win32_getstring(WIN32_progressbar_MMX),
		strlen(win32_getstring(WIN32_progressbar_MMX)) , &size);
    widths[3] = size.cx;
    GetTextExtentPoint32(hDC, win32_getstring(WIN32_progressbar_games),
		strlen(win32_getstring(WIN32_progressbar_games)), &size);
    widths[2] = size.cx;
    GetTextExtentPoint32(hDC, win32_getstring(WIN32_progressbar_imperfectcolors),
		strlen(win32_getstring(WIN32_progressbar_imperfectcolors)), &size);
    widths[1] = size.cx;

    ReleaseDC(hProgWnd, hDC);

    widths[0] = -1;
    SendMessage(hStatusBar, SB_SETPARTS, (WPARAM) 1, (LPARAM) (LPINT) widths);
    StatusBar_GetItemRect(hStatusBar, 0, &rect);

    widths[0] = (rect.right - rect.left) - (widths[1] + widths[2] + widths[3]);
    widths[1] += widths[0];
    widths[2] += widths[1];
    widths[3] = -1;

    if (! MAME32App.m_bMMXDetected)
    {
        numParts = 3;
        widths[2] = -1;
    }

    SendMessage(hStatusBar, SB_SETPARTS, (WPARAM) numParts, (LPARAM) (LPINT) widths);
    UpdateStatusBar();

    bProgressShown = FALSE;
}

static void ResizeProgressBar(void)
{
    if (bProgressShown)
    {
        RECT rect;
        int  widths[2] = {150, -1};

        SendMessage(hStatusBar, SB_SETPARTS, (WPARAM) 2, (LPARAM) (LPINT) widths);
        StatusBar_GetItemRect(hStatusBar, 1, &rect);
        MoveWindow(hProgWnd, rect.left, rect.top,
            rect.right - rect.left, rect.bottom - rect.top, TRUE);
    }
    else
    {
        ProgressBarHide();
    }
}

static void ProgressBarStep(void)
{
    char tmp[80];
    sprintf(tmp, "%s %d%% %s",
		win32_getstring(WIN32_progressbar_gamesearch),
        ((game_index + 1) * 100) / game_count,
		win32_getstring(WIN32_progressbar_complete));
    SendMessage(hStatusBar, SB_SETTEXT, 0, (LPARAM)tmp);
    if (game_index == 0)
        ShowWindow(hProgWnd, SW_SHOW);
    SendMessage(hProgWnd, PBM_STEPIT, 0, 0);
}

static HWND InitProgressBar(HWND hParent)
{
    RECT rect;

    StatusBar_GetItemRect(hStatusBar, 0, &rect);

    rect.left += 150;

    return CreateWindowEx(WS_EX_STATICEDGE,
            PROGRESS_CLASS,
            "Progress Bar",
            WS_CHILD | PBS_SMOOTH,
            rect.left,
            rect.top,
            rect.right - rect.left,
            rect.bottom - rect.top,
            hParent,
            NULL,
            hInst,
            NULL);
}

static void CopyToolTipText (LPTOOLTIPTEXT lpttt)
{
    int   i;
    int   iButton = lpttt->hdr.idFrom;
    LPSTR pString;
    LPSTR pDest = lpttt->lpszText;

    // Map command ID to string index
    for (i = 0 ; CommandToString[i] != -1; i++)
    {
        if (CommandToString[i] == iButton)
        {
            iButton = i;
            break;
        }
    }

	for (i = 0 ; i <= NUM_TOOLTIPS ; i++)
		szTbStrings[i] = win32_getstring(i + WIN32_icon_togglefolderlist);

    // Check for valid parameter
    pString = (iButton > NUM_TOOLTIPS) ? (LPSTR)win32_getstring(WIN32_icon_invalidbutton) : szTbStrings[iButton];

    lstrcpy (pDest, pString);
}

static HWND InitToolbar(HWND hParent)
{
    return CreateToolbarEx(hParent,
        WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS |
        CCS_TOP | TBSTYLE_FLAT | TBSTYLE_TOOLTIPS,
        1,
        8,
        hInst,
        IDB_TOOLBAR,
        tbb,
        NUM_TOOLBUTTONS,
        18,
        18,
        0,
        0,
        sizeof(TBBUTTON));
}

static HWND InitStatusBar(HWND hParent)
{
    HMENU hMenu = GetMenu(hParent);

    popstr[0].hMenu = 0;
    popstr[0].uiString = 0;
    popstr[1].hMenu = hMenu;
    popstr[1].uiString = IDS_UI_FILE;
    popstr[2].hMenu = GetSubMenu(hMenu,1);
    popstr[2].uiString = IDS_UI_TOOLBAR;
    popstr[3].hMenu = 0;
    popstr[3].uiString = 0;

    return CreateStatusWindow(
        WS_CHILD | WS_VISIBLE | WS_CLIPSIBLINGS |
        CCS_BOTTOM | SBARS_SIZEGRIP,
		win32_getstring(WIN32_statusbar_ready),
        hParent,
        2);
}


static LRESULT Statusbar_MenuSelect (HWND hwnd, WPARAM wParam, LPARAM lParam)
{
    UINT fuFlags = (UINT) HIWORD (wParam);
    HMENU hMainMenu = NULL ;
    int iMenu = 0 ;

    // Handle non-system popup menu descriptions.
    if ((fuFlags & MF_POPUP) &&
        (!(fuFlags & MF_SYSMENU)))
    {
        for (iMenu = 1 ; iMenu < MAX_MENUS ; iMenu++)
        {
            if ((HMENU) lParam == popstr[iMenu].hMenu)
            {
                hMainMenu = (HMENU) lParam;
                break;
            }
        }
    }

    // Display helpful text in status bar
    MenuHelp (WM_MENUSELECT, wParam, lParam, hMainMenu, hInst,
        hStatusBar, (UINT *) &popstr[iMenu]);

    return 0 ;
}

static void UpdateStatusBar(void)
{
    LPTREEFOLDER    lpFolder = GetCurrentFolder();
    int             games_shown = 0;
    char            game_text[20];
    int             i = -1;

    if (! lpFolder)
        return;

    do
    {
        if ((i = FindGame(lpFolder, i + 1)) != -1)
        {
            if (!GameFiltered(i, lpFolder->m_dwFlags))
                games_shown++;
        }
    } while (i != -1);

    /* Show number of games in the current 'View' in the status bar */
    sprintf(game_text, "%d %s", games_shown, win32_getstring(WIN32_statusbar_games));
    SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 2, (LPARAM) game_text);

    i = GetSelectedPickItem();

    if (games_shown == 0)
        DisableSelection();
    else
        SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 1, (LPARAM) GameInfoStatus(i));
    if (MAME32App.m_bMMXDetected)
   {
       options_type *o = GetGameOptions(i);

        if (games_shown && o->disable_mmx == FALSE)
            SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 3, (LPARAM) "MMX");
        else
            SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 3, (LPARAM) "");
    }
}

static void UpdateHistory(int gameNum)
{
    int show = (have_history && have_selection) ? SW_SHOW : SW_HIDE;

    if (gameNum != -1)
    {
        char *histText = GameHistory(gameNum);

        have_history = (histText && histText[0]) ? TRUE : FALSE;
        Edit_SetText(GetDlgItem(hPicker, IDC_HISTORY), histText);
    }
    if (GetShowScreenShot())
    {
        show = (!nPictType && have_selection) ? show : SW_HIDE;
        ShowWindow(GetDlgItem(hPicker, IDC_HISTORY), show);
    }
}

static void DisableSelection(void)
{
    MENUITEMINFO    mmi;
    HMENU           hMenu = GetMenu(hMain);
    BOOL            last_selection = have_selection;;

    mmi.cbSize = sizeof(mmi);
    mmi.fMask = MIIM_TYPE;
    mmi.fType = MFT_STRING;
    mmi.dwTypeData = (LPTSTR)win32_getstring(WIN32_menubar_disableplay);
    mmi.cch = strlen(mmi.dwTypeData);
    SetMenuItemInfo(hMenu,ID_FILE_PLAY,FALSE,&mmi);

    EnableMenuItem(hMenu, ID_FILE_PLAY, MF_GRAYED);
    EnableMenuItem(hMenu, ID_FILE_PLAY_RECORD, MF_GRAYED);
    EnableMenuItem(hMenu, ID_FILE_PLAY_RECORD_AVI, MF_GRAYED);
    EnableMenuItem(hMenu, ID_GAME_PROPERTIES, MF_GRAYED);
    EnableMenuItem(hMenu, ID_FILE_NETPLAY, MF_GRAYED);

#ifdef BUILD_ENGLISH
    SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 0, (LPARAM) "No Selection");
#else
    SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 0, (LPARAM) "IĂ܂");
#endif
    SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 1, (LPARAM) "");
    SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 3, (LPARAM) "");

    have_selection = FALSE;

    if (last_selection != have_selection)
        UpdateScreenShot();
}

static void EnableSelection(int nGame)
{
    char            buf[200];
    MENUITEMINFO    mmi;
    HMENU           hMenu = GetMenu(hMain);
    options_type    *o = GetGameOptions(nGame);

#ifdef BUILD_ENGLISH
    sprintf(buf, win32_getstring(WIN32_menubar_enableplay),
        ConvertAmpersandString(ModifyThe(drivers[nGame]->description)));
#else
    if (UseJapaneseList())
	{
	    sprintf(buf, win32_getstring(WIN32_menubar_enableplay),
		    ConvertAmpersandString(drivers[nGame]->description));
	}
	else
	{
	    sprintf(buf, win32_getstring(WIN32_menubar_enableplay),
		    ConvertAmpersandString(ModifyThe(drivers[nGame]->descriptione)));
	}
#endif

    mmi.cbSize = sizeof(mmi);
    mmi.fMask = MIIM_TYPE;
    mmi.fType = MFT_STRING;
    mmi.dwTypeData = buf;
    mmi.cch = strlen(mmi.dwTypeData);
    SetMenuItemInfo(hMenu,ID_FILE_PLAY,FALSE,&mmi);

#ifdef BUILD_ENGLISH
    SendMessage(hStatusBar, SB_SETTEXT, 0, (LPARAM)ModifyThe(drivers[nGame]->description));
#else
    if (UseJapaneseList())
		SendMessage(hStatusBar, SB_SETTEXT, 0, (LPARAM)drivers[nGame]->description);
	else
		SendMessage(hStatusBar, SB_SETTEXT, 0, (LPARAM)ModifyThe(drivers[nGame]->descriptione));
#endif
    /* Add this game's status to the status bar */
    SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 1, (LPARAM) GameInfoStatus(nGame));
    if (MAME32App.m_bMMXDetected && o->disable_mmx == FALSE)
        SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 3, (LPARAM) "MMX");
    else
        SendMessage(hStatusBar, SB_SETTEXT, (WPARAM) 3, (LPARAM) "");

    /* If doing updating game status and the game name is NOT pacman.... */

    EnableMenuItem(hMenu, ID_FILE_PLAY, MF_ENABLED);
    EnableMenuItem(hMenu, ID_FILE_PLAY_RECORD, MF_ENABLED);
    EnableMenuItem(hMenu, ID_FILE_PLAY_RECORD_AVI, MF_ENABLED);
#ifdef KAILLERA
    EnableMenuItem(hMenu, ID_FILE_NETPLAY, MF_ENABLED);
#else
	EnableMenuItem(hMenu, ID_FILE_NETPLAY, MF_GRAYED);
#endif /* KAILLERA */

    if (!oldControl)
        EnableMenuItem(hMenu, ID_GAME_PROPERTIES, MF_ENABLED);

    if (GetIsFavorite(nGame))
    {
        EnableMenuItem(hMenu, ID_CONTEXT_ADD_FAVORITE, MF_GRAYED);
        EnableMenuItem(hMenu, ID_CONTEXT_DEL_FAVORITE, MF_ENABLED);
    }
    else
    {
        EnableMenuItem(hMenu, ID_CONTEXT_ADD_FAVORITE, MF_ENABLED);
        EnableMenuItem(hMenu, ID_CONTEXT_DEL_FAVORITE, MF_GRAYED);
    }

	if (GameOnlyJukeBox(nGame))
	{
        EnableMenuItem(hMenu, ID_FILE_PLAY_RECORD, MF_GRAYED);
        EnableMenuItem(hMenu, ID_FILE_PLAY_BACK, MF_GRAYED);
        EnableMenuItem(hMenu, ID_FILE_PLAY_RECORD_AVI, MF_GRAYED);
        EnableMenuItem(hMenu, ID_FILE_NETPLAY, MF_GRAYED);
	}

	if (bProgressShown && bListReady == TRUE)
    {
#ifdef BUILD_ENGLISH
        SetDefaultGame(ModifyThe(drivers[nGame]->description));
#else
    if (UseJapaneseList())
        SetDefaultGame(drivers[nGame]->description);
	else
        SetDefaultGame(ModifyThe(drivers[nGame]->descriptione));
#endif
    }
    have_selection = TRUE;

    UpdateScreenShot();
}

static void PaintControl(HWND hWnd, BOOL useBitmap)
{
    RECT        rcClient;
    HRGN        rgnBitmap;
    HPALETTE    hPAL;
    HDC         hDC = GetDC(hWnd);

    /* So we don't paint over the controls border */
    GetClientRect(hWnd, &rcClient);
    InflateRect(&rcClient, -2, -2);

    if (hBitmap && useBitmap)
    {
        int         i, j;
        HDC         htempDC;
        HBITMAP     oldBitmap;

        htempDC = CreateCompatibleDC(hDC);
        oldBitmap = SelectObject(htempDC, hBitmap);

        rgnBitmap = CreateRectRgnIndirect(&rcClient);
        SelectClipRgn(hDC, rgnBitmap);
        DeleteObject(rgnBitmap);

        hPAL = (! hPALbg) ? CreateHalftonePalette(hDC) : hPALbg;

        if(GetDeviceCaps(htempDC, RASTERCAPS) & RC_PALETTE && hPAL != NULL )
        {
            SelectPalette(htempDC, hPAL, FALSE );
            RealizePalette(htempDC);
        }

        for( i = rcClient.left; i < rcClient.right; i += bmDesc.bmWidth )
            for( j = rcClient.top; j < rcClient.bottom; j += bmDesc.bmHeight )
                BitBlt(hDC, i, j, bmDesc.bmWidth, bmDesc.bmHeight, htempDC, 0, 0, SRCCOPY );

        SelectObject(htempDC, oldBitmap);
        DeleteDC(htempDC);

        if (! bmDesc.bmColors)
        {
            DeleteObject(hPAL);
            hPAL = 0;
        }
    }
    else    /* Use standard control face color */
    {
        HBRUSH  hBrush = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
        HBRUSH  oldBrush;

        oldBrush = SelectObject(hDC, hBrush);
        FillRect(hDC, &rcClient, hBrush);
        SelectObject(hDC, oldBrush);
        DeleteObject(hBrush);
    }
    ReleaseDC(hWnd, hDC);
}

static LPCSTR GetCloneParent(int nItem)
{
    if (drivers[nItem]->clone_of != 0
    &&  drivers[nItem]->clone_of != DRIVER_ROOT
    &&  drivers[nItem]->clone_of != DRIVER_NEOGEO)
#ifdef BUILD_ENGLISH
        return ModifyThe(drivers[nItem]->clone_of->description);
#else
        if (UseJapaneseList())
	        return drivers[nItem]->clone_of->description;
		else
	        return ModifyThe(drivers[nItem]->clone_of->descriptione);
#endif
    return "";
}

static BOOL PickerHitTest(HWND hWnd)
{
    RECT            rect;
    POINTS          p;
    DWORD           res = GetMessagePos();
    LVHITTESTINFO   htInfo;

    memset(&htInfo,'\0', sizeof(LVHITTESTINFO));
    p = MAKEPOINTS(res);
    GetWindowRect(hWnd, &rect);
    htInfo.pt.x = p.x - rect.left;
    htInfo.pt.y = p.y - rect.top;
    ListView_HitTest(hWnd, &htInfo);

    return (! (htInfo.flags & LVHT_NOWHERE));
}

static BOOL MamePickerNotify(NMHDR *nm)
{
    NM_LISTVIEW *pnmv;
    static int nLastState = -1;
    static int nLastItem  = -1;

    switch (nm->code)
    {
    case NM_RCLICK:
    case NM_CLICK:
        /* don't allow selection of blank spaces in the listview */
        if (!PickerHitTest(hwndList))
        {
            /* we have no current game selected */
            if (nLastItem != -1);
                SetSelectedPickItem(nLastItem);
            return TRUE;
        }
        break;

    case NM_DBLCLK:
        /* Check here to make sure an item was selected */
        if (!PickerHitTest(hwndList))
        {
            if (nLastItem != -1);
                SetSelectedPickItem(nLastItem);
            return TRUE;
        }
        else
            MamePlayGame();
        return TRUE;

    case LVN_GETDISPINFO:
        {
            LV_DISPINFO* pnmv = (LV_DISPINFO*)nm;
            int nItem = pnmv->item.lParam;

            if (pnmv->item.mask & LVIF_IMAGE)
            {
               pnmv->item.iImage = WhichIcon(nItem);
            }

            if (pnmv->item.mask & LVIF_STATE)
                pnmv->item.state = 0;

            if (pnmv->item.mask & LVIF_TEXT)
            {
                switch (realColumn[pnmv->item.iSubItem])
                {
                case COLUMN_GAMES:
                    // Driver description
#ifdef BUILD_ENGLISH
                    pnmv->item.pszText = (char *)ModifyThe(drivers[nItem]->description);
#else
                    if (UseJapaneseList())
	                    pnmv->item.pszText = (char *)drivers[nItem]->description;
					else
	                    pnmv->item.pszText = (char *)ModifyThe(drivers[nItem]->descriptione);
#endif
                    break;

                case COLUMN_DIRECTORY:
                    // Driver name (directory)
                    pnmv->item.pszText = (char*)drivers[nItem]->name;
                    break;

                case COLUMN_TRACKBALL:
                    // Trackball
                    if (GameUsesTrackball(nItem))
                        pnmv->item.pszText = "";
                    else
                        pnmv->item.pszText = "~";
                    break;

                case COLUMN_PLAYED:
                    // times played
                    {
                        static char buf[100];
                        sprintf(buf,"%i",GetPlayCount(nItem));
                        pnmv->item.pszText = buf;
                    }
                    break;

                case COLUMN_MANUFACTURER:
                    // Manufacturer
                    pnmv->item.pszText = (char *)drivers[nItem]->manufacturer;
                    break;

                case COLUMN_YEAR:
                    // Year
                    pnmv->item.pszText = (char *)drivers[nItem]->year;
                    break;

                case COLUMN_GENRE:
                    // Genre
#ifdef BUILD_ENGLISH
				    pnmv->item.pszText = (char *)win32_getstring(WIN32_genre_action + drivers[nItem]->genre - 1);
#else
				    if (UseJapaneseList())
					    pnmv->item.pszText = (char *)win32_getstring(WIN32_genre_action + drivers[nItem]->genre - 1);
					else
                        pnmv->item.pszText = (char *)win32_getstring(WIN32_genreeng_action + drivers[nItem]->genre - 1);
#endif
                    break;

                case COLUMN_CLONE:
                    pnmv->item.pszText = (char *)GetCloneParent(nItem);
                    break;

                }
            }
        }
        return TRUE;

    case LVN_ITEMCHANGED :
        pnmv = (NM_LISTVIEW *)nm;

        if ((pnmv->uOldState & LVIS_SELECTED)
            && !(pnmv->uNewState & LVIS_SELECTED))
        {
            if (pnmv->lParam != -1)
                nLastItem = pnmv->lParam;
            /* leaving item */
            /* printf("leaving %s\n",drivers[pnmv->lParam]->name); */
        }

        if (!(pnmv->uOldState & LVIS_SELECTED)
            && (pnmv->uNewState & LVIS_SELECTED))
        {
            int  nState;

            /* We load the screen shot DIB here, instead
             * of loading everytime we want to display it.
             */
            nState = LoadScreenShot(pnmv->lParam, nPictType);
            bScreenShotAvailable = nState;
            if (nState == TRUE || nState != nLastState)
            {
                HWND hWnd;

                if (GetShowScreenShot() &&
                    (hWnd = GetDlgItem(hPicker, IDC_SSFRAME)))
                {
                    RECT    rect;
                    HWND    hParent;
                    POINT   p = {0, 0};

                    hParent = GetParent(hWnd);

                    GetWindowRect(hWnd,&rect);
                    ClientToScreen(hParent, &p);
                    OffsetRect(&rect, -p.x, -p.y);
                    InvalidateRect(hParent, &rect, FALSE);
                    UpdateWindow(hParent);
                    nLastState = nState;
                }
            }

            /* printf("entering %s\n",drivers[pnmv->lParam]->name); */
            if (bDoBroadcast == TRUE)
            {
                ATOM a = GlobalAddAtom(drivers[pnmv->lParam]->description);
                SendMessage(HWND_BROADCAST, mame32_message, a, a);
                GlobalDeleteAtom(a);
            }

            EnableSelection(pnmv->lParam);
            UpdateHistory(pnmv->lParam);
        }
        return TRUE;

    case LVN_COLUMNCLICK :
        pnmv = (NM_LISTVIEW *)nm;
        ColumnSort(pnmv->iSubItem, FALSE);
        return TRUE;
    }
    return FALSE;
}


static BOOL TreeViewNotify(LPNMHDR nm)
{
    LPNMTREEVIEW lpNmtv;

    switch (nm->code)
    {
    case TVN_SELCHANGED :
        lpNmtv = (LPNMTREEVIEW)nm;
        {
            HTREEITEM hti = TreeView_GetSelection(hTreeView);
            TVITEM  tvi;

            tvi.mask = TVIF_PARAM | TVIF_HANDLE;
            tvi.hItem = hti;

            if (TreeView_GetItem( hTreeView, &tvi ))
            {
                SetCurrentFolder((LPTREEFOLDER)tvi.lParam);
                if (bListReady)
                    ResetListView();
            }
            return TRUE;
        }
        break;
    }

    return FALSE;
}

static BOOL HeaderOnContextMenu(HWND hWnd, WPARAM wParam, LPARAM lParam)
{
    HMENU hMenuLoad;
    HMENU hMenu;
    // Right button was clicked on header
    POINT   pt;
    RECT    rcCol;
    int     index;
    int     i;
    BOOL    found = FALSE;
    HWND    hwndHeader;

    if ((HWND)wParam != GetDlgItem(hWnd, IDC_LIST))
        return FALSE;

    hwndHeader = GetDlgItem(hwndList, 0);

    pt.x = LOWORD(lParam);
    pt.y = HIWORD(lParam);

    ScreenToClient(hwndHeader, &pt);

    // Determine the column index
    for( i = 0; Header_GetItemRect(hwndHeader, i, &rcCol); i++ )
    {
        if( PtInRect(&rcCol, pt ) )
        {
            index = i;
            found = TRUE;
            break;
        }
    }

    // Do the context menu
    if (found)
    {
        hMenuLoad = LoadMenu(hInst, MAKEINTRESOURCE(IDR_CONTEXT_HEADER));
        hMenu = GetSubMenu(hMenuLoad, 0);
        lastColumnClick = i;
        TrackPopupMenu(hMenu,
            TPM_LEFTALIGN | TPM_RIGHTBUTTON,
            LOWORD(lParam),
            HIWORD(lParam),
            0,
            hWnd,
            NULL);

        DestroyMenu(hMenuLoad);

        return TRUE;
    }
    return FALSE;
}


static char * ConvertAmpersandString(const char *s)
{
    /* takes a string and changes any ampersands to double ampersands,
       for setting text of window controls that don't allow us to disable
       the ampersand underlining.
     */
    /* returns a static buffer--use before calling again */

    static char buf[200];
    char *ptr;

    ptr = buf;
    while (*s)
    {
        if (*s == '&')
            *ptr++ = *s;
            *ptr++ = *s++;
    }
    *ptr = 0;

    return buf;
}

static void PollGUIJoystick(void)
{
    if (in_emulation)
        return;

    joygui->poll_joysticks();

    if (joygui->is_joy_pressed(JOYCODE(1,JOYCODE_STICK_AXIS,2,JOYCODE_DIR_NEG)))
    {
        SetFocus(hwndList);
        PressKey(hwndList,VK_UP);

    }
    if (joygui->is_joy_pressed(JOYCODE(1,JOYCODE_STICK_AXIS,2,JOYCODE_DIR_POS)))
    {
        SetFocus(hwndList);
        PressKey(hwndList,VK_DOWN);
    }
    if (joygui->is_joy_pressed(JOYCODE(1,JOYCODE_STICK_AXIS,1,JOYCODE_DIR_NEG)))
    {
        SetFocus(hwndList);
        PressKey(hwndList,VK_LEFT);

    }
    if (joygui->is_joy_pressed(JOYCODE(1,JOYCODE_STICK_AXIS,1,JOYCODE_DIR_POS)))
    {
        SetFocus(hwndList);
        PressKey(hwndList,VK_RIGHT);
    }
    if (joygui->is_joy_pressed(JOYCODE(1,JOYCODE_STICK_BTN,1,JOYCODE_DIR_BTN)))
    {
        SetFocus(hwndList);
        MamePlayGame();
    }
    if (joygui->is_joy_pressed(JOYCODE(1,JOYCODE_STICK_BTN,2,JOYCODE_DIR_BTN)))
    {
        SetFocus(hwndList);
        PressKey(hwndList,VK_NEXT);
    }
    if (joygui->is_joy_pressed(JOYCODE(1,JOYCODE_STICK_BTN,5,JOYCODE_DIR_BTN)))
    {
        SetFocus(hwndList);
        PressKey(hwndList,VK_PRIOR);
    }
    if (joygui->is_joy_pressed(JOYCODE(1,JOYCODE_STICK_BTN,3,JOYCODE_DIR_BTN)))
    {
        SetFocus(hwndList);
        PressKey(hwndList,VK_END);
    }
    if (joygui->is_joy_pressed(JOYCODE(1,JOYCODE_STICK_BTN,6,JOYCODE_DIR_BTN)))
    {
        SetFocus(hwndList);
        PressKey(hwndList,VK_HOME);
    }
    if (joygui->is_joy_pressed(JOYCODE(1,JOYCODE_STICK_BTN,1,JOYCODE_DIR_BTN)))
    {
        SetFocus(hwndList);
        MamePlayGame();
    }
}

static void PressKey(HWND hwnd,UINT vk)
{
    SendMessage(hwnd,WM_KEYDOWN,vk,0);
    SendMessage(hwnd,WM_KEYUP,vk,0xc0000000);
}

static void SetView(int menu_id,int listview_style)
{
    /* first uncheck previous menu item */
    CheckMenuRadioItem(GetMenu(hMain), ID_VIEW_ICON, ID_VIEW_DETAIL,
        menu_id, MF_CHECKED);
    ToolBar_CheckButton(hToolBar, menu_id, MF_CHECKED);
    SetWindowFont(hwndList, hFont, FALSE);
    SetWindowLong(hwndList,GWL_STYLE,
        (GetWindowLong(hwndList,GWL_STYLE) & ~LVS_TYPEMASK) | listview_style);

    current_view_id = menu_id;
    SetViewMode(menu_id - ID_VIEW_ICON);

    switch(menu_id)
    {
    case ID_VIEW_ICON:
        ResetListView();
        break;
    case ID_VIEW_SMALL_ICON:
        ListView_Arrange(hwndList, LVA_ALIGNTOP);
        break;
    }
}

static void ResetListView(void)
{
    RECT    rect;
    int     i;
    int     current_game;
    LV_ITEM lvi;
    int     column;
    BOOL    no_selection = FALSE;
    LPTREEFOLDER lpFolder = GetCurrentFolder();

    if (! lpFolder)
        return;

    SetWindowRedraw(hwndList,FALSE);

    /* If the last folder was empty, no_selection is TRUE */
    if (have_selection == FALSE)
        no_selection = TRUE;

    current_game = GetSelectedPickItem();

    ListView_DeleteAllItems(hwndList);

    ListView_SetItemCount(hwndList, game_count);

    lvi.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM;
    lvi.stateMask = 0;

    i = -1;

    do
    {
        /* Add the games that are in this folder */
        if ((i = FindGame(lpFolder, i + 1)) != -1)
        {
            if (GameFiltered(i, lpFolder->m_dwFlags))
                continue;

            lvi.iItem    = i;
            lvi.iSubItem = 0;
            lvi.lParam   = i;
            // Do not set listview to LVS_SORTASCENDING or LVS_SORTDESCENDING
            lvi.pszText  = LPSTR_TEXTCALLBACK;
            lvi.iImage   = I_IMAGECALLBACK;
            ListView_InsertItem(hwndList,&lvi);
        }
    } while (i != -1);

    reverse_sort = FALSE;

    if ((column = GetSortColumn()) > 0)
    {
        last_sort = -1;
    }
    else
    {
        column = abs(column);
        last_sort = column - 1;
    }

    ColumnSort(column - 1, TRUE);

    /* If last folder was empty, select the first item in this folder */
    if (no_selection)
    {
        SetSelectedPick(0);
    }
    else
        SetSelectedPickItem(current_game);

    GetClientRect(hwndList, &rect);
    InvalidateRect(hwndList, &rect, TRUE);
    // if (current_view_id == ID_VIEW_SMALL_ICON)
        ListView_Arrange(hwndList, LVA_DEFAULT);// LVA_ALIGNTOP);

    UpdateStatusBar();

    SetWindowRedraw(hwndList,TRUE);
}


static void UpdateGameList(void)
{
    int i;

    for (i = 0; i < game_count; i++)
        SetHasRoms(i, UNKNOWN);

    idle_work  = TRUE;
    bDoGameCheck = TRUE;
    game_index = 0;

    /* Let REFRESH also load new background if found */
    ReloadIcons(hwndList);
    LoadListBitmap();
    ResetListView();

#ifdef BUILD_ENGLISH
    SetDefaultGame(ModifyThe(drivers[GetSelectedPickItem()]->description));
#else
    if (UseJapaneseList())
	    SetDefaultGame(drivers[GetSelectedPickItem()]->description);
	else
	    SetDefaultGame(ModifyThe(drivers[GetSelectedPickItem()]->descriptione));
#endif
    //SetSelectedPick(0); /* To avoid flickering. */
}

static void PickFont(void)
{
    LOGFONT font;
    CHOOSEFONT cf;

    GetListFont(&font);
    cf.lStructSize = sizeof(CHOOSEFONT);
    cf.hwndOwner = hMain;
    cf.lpLogFont = &font;
    cf.rgbColors = GetListFontColor();
    cf.Flags = CF_SCREENFONTS | CF_INITTOLOGFONTSTRUCT | CF_EFFECTS;
    if (!ChooseFont(&cf))
      return;

    SetListFont(&font);
    if (hFont != NULL)
        DeleteObject(hFont);
    hFont = CreateFontIndirect(&font);
    if (hFont != NULL)
    {
        COLORREF textColor = cf.rgbColors;

        if (textColor == RGB(255,255,255))
            textColor = RGB(240, 240, 240);
        SetWindowFont(hwndList, hFont, TRUE);
        SetWindowFont(hTreeView, hFont, TRUE);
        ListView_SetTextColor(hwndList, textColor);
        TreeView_SetTextColor(hTreeView,textColor);
        SetListFontColor(cf.rgbColors);
        SetWindowFont(GetDlgItem(hPicker, IDC_HISTORY), hFont, FALSE);
        ResetListView();
    }
}

/* This centers a window on another window */
static BOOL CenterSubDialog(HWND hParent, HWND hWndCenter, BOOL in_client_coords)
{
    RECT rect, wRect;
    int x,y;

    if (in_client_coords)
       GetClientRect(hParent, &rect);
    else
       GetWindowRect(hParent, &rect);

    GetClientRect(hWndCenter, &wRect);

    x = (rect.left + rect.right)/2 - (wRect.left + wRect.right)/2;
    y = (rect.bottom + rect.top)/2 - (wRect.bottom + wRect.top)/2;

    // map parent coordinates to child coordinates
    return SetWindowPos(hWndCenter, NULL, x, y, -1, -1,
        SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
}

#ifdef KAILLERA
int WINAPI kGameCallback(char *game, int player, int maxplayers)
{
    LV_FINDINFO lvfi;
    int         i;

	if (kPlay)	return 0;

    SetCurrentFolder(GetFolderByID(FOLDER_AVAILABLE));
    ResetListView();

    lvfi.flags = LVFI_STRING;
    lvfi.psz   = game;
    i = ListView_FindItem(hwndList, -1, &lvfi);

    if (i == -1)
    {
        kailleraEndGame();
        return 0;
    }

    SetSelectedPick(i);

	srand(0xbadc0de);

  // sound HAS to be enabled otherwise some games (a lot actually) in MAME will desynch
	{
		options_type *o = GetDefaultOptions();
		if (o->sound == SOUND_NONE)
		{
			o->sound = SOUND_MIDAS;
			o->sample_rate = 11050;
			o->stereo = 0;
		}

	    o = GetGameOptions(GetSelectedPick());
	    if (o->sound == SOUND_NONE)
	    {
			o->sound = SOUND_MIDAS;
			o->sample_rate = 11050;
			o->stereo = 0;
	    }
	}

	kPlay = 1;

	MamePlayGame();

	kailleraEndGame();
	kPlay = 0;

	return 0;
}

int WINAPI kChatCallback(char *nick, char *text)
{
	char tmp[512];

    wsprintf(tmp, "<%s> %s", nick, text);
    KailleraChateReceive(tmp);

	return 0;
}

void WINAPI kDropCallback(char *nick, int playernb)
{
	char tmp[512];

	wsprintf(tmp,"* Player %i (%s) dropped from the current game.",playernb,nick);
    KailleraChateReceive(tmp);
}

void WINAPI kInfosCallback(char *gamename)
{
	int i, f = -1;

	for(i = 0 ; i < game_count && f == -1 ; i++)
		if(!strcmp(drivers[i]->description,gamename)) f = i;

	if(f != -1)
	{
		char tmpurl[512];

		wsprintf(tmpurl, "http://www.mame.dk/gameinfo/%s/?mamever=37b15&style=kil", drivers[f]->name);
		ShellExecute(NULL, "open", tmpurl, NULL, ".", 0);
	}
}

static kailleraInfos kInfos = {
	"",
	"",
	kGameCallback,
	kChatCallback,
	kDropCallback,
	kInfosCallback
};
#endif /* KAILLERA */

static BOOL MameCommand(HWND hwnd,int id, HWND hwndCtl, UINT codeNotify)
{
    switch (id)
    {
    case ID_FILE_PLAY :
        MamePlayGame();
        return TRUE;
    case ID_FILE_PLAY_RECORD :
        MamePlayRecordGame();
        return TRUE;
    case ID_FILE_PLAY_BACK :
        MamePlayBackGame();
        return TRUE;
    case ID_FILE_PLAY_RECORD_AVI :
        MamePlayRecordAVI();
        return TRUE;

#ifdef KAILLERA
    case ID_FILE_NETPLAY:
        {
			char mameVer[512];
			char *tmpGames = NULL;
			int gamesSize = 8192;
			LVITEM i;
			int j, s;
			BOOL tmpjpn;

			tmpGames = (char *)MyMalloc(gamesSize);
			tmpGames[0] = 0;
			tmpGames[1] = 0;
			s = 0;

			SetCurrentFolder(GetFolderByID(FOLDER_AVAILABLE));

			tmpjpn = bUseJapanese;
			bUseJapanese = FALSE;
#ifndef BUILD_ENGLISH
	        SetUseJapaneseList(bUseJapanese);
#endif
	        ToolBar_CheckButton(hToolBar, ID_USE_JLIST, (bUseJapanese) ? MF_CHECKED : MF_UNCHECKED);

			last_sort = -1;
            ColumnSort(0, FALSE);
            ResetListView();

            memset(&i, 0, sizeof(LVITEM));

			for(j = 0 ; j < ListView_GetItemCount(hwndList) ; j++)
			{
				char *p;
				char tmptxt[512];
				int l;

				i.mask = LVIF_TEXT | LVIF_PARAM;
				i.iItem = j;
				i.iSubItem = 0;
				i.pszText = tmptxt;
				i.cchTextMax = 511;
				ListView_GetItem(hwndList, &i);

				while((s + strlen(tmptxt) + 4) > gamesSize)
				{
					gamesSize *= 2;
					tmpGames = (char *)MyRealloc(tmpGames, gamesSize);
				}

				p = tmpGames + s;
				strcpy(p,tmptxt);
				l = strlen(p) + 1;
				p[l] = 0;
				s += l;
			}

            {
                char *p;
                char tmptxt[512];
                int l;

                strcpy(tmptxt, "*Chat (not game)");
                while ((s + strlen(tmptxt) + 4) > gamesSize)
                {
                    gamesSize *= 2;
                    tmpGames = (char *)realloc(tmpGames, gamesSize);
                }
                p = tmpGames + s;
                strcpy(p, tmptxt);
                l = strlen(p) + 1;
                p[l] = 0;
                s += l;
            }

			if (version_name[0])
				strcpy(mameVer, version_name);
			else
				sprintf(mameVer,"%s %s", MAME32NAME, build_version);

			kInfos.appName = mameVer;
			kInfos.gameList = tmpGames;

			kailleraSetInfos(&kInfos);

			MyFree(tmpGames);

			kailleraSelectServerDialog(hMain);

			bUseJapanese = tmpjpn;
#ifndef BUILD_ENGLISH
	        SetUseJapaneseList(bUseJapanese);
#endif
	        ToolBar_CheckButton(hToolBar, ID_USE_JLIST, (bUseJapanese) ? MF_CHECKED : MF_UNCHECKED);
            ResetListView();
		}

		return TRUE;
#endif /* KAILLERA */

    case ID_FILE_AUDIT :
        AuditDialog(hMain);
        InitGames(game_count);
        ResetListView();
        SetFocus(hwndList);
        return TRUE;
    case ID_FILE_EXIT :
        PostMessage(hMain,WM_CLOSE,0,0);
        return TRUE;
    case ID_VIEW_ICON :
        SetView(ID_VIEW_ICON,LVS_ICON);
        return TRUE;
    case ID_VIEW_DETAIL :
        SetView(ID_VIEW_DETAIL,LVS_REPORT);
        return TRUE;
    case ID_VIEW_SMALL_ICON :
        SetView(ID_VIEW_SMALL_ICON,LVS_SMALLICON);
        return TRUE;
    case ID_VIEW_LIST_MENU :
        SetView(ID_VIEW_LIST_MENU,LVS_LIST);
        return TRUE;
    // Arrange Icons submenu
    case ID_VIEW_BYGAME:
        ColumnSort(COLUMN_GAMES, TRUE);
        break;
    case ID_VIEW_BYDIRECTORY:
        ColumnSort(COLUMN_DIRECTORY, TRUE);
        break;
    case ID_VIEW_BYMANUFACTURER:
        ColumnSort(COLUMN_MANUFACTURER, TRUE);
        break;
    case ID_VIEW_BYTIMESPLAYED:
        ColumnSort(COLUMN_PLAYED, TRUE);
        break;
    case ID_VIEW_BYYEAR:
        ColumnSort(COLUMN_YEAR, TRUE);
        break;
    case ID_VIEW_FOLDERS:
        bShowTree = !bShowTree;
        SetShowFolderList(bShowTree);
        CheckMenuItem(GetMenu(hMain), ID_VIEW_FOLDERS, (bShowTree) ? MF_CHECKED : MF_UNCHECKED);
        ToolBar_CheckButton(hToolBar, ID_VIEW_FOLDERS, (bShowTree) ? MF_CHECKED : MF_UNCHECKED);
        UpdateScreenShot();
        break;
    case ID_VIEW_TOOLBARS:
        bShowToolBar = !bShowToolBar;
        SetShowToolBar(bShowToolBar);
        CheckMenuItem(GetMenu(hMain), ID_VIEW_TOOLBARS, (bShowToolBar) ? MF_CHECKED : MF_UNCHECKED);
        ToolBar_CheckButton(hToolBar, ID_VIEW_TOOLBARS, (bShowToolBar) ? MF_CHECKED : MF_UNCHECKED);
        ShowWindow(hToolBar, (bShowToolBar) ? SW_SHOW : SW_HIDE);
        ResizePickerControls(hPicker);
        UpdateScreenShot();
        break;

    case ID_VIEW_STATUS:
        bShowStatusBar = !bShowStatusBar;
        SetShowStatusBar(bShowStatusBar);
        CheckMenuItem(GetMenu(hMain), ID_VIEW_STATUS, (bShowStatusBar) ? MF_CHECKED : MF_UNCHECKED);
        ToolBar_CheckButton(hToolBar, ID_VIEW_STATUS, (bShowStatusBar) ? MF_CHECKED : MF_UNCHECKED);
        ShowWindow(hStatusBar, (bShowStatusBar) ? SW_SHOW : SW_HIDE);
        ResizePickerControls(hPicker);
        UpdateScreenShot();
        break;

#ifndef BUILD_ENGLISH
    case ID_USE_JLIST:
        bUseJapanese = !bUseJapanese;
        SetUseJapaneseList(bUseJapanese);
        ToolBar_CheckButton(hToolBar, ID_USE_JLIST, (bUseJapanese) ? MF_CHECKED : MF_UNCHECKED);
        ResetListView();
        break;
#endif

    case ID_MODIFY_THE:
        bUseModifyThe = !bUseModifyThe;
        SetUseModifyThe(bUseModifyThe);
        if (!bUseJapanese)
        {
            ToolBar_CheckButton(hToolBar, ID_MODIFY_THE, (bUseModifyThe) ? MF_CHECKED : MF_UNCHECKED);
            ResetListView();
        }
        break;

#ifdef KAILLERA
#ifndef BUILD_ENGLISH
    case ID_USE_IME_IN_CHAT:
        bUseImeInChat = !bUseImeInChat;
        SetUseImeInChat(bUseImeInChat);
        ToolBar_CheckButton(hToolBar, ID_USE_IME_IN_CHAT, (bUseImeInChat) ? MF_CHECKED : MF_UNCHECKED);
        break;
#endif
    case ID_USE_FIXED_FORM:
        bUseFixedForm = !bUseFixedForm;
        SetUseFixedForm(bUseFixedForm);
        ToolBar_CheckButton(hToolBar, ID_USE_FIXED_FORM, (bUseFixedForm) ? MF_CHECKED : MF_UNCHECKED);
        break;
#endif
    case ID_GAME_AUDIT :
        InitGameAudit(0);
        if (!oldControl)
            InitPropertyPageToPage(hInst, hwnd, GetSelectedPickItem(),AUDIT_PAGE);
        /* Just in case the toggle MMX on/off */
        UpdateStatusBar();
       break;

    // ListView Context Menu
    case ID_CONTEXT_ADD_FAVORITE:
    case ID_CONTEXT_DEL_FAVORITE:
        SetIsFavorite(GetSelectedPickItem(), (id == ID_CONTEXT_ADD_FAVORITE) ? TRUE : FALSE);
        InitGames(game_count);
        if (GetFolderByID(FOLDER_FAVORITE) == GetCurrentFolder())
            ResetListView();
        break;

    // Tree Context Menu
    case ID_CONTEXT_FILTERS:
        if (DialogBox(GetModuleHandle(NULL),
            MAKEINTRESOURCE(IDD_FILTERS), hMain, FilterDialogProc) == TRUE)
            ResetListView();
        SetFocus(hwndList);
        return TRUE;

    // Header Context Menu
    case ID_SORT_ASCENDING:
        last_sort = -1;
        ColumnSort(lastColumnClick, FALSE);
        break;
    case ID_SORT_DESCENDING:
        last_sort = lastColumnClick;
        ColumnSort(lastColumnClick, FALSE);
        break;
    case ID_CUSTOMIZE_FIELDS:
        if (DialogBox(GetModuleHandle(NULL),
            MAKEINTRESOURCE(IDD_COLUMNS), hMain, ColumnDialogProc) == TRUE)
            ResetColumnDisplay(FALSE);
        SetFocus(hwndList);
        return TRUE;

    // View Menu
    case ID_VIEW_LINEUPICONS:
        ResetListView();
        break;
    case ID_GAME_PROPERTIES:
        if (!oldControl)
        {
            InitPropertyPage(hInst, hwnd, GetSelectedPickItem());
            SaveGameOptions(GetSelectedPickItem());
        }
        /* Just in case the toggle MMX on/off */
        UpdateStatusBar();
        break;
    case ID_VIEW_SCREEN_SHOT:
        ToggleScreenShot();
        if (current_view_id == ID_VIEW_ICON)
            ResetListView();
        break;
    case ID_UPDATE_GAMELIST:
        UpdateGameList();
        break;

    case ID_OPTIONS_FONT:
        PickFont();
        SetFocus(hwndList);
        return TRUE;

    case ID_OPTIONS_DEFAULTS :
        /* Check the return value to see if changes were applied */
        if (!oldControl)
        {
            InitDefaultPropertyPage(hInst, hwnd);
            SaveDefaultOptions();
        }
        SetFocus(hwndList);
        return TRUE;

    case ID_OPTIONS_DIR:
        {
            int  nResult;
            BOOL bUpdateRoms;

            nResult = DialogBox(GetModuleHandle(NULL),
                                MAKEINTRESOURCE(IDD_DIRECTORIES),
                                hMain,
                                DirectoriesDialogProc);
            bUpdateRoms    = ((nResult & DIRDLG_ROMS)    == DIRDLG_ROMS)    ? TRUE : FALSE;

            /* update file code and game list */
            if (bUpdateRoms == TRUE)
            {
                File_UpdateRomPath(GetRomDirs());
                UpdateGameList();
            }
            SetFocus(hwndList);
        }
        return TRUE;
    case ID_OPTIONS_RESET_DEFAULTS:
        if (DialogBox(GetModuleHandle(NULL),
            MAKEINTRESOURCE(IDD_RESET), hMain, ResetDialogProc) == TRUE)
            PostQuitMessage(0);
        return TRUE;
    case ID_OPTIONS_STARTUP:
        DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_START_OPTIONS),
            hMain, StartupDialogProc);
        return TRUE;

    case ID_HELP_RELEASE:
        DisplayTextFile(hMain, "Readme.txt");
        break;
    case ID_HELP_WHATS_NEW:
        DisplayTextFile(hMain, "Whatsnew.txt");
        break;
    case ID_HELP_JUKEBOX:
        DisplayTextFile(hMain, "Jukebox.txt");
        break;
    case ID_HELP_OBJECT:
        DisplayTextFile(hMain, "Object.txt");
        break;
    case ID_HELP_KAILLERA:
        DisplayTextFile(hMain, "Kaillera.txt");
        break;

    case ID_HELP_ABOUT :
        DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_ABOUT),
            hMain, AboutDialogProc);
        SetFocus(hwndList);
        return TRUE;

    case IDC_PLAY_GAME:
        if (have_selection)
            MamePlayGame();
        break;

    case IDC_SSPICTURE:
    case IDC_SSDEFPIC:
        nPictType = (nPictType + 1) % MAX_PICT_TYPES;
        SetShowPictType(nPictType);
        bScreenShotAvailable = LoadScreenShot(GetSelectedPickItem(), nPictType);
        UpdateScreenShot();
        SendMessage(hPicker, WM_PAINT, 0, 0);
        break;

    case ID_CONTEXT_SELECT_RANDOM:
        SetRandomPickItem();
        break;

    // Kaillera Version Setting
    case ID_KAILLERA_VERSION:
        DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_KAILLERA_VERSION),
			hMain, VersionDialogProc);
        return TRUE;
    }

    SetFocus(hwndList);
    return FALSE;
}


static void LoadListBitmap(void)
{
    HGLOBAL     hDIBbg;
    char*       pFileName = 0;
    UINT        nResource = 0;

    if (hBitmap)
    {
        DeleteObject(hBitmap);
        hBitmap = 0;
    }

    if (hPALbg)
    {
        DeleteObject(hPALbg);
        hPALbg = 0;
    }

    /* Pick images based on number of colors avaliable. */
    if (GetDepth(hwndList) <= 8)
    {
        pFileName = "bkgnd16";
        //nResource = IDB_BKGROUND16;
    }
    else
    {
        pFileName = "bkground";
        //nResource = IDB_BKGROUND;
    }

    if (LoadDIB(pFileName, &hDIBbg, &hPALbg, FALSE))
    {
        HDC hDC = GetDC(hwndList);
        hBitmap = DIBToDDB(hDC, hDIBbg, &bmDesc);
        GlobalFree(hDIBbg);
        ReleaseDC(hwndList, hDC);
    }
#ifdef INTERNAL_BKBITMAP
    else
        hBitmap = LoadBitmapAndPalette( hInst, MAKEINTRESOURCE(nResource), &hPALbg, &bmDesc);
#endif
}

/* Initialize the Picker and List controls */
static void InitPicker(void)
{
    int order[COLUMN_MAX];
    int shown[COLUMN_MAX];
    int i;

    Header_Initialize(hwndList);

    // Disabled column customize with old Control
    if (oldControl)
    {
        for (i = 0; i < COLUMN_MAX ; i++)
        {
            order[i] = i;
            shown[i] = TRUE;
        }
        SetColumnOrder(order);
        SetColumnShown(shown);
    }

    /* Create IconsList for ListView Control */
    CreateIcons(hwndList);
    GetColumnOrder(realColumn);

    ResetColumnDisplay(TRUE);

    /* Allow selection to change the default saved game */
    bListReady = TRUE;
}

/* Re/initialize the ListControl Columns */
static void ResetColumnDisplay(BOOL firstime)
{
    LV_FINDINFO lvfi;
    LV_COLUMN   lvc;
    int         i;
    int         nColumn;
    int         widths[COLUMN_MAX];
    int         order[COLUMN_MAX];
    int         shown[COLUMN_MAX];
    int         nGame = GetSelectedPickItem();

    GetColumnWidths(widths);
    GetColumnOrder(order);
    GetColumnShown(shown);

    if (!firstime)
    {
        nColumn = GetNumColumns(hwndList);

        // The first time thru this won't happen, on purpose
        for (i = 0; i < nColumn; i++)
        {
            widths[realColumn[i]] = ListView_GetColumnWidth(hwndList, i);
        }

        SetColumnWidths(widths);

        for (i = nColumn; i > 0; i--)
        {
            ListView_DeleteColumn(hwndList, i - 1);
        }
    }

    ListView_SetExtendedListViewStyle(hwndList,
        LVS_EX_FULLROWSELECT | LVS_EX_HEADERDRAGDROP);

    lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;

    lvc.fmt = LVCFMT_LEFT;

    nColumn = 0;
    for (i = 0; i < COLUMN_MAX; i++)
    {
        if (shown[order[i]])
        {
            lvc.mask = LVCF_FMT | LVCF_WIDTH | LVCF_SUBITEM;

            lvc.mask |= LVCF_TEXT;
            lvc.pszText = column_names[order[i]];
            lvc.iSubItem = nColumn;
            lvc.cx = widths[order[i]];
            ListView_InsertColumn(hwndList, nColumn, &lvc);
            realColumn[nColumn] = order[i];
            nColumn++;
        }
    }

    // Fill this in so we can still sort on columns NOT shown
    for (i = 0; i < COLUMN_MAX && nColumn < COLUMN_MAX; i++)
    {
        if (!shown[order[i]])
        {
            realColumn[nColumn] = order[i];
            nColumn++;
        }
    }

    if (GetListFontColor() == RGB(255, 255, 255))
        ListView_SetTextColor(hwndList, RGB(240,240,240));
    else
        ListView_SetTextColor(hwndList, GetListFontColor());

    ResetListView();

    lvfi.flags = LVFI_STRING;
    lvfi.psz = GetDefaultGame();
    i = ListView_FindItem(hwndList, -1, &lvfi);
    SetSelectedPick(i);
}

static void AddIcon(int index)
{
    HICON hIcon = 0;

    if (icon_index[index] != 0)
        return;

    if ((hIcon = LoadIconFromFile((char *)drivers[index]->name)) == 0)
    {
        if (drivers[index]->clone_of != 0)
        {
            hIcon = LoadIconFromFile((char *)drivers[index]->clone_of->name);
            if (!hIcon && drivers[index]->clone_of->clone_of)
                hIcon = LoadIconFromFile((char *)drivers[index]->clone_of->clone_of->name);
        }
    }

    if (hIcon)
    {
        int nIconPos = ImageList_AddIcon(hSmall, hIcon);
        ImageList_AddIcon(hLarge, hIcon);
        if (nIconPos != -1)
            icon_index[index] = nIconPos;
    }
    if (icon_index[index] == 0)
        icon_index[index] = 1;
}

static void ReloadIcons(HWND hWnd)
{
    if ((hSmall = ListView_GetImageList(hWnd, LVSIL_SMALL)) != NULL)
        ImageList_Destroy(hSmall);

    if ((hLarge = ListView_GetImageList(hWnd, LVSIL_NORMAL)) != NULL)
        ImageList_Destroy(hLarge);

    hSmall = 0;
    hLarge = 0;

    CreateIcons(hWnd);
}

static DWORD GetShellLargeIconSize( void )
{
    DWORD   dwSize, dwLength = 512, dwType = REG_SZ;
    TCHAR   szBuffer[512];
    HKEY   hKey;

    // Get the Key
    RegOpenKey( HKEY_CURRENT_USER, "Control Panel\\desktop\\WindowMetrics", &hKey);
    // Save the last size
    RegQueryValueEx( hKey, "Shell Icon Size", NULL, &dwType, szBuffer,  &dwLength );
    dwSize = atol( szBuffer );

    if (dwSize < 32)
        dwSize = 32;

    if (dwSize > 48)
        dwSize = 48;

    // Clean up
    RegCloseKey( hKey );
    return dwSize;
}

/* create iconlist and Listview control */
static BOOL CreateIcons(HWND hWnd)
{
    HICON   hIcon;
    INT     i;
    DWORD   dwLargeIconSize = GetShellLargeIconSize();

    if (!icon_index)
        icon_index = MyMalloc(sizeof(int) * game_count);

    if (icon_index != NULL)
    {
        for (i = 0; i < game_count; i++)
            icon_index[i] = 0;
    }

    hSmall = ImageList_Create (ICONMAP_WIDTH, ICONMAP_HEIGHT,
        ILC_COLORDDB | ILC_MASK, NUM_ICONS, NUM_ICONS + game_count);

    hLarge = ImageList_Create (dwLargeIconSize, dwLargeIconSize,
        ILC_COLORDDB | ILC_MASK, NUM_ICONS, NUM_ICONS + game_count);

    for (i = IDI_WIN_NOROMS; i <= IDI_WIN_REDX; i++)
    {
        if ((hIcon = LoadIconFromFile(icon_names[i - IDI_WIN_NOROMS])) == 0)
            hIcon = LoadIcon (hInst, MAKEINTRESOURCE(i));

        if ((ImageList_AddIcon (hSmall, hIcon) == -1)
            || (ImageList_AddIcon (hLarge, hIcon) == -1))
            return FALSE;
    }

    // Be sure that all the small icons were added.
    if (ImageList_GetImageCount (hSmall) < NUM_ICONS)
        return FALSE;

    // Be sure that all the large icons were added.
    if (ImageList_GetImageCount (hLarge) < NUM_ICONS)
        return FALSE;

    // Associate the image lists with the list view control.
    ListView_SetImageList (hWnd, hSmall, LVSIL_SMALL);

    ListView_SetImageList (hWnd, hLarge, LVSIL_NORMAL);

    return TRUE;
}


/* Sort subroutine to sort the list control */
static int CALLBACK ListCompareFunc(LPARAM index1, LPARAM index2, int sort_subitem)
{
    int value;
    const char *name1 = NULL;
    const char *name2 = NULL;

    switch (sort_subitem)
    {
    case COLUMN_GAMES :
#ifdef BUILD_ENGLISH
        value = stricmp(ModifyThe(drivers[index1]->description),ModifyThe(drivers[index2]->description));
#else
	    if (UseJapaneseList())
	        value = stricmp(drivers[index1]->sortname, drivers[index2]->sortname);
		else
	        value = stricmp(ModifyThe(drivers[index1]->descriptione), ModifyThe(drivers[index2]->descriptione));
#endif
        break;

    case COLUMN_DIRECTORY :
        value = stricmp(drivers[index1]->name, drivers[index2]->name);
        break;

    case COLUMN_TRACKBALL :
        if (GameUsesTrackball(index1) == GameUsesTrackball(index2))
            return ListCompareFunc(index1,index2,COLUMN_GAMES);

        value = GameUsesTrackball(index1) - GameUsesTrackball(index2);
        break;

    case COLUMN_PLAYED :
        value = GetPlayCount(index1) - GetPlayCount(index2);
        if (value == 0)
            return ListCompareFunc(index1,index2,COLUMN_GAMES);

        break;

    case COLUMN_MANUFACTURER :
        if (stricmp(drivers[index1]->manufacturer,drivers[index2]->manufacturer) == 0)
            return ListCompareFunc(index1,index2,COLUMN_GAMES);

        value = stricmp(drivers[index1]->manufacturer,drivers[index2]->manufacturer);
        break;

    case COLUMN_YEAR :
        if (stricmp(drivers[index1]->year,drivers[index2]->year) == 0)
            return ListCompareFunc(index1,index2,COLUMN_GAMES);

        value = stricmp(drivers[index1]->year,drivers[index2]->year);
        break;

    case COLUMN_GENRE :
        value = drivers[index1]->genre - drivers[index2]->genre;
        if (value == 0)
            return ListCompareFunc(index1,index2,COLUMN_GAMES);

        break;

    case COLUMN_CLONE :
        name1 = GetCloneParent(index1);
        name2 = GetCloneParent(index2);

        if (*name1 == '\0')
            name1 = NULL;
        if (*name2 == '\0')
            name2 = NULL;

        if (name1 == name2)
            return ListCompareFunc(index1, index2, COLUMN_GAMES);

        if (name2 == NULL)
            value = -1;
        else if (name1 == NULL)
            value = 1;
        else
            value = stricmp(name1, name2);
        break;

    default :
        return ListCompareFunc(index1,index2,COLUMN_GAMES);
    }

    if (reverse_sort && value != 0)
        value = -(value);

    return value;
}

static int GetSelectedPick(void)
{
    // returns index of listview selected item
    // This will return -1 if not found
    return ListView_GetNextItem(hwndList, -1, LVIS_SELECTED | LVIS_FOCUSED);
}

static int GetSelectedPickItem(void)
{
    // returns lParam (game index) of listview selected item
    LV_ITEM lvi;

    lvi.iItem = GetSelectedPick();
    if (lvi.iItem == -1)
        return 0;

    lvi.iSubItem = 0;
    lvi.mask = LVIF_PARAM;
    ListView_GetItem(hwndList, &lvi);

    return lvi.lParam;
}

static void SetSelectedPick(int new_index)
{
    if (new_index < 0)
        new_index = 0;

    ListView_SetItemState(hwndList, new_index, LVIS_FOCUSED | LVIS_SELECTED,
                          LVIS_FOCUSED | LVIS_SELECTED);
    ListView_EnsureVisible(hwndList, new_index, FALSE);
}

static void SetSelectedPickItem(int val)
{
    int i;
    LV_FINDINFO lvfi;

    if (val < 0)
        return;

    lvfi.flags = LVFI_PARAM;
    lvfi.lParam = val;
    i = ListView_FindItem(hwndList, -1, &lvfi);
    if (i == -1)
    {
        POINT p = {0,0};
        lvfi.flags = LVFI_NEARESTXY;
        lvfi.pt = p;
        i = ListView_FindItem(hwndList, -1, &lvfi);
    }
    SetSelectedPick((i == -1) ? 0 : i);
}

static void SetRandomPickItem(void)
{
    int nListCount;

    nListCount = ListView_GetItemCount(hwndList);

    if (nListCount > 0)
    {
        SetSelectedPick(rand() % nListCount);
    }
}

/* returns if we should immediately start the game */
static BOOL ParseCommandLine(char *command_line)
{
    BOOL help;
    BOOL cpu_detect = GetMMXCheck();
    BOOL ignore_badoptions = FALSE;
    int h = 0, w = 0;
    int i,j;
    int argc;
    char *argv[200];
    char buf[500];
    options_type *o, origOpts;
    int  nGame;

    argc = 1;
    argv[argc] = strtok(command_line, " \t\n");

    if (argv[1] != NULL)
    {
        argc++;
        while ((argv[argc] = strtok(NULL, " \t\n")) != NULL)
            argc++;
    }

    help = FALSE;

    for (i = 1; i < argc; i++)
    {
        if (argv[i][0] == '-' && argv[i][1] == 'i')
            if (strcmp(argv[i],"-ignorebadoptions") == 0)
                ignore_badoptions = TRUE;
    }

    nGame = GetSelectedPickItem();
    o     = GetGameOptions(nGame);

    if (argc >= 2 && argv[1][0] != '-')
    {
        i = 0;
        while (drivers[i])
        {
            if (!stricmp(argv[1], drivers[i]->name))
            {
                LV_FINDINFO lvfi;
                int         index;

                /* Get this games options */
                nGame = i;
                o = GetGameOptions(nGame);

                /* Set current folder to 'All Games' */
                SetCurrentFolder(GetFolderByID(0));
                ResetListView();

                /* Set this game as the current game */
                lvfi.flags  = LVFI_PARAM;
                lvfi.lParam = i;
                index       = ListView_FindItem(hwndList, -1, &lvfi);

                SetSelectedPick(index);
                break;
            }
            i++;
        }
        if (i >= game_count)
        {
            /* Game is not in our list!!! */
            sprintf(buf, win32_getstring(WIN32_commandline_nogame), argv[1]);
            MessageBox(NULL, buf, win32_getstring(WIN32_commandline_nogametitle), MB_OK | MB_ICONERROR);
        }
    }

    memcpy(&origOpts, o, sizeof(options_type));

    for (i = 1; i < argc; i++)
    {
        if (argv[i][0] != '-')
            continue;

        if (sscanf(argv[i], "-%dx%d", &w, &h) == 2
        ||  stricmp(argv[i], "-resolution") == 0)
        {
            struct tDisplayModes*   pDisplayModes;
            BOOL                    bFound = FALSE;

            if (stricmp(argv[i], "-resolution") == 0)
            {
                if (argc <= ++i)
                    break;

                if (sscanf(argv[i], "%dx%d", &w, &h) != 2)
                    break;
            }

            pDisplayModes = DirectDraw_GetDisplayModes();

            /* Verify that such a mode exists, 8 or 16 bit. */
            for (j = 0; j < pDisplayModes->m_nNumModes; j++)
            {
                if (w == (int)pDisplayModes->m_Modes[j].m_dwWidth
                &&  h == (int)pDisplayModes->m_Modes[j].m_dwHeight)
                {
                    bFound = TRUE;

                    /* Set to FULL_SCREEN */
                    o->width            = w;
                    o->height           = h;
                    o->is_window        = FALSE;
                    o->display_best_fit = FALSE;

                    break;
                }
            }

            if (!bFound)
            {
                sprintf(buf, win32_getstring(WIN32_commandline_noreso), w, h);
                MessageBox(NULL, buf, win32_getstring(WIN32_commandline_noresotitle), MB_OK | MB_ICONERROR);
            }
            continue;
        }

        if (stricmp(argv[i], "-gamma") == 0)
        {
            double dGamma;

            if (argc <= ++i)
                break;

            dGamma = atof(argv[i]);
            if (0.50 < dGamma || dGamma > 2.00)
                continue;

            o->gamma = dGamma;
            continue;
        }

        if (stricmp(argv[i], "-brightness") == 0)
        {
            if (argc <= ++i)
                break;

            o->brightness = atoi(argv[i]);
            continue;
        }

        if (stricmp(argv[i], "-transparency") == 0)
        {
            if (argc <= ++i)
                break;

            o->transparency = atoi(argv[i]);
            continue;
        }

        if (stricmp(argv[i], "-randomtime") == 0)
        {
            if (argc <= ++i)
                break;

            o->randomtime = atoi(argv[i]);
            continue;
        }

        if (stricmp(argv[i], "-dsbuffersize") == 0)
        {
            if (argc <= ++i)
                break;

            o->dsound_buffer = atoi(argv[i]);
            continue;
        }

        if (stricmp(argv[i], "-render_fullscreen") == 0)
        {
            if (argc <= ++i)
                break;

            o->render_fullscreen = atoi(argv[i]);

            if (o->render_fullscreen < 0
            ||  o->render_fullscreen > 8)
                o->render_fullscreen = 0;
            continue;
        }

        if (stricmp(argv[i], "-render_window") == 0)
        {
            if (argc <= ++i)
                break;

            o->render_window = atoi(argv[i]);

            if (o->render_window < 0
            ||  o->render_window > 7)
                o->render_window= 0;
            continue;
        }

        if (stricmp(argv[i], "-triplebuffer") == 0)
        {
            o->triple_buffer = TRUE;
            if (o->vsync_fullscreen == TRUE)
                o->vsync_fullscreen = FALSE;
            continue;
        }

        if (stricmp(argv[i], "-notriplebuffer") == 0)
        {
            o->triple_buffer = FALSE;
            continue;
        }

        if (stricmp(argv[i], "-vsync_fullscreen") == 0)
        {
            if (o->triple_buffer == FALSE)
                o->vsync_fullscreen = TRUE;
            continue;
        }

        if (stricmp(argv[i], "-novsync_fullscreen") == 0)
        {
            o->vsync_fullscreen = FALSE;
            continue;
        }

        if (stricmp(argv[i], "-vsync_window") == 0)
        {
            o->vsync_window = TRUE;
            continue;
        }

        if (stricmp(argv[i], "-novsync_fullscreen") == 0)
        {
            o->vsync_window = FALSE;
            continue;
        }

        if (stricmp(argv[i], "-screen") == 0)
        {
            o->is_window = FALSE;
            continue;
        }
        if (stricmp(argv[i], "-window") == 0)
        {
            o->is_window = TRUE;
            continue;
        }

        if (stricmp(argv[i], "-nommx") == 0)
        {
            o->disable_mmx = TRUE;
            continue;
        }

        if (stricmp(argv[i], "-frameskip") == 0)
        {
            i++;
            if (i < argc && argv[i][0] != '-')
            {
                int skip = atoi(argv[i]) + 1;

                /* the Main mame code allows from 1 to 12 for this value */
                if (skip < 1)  skip = 1;
                if (skip > 13) skip = 12;
                o->frame_skip = skip;
                continue;
            }
            else
            {
                i--;
            }
            if (!ignore_badoptions)
                help = 1;
            continue;
        }

        /* Sound */
        if (stricmp(argv[i], "-nosound") == 0)
        {
            o->sound = SOUND_NONE;
            continue;
        }
#ifndef NOMIDAS
        if (stricmp(argv[i], "-midas") == 0)
        {
            o->sound = SOUND_MIDAS;
            continue;
        }
#endif
        if (stricmp(argv[i], "-directsound") == 0)
        {
            o->sound = SOUND_DIRECT;
            continue;
        }

        if (stricmp(argv[i], "-sr") == 0)
        {
            if (argc <= ++i)
                break;

            if (stricmp(argv[i], "11025") == 0)
                o->sample_rate = 0;
            else
            if (stricmp(argv[i], "22050") == 0)
                o->sample_rate = 1;
            else
            if (stricmp(argv[i], "44100") == 0)
                o->sample_rate = 2;
            else
            if (stricmp(argv[i], "48000") == 0)
                o->sample_rate = 3;

            continue;
        }
        if (stricmp(argv[i], "-stereo") == 0)
        {
            o->stereo = TRUE;
            continue;
        }
        if (stricmp(argv[i], "-nostereo") == 0)
        {
            o->stereo = FALSE;
            continue;
        }
        if (stricmp(argv[i], "-directsoundfix") == 0)
        {
            o->directsound_fix = TRUE;
            continue;
        }
        if (stricmp(argv[i], "-nodirectsoundfix") == 0)
        {
            o->directsound_fix = FALSE;
            continue;
        }
        if (stricmp(argv[i], "-volume") == 0)
        {
            int vol;

            if (argc <= ++i)
                break;

            vol = atoi(argv[i]);
            if (-32 <= vol && vol <= 0)
                o->volume = - vol;

            continue;
        }

        /* Input */
        if (stricmp(argv[i], "-mouse") == 0)
        {
            o->use_ai_mouse = TRUE;
            continue;
        }
        if (stricmp(argv[i], "-nomouse") == 0)
        {
            o->use_ai_mouse = FALSE;
            continue;
        }
        if (stricmp(argv[i], "-joy") == 0)
        {
            o->use_joystick = TRUE;
            continue;
        }
        if (stricmp(argv[i], "-nojoy") == 0)
        {
            o->use_joystick = FALSE;
            continue;
        }
        if (stricmp(argv[i], "-dikbd") == 0)
        {
            if (DIKeyboard_Available())
                o->di_keyboard = TRUE;
            continue;
        }
        if (stricmp(argv[i], "-dijoy") == 0)
        {
            if (DIJoystick.Available(1))
                o->di_joystick = TRUE;
            continue;
        }
        if (stricmp(argv[i], "-hotrod") == 0)
        {
            o->default_input = INPUT_LAYOUT_HR;
            continue;
        }
        if (stricmp(argv[i], "-hotrodse") == 0)
        {
            o->default_input = INPUT_LAYOUT_HRSE;
            continue;
        }

        /* Misc */
        if (stricmp(argv[i], "-cheat") == 0)
        {
            o->cheat = TRUE;
            continue;
        }

#ifdef MAME_DEBUG
        if (stricmp(argv[i], "-win32debug") == 0)
        {
            win32_debug = 1;
            continue;
        }

        if (stricmp(argv[i], "-debug") == 0)
        {
            mame_debug = 1;
            continue;
        }
#endif

        if (stricmp(argv[i], "-log") == 0)
        {
            o->error_log = TRUE;
            continue;
        }

        if (stricmp(argv[i], "-oslog") == 0)
        {
            extern BOOL g_bOsDebug;
            g_bOsDebug = TRUE;
            continue;
        }

        if (stricmp(argv[i], "-nopatch") == 0)
        {
            o->patch_roms = TRUE;
            continue;
        }

        if (stricmp(argv[i], "-nousenglo") == 0)
        {
            o->nglo_disabled = TRUE;
            continue;
        }

        if (stricmp(argv[i], "-noautopause") == 0)
        {
            o->auto_pause = FALSE;
            continue;
        }

        if (stricmp(argv[i], "-quit") == 0)
        {
            quit = TRUE;
            continue;
        }

        if (stricmp(argv[i], "-joygui") == 0)
        {
            joygui = &Joystick;
            continue;
        }

        if (stricmp(argv[i], "-dijoygui") == 0)
        {
            joygui = &DIJoystick;
             continue;
        }

        if (stricmp(argv[i], "-nocpudetect") == 0)
        {
            cpu_detect = FALSE;
            continue;
        }

        if (stricmp(argv[i], "-broadcast") == 0)
        {
            bDoBroadcast = TRUE;
            continue;
        }

        if (stricmp(argv[i], "-nobroadcast") == 0)
        {
            bDoBroadcast = FALSE;
            continue;
        }

        /* Are they asking for help? */
        if (stricmp(argv[i], "-help") == 0 || stricmp(argv[i], "-?") == 0)
        {
            help = 1;
            break;
        }

        /* Let the user decide if they want to contunue
         * if we encounter an Unknown command line option.
         *
         * This allows people who want to use a different
         * front end, to continue to do so.
         *
         * In addition,there is a -ignorebadoptions command
         * line flag.
         */
        if (argv[i][0] == '-' && ignore_badoptions == FALSE)
        {
            char cBuf[200];

            sprintf(cBuf, win32_getstring(WIN32_commandline_unknown), argv[i]);

            if (IDNO == MessageBox(NULL, cBuf, win32_getstring(WIN32_commandline_unknowntitle),
                MB_YESNO | MB_ICONQUESTION))
            {
                ExitProcess(0);
            }
            else
            {
                ignore_badoptions = TRUE;
            }
            continue;
        }
    }

    if (cpu_detect)
    {
        MAME32App.m_bMMXDetected = MAME32App.DetectMMX();
    }
    if (help)
    {
        DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(IDD_HELP), NULL, MameHelpDialogProc);
        ExitProcess(0);
    }

    if (memcmp(&origOpts, o, sizeof(options_type)) != 0)
        o->use_default = FALSE;

    if (argc >= 2 && argv[1][0] != '-')
        return TRUE;

    return FALSE;
}

static void HelpPrintln(char *s)
{
    HWND hEdit;

    hEdit = GetDlgItem(hHelp, IDC_HELP_TEXT);

    Edit_SetSel(hEdit, Edit_GetTextLength(hEdit), Edit_GetTextLength(hEdit));
    Edit_ReplaceSel(hEdit, s);
    Edit_SetSel(hEdit, Edit_GetTextLength(hEdit), Edit_GetTextLength(hEdit));
    Edit_ReplaceSel(hEdit, "\r\n");
}

static INT_PTR CALLBACK AboutDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    switch (Msg)
    {
    case WM_INITDIALOG:
        {
			char stuff[4096];

			connect_strings(stuff, WIN32_about01, WIN32_about54);

            Edit_SetText(GetDlgItem(hDlg, IDC_EDIT_ABOUT), stuff);
        }
        return 1;

    case WM_COMMAND :
        if (LOWORD(wParam) == IDC_NMX_WEB_PAGE)
            ShellExecute(hMain,NULL,"http://go.to/nmamex/",NULL,NULL,SW_SHOWNORMAL);

		if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL || LOWORD(wParam) == IDC_NMX_WEB_PAGE)
        {
            EndDialog(hDlg, 0);
            return 1;
        }
    }
    return 0;
}


static INT_PTR CALLBACK DirectXDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
	HWND hEdit;

	char directx_help[512];

	connect_strings(directx_help, WIN32_directxhelp01, WIN32_directxhelp04);

    switch (Msg)
    {
    case WM_INITDIALOG:
        hEdit = GetDlgItem(hDlg,IDC_DIRECTX_HELP);
        Edit_SetSel(hEdit, Edit_GetTextLength(hEdit), Edit_GetTextLength(hEdit));
        Edit_ReplaceSel(hEdit,directx_help);
        return 1;

    case WM_COMMAND :
        if (LOWORD(wParam) == IDC_WEB_PAGE)
#ifdef BUILD_ENGLISH
            ShellExecute(hMain,NULL,"http://www.microsoft.com/directx",NULL,NULL,SW_SHOWNORMAL);
#else
            ShellExecute(hMain,NULL,"http://www.microsoft.com/japan/directx",NULL,NULL,SW_SHOWNORMAL);
#endif

        if (LOWORD(wParam) == IDCANCEL || LOWORD(wParam) == IDC_WEB_PAGE)
            EndDialog(hDlg, 0);
        return 1;
    }
    return 0;
}

static INT_PTR CALLBACK MameHelpDialogProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch (uMsg)
    {
        case WM_INITDIALOG :
            hHelp = hwndDlg;
            HelpPrintln((char *)win32_getstring(WIN32_help01));
            HelpPrintln((char *)win32_getstring(WIN32_help02));
            HelpPrintln("");
            HelpPrintln((char *)win32_getstring(WIN32_help03));
            HelpPrintln((char *)win32_getstring(WIN32_help04));
            HelpPrintln("");
            HelpPrintln((char *)win32_getstring(WIN32_help05));
            HelpPrintln((char *)win32_getstring(WIN32_help06));
            HelpPrintln("");
            HelpPrintln((char *)win32_getstring(WIN32_help07));
            HelpPrintln("");
            HelpPrintln((char *)win32_getstring(WIN32_help08));

            return TRUE;

    case WM_COMMAND :
        switch (LOWORD(wParam))
        {
            case IDOK :
                EndDialog(hwndDlg, 0);
                return TRUE;
            case IDCANCEL :
                EndDialog(hwndDlg, 1);
                return TRUE;
        }
        break;
    }
    return FALSE;
}

static INT_PTR CALLBACK VersionDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    switch (Msg)
    {
    case WM_INITDIALOG:
		GetVersionName(version_name);
        Edit_SetText(GetDlgItem(hDlg, IDC_VERSION_EDIT), version_name);
        Edit_SetSel(GetDlgItem(hDlg, IDC_VERSION_EDIT), 0, -1);
        SetFocus(GetDlgItem(hDlg, IDC_VERSION_EDIT));
        return FALSE;

    case WM_COMMAND :
        {
            WORD wID         = GET_WM_COMMAND_ID(wParam, lParam);

            switch (wID)
            {
	            case IDOK :
	                Edit_GetText(GetDlgItem(hDlg, IDC_VERSION_EDIT), version_name, VERSIONTEXT_LEN);
					SetVersionName(version_name);
	                EndDialog(hDlg, 1);
	                return TRUE;

	            case IDCANCEL :
	                EndDialog(hDlg, 0);
	                return TRUE;
            }
        }
        break;
	}
    return 0;
}

options_type * GetPlayingGameOptions(void)
{
    return &playing_game_options;
}

static void MamePlayRecordAVI(void)
{
    char filename[MAX_PATH];
    *filename = 0;

    if (CommonFileDialogAVI(GetOpenFileName,filename))
    {
        int index = GetSelectedPickItem();

        memcpy(&playing_game_options, GetGameOptions(index), sizeof(options_type));

        options.record     = NULL;
        options.playback   = NULL;

        if (!AviStartCapture(filename,
#if 0
                             drivers[index]->drv->default_visible_area.min_x,
                             drivers[index]->drv->default_visible_area.min_y,
                             drivers[index]->drv->default_visible_area.max_x,
                             drivers[index]->drv->default_visible_area.max_y))
#else
                             8, 16, 311, 239))
#endif
            return;

        MamePlayGameWithOptions();

        AviEndCapture();
    }
}

static void MamePlayRecordGame(void)
{
    INP_HEADER inpHeader;
    char filename[MAX_PATH];
    *filename = 0;

    memset(&inpHeader,'\0', sizeof(INP_HEADER));

    strcpy(filename, drivers[GetSelectedPickItem()]->name);

    if (CommonFileDialog(GetSaveFileName,filename, FALSE))
    {
        options.record = osd_fopen(filename, 0, OSD_FILETYPE_INPUTLOG, 1);
        options.playback = NULL;

        if (options.record == NULL)
        {
            char buf[1024];

            sprintf(buf, win32_getstring(WIN32_record_validinp), filename);
            MessageBox(NULL, buf, win32_getstring(WIN32_record_validinptitle), MB_OK | MB_ICONERROR);
            return;
        }

        strcpy(inpHeader.name,drivers[GetSelectedPickItem()]->name);
        inpHeader.version[1] = VERSION;
#ifdef BETA_VERSION
        inpHeader.version[2] = BETA_VERSION;
#endif
        osd_fwrite(options.record, &inpHeader, sizeof(INP_HEADER));

        MamePlayGameWithOptions();

        osd_fclose(options.record);
    }
}

static void MamePlayBackGame(void)
{
    INP_HEADER inpHeader;
    char filename[MAX_PATH];
    LPTREEFOLDER lpOldFolder = GetCurrentFolder();
    LPTREEFOLDER lpNewFolder = GetFolder(0);
    DWORD        dwOldFilters = 0;
    int          nOldPick = GetSelectedPickItem();

    *filename = 0;
    memset(&inpHeader,'\0', sizeof(INP_HEADER));

    if (CommonFileDialog(GetOpenFileName, filename, TRUE))
    {
        options.record = NULL;
        options.playback = osd_fopen(filename, 0, OSD_FILETYPE_INPUTLOG, 0);

        if (options.playback == NULL)
        {
            char buf[1024];

            sprintf(buf, win32_getstring(WIN32_record_validinp), filename);
            MessageBox(NULL, buf, win32_getstring(WIN32_record_validinptitle), MB_OK | MB_ICONERROR);
            return;
        }

        /* find out which version of Mame wrote the inpfile and which game */
        if ((osd_fread(options.playback, (LPSTR)&inpHeader,
            sizeof(INP_HEADER)) != sizeof(INP_HEADER)) ||
            inpHeader.name[0] < '\x20')
        {
            osd_fseek(options.playback, 0, SEEK_SET);
        }
        else
        {
            LVFINDINFO   lvfi;
            int          i;

            /* Set the current folder to 'All Games'
             * and remove any filters, so all games
             * are in the tree view.
             */
            if (lpNewFolder)
            {
                dwOldFilters = lpNewFolder->m_dwFlags & F_MASK;
                lpNewFolder->m_dwFlags &= ~F_MASK;
                SetCurrentFolder(lpNewFolder);
                ResetListView();
            }

            /* Find the game and select it */
            for (i = 0; i < game_count; i++)
            {
                if (strcmp(drivers[i]->name, inpHeader.name) == 0)
                {
                    lvfi.flags  = LVFI_PARAM;
                    lvfi.lParam = i;
                    SetSelectedPick(ListView_FindItem(hwndList, -1, &lvfi));
                    break;
                }
            }
        }

        MamePlayGameWithOptions();

        osd_fclose(options.playback);

        check_inp_type();

        /* Restore folder filters */
        if (lpNewFolder)
        {
            lpNewFolder->m_dwFlags |= dwOldFilters;
        }

        /* Restore old folder */
        if (lpOldFolder)
        {
            SetCurrentFolder(lpOldFolder);
            ResetListView();
            SetSelectedPickItem(nOldPick);
        }
    }
}

static BOOL CommonFileDialog(common_file_dialog_proc cfd, char *filename, BOOL bZip)
{
    BOOL success;

    OPENFILENAME of;

    of.lStructSize = sizeof(of);
    of.hwndOwner = hMain;
    of.hInstance = NULL;
    if (bZip == TRUE)
        of.lpstrFilter = win32_getstring(WIN32_record_inpzipfile);
    else
        of.lpstrFilter = win32_getstring(WIN32_record_inpfile);
    of.lpstrCustomFilter = NULL;
    of.nMaxCustFilter = 0;
    of.nFilterIndex = 1;
    of.lpstrFile = filename;
    of.nMaxFile = MAX_PATH;
    of.lpstrFileTitle = NULL;
    of.nMaxFileTitle = 0;
    of.lpstrInitialDir = last_directory;
    of.lpstrTitle = NULL;
    of.Flags = OFN_EXPLORER | OFN_NOCHANGEDIR | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY;
    of.nFileOffset = 0;
    of.nFileExtension = 0;
    of.lpstrDefExt = "inp";
    of.lCustData = 0;
    of.lpfnHook = NULL;
    of.lpTemplateName = NULL;

    success = cfd(&of);
    if (success)
    {
        //GetDirectory(filename,last_directory,sizeof(last_directory));
    }

    return success;
}

static BOOL CommonFileDialogAVI(common_file_dialog_proc cfd,char *filename)
{
    BOOL success;

    OPENFILENAME of;

    of.lStructSize = sizeof(of);
    of.hwndOwner = hMain;
    of.hInstance = NULL;
    of.lpstrFilter = win32_getstring(WIN32_record_avifile);
    of.lpstrCustomFilter = NULL;
    of.nMaxCustFilter = 0;
    of.nFilterIndex = 1;
    of.lpstrFile = filename;
    of.nMaxFile = MAX_PATH;
    of.lpstrFileTitle = NULL;
    of.nMaxFileTitle = 0;
    of.lpstrInitialDir = NULL;
    of.lpstrTitle = NULL;
    of.Flags = OFN_EXPLORER | OFN_NOCHANGEDIR | OFN_PATHMUSTEXIST | OFN_HIDEREADONLY;
    of.nFileOffset = 0;
    of.nFileExtension = 0;
    of.lpstrDefExt = "avi";
    of.lCustData = 0;
    of.lpfnHook = NULL;
    of.lpTemplateName = NULL;

    success = cfd(&of);

    return success;
}

static void MamePlayGame(void)
{
    options.record   = NULL;
    options.playback = NULL;

    MamePlayGameWithOptions();
}

static void MamePlayGameWithOptions(void)
{
    int index;
    RECT rect;
    char buf[200];

    index = GetSelectedPickItem();

    memcpy(&playing_game_options, GetGameOptions(index), sizeof(options_type));

    /* Deal with options that can be disabled. */
    EnablePlayOptions(index, &playing_game_options);

    /*
        based on our raw options, convert the ones that we can into
        regular MAME config options
    */

#ifdef MAME_DEBUG
    options.mame_debug = mame_debug;
#endif
    options.cheat           = playing_game_options.cheat;
    options.gui_host        = 1;
    options.core_type       = playing_game_options.core_type;
    options.driver_type     = playing_game_options.driver_type;
    options.bios_type       = playing_game_options.bios_type;
    options.disable_patch   = playing_game_options.patch_roms;
    options.disable_nglo    = playing_game_options.nglo_disabled;
	options.ui_transparency = playing_game_options.transparency;
	options.random_time     = playing_game_options.randomtime;

	if (playing_game_options.cyclecpu == TRUE)	options.cyclecpu = 0;
	else	options.cyclecpu = 1;

    /*
        Need sample rate initialised _before_ rom loading
        for optional rom regions.
    */
    if (playing_game_options.sound == SOUND_NONE)
    {
        /* update the Machine structure to show that sound is disabled */
        Machine->sample_rate = 0;
        options.samplerate = 0;
    }
    else
    {
        options.samplerate = playing_game_options.sample_rate;
    }
    options.stereo = (playing_game_options.stereo == TRUE);

#ifdef MAME_DEBUG
    options.debug_width  = 640;
    options.debug_height = 480;
#endif

	if (GameUsesOverClock(index))
		options.use_overclock = 1;
	else
		options.use_overclock = 0;

	if (GameOnlyJukeBox(index))
		options.only_jukebox = 1;
	else
		options.only_jukebox = 0;

    if (joygui != NULL)
        KillTimer(hMain,0);

#ifdef BUILD_ENGLISH
    sprintf(buf, "%s - " MAME32NAME, ModifyThe(drivers[index]->description));
#else
    if (UseJapaneseList())
	    sprintf(buf, "%s - " MAME32NAME, drivers[index]->description);
	else
	    sprintf(buf, "%s - " MAME32NAME, ModifyThe(drivers[index]->descriptione));
#endif
    SetWindowText(hMain, buf);

    abort_loading = FALSE;

    if (drivers[index]->rom != NULL)
    {
        hLoad = CreateDialog(GetModuleHandle(NULL),
                             MAKEINTRESOURCE(IDD_LOAD_PROGRESS),0,
                             LoadProgressDialogProc);
    }
    else
        hLoad = NULL;

    if (quit && hLoad != NULL)
    {
        RECT r;
        int xSize, ySize;
        int x, y;

        xSize = GetSystemMetrics(SM_CXSCREEN);     // Screen Width
        ySize = GetSystemMetrics(SM_CYSCREEN);     // Screen Height
        GetWindowRect(hLoad, &r);
        x = (xSize - (r.right  - r.left)) / 2;
        y = (ySize - (r.bottom - r.top))  / 2;

        // Center in the middle of the screen
        SetWindowPos(hLoad, NULL, x, y, -1, -1,
                     SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE);
    }

    if (hLoad != NULL)
    {
#ifdef BUILD_ENGLISH
        sprintf(buf, "%s %s",
			win32_getstring(WIN32_loading),
			drivers[index]->description);
#else
        if (UseJapaneseList())
	        sprintf(buf, "%s %s",
			win32_getstring(WIN32_loading),
			drivers[index]->description);
		else
	        sprintf(buf, "%s %s",
			win32_getstring(WIN32_loading),
			drivers[index]->descriptione);
#endif
        SetWindowText(hLoad, buf);
    }

    GetWindowRect(hMain, &rect);

    in_emulation = TRUE;

#ifndef BUILD_ENGLISH
    WINNLSEnableIME(NULL, FALSE);
#endif

    if (run_game(index) == 0)
    {
       IncrementPlayCount(index);
       ListView_RedrawItems(hwndList,GetSelectedPickItem(),GetSelectedPickItem());
    }
    else
    {
        if (abort_loading == FALSE)
        {
            if (!quit)
                ShowWindow(hMain, SW_SHOW);
            if (hLoad != NULL)
            {
                ShowWindow(hLoad, SW_HIDE);
                EndDialog(hLoad, 0);
                hLoad = NULL;
            }
            DialogBox(GetModuleHandle(NULL),MAKEINTRESOURCE(IDD_ROM_ERROR), NULL, LoadErrorDialogProc);
        }
    }

#ifndef BUILD_ENGLISH
    WINNLSEnableIME(NULL, TRUE);
#endif

    in_emulation = FALSE;

	if (hLoad != NULL)	// test code
	{
        EndDialog(hLoad, 0);
        hLoad = NULL;
	}

    if (hAbort != NULL)
    {
        ShowWindow(hAbort, SW_HIDE);
        EndDialog(hAbort, 0);
        hAbort = NULL;
        EnableWindow(hMain, TRUE);
    }

    if (quit)
    {
       PostMessage(hMain, WM_CLOSE, 0, 0);
       return;
    }

    // resort if sorting on # of times played
    if (abs(GetSortColumn()) - 1 == COLUMN_PLAYED)
    {
        int column = GetSortColumn();

        reverse_sort = FALSE;

        if (column > 0)
        {
            last_sort = -1;
        }
        else
        {
            column = abs(column);
            last_sort = column - 1;
        }

        ColumnSort(column - 1, TRUE);
    }

    SetWindowText(hMain, MAME32NAME);
    UpdateStatusBar();

    if (!quit)
    {
        SetWindowFont(hMain, hFont, FALSE);
        SetWindowPos(hMain, 0, rect.left, rect.top,
                     rect.right - rect.left,
                     rect.bottom - rect.top,
                     SWP_NOZORDER | SWP_NOCOPYBITS);
        ShowWindow(hMain, SW_SHOW);
        SetFocus(hwndList);
        InvalidateRect(hMain, NULL, FALSE);
    }

    if (joygui != NULL)
        SetTimer(hMain,0,50,NULL);
}

static char * GetRomList(int iGame)
{
    static char buf[5000];
    const struct RomModule *region, *rom, *chunk;

    buf[0] = '\0';

    for (region = rom_first_region(drivers[iGame]); region; region = rom_next_region(region))
    {
        for (rom = rom_first_file(region); rom; rom = rom_next_file(rom))
        {
            int length = 0;

            for (chunk = rom_first_chunk(rom); chunk; chunk = rom_next_chunk(chunk))
                length += ROM_GETLENGTH(chunk);

            sprintf(buf + strlen(buf), "%s:\t%-12s\t%u %s\r\n",
                    drivers[iGame]->name, ROM_GETNAME(rom), length, win32_getstring(WIN32_romlist_byte));
        }
    }

    return buf;
}

static INT_PTR CALLBACK LoadErrorDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    int iGame;

    switch (Msg)
    {
    case WM_INITDIALOG:
        iGame = GetSelectedPickItem();
        CenterSubDialog(hMain,hDlg,FALSE);
        Static_SetText(GetDlgItem(hDlg, IDC_PROP_TITLE), GameInfoTitle(iGame));
        Edit_SetText(GetDlgItem(hDlg, IDC_ERROR_DETAILS), GetRomList(iGame));
        return 1;

    case WM_COMMAND :
       if (LOWORD(wParam) == IDOK)
       {
            EndDialog(hDlg,1);
       }
       return 1;
    }
    return 0;
}

static INT_PTR CALLBACK LoadProgressDialogProc(HWND hDlg, UINT Msg, WPARAM wParam, LPARAM lParam)
{
    switch (Msg)
    {
    case WM_INITDIALOG:
        if (!quit)
            CenterSubDialog(hMain,hDlg,FALSE);
        return 1;

    case WM_COMMAND :

       if (LOWORD(wParam) == IDCANCEL)
       {
           abort_loading = TRUE;

           if (hAbort == NULL)
           {
               RECT r1, r2;
               int x, y;

               hAbort = CreateDialog(GetModuleHandle(NULL),MAKEINTRESOURCE(IDD_WAIT_EMULATION_FINISH), NULL, NULL);

               GetWindowRect(hMain, &r1);
               GetWindowRect(hAbort, &r2);
               x = ((r1.right - r1.left) - (r2.right  - r2.left)) / 2;
               y = ((r1.bottom - r1.top) - (r2.bottom - r2.top))  / 2;

               // Center in the middle of the screen
               SetWindowPos(hAbort, NULL, r1.left + x, r1.top + y, -1, -1, SWP_NOSIZE | SWP_NOZORDER);
               ShowWindow(hAbort, SW_SHOW);
               InvalidateRect(hAbort, NULL, FALSE);
               SetAbsoluteForegroundWindow(hAbort);
               EnableWindow(hMain, FALSE);
           }

           EndDialog(hDlg,1);
       }
       return 1;
    }
    return 0;
}

/* Toggle ScreenShot ON/OFF */
static void ToggleScreenShot(void)
{
    BOOL showScreenShot = GetShowScreenShot();

    SetShowScreenShot((showScreenShot) ? FALSE : TRUE);
    UpdateScreenShot();

    /* Invalidate the List control */
    if( hBitmap != NULL && showScreenShot)
    {
        RECT rcClient;
        GetClientRect( hwndList, &rcClient );
        InvalidateRect( hwndList, &rcClient, TRUE );
    }
}

static void AdjustMetrics(void)
{
    HDC hDC;
    TEXTMETRIC tm;
    int xtraX, xtraY;
    AREA area;
    int  offX, offY;
    int  maxX, maxY;
    COLORREF textColor;

    // WM_SETTINGCHANGE also
    xtraX =  GetSystemMetrics(SM_CXFIXEDFRAME); // Dialog frame width
    xtraY =  GetSystemMetrics(SM_CYFIXEDFRAME); // Dialog frame height
    xtraY += GetSystemMetrics(SM_CYMENUSIZE);   // Menu height
    xtraY += GetSystemMetrics(SM_CYCAPTION);    // Caption Height
    maxX  =  GetSystemMetrics(SM_CXSCREEN);     // Screen Width
    maxY  =  GetSystemMetrics(SM_CYSCREEN);     // Screen Height

    hDC = GetDC(hMain);
    GetTextMetrics (hDC, &tm);

    // Convert MIN Width/Height from Dialog Box Units to pixels.
    MIN_WIDTH  = (int)((tm.tmAveCharWidth / 4.0) * DBU_MIN_WIDTH) + xtraX;
    MIN_HEIGHT = (int)((tm.tmHeight / 8.0) * DBU_MIN_HEIGHT) + xtraY;
    ReleaseDC(hMain, hDC);

    ListView_SetBkColor(hwndList, GetSysColor(COLOR_WINDOW));

    if ((textColor = GetListFontColor()) == RGB(255,255,255))
        textColor = RGB(240, 240, 240);

    ListView_SetTextColor(hwndList, textColor);
    TreeView_SetBkColor(hTreeView, GetSysColor(COLOR_WINDOW));
    TreeView_SetTextColor(hTreeView, textColor);
    GetWindowArea(&area);

    offX = area.x + area.width;
    offY = area.y + area.height;

    if (offX > maxX)
    {
        offX = maxX;
        area.x = (offX - area.width > 0) ? (offX - area.width) : 0;
    }
    if (offY > maxY)
    {
        offY = maxY;
        area.y = (offY - area.height > 0) ? (offY - area.height) : 0;
    }

    SetWindowArea(&area);
    if (!quit)
        SetWindowPos(hMain, 0, area.x, area.y, area.width, area.height, SWP_NOZORDER | SWP_SHOWWINDOW);
}


/* Adjust options - tune them to the currently selected game */
static void EnablePlayOptions(int nIndex, options_type *o)
{
    if (!MAME32App.m_bMMXDetected)
        o->disable_mmx = TRUE;

    if (!Joystick.Available(1))
        o->use_joystick = FALSE;

    if (!DIKeyboard_Available())
        o->di_keyboard = FALSE;

    if (!DIJoystick.Available(1))
        o->di_joystick = FALSE;

    if (!GameUsesTrackball(nIndex))
        o->use_ai_mouse = FALSE;
}

static int WhichIcon(int nItem)
{
    int iconRoms;

    iconRoms = GetHasRoms(nItem);

    /* Show Red-X if the ROMs are present and flaged as NOT WORKING */
    if (iconRoms == 1 && drivers[nItem]->flags & GAME_BROKEN)
        iconRoms = 3;

    /* Use custom icon if found */
    if (iconRoms == 1 && icon_index != NULL)
    {
        if (icon_index[nItem] == 0)
            AddIcon(nItem);
        iconRoms = icon_index[nItem];
    }

    return iconRoms;
}

static BOOL HandleTreeContextMenu( HWND hWnd, WPARAM wParam, LPARAM lParam)
{
    HMENU hMenuLoad;
    HMENU hMenu;

    if((HWND)wParam != GetDlgItem(hWnd, IDC_TREE))
        return FALSE;

    hMenuLoad = LoadMenu(hInst, MAKEINTRESOURCE(IDR_CONTEXT_TREE));
    hMenu = GetSubMenu(hMenuLoad, 0);

    UpdateMenu(GetDlgItem(hWnd, IDC_TREE), hMenu);

    TrackPopupMenu( hMenu,
        TPM_LEFTALIGN | TPM_RIGHTBUTTON,
        LOWORD(lParam),
        HIWORD(lParam),
        0,
        hWnd,
        NULL);

    DestroyMenu(hMenuLoad);

    return TRUE;
}


static BOOL HandleContextMenu( HWND hWnd, WPARAM wParam, LPARAM lParam)
{
    HMENU hMenuLoad;
    HMENU hMenu;

    if((HWND)wParam != GetDlgItem(hWnd, IDC_LIST))
        return FALSE;

    if (current_view_id == ID_VIEW_DETAIL &&
        HeaderOnContextMenu(hWnd, wParam, lParam) == TRUE)
        return TRUE;

    hMenuLoad = LoadMenu(hInst, MAKEINTRESOURCE(IDR_CONTEXT_MENU));
    hMenu = GetSubMenu(hMenuLoad, 0);

    UpdateMenu(GetDlgItem(hWnd, IDC_LIST), hMenu);

    TrackPopupMenu( hMenu,
        TPM_LEFTALIGN | TPM_RIGHTBUTTON,
        LOWORD(lParam),
        HIWORD(lParam),
        0,
        hWnd,
        NULL);

    DestroyMenu(hMenuLoad);

    return TRUE;
}

static void UpdateMenu(HWND hwndList, HMENU hMenu)
{
    char            buf[200];
    MENUITEMINFO    mItem;
    int             nGame = GetSelectedPickItem();

    if (have_selection)
    {
#ifdef BUILD_ENGLISH
        sprintf(buf, win32_getstring(WIN32_menubar_enableplay), ConvertAmpersandString(ModifyThe(drivers[nGame]->description)));
#else
        if (UseJapaneseList())
	        sprintf(buf, win32_getstring(WIN32_menubar_enableplay), ConvertAmpersandString(drivers[nGame]->description));
		else
	        sprintf(buf, win32_getstring(WIN32_menubar_enableplay), ConvertAmpersandString(ModifyThe(drivers[nGame]->descriptione)));
#endif

        mItem.cbSize     = sizeof(mItem);
        mItem.fMask      = MIIM_TYPE;
        mItem.fType      = MFT_STRING;
        mItem.dwTypeData = buf;
        mItem.cch        = strlen(mItem.dwTypeData);

        SetMenuItemInfo(hMenu, ID_FILE_PLAY, FALSE, &mItem);

        if (GetIsFavorite(nGame))
            EnableMenuItem(hMenu, ID_CONTEXT_ADD_FAVORITE, MF_GRAYED);
        else
            EnableMenuItem(hMenu, ID_CONTEXT_DEL_FAVORITE, MF_GRAYED);
#ifndef KAILLERA
        EnableMenuItem(hMenu, ID_FILE_NETPLAY, MF_GRAYED);
#endif
        EnableMenuItem(hMenu, ID_CONTEXT_SELECT_RANDOM, MF_ENABLED);

		if (GameOnlyJukeBox(nGame))
		{
            EnableMenuItem(hMenu, ID_FILE_PLAY_RECORD, MF_GRAYED);
            EnableMenuItem(hMenu, ID_FILE_PLAY_BACK, MF_GRAYED);
            EnableMenuItem(hMenu, ID_FILE_PLAY_RECORD_AVI, MF_GRAYED);
            EnableMenuItem(hMenu, ID_FILE_NETPLAY, MF_GRAYED);
		}
    }
    else
    {
        EnableMenuItem(hMenu, ID_FILE_PLAY, MF_GRAYED);
        EnableMenuItem(hMenu, ID_FILE_PLAY_RECORD, MF_GRAYED);
        EnableMenuItem(hMenu, ID_FILE_PLAY_RECORD_AVI, MF_GRAYED);
        EnableMenuItem(hMenu, ID_GAME_PROPERTIES, MF_GRAYED);
        EnableMenuItem(hMenu, ID_CONTEXT_ADD_FAVORITE, MF_GRAYED);
        EnableMenuItem(hMenu, ID_CONTEXT_DEL_FAVORITE, MF_GRAYED);
        EnableMenuItem(hMenu, ID_FILE_NETPLAY, MF_GRAYED);
        EnableMenuItem(hMenu, ID_CONTEXT_SELECT_RANDOM, MF_GRAYED);
    }

    if (oldControl)
    {
        EnableMenuItem(hMenu, ID_CUSTOMIZE_FIELDS, MF_GRAYED);
        EnableMenuItem(hMenu, ID_GAME_PROPERTIES,  MF_GRAYED);
        EnableMenuItem(hMenu, ID_OPTIONS_DEFAULTS, MF_GRAYED);
    }
}

#define ISSJIS1(c) ((c >= 0x81 && c <= 0x9f) || (c >= 0xe0 && c <= 0xfc))
#define ISSJIS2(c) ((c >= 0x40 && c <= 0xfc) && (c != 0x7f))

/* Add ... to Items in ListView if needed */
static LPCTSTR MakeShortString(HDC hDC, LPCTSTR lpszLong, int nColumnLen, int nOffset)
{
    static const CHAR szThreeDots[]="...";
    static CHAR szShort[MAX_PATH];
    int nStringLen=lstrlen(lpszLong);
    int nAddLen;
    SIZE size;
    int i;
    unsigned char c1, c2;

    GetTextExtentPoint32(hDC,lpszLong, nStringLen, &size);
    if(nStringLen==0 || size.cx + nOffset<=nColumnLen)
        return(lpszLong);

    lstrcpy(szShort,lpszLong);
    GetTextExtentPoint32(hDC,szThreeDots,sizeof(szThreeDots),&size);
    nAddLen=size.cx;

    for( i = nStringLen-1; i > 0; i--)
    {
        c1 = (unsigned char)szShort[i - 1];
        c2 = (unsigned char)szShort[i];

        if (ISSJIS1(c1) && ISSJIS2(c2))
        {
           i--;
           if (i <= 0)
               break;
        }

        szShort[i]=0;
        GetTextExtentPoint32(hDC, szShort,i, &size);
        if(size.cx + nOffset + nAddLen <= nColumnLen)
            break;
    }

    lstrcat(szShort,szThreeDots);

    return(szShort);
}

#undef ISSJIS1
#undef ISSJIS2

/* DrawItem for ListView */
static void DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
    HDC         hDC = lpDrawItemStruct->hDC;
    RECT        rcItem = lpDrawItemStruct->rcItem;
    UINT        uiFlags = ILD_TRANSPARENT;
    HIMAGELIST  hImageList;
    int         nItem = lpDrawItemStruct->itemID;
    COLORREF    clrTextSave, clrBkSave;
    COLORREF    clrImage = GetSysColor(COLOR_WINDOW);
    static CHAR szBuff[MAX_PATH];
    BOOL        bFocus = (GetFocus() == hwndList);
    LPCTSTR     pszText;
    UINT        nStateImageMask;
    BOOL        bSelected;
    LV_COLUMN   lvc;
    LV_ITEM     lvi;
    RECT        rcAllLabels;
    RECT        rcLabel;
    RECT        rcIcon;
    int         offset;
    SIZE        size;
    int         i;
    int         nColumn;
    int         nColumnMax = 0;
    int         nResults = 0;
    int         order[COLUMN_MAX];


    nColumnMax = GetNumColumns(hwndList);

    if (oldControl)
    {
        GetColumnOrder(order);
    }
    else
    {
        /* Get the Column Order and save it */
        ListView_GetColumnOrderArray(hwndList, nColumnMax, order);

        /* Disallow moving column 0 */
        if (order[0] != 0)
        {
            for (i = 0; i < nColumnMax; i++)
            {
                if (order[i] == 0)
                {
                    order[i] = order[0];
                    order[0] = 0;
                }
            }
            ListView_SetColumnOrderArray(hwndList, nColumnMax, order);
        }
    }

    // Labels are offset by a certain amount
    // This offset is related to the width of a space character
    GetTextExtentPoint32(hDC, " ", 1 , &size);
    offset = size.cx * 2;

    lvi.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_STATE;
    lvi.iItem = nItem;
    lvi.iSubItem = order[0];
    lvi.pszText = szBuff;
    lvi.cchTextMax = sizeof(szBuff);
    lvi.stateMask = 0xFFFF;       // get all state flags
    ListView_GetItem(hwndList, &lvi);

    // This makes NO sense, but doesn't work without it?
    strcpy(szBuff, lvi.pszText);

    bSelected =((lvi.state & LVIS_DROPHILITED) || ( (lvi.state & LVIS_SELECTED)
        && ((bFocus) || (GetWindowLong(hwndList, GWL_STYLE) & LVS_SHOWSELALWAYS))));

    ListView_GetItemRect(hwndList, nItem, &rcAllLabels, LVIR_BOUNDS);

    ListView_GetItemRect(hwndList, nItem, &rcLabel, LVIR_LABEL);
    rcAllLabels.left = rcLabel.left;

    if( hBitmap != NULL )
    {
        RECT        rcClient;
        HRGN        rgnBitmap;
        RECT        rcTmpBmp = rcItem;
        RECT        rcFirstItem;
        int         i, j;
        HPALETTE    hPAL;
        HDC         htempDC;
        HBITMAP     oldBitmap;

        htempDC = CreateCompatibleDC(hDC);

        oldBitmap = SelectObject(htempDC, hBitmap);

        GetClientRect(hwndList, &rcClient);
        rcTmpBmp.right = rcClient.right;
        // We also need to check whether it is the last item
        // The update region has to be extended to the bottom if it is
        if( nItem == ListView_GetItemCount(hwndList) - 1 )
            rcTmpBmp.bottom = rcClient.bottom;

        rgnBitmap = CreateRectRgnIndirect(&rcTmpBmp);
        SelectClipRgn(hDC, rgnBitmap);
        DeleteObject(rgnBitmap);

        hPAL = (! hPALbg) ? CreateHalftonePalette(hDC) : hPALbg;

        if(GetDeviceCaps(htempDC, RASTERCAPS) & RC_PALETTE && hPAL != NULL )
        {
            SelectPalette(htempDC, hPAL, FALSE );
            RealizePalette(htempDC);
        }

        ListView_GetItemRect(hwndList, 0, &rcFirstItem, LVIR_BOUNDS);

        for( i = rcFirstItem.left; i < rcClient.right; i += bmDesc.bmWidth )
            for( j = rcFirstItem.top; j < rcClient.bottom; j += bmDesc.bmHeight )
                BitBlt(hDC, i, j, bmDesc.bmWidth, bmDesc.bmHeight, htempDC, 0, 0, SRCCOPY );

        SelectObject(htempDC, oldBitmap);
        DeleteDC(htempDC);

        if (! bmDesc.bmColors)
        {
            DeleteObject(hPAL);
            hPAL = 0;
        }
    }

    SetTextColor(hDC, GetListFontColor());

    if(bSelected)
    {
        HBRUSH hBrush;
        HBRUSH hOldBrush;

        if (bFocus)
        {
            clrTextSave = SetTextColor(hDC, GetSysColor(COLOR_HIGHLIGHTTEXT));
            clrBkSave = SetBkColor(hDC, GetSysColor(COLOR_HIGHLIGHT));
            hBrush = CreateSolidBrush(GetSysColor(COLOR_HIGHLIGHT));
        }
        else
        {
            clrTextSave = SetTextColor(hDC, GetSysColor(COLOR_BTNTEXT));
            clrBkSave = SetBkColor(hDC, GetSysColor(COLOR_BTNFACE));
            hBrush = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
        }

        hOldBrush = SelectObject(hDC, hBrush);
        FillRect(hDC, &rcAllLabels, hBrush);
        SelectObject(hDC, hOldBrush);
        DeleteObject(hBrush);
    }
    else if( hBitmap == NULL )
    {
        HBRUSH hBrush;

        hBrush = CreateSolidBrush( GetSysColor(COLOR_WINDOW));
        FillRect(hDC, &rcAllLabels, hBrush);
        DeleteObject(hBrush);
    }

    if(lvi.state & LVIS_CUT)
    {
        clrImage = GetSysColor(COLOR_WINDOW);
        uiFlags |= ILD_BLEND50;
    }
    else if(bSelected)
    {
        if (bFocus)
            clrImage = GetSysColor(COLOR_HIGHLIGHT);
        else
            clrImage = GetSysColor(COLOR_BTNFACE);

        uiFlags |= ILD_BLEND50;
    }

    nStateImageMask = lvi.state & LVIS_STATEIMAGEMASK;

    if(nStateImageMask)
    {
        int nImage = (nStateImageMask >> 12) - 1;
        hImageList = ListView_GetImageList(hwndList, LVSIL_STATE);
        if(hImageList)
            ImageList_Draw(hImageList, nImage, hDC, rcItem.left, rcItem.top, ILD_TRANSPARENT);
    }

    ListView_GetItemRect(hwndList, nItem, &rcIcon, LVIR_ICON);

    hImageList = ListView_GetImageList(hwndList, LVSIL_SMALL);
    if(hImageList)
    {
        UINT nOvlImageMask = lvi.state & LVIS_OVERLAYMASK;
        if(rcItem.left < rcItem.right-1)
            ImageList_DrawEx(hImageList, lvi.iImage, hDC, rcIcon.left, rcIcon.top,
                16, 16, GetSysColor(COLOR_WINDOW), clrImage, uiFlags | nOvlImageMask);
    }

    ListView_GetItemRect(hwndList, nItem, &rcItem,LVIR_LABEL);

    pszText = MakeShortString(hDC, szBuff,rcItem.right - rcItem.left,  2 * offset);

    rcLabel = rcItem;
    rcLabel.left += offset;
    rcLabel.right -= offset;

    DrawText(hDC, pszText,-1, &rcLabel,
        DT_LEFT | DT_SINGLELINE | DT_NOPREFIX | DT_NOCLIP | DT_VCENTER);

    for(nColumn = 1; nColumn < nColumnMax ; nColumn++)
    {
        int nRetLen;
        UINT nJustify;
        LV_ITEM lvi;

        lvc.mask = LVCF_FMT | LVCF_WIDTH;
        ListView_GetColumn(hwndList, order[nColumn] , &lvc);

        lvi.mask = LVIF_TEXT;
        lvi.iItem = nItem;
        lvi.iSubItem = order[nColumn];
        lvi.pszText = szBuff;
        lvi.cchTextMax = sizeof(szBuff);

        if (ListView_GetItem(hwndList, &lvi) == FALSE)
            continue;

        /* This shouldn't oughtta be, but it's needed!!! */
        strcpy(szBuff, lvi.pszText);

        rcItem.left = rcItem.right;
        rcItem.right += lvc.cx;

        nRetLen = strlen(szBuff);
        if(nRetLen == 0)
            continue;

        pszText = MakeShortString(hDC, szBuff, rcItem.right - rcItem.left, 2 * offset);

        nJustify = DT_LEFT;

        if(pszText == szBuff)
        {
            switch(lvc.fmt & LVCFMT_JUSTIFYMASK)
            {
            case LVCFMT_RIGHT:
                nJustify = DT_RIGHT;
                break;
            case LVCFMT_CENTER:
                nJustify = DT_CENTER;
                break;
            default:
                break;
            }
        }

        rcLabel = rcItem;
        rcLabel.left += offset;
        rcLabel.right -= offset;
        DrawText(hDC, pszText, -1, &rcLabel,
            nJustify | DT_SINGLELINE | DT_NOPREFIX | DT_NOCLIP | DT_VCENTER);
    }

    if(lvi.state & LVIS_FOCUSED && bFocus)
        DrawFocusRect(hDC, &rcAllLabels);

    if(bSelected)
    {
        SetTextColor(hDC, clrTextSave);
        SetBkColor(hDC, clrBkSave);
    }
}

/* Header code - Directional Arrows */
static LRESULT CALLBACK HeaderWndProc( HWND hwnd,  UINT uMsg,  WPARAM wParam,  LPARAM lParam)
{
    switch(uMsg)
    {
    case WM_ERASEBKGND:
        return ListCtrlOnErase( hwnd, (HDC)wParam);

    case WM_PAINT:
        if (current_view_id != ID_VIEW_DETAIL)
            if (ListCtrlOnPaint( hwnd, uMsg ) == 0)
                return 0;
        break;

    case WM_DRAWITEM:
        {
            /* UINT idCtl = (UINT)wParam; */
            LPDRAWITEMSTRUCT lpdis = (LPDRAWITEMSTRUCT)lParam;
            RECT rcLabel;
            int nSavedDC;
            HRGN hrgn;
            int nOffset;
            SIZE size;
            TCHAR tszText[256];
            HD_ITEM hdi;
            UINT uFormat;
            HDC hdc;

            hdc = lpdis->hDC;

            CopyRect(&rcLabel, &(lpdis->rcItem));

            /* save the DC */
            nSavedDC = SaveDC(hdc);

            /* set clip region to column */
            hrgn = CreateRectRgnIndirect(&rcLabel);
            SelectObject(hdc, hrgn);
            DeleteObject(hrgn);

            /* draw the background */
            FillRect(hdc, &rcLabel, GetSysColorBrush(COLOR_3DFACE));

            /* offset the label */
            GetTextExtentPoint32(hdc, _T(" "), 1, &size);
            nOffset = size.cx * 2;

            /* get the column text and format */
            hdi.mask = HDI_TEXT | HDI_FORMAT;
            hdi.pszText = tszText;
            hdi.cchTextMax = 255;
            Header_GetItem(GetDlgItem(hwnd, 0), lpdis->itemID, &hdi);

            /* determine format for drawing label */
            uFormat = DT_SINGLELINE | DT_NOPREFIX | DT_NOCLIP |
                DT_VCENTER | DT_END_ELLIPSIS;

            /* determine justification */
            if (hdi.fmt & HDF_CENTER)
                uFormat |= DT_CENTER;
            else if (hdi.fmt & HDF_RIGHT)
                uFormat |= DT_RIGHT;
            else
                uFormat |= DT_LEFT;

            /* adjust the rect if selected */
            if (lpdis->itemState & ODS_SELECTED)
            {
                rcLabel.left++;
                rcLabel.top += 2;
                rcLabel.right++;
            }

            /* adjust rect for sort arrow */
            if (lpdis->itemID == m_uHeaderSortCol)
            {
                rcLabel.right -= (3 * nOffset);
            }

            rcLabel.left += nOffset;
            rcLabel.right -= nOffset;

            /* draw label */
            if (rcLabel.left < rcLabel.right)
                DrawText(hdc, tszText, -1, &rcLabel, uFormat);

            /* draw the arrow */
            if (lpdis->itemID == m_uHeaderSortCol)
            {
                RECT rcIcon;
                HPEN hpenLight, hpenShadow, hpenOld;

                CopyRect(&rcIcon, &(lpdis->rcItem));

                hpenLight = CreatePen(PS_SOLID, 1, GetSysColor(COLOR_3DHILIGHT));
                hpenShadow = CreatePen(PS_SOLID, 1, GetSysColor(COLOR_3DSHADOW));
                hpenOld = SelectObject(hdc, hpenLight);

                if (m_fHeaderSortAsc)
                {
                    /* draw triangle pointing up */
                    MoveToEx(hdc, rcIcon.right - 2 * nOffset, nOffset - 1, NULL);
                    LineTo(hdc, rcIcon.right - 3 * nOffset / 2,
                        rcIcon.bottom - nOffset);
                    LineTo(hdc, rcIcon.right - 5 * nOffset / 2 - 2,
                        rcIcon.bottom - nOffset);
                    MoveToEx(hdc, rcIcon.right - 5 * nOffset / 2 - 1,
                        rcIcon.bottom - nOffset, NULL);

                    SelectObject(hdc, hpenShadow);
                    LineTo(hdc, rcIcon.right - 2 * nOffset, nOffset - 2);
                }
                else
                {
                    /* draw triangle pointing down */
                    MoveToEx(hdc, rcIcon.right - 3 * nOffset / 2, nOffset - 1,
                        NULL);
                    LineTo(hdc, rcIcon.right - 2 * nOffset - 1,
                        rcIcon.bottom - nOffset);
                    LineTo(hdc, rcIcon.right - 2 * nOffset - 1,
                        rcIcon.bottom - nOffset);
                    MoveToEx(hdc, rcIcon.right - 2 * nOffset - 1,
                        rcIcon.bottom - nOffset, NULL);

                    SelectObject(hdc, hpenShadow);
                    LineTo(hdc, rcIcon.right - 5 * nOffset / 2 - 1,
                        nOffset - 1);
                    LineTo(hdc, rcIcon.right - 3 * nOffset / 2,
                        nOffset - 1);
                }
                SelectObject(hdc, hpenOld);
                DeletePen(hpenLight);
                DeletePen(hpenShadow);
            }
            RestoreDC(hdc, nSavedDC);
            return 1;
        }

    case WM_NOTIFY:
        {
            HD_NOTIFY   *pHDN = (HD_NOTIFY*)lParam;

            // This code is for using bitmap in the background
            // Invalidate the right side of the control when a column is resized
            if (pHDN->hdr.code == HDN_ITEMCHANGINGW ||
                pHDN->hdr.code == HDN_ITEMCHANGINGA)
            {
                if( hBitmap != NULL )
                {
                    RECT rcClient;
                    DWORD dwPos;
                    POINT pt;

                    dwPos = GetMessagePos();
                    pt.x = LOWORD(dwPos);
                    pt.y = HIWORD(dwPos);

                    GetClientRect(hwnd, &rcClient );
                    ScreenToClient( hwnd, &pt );
                    rcClient.left = pt.x;
                    InvalidateRect( hwnd, &rcClient, FALSE );
                }
            }
        }
        break;
    }

    /* message not handled */
    return CallWindowProc(g_lpHeaderWndProc, hwnd, uMsg, wParam, lParam);
}

/* Subclass the Listview Header */
static void Header_Initialize(HWND hwndList)
{
    /* this will subclass the listview (where WM_DRAWITEM gets sent for
       the header control) */
    g_lpHeaderWndProc = (WNDPROC)GetWindowLong(hwndList, GWL_WNDPROC);
    SetWindowLong(hwndList, GWL_WNDPROC, (LONG)HeaderWndProc);
}


/* Set Sort information needed by Header */
static void Header_SetSortInfo(HWND hwndList, int nCol, BOOL bAsc)
{
    HWND hwndHeader;
    HD_ITEM hdi;

    m_uHeaderSortCol = nCol;
    m_fHeaderSortAsc = bAsc;

    hwndHeader = GetDlgItem(hwndList, 0);

    /* change this item to owner draw */
    hdi.mask = HDI_FORMAT;
    Header_GetItem(hwndHeader, nCol, &hdi);
    hdi.fmt |= HDF_OWNERDRAW;
    Header_SetItem(hwndHeader, nCol, &hdi);

    /* repaint the header */
    InvalidateRect(hwndHeader, NULL, TRUE);
}

static BOOL ListCtrlOnErase(HWND hWnd, HDC hDC)
{
    RECT        rcClient;
    HRGN        rgnBitmap;
    HPALETTE    hPAL;

    if (hWnd != hwndList)
        return 1;

    GetClientRect(hWnd, &rcClient);

    if (hBitmap)
    {
        int         i, j;
        HDC         htempDC;
        POINT       ptOrigin;
        POINT       pt = {0,0};
        HBITMAP     hOldBitmap;

        htempDC = CreateCompatibleDC(hDC);
        hOldBitmap = SelectObject(htempDC, hBitmap);

        rgnBitmap = CreateRectRgnIndirect(&rcClient);
        SelectClipRgn(hDC, rgnBitmap);
        DeleteObject(rgnBitmap);

        hPAL = (! hPALbg) ? CreateHalftonePalette(hDC) : hPALbg;

        if(GetDeviceCaps(htempDC, RASTERCAPS) & RC_PALETTE && hPAL != NULL )
        {
            SelectPalette(htempDC, hPAL, FALSE );
            RealizePalette(htempDC);
        }

        // Get x and y offset
        ClientToScreen(hWnd, &pt);
        GetDCOrgEx(hDC, &ptOrigin);
        ptOrigin.x -= pt.x;
        ptOrigin.y -= pt.y;
        ptOrigin.x = -GetScrollPos( hWnd, SB_HORZ );
        ptOrigin.y = -GetScrollPos( hWnd, SB_VERT );

        for( i = ptOrigin.x; i < rcClient.right; i += bmDesc.bmWidth )
            for( j = ptOrigin.y; j < rcClient.bottom; j += bmDesc.bmHeight )
                BitBlt(hDC, i, j, bmDesc.bmWidth, bmDesc.bmHeight, htempDC, 0, 0, SRCCOPY );

        SelectObject(htempDC, hOldBitmap);
        DeleteDC(htempDC);

        if (! bmDesc.bmColors)
        {
            DeleteObject(hPAL);
            hPAL = 0;
        }
    }
    else    /* Use standard control face color */
    {
        HBRUSH hBrush = CreateSolidBrush(GetSysColor(COLOR_WINDOW));
        HBRUSH hOldBrush;

        hOldBrush = SelectObject(hDC, hBrush);
        FillRect(hDC, &rcClient, hBrush);
        SelectObject(hDC, hOldBrush);
        DeleteObject(hBrush);
    }

    return TRUE;
}

static BOOL ListCtrlOnPaint(HWND hWnd, UINT uMsg)
{
    PAINTSTRUCT ps;
    HDC         hDC;
    RECT        rcClip, rcClient;
    HDC         memDC;
    HBITMAP     bitmap;
    HBITMAP     hOldBitmap;

    if (hWnd != hwndList)
        return 1;

    hDC = BeginPaint(hWnd, &ps);
    rcClient = ps.rcPaint;

    GetClipBox(hDC, &rcClip );
    GetClientRect(hWnd, &rcClient);

    // Create a compatible memory DC
    memDC = CreateCompatibleDC( hDC );

    // Select a compatible bitmap into the memory DC
    bitmap = CreateCompatibleBitmap( hDC, rcClient.right - rcClient.left,
                                     rcClient.bottom - rcClient.top );
    hOldBitmap = SelectObject( memDC, bitmap );

    BitBlt( memDC, rcClip.left, rcClip.top,
            rcClip.right - rcClip.left, rcClip.bottom - rcClip.top,
            hDC, rcClip.left, rcClip.top, SRCCOPY );

    // First let the control do its default drawing.
    CallWindowProc(g_lpHeaderWndProc, hWnd, uMsg, (WPARAM)memDC, 0);

    // Draw bitmap in the background if one has been set
    if(  hBitmap != NULL )
    {
        HPALETTE    hPAL;
        HDC maskDC;
        HBITMAP maskBitmap;
        HDC tempDC;
        HDC imageDC;
        HBITMAP bmpImage;
        HBITMAP hOldBmpImage;
        HBITMAP hOldMaskBitmap;
        HBITMAP hOldHBitmap;
        int i, j;
        POINT ptOrigin;
        POINT pt = {0,0};

        // Now create a mask
        maskDC = CreateCompatibleDC(hDC);

        // Create monochrome bitmap for the mask
        maskBitmap = CreateBitmap( rcClient.right - rcClient.left,
                                   rcClient.bottom - rcClient.top,
                                   1, 1, NULL );

        hOldMaskBitmap = SelectObject( maskDC, maskBitmap );
        SetBkColor( memDC, GetSysColor( COLOR_WINDOW ) );

        // Create the mask from the memory DC
        BitBlt( maskDC, 0, 0, rcClient.right - rcClient.left,
                rcClient.bottom - rcClient.top, memDC,
                rcClient.left, rcClient.top, SRCCOPY );

        tempDC = CreateCompatibleDC(hDC);
        hOldHBitmap = SelectObject(tempDC, hBitmap );

        imageDC = CreateCompatibleDC( hDC );
        bmpImage = CreateCompatibleBitmap( hDC,
                                           rcClient.right - rcClient.left,
                                           rcClient.bottom - rcClient.top );
        hOldBmpImage = SelectObject(imageDC, bmpImage );

        hPAL = (! hPALbg) ? CreateHalftonePalette(hDC) : hPALbg;

        if( GetDeviceCaps(hDC, RASTERCAPS) & RC_PALETTE && hPAL != NULL )
        {
            SelectPalette(hDC, hPAL, FALSE );
            RealizePalette(hDC);
            SelectPalette(imageDC, hPAL, FALSE );
        }

        // Get x and y offset
        ClientToScreen(hWnd, &pt);
        GetDCOrgEx(hDC, &ptOrigin);
        ptOrigin.x -= pt.x;
        ptOrigin.y -= pt.y;
        ptOrigin.x = -GetScrollPos( hWnd, SB_HORZ );
        ptOrigin.y = -GetScrollPos( hWnd, SB_VERT );

        // Draw bitmap in tiled manner to imageDC
        for( i = ptOrigin.x; i < rcClient.right; i += bmDesc.bmWidth )
            for( j = ptOrigin.y; j < rcClient.bottom; j += bmDesc.bmHeight )
                BitBlt( imageDC,  i, j, bmDesc.bmWidth, bmDesc.bmHeight,
                        tempDC, 0, 0, SRCCOPY );

        // Set the background in memDC to black. Using SRCPAINT with black
        // and any other color results in the other color, thus making black
        // the transparent color
        SetBkColor(memDC, RGB(0,0,0));
        SetTextColor(memDC, RGB(255,255,255));
        BitBlt( memDC, rcClip.left, rcClip.top, rcClip.right - rcClip.left,
                rcClip.bottom - rcClip.top,
                maskDC, rcClip.left, rcClip.top, SRCAND);

        // Set the foreground to black. See comment above.
        SetBkColor(imageDC, RGB(255,255,255));
        SetTextColor(imageDC, RGB(0,0,0));
        BitBlt( imageDC, rcClip.left, rcClip.top, rcClip.right - rcClip.left,
                rcClip.bottom - rcClip.top,
                maskDC, rcClip.left, rcClip.top, SRCAND);

        // Combine the foreground with the background
        BitBlt( imageDC, rcClip.left, rcClip.top, rcClip.right - rcClip.left,
                rcClip.bottom - rcClip.top,
                memDC, rcClip.left, rcClip.top, SRCPAINT);

        // Draw the final image to the screen
        BitBlt( hDC, rcClip.left, rcClip.top, rcClip.right - rcClip.left,
                rcClip.bottom - rcClip.top,
                imageDC, rcClip.left, rcClip.top, SRCCOPY );

        SelectObject( maskDC, hOldMaskBitmap );
        SelectObject( tempDC, hOldHBitmap );
        SelectObject( imageDC, hOldBmpImage );

        DeleteDC(maskDC);
        DeleteDC(imageDC);
        DeleteDC(tempDC);
        DeleteObject(bmpImage);
        DeleteObject(maskBitmap);
        if (!hPALbg)
        {
            DeleteObject(hPAL);
        }
    }
    else
    {
        BitBlt( hDC,  rcClip.left, rcClip.top, rcClip.right - rcClip.left,
                rcClip.bottom - rcClip.top,
                memDC, rcClip.left, rcClip.top, SRCCOPY );
    }
    SelectObject( memDC, hOldBitmap );
    DeleteObject(bitmap);
    DeleteDC(memDC);
    EndPaint(hWnd, &ps);
    return 0;
}
