/*********************************************************************

  drawgfx.h

  Generic graphic functions.

*********************************************************************/

#ifndef DRAWGFX_H
#define DRAWGFX_H

#define MAX_GFX_PLANES	8
#define MAX_GFX_SIZE	64

#define RGN_FRAC(num,den)   (0x80000000 | (((num) & 0x0f) << 27) | (((den) & 0x0f) << 23))
#define IS_FRAC(offset)     ((offset) & 0x80000000)
#define FRAC_NUM(offset)    (((offset) >> 27) & 0x0f)
#define FRAC_DEN(offset)    (((offset) >> 23) & 0x0f)
#define FRAC_OFFSET(offset) ((offset) & 0x007fffff)


struct GfxLayout
{
	UINT16 width,height;				/* width and height (in pixels) of chars/sprites */
	UINT32 total;						/* total numer of chars/sprites in the rom */
	UINT16 planes;						/* number of bitplanes */
	UINT32 planeoffset[MAX_GFX_PLANES];	/* start of every bitplane (in bits) */
	UINT32 xoffset[MAX_GFX_SIZE];		/* position of the bit corresponding to the pixel */
	UINT32 yoffset[MAX_GFX_SIZE];		/* of the given coordinates */
	UINT16 charincrement;				/* distance between two consecutive characters/sprites (in bits) */
};

struct GfxElement
{
	int width,height;

	unsigned int total_elements;	/* total number of characters/sprites */
	int color_granularity;			/* number of colors for each color code */
									/* (for example, 4 for 2 bitplanes gfx) */
	unsigned short *colortable;		/* map color codes to screen pens */
									/* if this is 0, the function does a verbatim copy */
	int total_colors;
	unsigned int *pen_usage;		/* an array of total_elements ints. */
									/* It is a table of the pens each character uses */
									/* (bit 0 = pen 0, and so on). This is used by */
									/* drawgfgx() to do optimizations like skipping */
									/* drawing of a totally transparent characters */
	unsigned char *gfxdata;			/* pixel data */
	int line_modulo;				/* amount to add to get to the next line (usually = width) */
	int char_modulo;				/* = line_modulo * height */
};

struct GfxDecodeInfo
{
	int memory_region;		/* memory region where the data resides (usually 1) */
							/* -1 marks the end of the array */
	int start;				/* beginning of data to decode */
	struct GfxLayout *gfxlayout;
	int color_codes_start;	/* offset in the color lookup table where color codes start */
	int total_color_codes;	/* total number of color codes */
};


struct rectangle
{
	int min_x,max_x;
	int min_y,max_y;
};



enum
{
	TRANSPARENCY_NONE,			/* opaque with remapping */
	TRANSPARENCY_PEN,			/* single pen transparency with remapping */
	TRANSPARENCY_FONT,
	TRANSPARENCY_FONT2,
	TRANSPARENCY_MODES			/* total number of modes; must be last */
};

void plot_pixel(struct osd_bitmap *bitmap,int x,int y,int pen);
int read_pixel(struct osd_bitmap *bitmap,int x,int y);
void plot_box(struct osd_bitmap *bitmap,int x,int y,int width,int height,int pen,int trans);

void decodechar(struct GfxElement *gfx,int num,const unsigned char *src,const struct GfxLayout *gl);
struct GfxElement *decodegfx(const unsigned char *src,const struct GfxLayout *gl);
void freegfx(struct GfxElement *gfx);
void drawgfx(struct osd_bitmap *dest,const struct GfxElement *gfx,
		unsigned int code,unsigned int color,int flipx,int flipy,int sx,int sy,
		const struct rectangle *clip,int transparency,int transparent_color);
void copybitmap(struct osd_bitmap *dest,struct osd_bitmap *src,int flipx,int flipy,int sx,int sy,
		const struct rectangle *clip,int transparency,int transparent_color);
void fillbitmap(struct osd_bitmap *dest,UINT32 pen);

#endif
