/***************************************************************************

  inptport.c

  Input ports handling

TODO:	remove the 1 analog device per port limitation
		support for inputports producing interrupts
		support for extra "real" hardware (PC throttle's, spinners etc)

***************************************************************************/

#include "driver.h"
#include <math.h>
#include "ui_text.h"
#include "drivers/neogeo.h"

#define NOLEGACY

#ifdef KAILLERA
#include "kailleraclient.h"
#include "KailleraChat.h"
#include "uclock.h"
int kBitsPlayValues[2][16][3];
int kFirstMsg;
static unsigned short kDefValues[MAX_INPUT_PORTS];
#endif /* KAILLERA */

/* Use the MRU code for 4way joysticks */
#define MRU_JOYSTICK

/* header identifying the version of the game.cfg file */
/* mame 0.36b11 */
#define MAMECFGSTRING_V5 "MAMECFG\5"
#define MAMEDEFSTRING_V5 "MAMEDEF\4"

/* mame 0.36b12 with multi key/joy extension */
#define MAMECFGSTRING_V6 "MAMECFG\6"
#define MAMEDEFSTRING_V6 "MAMEDEF\5"

/* mame 0.36b13 with and/or/not combination */
#define MAMECFGSTRING_V7 "MAMECFG\7"
#define MAMEDEFSTRING_V7 "MAMEDEF\6"

/* mame 0.36b16 with key/joy merge */
#define MAMECFGSTRING_V8 "MAMECFG\x8"
#define MAMEDEFSTRING_V8 "MAMEDEF\7"

extern void *record;
extern void *playback;
extern short cpu_clocks[];

extern unsigned int dispensed_tickets;
extern unsigned int coins[COIN_COUNTERS];
extern unsigned int lastcoin[COIN_COUNTERS];
extern unsigned int coinlockedout[COIN_COUNTERS];

static unsigned short input_port_value[MAX_INPUT_PORTS];

/* Assuming a maxium of one analog input device per port BW 101297 */
static struct InputPort *input_analog[MAX_INPUT_PORTS];
static int input_analog_current_value[MAX_INPUT_PORTS],input_analog_previous_value[MAX_INPUT_PORTS];
static int input_analog_init[MAX_INPUT_PORTS];

static int mouse_delta_x[OSD_MAX_JOY_ANALOG], mouse_delta_y[OSD_MAX_JOY_ANALOG];
static int analog_current_x[OSD_MAX_JOY_ANALOG], analog_current_y[OSD_MAX_JOY_ANALOG];
static int analog_previous_x[OSD_MAX_JOY_ANALOG], analog_previous_y[OSD_MAX_JOY_ANALOG];

static int _inputport_max = MAX_INPUT_PORTS;
static int warming_up = 0;

int show_input = 0;


/***************************************************************************/

const static UINT16 button_mask[] =
{
	0x00,    /* None    */
	0x30,    /* A+B     */
	0x50,    /* A+C     */
	0x90,    /* A+D     */
	0x60,    /* B+C     */
	0xa0,    /* B+D     */
	0xc0,    /* C+D     */
	0x70,    /* A+B+C   */
	0xd0,    /* A+B+D   */
	0xe0,    /* B+C+D   */
	0xf0     /* A+B+C+D */
};

const static int start[MAX_PLAYERS] = { 8, 20 };
const static int left[MAX_PLAYERS]  = { 2, 14 };

UINT16 custom_button[MAX_PLAYERS][MAX_CUSTOM_BUTTONS];
UINT16 autofire_mask[MAX_PLAYERS][MAX_CUSTOM_BUTTONS];
UINT16 autofire_delay[MAX_PLAYERS][MAX_CUSTOM_BUTTONS];
UINT16 combocount[MAX_PLAYERS];

static UINT16 shadow[MAX_PLAYERS][MAX_CUSTOM_BUTTONS];
static UINT16 autofire_skip;
static UINT16 comboentries;
static UINT16 *comboa;
static UINT16 *rcomboa;
static UINT16 *combo[MAX_PLAYERS];
static UINT16 setcombo[MAX_PLAYERS][MAX_CUSTOM_BUTTONS];
static INT16  combondx[MAX_PLAYERS * MAX_CUSTOM_BUTTONS];
static int    docombo[2] = { 0, 0 };

struct BRANCH_ITEM
{
	int begin;
	int which_bh;
	int bh_mask;
};
struct BRANCH_ITEM branch[MAX_PLAYERS * MAX_CUSTOM_BUTTONS * MAX_BRANCH_NUM];

static UINT16 branchentries = 0;
static UINT16 branchndx[MAX_PLAYERS * MAX_CUSTOM_BUTTONS][MAX_BRANCH_NUM];

int record_combo = 0;

UINT8 command_buf[COMMAND_LOG_BUFSIZE];
int command_counter = 0;

/***************************************************************************/

char *GetComboDir(void);

static void make_input_log(void);

static void do_autofire(UINT16 inparr[], int pressed[]);
static void do_macro_ex(UINT16 inparr[], int pressed[]);
static void save_combo(void);

/***************************************************************************

  Configuration load/save

***************************************************************************/

struct ipd inputport_defaults[] =
{
	{ IPT_UI_CONFIGURE,                              portname(ConfigMenu),          SEQ_DEF_1(KEYCODE_TAB) },
	{ IPT_UI_ON_SCREEN_DISPLAY,                      portname(OnScreenDisplay),     SEQ_DEF_1(KEYCODE_OPENBRACE) },
	{ IPT_UI_PAUSE,                                  portname(Pause),               SEQ_DEF_1(KEYCODE_P) },
	{ IPT_UI_SINGLE_STEP,                            portname(SingleStep),          SEQ_DEF_1(KEYCODE_O) },
	{ IPT_UI_RESET_MACHINE,                          portname(ResetGame),           SEQ_DEF_1(KEYCODE_F3) },
	{ IPT_UI_SHOW_GFX,                               portname(ShowGfx),             SEQ_DEF_1(KEYCODE_F4) },
	{ IPT_UI_FRAMESKIP_DEC,                          portname(FrameskipDec),        SEQ_DEF_1(KEYCODE_F8) },
	{ IPT_UI_FRAMESKIP_INC,                          portname(FrameskipInc),        SEQ_DEF_1(KEYCODE_F9) },
	{ IPT_UI_THROTTLE,                               portname(Throttole),           SEQ_DEF_1(KEYCODE_F10) },
	{ IPT_UI_SHOW_FPS,                               portname(ShowFPS),             SEQ_DEF_3(KEYCODE_F11, CODE_NOT, KEYCODE_LSHIFT) },
	{ IPT_UI_SAVE_STATE,                             portname(SaveState),           SEQ_DEF_3(KEYCODE_F6, CODE_NOT, KEYCODE_LSHIFT) },
	{ IPT_UI_LOAD_STATE,                             portname(LoadState),           SEQ_DEF_3(KEYCODE_F7, CODE_NOT, KEYCODE_LSHIFT) },
	{ IPT_UI_SAVE_COMBO,                             portname(SaveCombo),           SEQ_DEF_2(KEYCODE_F6, KEYCODE_LSHIFT) },
	{ IPT_UI_LOAD_COMBO,                             portname(LoadCombo),           SEQ_DEF_2(KEYCODE_F7, KEYCODE_LSHIFT) },
	{ IPT_UI_SHOW_PROFILER,                          portname(ShowProfiler),        SEQ_DEF_2(KEYCODE_F11, KEYCODE_LSHIFT) },
	{ IPT_UI_SNAPSHOT,                               portname(SaveSnapshot),        SEQ_DEF_1(KEYCODE_F12) },
	{ IPT_UI_TOGGLE_CHEAT,                           portname(ToggleCheat),         SEQ_DEF_1(KEYCODE_F5) },
	{ IPT_UI_TOGGLE_DISPLAY,                         portname(ChangeScreenMode),    SEQ_DEF_2(KEYCODE_LALT, KEYCODE_ENTER) },
	{ IPT_UI_COMMAND,                                portname(ShowCommandList),     SEQ_DEF_1(KEYCODE_END) },
	{ IPT_UI_CHEAT,                                  portname(ShowCheatList),       SEQ_DEF_1(KEYCODE_RCONTROL) },
	{ IPT_UI_TIME,                                   portname(ShowTime),            SEQ_DEF_1(KEYCODE_CLOSEBRACE) },
	{ IPT_UI_SHOW_INPUT_LOG,                         portname(ShowInputLog),        SEQ_DEF_1(KEYCODE_DEL) },
	{ IPT_UI_WAVE_RECORD,                            portname(ToggleWaveRecording), SEQ_DEF_1(KEYCODE_BACKSPACE) },
	{ IPT_UI_SOUND_TEST,                             portname(SoundTest),           SEQ_DEF_1(KEYCODE_INSERT) },
	{ IPT_UI_OPERATION_OBJ,                          portname(OperationObj),        SEQ_DEF_1(KEYCODE_HOME) },
#ifdef KAILLERA
	{ IPT_UI_KAILLERA_CHAT,                          portname(KailleraChatOpen),    SEQ_DEF_1(KEYCODE_PGUP) },
	{ IPT_UI_CHAT_MOVE,	                             portname(KailleraChatMove),    SEQ_DEF_1(KEYCODE_PGDN) },
#endif
	{ IPT_UI_UP,                                     portname(UIUp),                SEQ_DEF_3(KEYCODE_UP, CODE_OR, JOYCODE_1_UP) },
	{ IPT_UI_DOWN,                                   portname(UIDown),              SEQ_DEF_3(KEYCODE_DOWN, CODE_OR, JOYCODE_1_DOWN) },
	{ IPT_UI_LEFT,                                   portname(UILeft),              SEQ_DEF_3(KEYCODE_LEFT, CODE_OR, JOYCODE_1_LEFT) },
	{ IPT_UI_RIGHT,                                  portname(UIRight),             SEQ_DEF_3(KEYCODE_RIGHT, CODE_OR, JOYCODE_1_RIGHT) },
	{ IPT_UI_SELECT,                                 portname(UISelect),            SEQ_DEF_3(KEYCODE_ENTER, CODE_OR, JOYCODE_1_BUTTON1) },
	{ IPT_UI_CANCEL,                                 portname(UICancel),            SEQ_DEF_1(KEYCODE_ESC) },

	{ IPT_COIN1,                                     portname(Coin1),               SEQ_DEF_3(KEYCODE_3, CODE_OR, JOYCODE_1_BUTTON9) },
	{ IPT_COIN2,                                     portname(Coin2),               SEQ_DEF_3(KEYCODE_4, CODE_OR, JOYCODE_2_BUTTON9) },
	{ IPT_COIN3,                                     portname(Coin3),               SEQ_DEF_1(KEYCODE_5) },
	{ IPT_START1,                                    portname(P1Start),             SEQ_DEF_3(KEYCODE_1, CODE_OR, JOYCODE_1_BUTTON10) },
	{ IPT_START2,                                    portname(P2Start),             SEQ_DEF_3(KEYCODE_2, CODE_OR, JOYCODE_2_BUTTON10) },
	{ IPT_SELECT1,                                   portname(P1Select),            SEQ_DEF_3(KEYCODE_6, CODE_OR, JOYCODE_1_BUTTON9) },
	{ IPT_SELECT2,                                   portname(P2Select),            SEQ_DEF_3(KEYCODE_7, CODE_OR, JOYCODE_2_BUTTON9) },

	{ IPT_JOYSTICK_UP    | IPF_PLAYER1,              portname(P1Up),                SEQ_DEF_3(KEYCODE_UP, CODE_OR, JOYCODE_1_UP) },
	{ IPT_JOYSTICK_DOWN  | IPF_PLAYER1,              portname(P1Down),              SEQ_DEF_3(KEYCODE_DOWN, CODE_OR, JOYCODE_1_DOWN) },
	{ IPT_JOYSTICK_LEFT  | IPF_PLAYER1,              portname(P1Left),              SEQ_DEF_3(KEYCODE_LEFT, CODE_OR, JOYCODE_1_LEFT) },
	{ IPT_JOYSTICK_RIGHT | IPF_PLAYER1,              portname(P1Right),             SEQ_DEF_3(KEYCODE_RIGHT, CODE_OR, JOYCODE_1_RIGHT) },
	{ IPT_BUTTON1        | IPF_PLAYER1,              portname(P1ButtonA),           SEQ_DEF_3(KEYCODE_A, CODE_OR, JOYCODE_1_BUTTON1) },
	{ IPT_BUTTON2        | IPF_PLAYER1,              portname(P1ButtonB),           SEQ_DEF_3(KEYCODE_S, CODE_OR, JOYCODE_1_BUTTON2) },
	{ IPT_BUTTON3        | IPF_PLAYER1,              portname(P1ButtonC),           SEQ_DEF_3(KEYCODE_D, CODE_OR, JOYCODE_1_BUTTON3) },
	{ IPT_BUTTON4        | IPF_PLAYER1,              portname(P1ButtonD),           SEQ_DEF_3(KEYCODE_F, CODE_OR, JOYCODE_1_BUTTON4) },
	{ IPT_BUTTON5        | IPF_PLAYER1,              portname(P1Custom1),           SEQ_DEF_3(KEYCODE_Z, CODE_OR, JOYCODE_1_BUTTON5) },
	{ IPT_BUTTON6        | IPF_PLAYER1,              portname(P1Custom2),           SEQ_DEF_3(KEYCODE_X, CODE_OR, JOYCODE_1_BUTTON6) },
	{ IPT_BUTTON7        | IPF_PLAYER1,              portname(P1Custom3),           SEQ_DEF_3(KEYCODE_C, CODE_OR, JOYCODE_1_BUTTON7) },
	{ IPT_BUTTON8        | IPF_PLAYER1,              portname(P1Custom4),           SEQ_DEF_3(KEYCODE_V, CODE_OR, JOYCODE_1_BUTTON8) },

	{ IPT_JOYSTICK_UP    | IPF_PLAYER2,              portname(P2Up),                SEQ_DEF_3(KEYCODE_8_PAD, CODE_OR, JOYCODE_2_UP) },
	{ IPT_JOYSTICK_DOWN  | IPF_PLAYER2,              portname(P2Down),              SEQ_DEF_3(KEYCODE_2_PAD, CODE_OR, JOYCODE_2_DOWN) },
	{ IPT_JOYSTICK_LEFT  | IPF_PLAYER2,              portname(P2Left),              SEQ_DEF_3(KEYCODE_4_PAD, CODE_OR, JOYCODE_2_LEFT) },
	{ IPT_JOYSTICK_RIGHT | IPF_PLAYER2,              portname(P2Right),             SEQ_DEF_3(KEYCODE_6_PAD, CODE_OR, JOYCODE_2_RIGHT) },
	{ IPT_BUTTON1        | IPF_PLAYER2,              portname(P2ButtonA),           SEQ_DEF_3(KEYCODE_G, CODE_OR, JOYCODE_2_BUTTON1) },
	{ IPT_BUTTON2        | IPF_PLAYER2,              portname(P2ButtonB),           SEQ_DEF_3(KEYCODE_H, CODE_OR, JOYCODE_2_BUTTON2) },
	{ IPT_BUTTON3        | IPF_PLAYER2,              portname(P2ButtonC),           SEQ_DEF_3(KEYCODE_J, CODE_OR, JOYCODE_2_BUTTON3) },
	{ IPT_BUTTON4        | IPF_PLAYER2,              portname(P2ButtonD),           SEQ_DEF_3(KEYCODE_K, CODE_OR, JOYCODE_2_BUTTON4) },
	{ IPT_BUTTON5        | IPF_PLAYER2,              portname(P2Custom1),           SEQ_DEF_1(JOYCODE_2_BUTTON5) },
	{ IPT_BUTTON6        | IPF_PLAYER2,              portname(P2Custom2),           SEQ_DEF_1(JOYCODE_2_BUTTON6) },
	{ IPT_BUTTON7        | IPF_PLAYER2,              portname(P2Custom3),           SEQ_DEF_1(JOYCODE_2_BUTTON7) },
	{ IPT_BUTTON8        | IPF_PLAYER2,              portname(P2Custom4),           SEQ_DEF_1(JOYCODE_2_BUTTON8) },

	{ IPT_TRACKBALL_Y | IPF_PLAYER1,                 portname(P1TrackYUp),          SEQ_DEF_3(KEYCODE_UP, CODE_OR, JOYCODE_1_UP) },
	{ (IPT_TRACKBALL_Y | IPF_PLAYER1)+IPT_EXTENSION, portname(P1TrackYDown),        SEQ_DEF_3(KEYCODE_DOWN, CODE_OR, JOYCODE_1_DOWN) },
	{ IPT_TRACKBALL_X | IPF_PLAYER1,                 portname(P1TrackXLeft),        SEQ_DEF_3(KEYCODE_LEFT, CODE_OR, JOYCODE_1_LEFT) },
	{ (IPT_TRACKBALL_X | IPF_PLAYER1)+IPT_EXTENSION, portname(P1TrackXRight),       SEQ_DEF_3(KEYCODE_RIGHT, CODE_OR, JOYCODE_1_RIGHT) },

	{ IPT_TRACKBALL_Y | IPF_PLAYER2,                 portname(P2TrackYUp),          SEQ_DEF_3(KEYCODE_8, CODE_OR, JOYCODE_2_UP) },
	{ (IPT_TRACKBALL_Y | IPF_PLAYER2)+IPT_EXTENSION, portname(P2TrackYDown),        SEQ_DEF_3(KEYCODE_2, CODE_OR, JOYCODE_2_DOWN) },
	{ IPT_TRACKBALL_X | IPF_PLAYER2,                 portname(P2TrackXLeft),        SEQ_DEF_3(KEYCODE_4, CODE_OR, JOYCODE_2_LEFT) },
	{ (IPT_TRACKBALL_X | IPF_PLAYER2)+IPT_EXTENSION, portname(P2TrackXRight),       SEQ_DEF_3(KEYCODE_6, CODE_OR, JOYCODE_2_RIGHT) },

	{ IPT_UNKNOWN,                                   portname(Unknown),             SEQ_DEF_0 },
	{ IPT_END,                                       0,                             SEQ_DEF_0 }
};

struct ipd inputport_defaults_backup[sizeof(inputport_defaults)/sizeof(struct ipd)];

/***************************************************************************/
/* Generic IO */

static int readint(void *f,UINT32 *num)
{
	unsigned i;

	*num = 0;
	for (i = 0;i < sizeof(UINT32);i++)
	{
		unsigned char c;


		*num <<= 8;
		if (osd_fread(f,&c,1) != 1)
			return -1;
		*num |= c;
	}

	return 0;
}

static void writeint(void *f,UINT32 num)
{
	unsigned i;

	for (i = 0;i < sizeof(UINT32);i++)
	{
		unsigned char c;


		c = (num >> 8 * (sizeof(UINT32)-1)) & 0xff;
		osd_fwrite(f,&c,1);
		num <<= 8;
	}
}

static int readword(void *f,UINT16 *num)
{
	unsigned i;
	int res;

	res = 0;
	for (i = 0;i < sizeof(UINT16);i++)
	{
		unsigned char c;


		res <<= 8;
		if (osd_fread(f,&c,1) != 1)
			return -1;
		res |= c;
	}

	*num = res;
	return 0;
}

static void writeword(void *f,UINT16 num)
{
	unsigned i;

	for (i = 0;i < sizeof(UINT16);i++)
	{
		unsigned char c;


		c = (num >> 8 * (sizeof(UINT16)-1)) & 0xff;
		osd_fwrite(f,&c,1);
		num <<= 8;
	}
}

#ifndef NOLEGACY
#include "legacy.h"
#endif

static int seq_read_ver_8(void* f, InputSeq* seq)
{
	int j,len;
	UINT32 i;
	UINT16 w;

	if (readword(f,&w) != 0)
		return -1;

	len = w;
	seq_set_0(seq);
	for(j=0;j<len;++j)
	{
		if (readint(f,&i) != 0)
 			return -1;
		(*seq)[j] = savecode_to_code(i);
 	}

 	return 0;
  }

static int seq_read(void* f, InputSeq* seq, int ver)
  {
#ifdef NOLEGACY
	if (ver==8)
		return seq_read_ver_8(f,seq);
#else
	switch (ver) {
		case 5 : return seq_read_ver_5(f,seq);
		case 6 : return seq_read_ver_6(f,seq);
		case 7 : return seq_read_ver_7(f,seq);
		case 8 : return seq_read_ver_8(f,seq);
	}
#endif
	return -1;
  }

static void seq_write(void* f, InputSeq* seq)
  {
	int j,len;
        for(len=0;len<SEQ_MAX;++len)
		if ((*seq)[len] == CODE_NONE)
			break;
	writeword(f,len);
	for(j=0;j<len;++j)
		writeint(f, code_to_savecode( (*seq)[j] ));
  }

/***************************************************************************/
/* Load */

static void load_default_keys(void)
{
	void *f;


	osd_customize_inputport_defaults(inputport_defaults);
	memcpy(inputport_defaults_backup,inputport_defaults,sizeof(inputport_defaults));

	if ((f = osd_fopen("default",0,OSD_FILETYPE_CONFIG,0)) != 0)
	{
		char buf[8];
		int version;

		/* read header */
		if (osd_fread(f,buf,8) != 8)
			goto getout;

#ifndef NOLEGACY
		if (memcmp(buf,MAMEDEFSTRING_V5,8) == 0)
			version = 5;
		else if (memcmp(buf,MAMEDEFSTRING_V6,8) == 0)
			version = 6;
		else if (memcmp(buf,MAMEDEFSTRING_V7,8) == 0)
			version = 7;
		else
#endif
		if (memcmp(buf,MAMEDEFSTRING_V8,8) == 0)
			version = 8;
		else
			goto getout;	/* header invalid */

		for (;;)
		{
			UINT32 type;
			InputSeq def_seq;
			InputSeq seq;
			int i;

			if (readint(f,&type) != 0)
				goto getout;

			if (seq_read(f,&def_seq,version)!=0)
				goto getout;
			if (seq_read(f,&seq,version)!=0)
				goto getout;

			i = 0;
			while (inputport_defaults[i].type != IPT_END)
			{
				if (inputport_defaults[i].type == type)
				{
					/* load stored settings only if the default hasn't changed */
					if (seq_cmp(&inputport_defaults[i].seq,&def_seq)==0)
						seq_copy(&inputport_defaults[i].seq,&seq);
				}

				i++;
			}
		}

getout:
		osd_fclose(f);
	}
}

static void save_default_keys(void)
{
	void *f;


	if ((f = osd_fopen("default",0,OSD_FILETYPE_CONFIG,1)) != 0)
	{
		int i;


		/* write header */
		osd_fwrite(f,MAMEDEFSTRING_V8,8);

		i = 0;
		while (inputport_defaults[i].type != IPT_END)
		{
			writeint(f,inputport_defaults[i].type);

			seq_write(f,&inputport_defaults_backup[i].seq);
			seq_write(f,&inputport_defaults[i].seq);

			i++;
		}

		osd_fclose(f);
	}
	memcpy(inputport_defaults,inputport_defaults_backup,sizeof(inputport_defaults_backup));
}

static int input_port_read_ver_8(void *f,struct InputPort *in)
{
	UINT32 i;
	UINT16 w;
	if (readint(f,&i) != 0)
		return -1;
	in->type = i;

	if (readword(f,&w) != 0)
		return -1;
	in->mask = w;

	if (readword(f,&w) != 0)
		return -1;
	in->default_value = w;

	if (seq_read_ver_8(f,&in->seq) != 0)
		return -1;

	return 0;
}

static int input_port_read(void *f,struct InputPort *in, int ver)
{
#ifdef NOLEGACY
	if (ver==8)
		return input_port_read_ver_8(f,in);
#else
	switch (ver) {
		case 5 : return	input_port_read_ver_5(f,in);
		case 6 : return	input_port_read_ver_6(f,in);
		case 7 : return	input_port_read_ver_7(f,in);
		case 8 : return	input_port_read_ver_8(f,in);
	}
#endif
	return -1;
}

static void input_port_write(void *f,struct InputPort *in)
{
	writeint(f,in->type);
	writeword(f,in->mask);
	writeword(f,in->default_value);
	seq_write(f,&in->seq);
}


int load_input_port_settings(void)
{
	void *f;
	int i;

	for (i = 0; i < MAX_CPU; i++)
	{
		if (options.use_overclock)
			cpu_clocks[i] = 200;
		else
			cpu_clocks[i] = 100;
	}

	load_default_keys();

	if ((f = osd_fopen(Machine->gamedrv->name,0,OSD_FILETYPE_CONFIG,0)) != 0)
	{
		struct InputPort *in;
		unsigned int total,savedtotal;
		char buf[8];
		int version;
		int player, button;

		in = Machine->input_ports_default;

		/* calculate the size of the array */
		total = 0;
		while (in->type != IPT_END)
		{
			total++;
			in++;
		}

		/* read header */
		if (osd_fread(f,buf,8) != 8)
			goto getout;

#ifndef NOLEGACY
		if (memcmp(buf,MAMECFGSTRING_V5,8) == 0)
			version = 5;
		else if (memcmp(buf,MAMECFGSTRING_V6,8) == 0)
			version = 6;
		else if (memcmp(buf,MAMECFGSTRING_V7,8) == 0)
			version = 7;
		else
#endif
		if (memcmp(buf,MAMECFGSTRING_V8,8) == 0)
			version = 8;
		else
			goto getout;	/* header invalid */

		/* read array size */
		if (readint(f,&savedtotal) != 0)
			goto getout;
		if (total != savedtotal)
			goto getout;	/* different size */

		/* read the original settings and compare them with the ones defined in the driver */
		in = Machine->input_ports_default;
		while (in->type != IPT_END)
		{
			struct InputPort saved;

			if (input_port_read(f,&saved,version) != 0)
				goto getout;

			if (in->mask != saved.mask ||
				in->default_value != saved.default_value ||
				in->type != saved.type ||
				seq_cmp(&in->seq,&saved.seq) !=0 )
			goto getout;	/* the default values are different */

			in++;
		}

		/* read the current settings */
		in = Machine->input_ports;
		while (in->type != IPT_END)
		{
#ifdef KAILLERA
			struct InputPort temp;
			extern int kPlay;

			if (kPlay && ((in->type & ~IPF_MASK) == IPT_DIPSWITCH_NAME)) {
				if (input_port_read(f,&temp,version) != 0)
					goto getout;
			} else
#endif /* KAILLERA */
			if (input_port_read(f,in,version) != 0)
				goto getout;
			in++;
		}

		/* Clear the coin & ticket counters/flags - LBO 042898 */
		for (i = 0; i < COIN_COUNTERS; i ++)
			coins[i] = lastcoin[i] = coinlockedout[i] = 0;
		dispensed_tickets = 0;

		/* read in the coin/ticket counters */
		for (i = 0; i < COIN_COUNTERS; i ++)
		{
			if (readint(f,&coins[i]) != 0)
				goto getout;
		}
		if (readint(f,&dispensed_tickets) != 0)
			goto getout;

		mixer_read_config(f);

		/* Custom button setting */
		for (player = 0; player < 2; player++)
		{
			for (button = 0; button < 4; button++)
			{
				if (readword(f, &custom_button[player][button]) != 0)
					goto getout;
			}

			for (button = 0; button < 4; button++)
			{
				if (readword(f, &autofire_mask[player][button]) != 0)
					goto getout;
			}

			for (button = 0; button < 4; button++)
			{
				if (readword(f, &autofire_delay[player][button]) != 0)
					goto getout;
			}
		}

		{
			short tmp_clock;

			for (i = 0; i < MAX_CPU; i++)
			{
				if (!readword(f, &tmp_clock) && tmp_clock >= 50 && tmp_clock <= 200)
					cpu_clocks[i] = tmp_clock;
			}
		}

getout:
		osd_fclose(f);
	}

	/* All analog ports need initialization */
	{
		int i;
		for (i = 0; i < MAX_INPUT_PORTS; i++)
			input_analog_init[i] = 1;
	}

#ifdef KAILLERA
	{
		struct InputPort *in;
		int port;
		
		memset(&kBitsPlayValues,-1,sizeof(kBitsPlayValues));
		
		in = Machine->input_ports;
		in++;
		
		port = 0;
		while (in->type != IPT_END && port < MAX_INPUT_PORTS)
		{
			while (in->type != IPT_END && in->type != IPT_PORT)
			{
				int player = -1, pos = -1;
				
				if ((in->type & IPF_PLAYERMASK) == IPF_PLAYER1) player = 0;
				if ((in->type & IPF_PLAYERMASK) == IPF_PLAYER2) player = 1;
				
				switch (in->type & ~IPF_MASK)
				{
					case IPT_BUTTON1:						pos =  0; break;
					case IPT_BUTTON2:						pos =  1; break;
					case IPT_BUTTON3:						pos =  2; break;
					case IPT_BUTTON4:						pos =  3; break;
					case IPT_JOYSTICK_UP:					pos =  7; break;
					case IPT_JOYSTICK_DOWN:					pos =  8; break;
					case IPT_JOYSTICK_LEFT:					pos =  9; break;
					case IPT_JOYSTICK_RIGHT:				pos = 10; break;
					
					case IPT_COIN1:				player = 0;	pos = 11; break;
					case IPT_START1:			player = 0;	pos = 12; break;
					
					case IPT_COIN2:				player = 1;	pos = 11; break;
					case IPT_START2:			player = 1;	pos = 12; break;
				}
				
				if (pos != -1 && player != -1)
				{
					kBitsPlayValues[player][pos][0] = port;
					kBitsPlayValues[player][pos][1] = in->mask;
					kBitsPlayValues[player][pos][2] = in->default_value;
				}
				
				in++;
			}
			port++;
			if (in->type == IPT_PORT) in++;
		}
	}
	kFirstMsg = 1;
#endif /* KAILLERA */

	update_input_ports();

	/* if we didn't find a saved config, return 0 so the main core knows that it */
	/* is the first time the game is run and it should diplay the disclaimer. */
	if (f) return 1;
	else return 0;
}

/***************************************************************************/
/* Save */

void save_input_port_settings(void)
{
	void *f;

	save_default_keys();

	if ((f = osd_fopen(Machine->gamedrv->name,0,OSD_FILETYPE_CONFIG,1)) != 0)
	{
		struct InputPort *in;
		int total;
		int i;
		int player, button;


		in = Machine->input_ports_default;

		/* calculate the size of the array */
		total = 0;
		while (in->type != IPT_END)
		{
			total++;
			in++;
		}

		/* write header */
		osd_fwrite(f,MAMECFGSTRING_V8,8);
		/* write array size */
		writeint(f,total);
		/* write the original settings as defined in the driver */
		in = Machine->input_ports_default;
		while (in->type != IPT_END)
		{
			input_port_write(f,in);
			in++;
		}
		/* write the current settings */
		in = Machine->input_ports;
		while (in->type != IPT_END)
		{
			input_port_write(f,in);
			in++;
		}

		/* write out the coin/ticket counters for this machine - LBO 042898 */
		for (i = 0; i < COIN_COUNTERS; i ++)
			writeint(f,coins[i]);
		writeint(f,dispensed_tickets);

		mixer_write_config(f);

		/* Custom button setting */
		for (player = 0; player < 2; player++)
		{
			for (button = 0; button < 4; button++)
			{
				writeword(f, custom_button[player][button]);
			}

			for (button = 0; button < 4; button++)
			{
				writeword(f, autofire_mask[player][button]);
			}

			for (button = 0; button < 4; button++)
			{
				writeword(f, autofire_delay[player][button]);
			}
		}

		writeword(f, cpu_clocks[0]);
		writeword(f, cpu_clocks[1]);

		if (options.use_overclock)
			cpu_clocks[0] = cpu_clocks[1] = 200;
		else
			cpu_clocks[0] = cpu_clocks[1] = 100;

		osd_fclose(f);
	}
}

void init_input_port(void)
{
	warming_up = 1;

	change_port_mask();
//	update_input_ports();
	memset(command_buf, 0, COMMAND_LOG_BUFSIZE);
	show_input = 0;
	warming_up = 0;
}


/* Note that the following 3 routines have slightly different meanings with analog ports */
const char *input_port_name(const struct InputPort *in)
{
	int i;
	unsigned type;

	if (in->name != IP_NAME_DEFAULT) return in->name;

	i = 0;

	if ((in->type & ~IPF_MASK) == IPT_EXTENSION)
		type = (in-1)->type & (~IPF_MASK | IPF_PLAYERMASK);
	else
		type = in->type & (~IPF_MASK | IPF_PLAYERMASK);

	while (inputport_defaults[i].type != IPT_END &&
			inputport_defaults[i].type != type)
		i++;

	if ((in->type & ~IPF_MASK) == IPT_EXTENSION)
		return inputport_defaults[i+1].name;
	else
		return inputport_defaults[i].name;
}

InputSeq* input_port_type_seq(int type)
{
	unsigned i;

	i = 0;

	while (inputport_defaults[i].type != IPT_END &&
			inputport_defaults[i].type != type)
		i++;

	return &inputport_defaults[i].seq;
}

InputSeq* input_port_seq(const struct InputPort *in)
{
	int i,type;

	static InputSeq ip_none = SEQ_DEF_1(CODE_NONE);

	while (seq_get_1((InputSeq*)&in->seq) == CODE_PREVIOUS) in--;

	if ((in->type & ~IPF_MASK) == IPT_EXTENSION)
	{
		type = (in-1)->type & (~IPF_MASK | IPF_PLAYERMASK);
		/* if port is disabled, or cheat with cheats disabled, return no key */
		if (((in-1)->type & IPF_UNUSED) || (!options.cheat && ((in-1)->type & IPF_CHEAT)))
			return &ip_none;
	}
	else
	{
		type = in->type & (~IPF_MASK | IPF_PLAYERMASK);
		/* if port is disabled, or cheat with cheats disabled, return no key */
		if ((in->type & IPF_UNUSED) || (!options.cheat && (in->type & IPF_CHEAT)))
			return &ip_none;
	}

	if (seq_get_1((InputSeq*)&in->seq) != CODE_DEFAULT)
		return (InputSeq*)&in->seq;

	i = 0;

	while (inputport_defaults[i].type != IPT_END &&
			inputport_defaults[i].type != type)
		i++;

	if ((in->type & ~IPF_MASK) == IPT_EXTENSION)
		return &inputport_defaults[i+1].seq;
	else
		return &inputport_defaults[i].seq;
}

void update_analog_port(int port)
{
	struct InputPort *in;
	int current, delta, type, sensitivity, min, max, default_value;
	int axis, is_stick, check_bounds;
	InputSeq* incseq;
	InputSeq* decseq;
	int keydelta;
	int player;

	/* get input definition */
	in = input_analog[port];

	/* if we're not cheating and this is a cheat-only port, bail */
	if (!options.cheat && (in->type & IPF_CHEAT)) return;
	type=(in->type & ~IPF_MASK);

	decseq = input_port_seq(in);
	incseq = input_port_seq(in+1);

	keydelta = IP_GET_DELTA(in);

	switch (type)
	{
		case IPT_TRACKBALL_X:
			axis = X_AXIS; is_stick = 0; check_bounds = 0; break;
		case IPT_TRACKBALL_Y:
			axis = Y_AXIS; is_stick = 0; check_bounds = 0; break;
		default:
			/* Use some defaults to prevent crash */
			axis = X_AXIS; is_stick = 0; check_bounds = 0;
			logerror("Oops, polling non analog device in update_analog_port()????\n");
	}


	sensitivity = IP_GET_SENSITIVITY(in);
	min = IP_GET_MIN(in);
	max = IP_GET_MAX(in);
	default_value = in->default_value * 100 / sensitivity;
	/* extremes can be either signed or unsigned */
	if (min > max)
	{
		if (in->mask > 0xff) min = min - 0x10000;
		else min = min - 0x100;
	}


	input_analog_previous_value[port] = input_analog_current_value[port];

	/* if IPF_CENTER go back to the default position */
	/* sticks are handled later... */
	if ((in->type & IPF_CENTER) && (!is_stick))
		input_analog_current_value[port] = in->default_value * 100 / sensitivity;

	current = input_analog_current_value[port];

	delta = 0;

	switch (in->type & IPF_PLAYERMASK)
	{
		case IPF_PLAYER2:          player = 1; break;
		case IPF_PLAYER1: default: player = 0; break;
	}

	if (axis == X_AXIS)
		delta = mouse_delta_x[player];
	else
		delta = mouse_delta_y[player];

	if (seq_pressed(decseq)) delta -= keydelta;
	if (seq_pressed(incseq)) delta += keydelta;

	if (in->type & IPF_REVERSE) delta = -delta;

	if (is_stick)
	{
		int new, prev;

		/* center stick */
		if ((delta == 0) && (in->type & IPF_CENTER))
		{
			if (current > default_value)
			delta = -100 / sensitivity;
			if (current < default_value)
			delta = 100 / sensitivity;
		}

		/* An analog joystick which is not at zero position (or has just */
		/* moved there) takes precedence over all other computations */
		/* analog_x/y holds values from -128 to 128 (yes, 128, not 127) */

		if (axis == X_AXIS)
		{
			new  = analog_current_x[player];
			prev = analog_previous_x[player];
		}
		else
		{
			new  = analog_current_y[player];
			prev = analog_previous_y[player];
		}

		if ((new != 0) || (new-prev != 0))
		{
			delta=0;

			if (in->type & IPF_REVERSE)
			{
				new  = -new;
				prev = -prev;
			}

			/* apply sensitivity using a logarithmic scale */
			if (in->mask > 0xff)
			{
				if (new > 0)
				{
					current = (pow(new / 32768.0, 100.0 / sensitivity) * (max-in->default_value)
							+ in->default_value) * 100 / sensitivity;
				}
				else
				{
					current = (pow(-new / 32768.0, 100.0 / sensitivity) * (min-in->default_value)
							+ in->default_value) * 100 / sensitivity;
				}
			}
			else
			{
				if (new > 0)
				{
					current = (pow(new / 128.0, 100.0 / sensitivity) * (max-in->default_value)
							+ in->default_value) * 100 / sensitivity;
				}
				else
				{
					current = (pow(-new / 128.0, 100.0 / sensitivity) * (min-in->default_value)
							+ in->default_value) * 100 / sensitivity;
				}
			}
		}
	}

	current += delta;

	if (check_bounds)
	{
		if ((current * sensitivity + 50) / 100 < min)
			current = (min * 100 + sensitivity/2) / sensitivity;
		if ((current * sensitivity + 50) / 100 > max)
			current = (max * 100 + sensitivity/2) / sensitivity;
	}

	input_analog_current_value[port] = current;
}

static void scale_analog_port(int port)
{
#if 0
	struct InputPort *in;
	int delta,current,sensitivity;

profiler_mark(PROFILER_INPUT);
	in = input_analog[port];
	sensitivity = IP_GET_SENSITIVITY(in);

	delta = cpu_scalebyfcount(input_analog_current_value[port] - input_analog_previous_value[port]);

	current = input_analog_previous_value[port] + delta;

	input_port_value[port] &= ~in->mask;
	input_port_value[port] |= ((current * sensitivity + 50) / 100) & in->mask;

	if (playback)
		readword(playback,&input_port_value[port]);
	if (record)
		writeword(record,input_port_value[port]);
profiler_mark(PROFILER_END);
#else
	struct InputPort *in;
	int delta,current,sensitivity;

profiler_mark(PROFILER_INPUT);
	in = input_analog[port];
	sensitivity = IP_GET_SENSITIVITY(in);

	/* apply scaling fairly in both positive and negative directions */
	delta = input_analog_current_value[port] - input_analog_previous_value[port];
	if (delta >= 0)
		delta = cpu_scalebyfcount(delta);
	else
		delta = -cpu_scalebyfcount(-delta);

	current = input_analog_previous_value[port] + delta;
	/* apply scaling fairly in both positive and negative directions */
	if (current >= 0)
		current = (current * sensitivity + 50) / 100;
	else
		current = (-current * sensitivity + 50) / -100;

	input_port_value[port] &= ~in->mask;
	input_port_value[port] |= current & in->mask;

	if (playback)
		readword(playback,&input_port_value[port]);
	if (record)
		writeword(record,input_port_value[port]);
profiler_mark(PROFILER_END);
#endif
}


void update_input_ports(void)
{
	int port,ib;
	struct InputPort *in;
	static int waspressed[512];
	extern int neogeo_has_trackball;

	profiler_mark(PROFILER_INPUT);

	/* clear all the values before proceeding */
	for (port = 0; port < MAX_INPUT_PORTS; port++)
	{
		input_port_value[port] = 0;
		input_analog[port] = 0;
	}

	in = Machine->input_ports;

	if (in->type == IPT_END) return; 	/* nothing to do */

	/* make sure the InputPort definition is correct */
	if (in->type != IPT_PORT)
	{
		logerror("Error in InputPort definition: expecting PORT_START\n");
		return;
	}
	else in++;

	/* scan all the input ports */
	port = 0;
	ib = 0;
	while (in->type != IPT_END && port < MAX_INPUT_PORTS)
	{
		struct InputPort *start;


		/* first of all, scan the whole input port definition and build the */
		/* default value. I must do it before checking for input because otherwise */
		/* multiple keys associated with the same input bit wouldn't work (the bit */
		/* would be reset to its default value by the second entry, regardless if */
		/* the key associated with the first entry was pressed) */
		start = in;
		while (in->type != IPT_END && in->type != IPT_PORT)
		{
			if ((in->type & ~IPF_MASK) != IPT_DIPSWITCH_SETTING &&	/* skip dipswitch definitions */
				(in->type & ~IPF_MASK) != IPT_EXTENSION)			/* skip analog extension fields */
			{
				input_port_value[port] =
						(input_port_value[port] & ~in->mask) | (in->default_value & in->mask);
			}

			in++;
		}

#ifdef KAILLERA
		if (!KailleraChatIsActive())
		{
#endif
		/* now get back to the beginning of the input port and check the input bits. */
		for (in = start;
			 in->type != IPT_END && in->type != IPT_PORT;
			 in++, ib++)
		{
			if ((in->type & ~IPF_MASK) != IPT_DIPSWITCH_SETTING &&	/* skip dipswitch definitions */
				(in->type & ~IPF_MASK) != IPT_EXTENSION)			/* skip analog extension fields */
			{
				/* If it's an analog control, handle it appropriately */
				if (((in->type & ~IPF_MASK) > IPT_ANALOG_START)
					  && ((in->type & ~IPF_MASK) < IPT_ANALOG_END  )) /* LBO 120897 */
				{
					input_analog[port]=in;
					/* reset the analog port on first access */
					if (input_analog_init[port])
					{
						input_analog_init[port] = 0;
						input_analog_current_value[port] = input_analog_previous_value[port]
							= in->default_value * 100 / IP_GET_SENSITIVITY(in);
					}
				}
				else
				{
					InputSeq* seq;

					seq = input_port_seq(in);

					if (seq_pressed(seq))
					{
						/* skip if coin input and it's locked out */
						if ((in->type & ~IPF_MASK) >= IPT_COIN1 &&
							(in->type & ~IPF_MASK) <= IPT_COIN3 &&
                            coinlockedout[(in->type & ~IPF_MASK) - IPT_COIN1])
						{
							continue;
						}

						/* if IPF_RESET set, reset the first CPU */
						if ((in->type & IPF_RESETCPU) && waspressed[ib] == 0)
							cpu_set_reset_line(0,PULSE_LINE);

						input_port_value[port] ^= in->mask;
						waspressed[ib] = 1;
					}
					else
						waspressed[ib] = 0;
				}
			}
		}
#ifdef KAILLERA
		}
#endif

		port++;
		if (in->type == IPT_PORT) in++;
	}

	if (!neogeo_has_trackball)
	{
		do_autofire(input_port_value, waspressed);
		do_macro_ex(input_port_value, waspressed);
	}

	if (!warming_up)
	{
		if (playback)
		{
			int i, err = 0;

			for (i = 0; i < _inputport_max; i ++)
				err = readword(options.playback, &input_port_value[i]);

			if (err)
			{
				fclose(options.playback);
				options.playback = playback = 0;
			}
		}
		else if (record)
		{
			int i;

			for (i = 0; i < MAX_INPUT_PORTS; i ++)
				writeword(options.record, input_port_value[i]);
		}

		if (!neogeo_has_trackball)
		{
			if (record_combo)
				save_combo();
		}
	}

	if (show_input)
		make_input_log();

#ifdef KAILLERA
	if (kFirstMsg) kFirstMsg = 0;
	else
	{
		extern LONGLONG uclock(void);
		extern void addClock(LONGLONG a);
		extern int kPlayerNb;
		unsigned short val[4];
		LONGLONG st;
		int i, b, recvLen;

		val[0] = 0;
		b = 1;

		for(i = 0 ; i < 16 ; i++)
		{
			if (kBitsPlayValues[0][i][0] != -1)
			if (~input_port_value[kBitsPlayValues[0][i][0]] & kBitsPlayValues[0][i][1]) val[0] |= b;
			b = b << 1;
	    }

	    st = uclock();
	    recvLen = kailleraModifyPlayValues((void *)val, 2);
	    addClock(uclock() - st);

	    if(recvLen > 0)
	    {
			if(recvLen > 8) recvLen = 8;
		    memcpy(input_port_value, kDefValues, sizeof(input_port_value));
		    for (i = 0 ; i < (recvLen / 2) ; i++)
			{
				int j;
				b = 1;

				for (j = 0 ; j < 16 ; j++)
		        {
		            if(kBitsPlayValues[i][j][0] != -1)
					    if(kBitsPlayValues[i][j][2] != IP_ACTIVE_LOW)
						{
			                if(!(val[i] & b))
								input_port_value[kBitsPlayValues[i][j][0]] |= kBitsPlayValues[i][j][1];
						}
						else
						{
							if(val[i] & b)
								input_port_value[kBitsPlayValues[i][j][0]] &= ~kBitsPlayValues[i][j][1];
						}
					b = b << 1;
		        }
			}
		}
	    if (recvLen == 0) memcpy(kDefValues, input_port_value, sizeof(input_port_value));
	}
#endif /* KAILLERA */

	profiler_mark(PROFILER_END);
}


/* used the the CPU interface to notify that VBlank has ended, so we can update */
/* IPT_VBLANK input ports. */
void inputport_vblank_end(void)
{
	int i;


profiler_mark(PROFILER_INPUT);

	/* update the analog devices */
	for (i = 0;i < OSD_MAX_JOY_ANALOG;i++)
	{
		/* update the analog joystick position */
		analog_previous_x[i] = analog_current_x[i];
		analog_previous_y[i] = analog_current_y[i];
		osd_analogjoy_read (i, &(analog_current_x[i]), &(analog_current_y[i]));

		/* update mouse/trackball position */
		osd_trak_read (i, &mouse_delta_x[i], &mouse_delta_y[i]);
	}

	for (i = 0;i < MAX_INPUT_PORTS;i++)
	{
		struct InputPort *in;

		in=input_analog[i];
		if (in)
		{
			update_analog_port(i);
		}
	}
profiler_mark(PROFILER_END);
}



int readinputport(int port)
{
	struct InputPort *in;

	/* Update analog ports on demand */
	in=input_analog[port];
	if (in)
	{
		scale_analog_port(port);
	}

	return input_port_value[port];
}



#ifdef MAME_DEBUG
READ_HANDLER( input_port_0_r ) { return readinputport(0); }
READ_HANDLER( input_port_1_r ) { return readinputport(1); }
READ_HANDLER( input_port_2_r ) { return readinputport(2); }
READ_HANDLER( input_port_3_r ) { return readinputport(3); }
READ_HANDLER( input_port_4_r ) { return readinputport(4); }
READ_HANDLER( input_port_5_r ) { return readinputport(5); }
READ_HANDLER( input_port_6_r ) { return readinputport(6); }
READ_HANDLER( input_port_7_r ) { return readinputport(7); }
READ_HANDLER( input_port_8_r ) { return readinputport(8); }
READ_HANDLER( input_port_9_r ) { return readinputport(9); }
READ_HANDLER( input_port_10_r ) { return readinputport(10); }
READ_HANDLER( input_port_11_r ) { return readinputport(11); }
READ_HANDLER( input_port_12_r ) { return readinputport(12); }
READ_HANDLER( input_port_13_r ) { return readinputport(13); }
READ_HANDLER( input_port_14_r ) { return readinputport(14); }
READ_HANDLER( input_port_15_r ) { return readinputport(15); }
READ_HANDLER( input_port_16_r ) { return readinputport(16); }
READ_HANDLER( input_port_17_r ) { return readinputport(17); }
READ_HANDLER( input_port_18_r ) { return readinputport(18); }
READ_HANDLER( input_port_19_r ) { return readinputport(19); }

READ16_HANDLER( input_port_0_word_r ) { return readinputport(0); }
READ16_HANDLER( input_port_1_word_r ) { return readinputport(1); }
READ16_HANDLER( input_port_2_word_r ) { return readinputport(2); }
READ16_HANDLER( input_port_3_word_r ) { return readinputport(3); }
READ16_HANDLER( input_port_4_word_r ) { return readinputport(4); }
READ16_HANDLER( input_port_5_word_r ) { return readinputport(5); }
READ16_HANDLER( input_port_6_word_r ) { return readinputport(6); }
READ16_HANDLER( input_port_7_word_r ) { return readinputport(7); }
READ16_HANDLER( input_port_8_word_r ) { return readinputport(8); }
READ16_HANDLER( input_port_9_word_r ) { return readinputport(9); }
READ16_HANDLER( input_port_10_word_r ) { return readinputport(10); }
READ16_HANDLER( input_port_11_word_r ) { return readinputport(11); }
READ16_HANDLER( input_port_12_word_r ) { return readinputport(12); }
READ16_HANDLER( input_port_13_word_r ) { return readinputport(13); }
READ16_HANDLER( input_port_14_word_r ) { return readinputport(14); }
READ16_HANDLER( input_port_15_word_r ) { return readinputport(15); }
READ16_HANDLER( input_port_16_word_r ) { return readinputport(16); }
READ16_HANDLER( input_port_17_word_r ) { return readinputport(17); }
READ16_HANDLER( input_port_18_word_r ) { return readinputport(18); }
READ16_HANDLER( input_port_19_word_r ) { return readinputport(19); }
#endif

/***************************************************************************/
/* InputPort conversion */

static unsigned input_port_count(const struct InputPortTiny *src)
{
	unsigned total;

	total = 0;
	while (src->type != IPT_END)
	{
		int type = src->type & ~IPF_MASK;
		if (type > IPT_ANALOG_START && type < IPT_ANALOG_END)
			total += 2;
		else if (type != IPT_EXTENSION)
			++total;
		++src;
	}

	++total; /* for IPT_END */

	return total;
}

struct InputPort* input_port_allocate(const struct InputPortTiny *src)
{
	struct InputPort* dst;
	struct InputPort* base;
	unsigned total;

	total = input_port_count(src);

	base = (struct InputPort*)MyMalloc(total * sizeof(struct InputPort));
	dst = base;

	while (src->type != IPT_END)
	{
		int type = src->type & ~IPF_MASK;
		const struct InputPortTiny *ext;
		const struct InputPortTiny *src_end;
		InputCode seq_default;

		if (type > IPT_ANALOG_START && type < IPT_ANALOG_END)
			src_end = src + 2;
		else
			src_end = src + 1;

		switch (type)
		{
			case IPT_END :
			case IPT_PORT :
			case IPT_DIPSWITCH_NAME :
			case IPT_DIPSWITCH_SETTING :
				seq_default = CODE_NONE;
			break;
			default:
				seq_default = CODE_DEFAULT;
		}

		ext = src_end;
		while (src != src_end)
		{
			dst->type = src->type;
			dst->mask = src->mask;
			dst->default_value = src->default_value;
			dst->name = src->name;

  			if (ext->type == IPT_EXTENSION)
  			{
				InputCode or1 =	IP_GET_CODE_OR1(ext);
				InputCode or2 =	IP_GET_CODE_OR2(ext);

				if (or1 < __code_max)
				{
					if (or2 < __code_max)
						seq_set_3(&dst->seq, or1, CODE_OR, or2);
					else
						seq_set_1(&dst->seq, or1);
				} else {
					if (or1 == CODE_NONE)
						seq_set_1(&dst->seq, or2);
					else
						seq_set_1(&dst->seq, or1);
				}

  				++ext;
  			} else {
				seq_set_1(&dst->seq,seq_default);
  			}

			++src;
			++dst;
		}

		src = ext;
	}

	dst->type = IPT_END;

	return base;
}

void input_port_free(struct InputPort* dst)
{
	MyFree(dst);
}

/***************************************************************************/
/* for old input log file                                                  */
/***************************************************************************/

void check_inp_type(void)
{
	UINT16 port0;
	int offset;

	offset = osd_ftell(options.playback);
	osd_fseek(options.playback, 32, SEEK_CUR);
	osd_fread(options.playback, &port0, 2);
	osd_fseek(options.playback, offset, SEEK_SET);

	if (port0)
		_inputport_max = 16;
	else
		_inputport_max = MAX_INPUT_PORTS;
}


/***************************************************************************/
/* Custom Button                                                           */
/***************************************************************************/

/* change input port mask */
void change_port_mask(void)
{
	int player, button;
	struct InputPort *in;
	int custom_start[2] = { 9, 22 };
	extern int neogeo_has_trackball;

	if (neogeo_has_trackball) return;

	docombo[0] = docombo[1] = 0;

	for (player = 0; player < 2; player++)
	{
		for (button = 0; button < 4; button++)
		{
			in = &Machine->input_ports[custom_start[player] + button];

			setcombo[player][button] = 0;
			autofire_mask[player][button] = 0;
			in->mask = 0;

			if (custom_button[player][button] <= BTN_ABCD)
			{
				in->mask = button_mask[custom_button[player][button]];
			}
			else if (custom_button[player][button] <= BTN_CMB8)
			{
				setcombo[player][button] = custom_button[player][button] - BTN_ABCD;
				if (combondx[setcombo[player][button] - 1] == -1)
					setcombo[player][button] = 0;
			}
			else
			{
				autofire_mask[player][button] = 0x10 << (custom_button[player][button] - BTN_AF_A);
			}
		}
	}
}


#define BT_U   0x01
#define BT_D   0x02
#define BT_L   0x04
#define BT_R   0x08
#define BT_UL  (BT_U | BT_L)
#define BT_UR  (BT_U | BT_R)
#define BT_DL  (BT_D | BT_L)
#define BT_DR  (BT_D | BT_R)
#define BT_1   0x10
#define BT_2   0x20
#define BT_3   0x40
#define BT_4   0x80


INLINE void copy_command_buf(int button)
{
	int i, len;

	len = strlen((char *)command_buf);

	if (len == MAX_COMMAND_LOG)
	{
		for (i = 0; i < MAX_COMMAND_LOG - 2; i++)
			command_buf[i] = command_buf[i + 2];
	}

	command_buf[MAX_COMMAND_LOG - 2] = '\0';
	command_buf[MAX_COMMAND_LOG - 1] = '\0';

	len = strlen((char *)command_buf);

	switch(button)
	{
		case BT_U:
			command_buf[len]   = 0x81;
			command_buf[len+1] = 0xd8;
			break;
		case BT_D:
			command_buf[len]   = 0x81;
			command_buf[len+1] = 0xd2;
			break;
		case BT_L:
			command_buf[len]   = 0x81;
			command_buf[len+1] = 0xd4;
			break;
		case BT_R:
			command_buf[len]   = 0x81;
			command_buf[len+1] = 0xd6;
			break;
		case BT_UL:
			command_buf[len]   = 0x81;
			command_buf[len+1] = 0xd7;
			break;
		case BT_UR:
			command_buf[len]   = 0x81;
			command_buf[len+1] = 0xd9;
			break;
		case BT_DL:
			command_buf[len]   = 0x81;
			command_buf[len+1] = 0xd1;
			break;
		case BT_DR:
			command_buf[len]   = 0x81;
			command_buf[len+1] = 0xd3;
			break;

		case BT_1:
			command_buf[len]   = 0x81;
			command_buf[len+1] = 0xc0;
			break;
		case BT_2:
			command_buf[len]   = 0x81;
			command_buf[len+1] = 0xc1;
			break;
		case BT_3:
			command_buf[len]   = 0x81;
			command_buf[len+1] = 0xc2;
			break;
		case BT_4:
			command_buf[len]   = 0x81;
			command_buf[len+1] = 0xc3;
			break;

		default:
			break;
	}
}


static void make_input_log(void)
{
	static int old_dir  = 0;
	static int old_btn  = 0;
	int now_dir, now_btn;

	now_dir = 0;
	if (!(input_port_value[0] & BT_U)) now_dir |= BT_U;
	if (!(input_port_value[0] & BT_D)) now_dir |= BT_D;
	if (!(input_port_value[0] & BT_L)) now_dir |= BT_L;
	if (!(input_port_value[0] & BT_R)) now_dir |= BT_R;

	if ((now_dir & BT_UL) == BT_UL)      now_dir = BT_UL;
	else if ((now_dir & BT_UR) == BT_UR) now_dir = BT_UR;
	else if ((now_dir & BT_DL) == BT_DL) now_dir = BT_DL;
	else if ((now_dir & BT_DR) == BT_DR) now_dir = BT_DR;
	else if (now_dir & BT_U)             now_dir = BT_U;
	else if (now_dir & BT_D)             now_dir = BT_D;
	else if (now_dir & BT_L)             now_dir = BT_L;
	else if (now_dir & BT_R)             now_dir = BT_R;

	if (old_dir != now_dir)
	{
		copy_command_buf(now_dir);
		old_dir = now_dir;
	}

	now_btn = 0;
	if (!(input_port_value[0] & BT_1)) now_btn |= BT_1;
	if (!(input_port_value[0] & BT_2)) now_btn |= BT_2;
	if (!(input_port_value[0] & BT_3)) now_btn |= BT_3;
	if (!(input_port_value[0] & BT_4)) now_btn |= BT_4;

	if ((now_btn & BT_1) != 0 && (old_btn & BT_1) == 0)
		copy_command_buf(BT_1);
	if ((now_btn & BT_2) != 0 && (old_btn & BT_2) == 0)
		copy_command_buf(BT_2);
	if ((now_btn & BT_3) != 0 && (old_btn & BT_3) == 0)
		copy_command_buf(BT_3);
	if ((now_btn & BT_4) != 0 && (old_btn & BT_4) == 0)
		copy_command_buf(BT_4);

	old_btn = now_btn;

	if (now_dir || now_btn)
		command_counter = 60;
}


int custom_button_init(void)
{
	int player, button;

	if (comboa)  MyFree(comboa);
	if (rcomboa) MyFree(rcomboa);

	for (player = 0; player < MAX_PLAYERS; player++)
	{
		for (button = 0; button < MAX_CUSTOM_BUTTONS; button++)
		{
			setcombo[player][button] = 0;
			combondx[player * MAX_CUSTOM_BUTTONS + button] = -1;
		}

		combo[player] = 0;
		combocount[player] = 0;
	}

	comboentries = 0;

	comboa  = (UINT16 *)MyMalloc(MAX_COMBO_SIZE * sizeof(UINT16));
	rcomboa = (UINT16 *)MyMalloc(MAX_COMBO_SIZE * sizeof(UINT16));

	if (!comboa || !rcomboa)
	{
		if (comboa)  MyFree(comboa);
		if (rcomboa) MyFree(rcomboa);
		comboa = 0;
		rcomboa = 0;
		return 0;
	}

	memset(comboa,  0xff, MAX_COMBO_SIZE * sizeof(UINT16));
	memset(rcomboa, 0xff, MAX_COMBO_SIZE * sizeof(UINT16));

	branchentries = 0;

	return 1;
}


void custom_button_end(void)
{
	int player, button;

	for (player = 0; player < MAX_PLAYERS; player++)
	{
		for (button = 0; button < MAX_CUSTOM_BUTTONS; button++)
		{
			setcombo[player][button] = 0;
			combondx[player * MAX_CUSTOM_BUTTONS + button] = -1;
			custom_button[player][button] = 0;
			autofire_mask[player][button] = 0;
			autofire_delay[player][button] = 0;
		}

		combo[player] = 0;
		combocount[player] = 0;
	}

	if (comboa)  MyFree(comboa);
	if (rcomboa) MyFree(rcomboa);

	comboentries = 0;
	comboa = rcomboa = 0;
}


/* Autofire routine */
static void do_autofire(UINT16 inparr[], int pressed[])
{
	unsigned short mask, player, button;
	static int autofire_counter = 0;

	autofire_counter++;

	for (player = 0; player < MAX_PLAYERS; player++)
	{
		for (button = 0; button < MAX_CUSTOM_BUTTONS; button++)
		{
			if (custom_button[player][button] >= BTN_AF_A)
			{
				if (pressed[start[player] + button])
				{
					mask = autofire_mask[player][button];

					if (autofire_counter % (autofire_delay[player][button] + 1) == 0)
						shadow[player][button] ^= 1;

					if (shadow[player][button])
						inparr[player] &= ~mask;
					else
						inparr[player] |= mask;
				}
			}
		}
	}
}

/*-----------------------
     Combo Recorder
-----------------------*/

int combosaveslot = 0;


static char *combo_buf;

static int pre_dir = 0;
static int pre_btn = 0;
static int dir_cnt = 0;
static int bt1_cnt = 0;
static int bt2_cnt = 0;
static int bt3_cnt = 0;
static int bt4_cnt = 0;
static int dir_idle = 1;


void save_combo(void)
{
	int now_dir, now_btn;

	int dir_cnt_reset = 0;
	int force_dir_break = 0;

	now_dir = 0;
	if (!(input_port_value[0] & BT_U)) now_dir |= BT_U;
	if (!(input_port_value[0] & BT_D)) now_dir |= BT_D;
	if (!(input_port_value[0] & BT_L)) now_dir |= BT_L;
	if (!(input_port_value[0] & BT_R)) now_dir |= BT_R;

	if ((now_dir & BT_UL) == BT_UL)      now_dir = BT_UL;
	else if ((now_dir & BT_UR) == BT_UR) now_dir = BT_UR;
	else if ((now_dir & BT_DL) == BT_DL) now_dir = BT_DL;
	else if ((now_dir & BT_DR) == BT_DR) now_dir = BT_DR;
	else if (now_dir & BT_U)             now_dir = BT_U;
	else if (now_dir & BT_D)             now_dir = BT_D;
	else if (now_dir & BT_L)             now_dir = BT_L;
	else if (now_dir & BT_R)             now_dir = BT_R;
	else now_dir = 0;

	now_btn = 0;
	if (!(input_port_value[0] & BT_1)) now_btn |= BT_1;
	if (!(input_port_value[0] & BT_2)) now_btn |= BT_2;
	if (!(input_port_value[0] & BT_3)) now_btn |= BT_3;
	if (!(input_port_value[0] & BT_4)) now_btn |= BT_4;

	if (pre_btn != now_btn)
	{
		force_dir_break = 1;
		pre_btn = now_btn;
	}


	dir_cnt++;

	if (pre_dir != now_dir || force_dir_break)
	{
		dir_idle = 0;
		switch(pre_dir)
		{
			case BT_U:
				sprintf(combo_buf + strlen(combo_buf), "%d u ", dir_cnt);
				dir_cnt = 0;
				break;
			case BT_D:
				sprintf(combo_buf + strlen(combo_buf), "%d d ", dir_cnt);
				dir_cnt = 0;
				break;
			case BT_L:
				sprintf(combo_buf + strlen(combo_buf), "%d b ", dir_cnt);
				dir_cnt = 0;
				break;
			case BT_R:
				sprintf(combo_buf + strlen(combo_buf), "%d f ", dir_cnt);
				dir_cnt = 0;
				break;
			case BT_UL:
				sprintf(combo_buf + strlen(combo_buf), "%d ub ", dir_cnt);
				dir_cnt = 0;
				break;
			case BT_UR:
				sprintf(combo_buf + strlen(combo_buf), "%d uf ", dir_cnt);
				dir_cnt = 0;
				break;
			case BT_DL:
				sprintf(combo_buf + strlen(combo_buf), "%d db ", dir_cnt);
				dir_cnt = 0;
				break;
			case BT_DR:
				sprintf(combo_buf + strlen(combo_buf), "%d df ", dir_cnt);
				dir_cnt = 0;
				break;
			case 0:
				sprintf(combo_buf + strlen(combo_buf), "%d n ", dir_cnt);
				dir_cnt = 0;
				break;
			default:
				break;
		};
	}
	else if (pre_dir == now_dir == 0)
	{
		dir_idle = 1;
	}
	pre_dir = now_dir;



	if ((now_btn & BT_1) != 0 && (bt1_cnt) >= 0)
	{
		bt1_cnt++;
	}
	else if ((now_btn & BT_1) == 0 && (bt1_cnt) > 0)
	{
		if (dir_idle == 1 && dir_cnt_reset == 0)
		{
			sprintf(combo_buf + strlen(combo_buf), "%d n %d + A ", dir_cnt, bt1_cnt);
			dir_cnt_reset = 1;
		}
		else
		{
			sprintf(combo_buf + strlen(combo_buf), "%d + A ", bt1_cnt);
		}
		bt1_cnt = 0;
	}
	
	if ((now_btn & BT_2) != 0 && (bt2_cnt) >= 0)
	{
		bt2_cnt++;
	}
	else if ((now_btn & BT_2) == 0 && (bt2_cnt) > 0)
	{
		if (dir_idle == 1 && dir_cnt_reset == 0)
		{
			sprintf(combo_buf + strlen(combo_buf), "%d n %d + B ", dir_cnt, bt2_cnt);
			dir_cnt_reset = 1;
		}
		else
		{
			sprintf(combo_buf + strlen(combo_buf), "%d + B ", bt2_cnt);
		}
		bt2_cnt = 0;
	}

	if ((now_btn & BT_3) != 0 && (bt3_cnt) >= 0)
	{
		bt3_cnt++;
	}
	else if ((now_btn & BT_3) == 0 && (bt3_cnt) > 0)
	{
		if (dir_idle == 1 && dir_cnt_reset == 0)
		{
			sprintf(combo_buf + strlen(combo_buf), "%d n %d + C ", dir_cnt, bt3_cnt);
			dir_cnt_reset = 1;
		}
		else
		{
			sprintf(combo_buf + strlen(combo_buf), "%d + C ", bt3_cnt);
		}
		bt3_cnt = 0;
	}

	if ((now_btn & BT_4) != 0 && (bt4_cnt) >= 0)
	{
		bt4_cnt++;
	}
	else if ((now_btn & BT_4) == 0 && (bt4_cnt) > 0)
	{
		if (dir_idle == 1 && dir_cnt_reset == 0)
		{
			sprintf(combo_buf + strlen(combo_buf), "%d n %d + D ", dir_cnt, bt4_cnt);
			dir_cnt_reset = 1;
		}
		else
		{
			sprintf(combo_buf + strlen(combo_buf), "%d + D ", bt4_cnt);
		}
		bt4_cnt = 0;
	}
	if (dir_cnt_reset == 1)	dir_cnt = 0;
}


/*-----------------------
         Combo
-----------------------*/

static void do_macro_ex(UINT16 inparr[], int pressed[])
{
	int player, button;
	int btn_offset[2] = { 0, 12 };
	static int pressed_btn = -1;
	static int branch_num = -1;
	static int next_branch_cnt = -1;

	if (comboa == 0) return;

	for (player = 0; player < MAX_PLAYERS; player++)
	{
		if (docombo[player] == 0)
		{
			for (button = 0; button < MAX_CUSTOM_BUTTONS; button++)
			{
				if (pressed[start[player] + button] && setcombo[player][button])
				{
					combocount[player] = combondx[setcombo[player][button] - 1];
					if (!pressed[left[player]])
						combo[player] = comboa;
					else
						combo[player] = rcomboa;

					docombo[player] = 1;
					branch_num = 1;
					pressed_btn = button;
					next_branch_cnt = branch[ branchndx[ setcombo[player][pressed_btn] - 1 ][ branch_num ] ].begin;

					break;
				}
			}
		}

		if (docombo[player] == 3
		&& !(pressed[start[player] + pressed_btn] && setcombo[player][pressed_btn]))
		    docombo[player] = 0;

		if (combocount[player] || docombo[player] == 1)
		{
			if(!(pressed[start[player] + pressed_btn] && setcombo[player][pressed_btn]))
			{
				docombo[player] = 2;
			}

			if(docombo[player] == 2 && pressed[start[player] + pressed_btn] && setcombo[player][pressed_btn])
			{
				combocount[player] = 0;
				branch_num = -1;
				docombo[player] = 3;
			}
			else
			{
				if( combocount[player] == next_branch_cnt )
				{
					int bh_base = branchndx[ setcombo[player][pressed_btn] - 1 ][ branch_num ];
					
					while( branch[bh_base].bh_mask != -1)
					{
						if( pressed[btn_offset[player] + branch[bh_base].bh_mask] )
						{
							branch_num = branch[bh_base].which_bh;
							combocount[player] = branch[branchndx[setcombo[player][pressed_btn]-1][branch_num]].begin;
							break;
						}
						bh_base++;
					}
					branch_num++;
					next_branch_cnt = branch[ branchndx[ setcombo[player][pressed_btn] - 1 ][branch_num] ].begin;
				}
				if (combo[player][combocount[player]] == 0xffff)
				{
					combocount[player] = 0;
					branch_num = -1;
					docombo[player] = 0;
					return;
				}

				inparr[player] = combo[player][combocount[player]];
				combocount[player]++;
			}
		}
	}
}


static int parse_branch(char *line)
{
	int i = 0;
	char temp[64], *p;

	p = strrchr(line, '\n');
	if (p != NULL)
		*p = '\0';

	for (i = 0; i < 10; i++)
	{
		sscanf(line, "%s", temp);

		if (temp[0] == '{')
		{
			line += 2;
			continue;
		}

		if (temp[strlen(temp) - 1] == '}') break;

		branch[branchentries].begin = comboentries;
		branch[branchentries].which_bh = atoi(temp + 2);

		switch(temp[0])
		{
			case 'u':
				branch[branchentries].bh_mask = 0;
				break;

			case 'd':
				branch[branchentries].bh_mask = 1;
				break;

			case 'b':
				branch[branchentries].bh_mask = 2;
				break;

			case 'f':
				branch[branchentries].bh_mask = 3;
				break;

			case 'A':
				branch[branchentries].bh_mask = 4;
				break;

			case 'B':
				branch[branchentries].bh_mask = 5;
				break;

			case 'C':
				branch[branchentries].bh_mask = 6;
				break;

			case 'D':
				branch[branchentries].bh_mask = 7;
				break;

			default:
				break;
		};
		branchentries++;
		line += strlen(temp) + 1;
	}

	branch[branchentries].begin = comboentries;
	branch[branchentries].which_bh = -1;
	branch[branchentries].bh_mask = -1;
	branchentries++;

	return branchentries;
}


static int parse_combo(FILE *f)
{
	char s[10];
	unsigned int speed = 0;
	unsigned int i;

	fscanf(f, "%d", &speed);

	while (!feof(f))
	{
		fscanf(f,"%s",s);

		if (!feof(f))
		{
			if( s == NULL) continue;
			if (!strcmp(s,"b"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x4;
					rcomboa[comboentries++] = ~0x8;
				}
			}

			if (!strcmp(s,"f"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x8;
					rcomboa[comboentries++] = ~0x4;
				}
			}

			if (!strcmp(s,"u"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x1;
					rcomboa[comboentries++] = ~0x1;
				}
			}

			if (!strcmp(s,"d"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x2;
					rcomboa[comboentries++] = ~0x2;
				}
			}

			if (!strcmp(s,"ub"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x5;
					rcomboa[comboentries++] = ~0x9;
				}
			}

			if (!strcmp(s,"uf"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x9;
					rcomboa[comboentries++] = ~0x5;
				}
			}

			if (!strcmp(s,"db"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x6;
					rcomboa[comboentries++] = ~0xa;
				}
			}

			if (!strcmp(s,"df"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0xa;
					rcomboa[comboentries++] = ~0x6;
				}
			}

			if (!strcmp(s,"n"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries]=0x00ff;
					rcomboa[comboentries++]=0x00ff;
				}
			}

			if (!strcmp(s,"A"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x10;
					rcomboa[comboentries++] = ~0x10;
				}
			}

			if (!strcmp(s,"B"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x20;
					rcomboa[comboentries++] = ~0x20;
				}
			}

			if (!strcmp(s,"C"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x40;
					rcomboa[comboentries++] = ~0x40;
				}
			}

			if (!strcmp(s,"D"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x80;
					rcomboa[comboentries++] = ~0x80;
				}
			}

			if (!strcmp(s,"+"))
			{
				comboentries -= speed;
				fscanf(f,"%s",s);

				for (i = 0; i < speed; i++)
				{
					if (!strcmp(s,"A"))
					{
						comboa[comboentries] &= ~0x10;
						rcomboa[comboentries++] &= ~0x10;
					}

					if (!strcmp(s,"B"))
					{
						comboa[comboentries] &= ~0x20;
						rcomboa[comboentries++] &= ~0x20;
					}

					if (!strcmp(s,"C"))
					{
						comboa[comboentries] &= ~0x40;
						rcomboa[comboentries++] &= ~0x40;
					}

					if (!strcmp(s,"D"))
					{
						comboa[comboentries] &= ~0x80;
						rcomboa[comboentries++] &=~0x80;
					}
				}
			}

			if (!strcmp(s,"qcf"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x2;
					rcomboa[comboentries++] = ~0x2;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0xa;
					rcomboa[comboentries++] = ~0x6;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x8;
					rcomboa[comboentries++] = ~0x4;
				}
			}

			if (!strcmp(s,"qcb"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x2;
					rcomboa[comboentries++] = ~0x2;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x6;
					rcomboa[comboentries++] = ~0xa;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x4;
					rcomboa[comboentries++]=~0x8;
				}
			}

			if (!strcmp(s,"hcf"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x4;
					rcomboa[comboentries++] = ~0x8;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x6;
					rcomboa[comboentries++] = ~0xa;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x2;
					rcomboa[comboentries++] = ~0x2;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0xa;
					rcomboa[comboentries++] = ~0x6;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x8;
					rcomboa[comboentries++] = ~0x4;
				}
			}

			if (!strcmp(s,"hcb"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x8;
					rcomboa[comboentries++] = ~0x4;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0xa;
					rcomboa[comboentries++] = ~0x6;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x2;
					rcomboa[comboentries++] = ~0x2;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x6;
					rcomboa[comboentries++] = ~0xa;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x4;
					rcomboa[comboentries++] = ~0x8;
				}
			}

			if (!strcmp(s,"dp"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x8;
					rcomboa[comboentries++] = ~0x4;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x2;
					rcomboa[comboentries++] = ~0x2;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0xa;
					rcomboa[comboentries++] = ~0x6;
				}
			}

			if (!strcmp(s,"rdp"))
			{
				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x4;
					rcomboa[comboentries++] = ~0x8;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x2;
					rcomboa[comboentries++] = ~0x2;
				}

				for (i = 0; i < speed; i++)
				{
					comboa[comboentries] = ~0x6;
					rcomboa[comboentries++] = ~0xa;
				}
			}

			if (!stricmp(s,"end"))
			{
				comboa[comboentries] = 0xffff;
				rcomboa[comboentries++] = 0xffff;
			}

			if (!strcmp(s, "{"))
			{
				fseek(f, (ftell(f) - 1), SEEK_SET);
				break;
			}
#if 0
			if(!strncmp(s, "$combo=", 7))
			{
				//fseek(f, (ftell(f) - strlen(s) - 1), SEEK_SET);
				fseek(f, last_pos, SEEK_SET);
				logerror("parse_combo(): $cobmo= found\n");

				return -1;
			}
#endif
			if (!strcmp(s, "$endcombo"))
			{
				return -1;
			}

			if (atoi(s))
				speed = atoi(s);
		}
		else if (!stricmp(s,"$endcombo"))
		{
			//comboa[comboentries] = 0xffff;
			//rcomboa[comboentries++] = 0xffff;
			return -1;
		}
	}

	return comboentries;
}


int check_combo_data(char combo_name[][MAX_COMBONAME_LEN])
{
	int i;
	char fname[80];
	char line[MAX_COMBO_SIZE];
	char *p;
	FILE *f;

	for (i = 0; i < 8; i++)
		strcpy(combo_name[i], ui_getstring(UI_emptyslot));

	sprintf(fname, "%s\\%s.cb%d", GetComboDir(), Machine->gamedrv->name, combosaveslot);
	if ((f = fopen(fname,"rt")) == NULL)
		return 0;

	if (fgets(line, MAX_COMBO_SIZE, f) == NULL)
		return 0;

	i = 0;
	while (!feof(f))
	{
		if (!memcmp(line, "$combo=", 7))
		{
			memset(combo_name[i], 0, MAX_COMBONAME_LEN);
			memcpy(combo_name[i], line + 7, MAX_COMBONAME_LEN - 1);

			p = strrchr(combo_name[i], '\n');
			if (p != NULL)
				*p = '\0';

			if (strlen(combo_name[i]) == 0)
				strcpy(combo_name[i], ui_getstring(UI_combononame));

			i++;
			if (i >= 8) break;
		}

		fgets(line, MAX_COMBO_SIZE, f);
	}

	fclose(f);

	return 1;
}


int read_combo_ex(void)
{
	char fname[80];
	FILE *f;
	unsigned short cnt = 0, start = 0;
	unsigned short b_cnt = 0;
	char line[MAX_COMBO_SIZE];

	branchentries = 0;
	memset(branchndx, 0, (8 * MAX_BRANCH_NUM * sizeof(UINT16)));

	sprintf(fname, "%s\\%s.cb%d", GetComboDir(), Machine->gamedrv->name, combosaveslot);

	if (!(f = fopen(fname,"rt")))
	{
		MyFree(comboa);
		MyFree(rcomboa);
		comboa = rcomboa = 0;
		return 0;
	}

	if (fgets(line, MAX_COMBO_SIZE, f) == NULL) return 0;

	while (!feof(f))
	{
		if (!memcmp(line, "$combo=", 7))
		{
			do
			{
				memset(line, 0, MAX_COMBO_SIZE);

				while (fgets(line, MAX_COMBO_SIZE, f) != NULL)
				{
					if (line[0] == '{')
					{
						branchndx[cnt][b_cnt++] = branchentries;
						parse_branch(line);
						break;
					}
#if 1
					else if (!memcmp(line, "$combo=", 7))
					{
						break;
					}
#endif
					else if (!memcmp(line, "$endcombo", 9))
					{
						break;
					}
				}
#if 0
				if(feof(f))
					break;

				if(!strcmp(line, "$endcombo"))
					break;

				if(parse_combo(f) == -1)
				{
					break;
				}
#endif
				if (feof(f) || (!memcmp(line, "$endcombo", 9)) || (parse_combo(f) == -1))
					break;
			} while (!feof(f));

			if (feof(f))
				break;

			combondx[cnt++] = start;
			start = comboentries;

			b_cnt = 0;

			if (fgets(line, MAX_COMBO_SIZE, f) == NULL) break;
		}
		else
		{
			if (fgets(line, MAX_COMBO_SIZE, f) == NULL) break;
		}

		if (cnt >= 8) break;
	}

	fclose(f);

	return cnt;
}


int combo_record_start(void)
{
	if (combo_buf)
		return 0;

	combo_buf = (char *)MyMalloc(MAX_COMBO_SIZE * sizeof(char));
	if (combo_buf == NULL)
		return 0;

	memset(combo_buf, 0, sizeof(combo_buf));

	pre_dir = 0;
	pre_btn = 0;
	dir_cnt = 0;
	bt1_cnt = 0;
	bt2_cnt = 0;
	bt3_cnt = 0;
	bt4_cnt = 0;
	dir_idle = 1;

	return 1;
}


int combo_record_end(void)
{
	FILE *fp = NULL;
	char fname[128];
	char temp[8];

	sprintf(fname,"%s\\%s.cb%d", GetComboDir(), Machine->gamedrv->name, combosaveslot);

	if ((fp = fopen(fname, "at")) == NULL)
	{
		MyFree(combo_buf);
		combo_buf = NULL;
		return 0;
	}

	fprintf(fp, "\n$combo=\n");
	fprintf(fp, "{ }\n");

	memset(temp, 0, sizeof(temp));
	if (sscanf(combo_buf, "%s n ", temp))
	{
		fwrite(combo_buf + strlen(temp) + 3, strlen(combo_buf) - (strlen(temp) + 3), 1, fp);
	}
	else
		fwrite(combo_buf, strlen(combo_buf), 1, fp);

	fprintf(fp, "end\n");
	fprintf(fp, "$endcombo\n");
	fclose(fp);

	MyFree(combo_buf);
	combo_buf = NULL;

	return 1;
}


int combo_record_cancel(void)
{
	MyFree(combo_buf);
	combo_buf = NULL;
	return 1;
}
