#ifndef MACHINE_H
#define MACHINE_H

#include "osdepend.h"
#include "drawgfx.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "memmngr.h"

extern char build_version[];

#define MAX_GFX_ELEMENTS 32
#define MAX_MEMORY_REGIONS 8

struct RegionInfo
{
	UINT8 *		base;
	size_t		length;
	UINT32		type;
	UINT32		flags;
};

struct RunningMachine
{
	struct RegionInfo memory_region[MAX_MEMORY_REGIONS];
	struct GfxElement *gfx[MAX_GFX_ELEMENTS];	/* graphic sets (chars, sprites) */
	struct osd_bitmap *scrbitmap;	/* bitmap to draw into */
	struct rectangle visible_area;
	unsigned short *pens;	/* remapped palette pen numbers. When you write */
							/* directly to a bitmap, never use absolute values, */
							/* use this array to get the pen number. For example, */
							/* if you want to use color #6 in the palette, use */
							/* pens[6] instead of just 6. */
	unsigned short *game_colortable;	/* lookup table used to map gfx pen numbers */
										/* to color numbers */
	const struct GameDriver *gamedrv;	/* contains the definition of the game machine */
	const struct MachineDriver *drv;	/* same as gamedrv->drv */
	int sample_rate;	/* the digital audio sample rate; 0 if sound is disabled. */
						/* This is set to a default value, or a value specified by */
						/* the user; osd_init() is allowed to change it to the actual */
						/* sample rate supported by the audio card. */
	struct InputPort *input_ports;	/* the input ports definition from the driver */
								/* is copied here and modified (load settings from disk, */
								/* remove cheat commands, and so on) */
	struct InputPort *input_ports_default; /* original input_ports without modifications */
	struct GfxElement *uifont;	/* font used by the user interface */
	struct GfxElement *uifont2;	/* font used by the user interface */
	struct GfxElement *uifont2p;	/* font used by the user interface */
	int uifontwidth,uifontheight;
	int uifontwidth2,uifontheight2;
	int uifontwidth2p,uifontheight2p;
	int uixmin,uiymin;
	int uiwidth,uiheight;
	int system_color[16];
	struct rectangle absolute_visible_area;	/* as passed to osd_set_visible_area() */

#ifdef MAME_DEBUG
	/* stuff for the debugger */
	struct osd_bitmap *debug_bitmap;
	UINT16 *debug_pens;
	UINT16 *debug_remapped_colortable;
	struct GfxElement *debugger_font;
#endif
};

/* The host platform should fill these fields with the preferences specified in the GUI */
/* or on the commandline. */
struct GameOptions {
	void *record;
	void *playback;

#ifdef MAME_DEBUG
	int mame_debug;
	int debug_width;	/* initial size of the debug_bitmap */
	int debug_height;
#endif
	int cheat;
	int gui_host;

	int samplerate;
	int stereo;

	int core_type;
	int driver_type;
	int bios_type;
	int disable_patch;
	int disable_nglo;
	int use_overclock;
	int only_jukebox;

	int ui_transparency; /* user_interface window transparency (%) */
	int random_time;
	int cyclecpu;
};

extern struct GameOptions options;
extern struct RunningMachine *Machine;

int run_game(int game);
int updatescreen(void);
void draw_screen(void);

/* next time vh_screenrefresh is called, full_refresh will be true,
   thus requesting a redraw of the entire screen */
void schedule_full_refresh(void);

void update_video_and_audio(void);
/* osd_fopen() must use this to know if high score files can be used */
int mame_highscore_enabled(void);
void set_led_status(int num,int on);

#endif
