/*********************************************************************

  ui_text.c

  Functions used to retrieve text used by MAME, to aid in
  translation.

*********************************************************************/

#include "driver.h"
#include "ui_text.h"

char keyname[KEY_last_entry + 1][MAX_KEY_LEN + 1];
char portname[PORT_last_entry + 1][MAX_PORT_LEN + 1];

#ifdef BUILD_ENGLISH
static const char *default_text[] =
{
	"NMAME X II",

	/* copyright stuff */
	"Usage of emulators in conjunction with ROMs you don't own is forbidden by copyright law.",
	"IF YOU ARE NOT LEGALLY ENTITLED TO PLAY \"%s\" ON THIS EMULATOR, PRESS CANCEL.",
	"Otherwise, press OK to continue",

	/* misc stuff */
	"Return to Main Menu",
	"Return to Prior Menu",
	"Press Any Key",
	"Press Select Key/Button",
	"On",
	"Off",
	"NA",
	"OK",
	"INVALID",
	"(none)",
	"CPU",
	"Address",
	"Value",
	"Sound",
	"sound",
	"stereo",
	"Vector Game",
	"Screen Resolution",
	"Text",
	"Volume",
	"Relative",
	"ALL CHANNELS",
	"Brightness",
	"Gamma",
	"Vector Intensity",
	"Overclock",
	"ALL CPUS",
	"Window Transparency",
	"History not available",
	"Command List not available",

	/* special characters */
	"\x11",
	"\x10",
	"\x18",
	"\x19",
	"\x1a",
	"\x1b",

	/* main menu */
	"Input (general)",
	"Dip Switches",
	"Analog Controls",
	"Calibrate Joysticks",
	"Bookkeeping Info",
	"Input (this game)",
	"Game Information",
	"Game History",
	"Show Command List",
	"Reset Game",
	"Return to Game",
	"Cheat",
	"Memory Card",
	"Custom Buttons",
	"Autofire Delay",

	/* input */
	"Key/Joy Speed",
	"Reverse",
	"Sensitivity",

	/* stats */
	"Tickets dispensed",
	"Coin",
	"(locked)",

	/* memory card */
	"Load Memory Card",
	"Eject Memory Card",
	"Create Memory Card",
	"Call Memory Card Manager (RESET)",
	"Failed To Load Memory Card!",
	"Load OK!",
	"Memory Card Ejected!",
	"Memory Card Created OK!",
	"Failed To Create Memory Card!",
	"(It already exists ?)",
	"DAMN!! Internal Error!",

	/* cheats */
	"Watchpoint",
	"Cheats",
	"Watchpoints",
	"More Info",
	"More Info for",
	"Set",
	"Cheat conflict found: disabling",
	"Select a value",

	/* custom button */
	"Custom Button ",
	"Combo",
	"Not Use",
	" Button",
	"Delay",
	"\x81\xc0\x81\xc4\x81\xc1",
	"\x81\xc0\x81\xc4\x81\xc2",
	"\x81\xc0\x81\xc4\x81\xc3",
	"\x81\xc1\x81\xc4\x81\xc2",
	"\x81\xc1\x81\xc4\x81\xc3",
	"\x81\xc2\x81\xc4\x81\xc3",
	"\x81\xc0\x81\xc4\x81\xc1\x81\xc4\x81\xc2",
	"\x81\xc0\x81\xc4\x81\xc1\x81\xc4\x81\xc3",
	"\x81\xc1\x81\xc4\x81\xc2\x81\xc4\x81\xc3",
	"\x81\xc0\x81\xc4\x81\xc1\x81\xc4\x81\xc2\x81\xc4\x81\xc3",
	"Auto Fire \x81\xc0 Button",
	"Auto Fire \x81\xc1 Button",
	"Auto Fire \x81\xc2 Button",
	"Auto Fire \x81\xc3 Button",
	
	/* save/load state */
	"Save State",
	"Load State",
	"Date",
	"Time",
	"Info",
	"State Saved.",
	"State Loaded.",

	/* save/load combo stuff */
	"Save Combo",
	"Load Combo",
	"Combo Saved.",
	"Combo Loaded.",
	"Start Recording Combo...",

	/* Wave recording */
	"WAV Recording...",

	/* Juke Box */
	"- JUKE BOX -",
	"Bank",
	"Code 1",
	"Code 2",
	"COMMAND",
	"PLAY",
	"STOP",
	"RECORD",
	"STOP: S",
	"RECORD: R",
	"CODE: C",
	"FADEOUT: F",
	"STOP: S",
	"NEXT: N",
	"STOP: S",
	"RECORD: R",
	"CODE: C",
	"FADEOUT: F",
	"Fade Out Timer",
	"s",
	"Random Play",
	"Go to Play List Menu",
	"Go to Manual Code Menu",
	"BGM",
	"SE1",
	"SE2",
	"SE3",
	"SE4",
	"SE5",
	"SE6",
	"SE7",

	/* object operation */
	"- OBJECT OPERATION -",
	"Object Number",
	"Object Output",
	"Front Layer",
	"Alt Backcolor",
	"All Output Clear",
	"Set Default",
	"OFF",
	" ON",

	/* audit.c */
	"bytes",
	"INCORRECT LENGTH",
	"NO GOOD DUMP KNOWN",
	"ROM NEEDS REDUMP",
	"INCORRECT CHECKSUM",
	"NOT FOUND",
	"ROM GOOD",
	"Out of memory reading ROM",

	/* inptport.c */
	"Empty Slot",
	"No Name",

	/* input.c */
	"n/a",
	"None",
	"not",
	"or",

	/* driver/neogeo.c */
	"raster effects",
	"enabled",
	"disabled",

	/* win32/DIJoystick.c */
	"Forward",
	"Backward",
	"Right",
	"Left",
	"MOUSE B%d",
	"Joy[%d] %s -",
	"Joy[%d] %s +",
	"Joy[%d] Button %d",
	"Joy[%d] POV%i %s",

	/* win32/Joystick.c */
	"X Axis",
	"Y Axis",
	"Z Axis",
	"R Axis",
	"U Axis",
	"V Axis",
	"POV Forward",
	"POV Right",
	"POV Backward",
	"POV Left",
	"MOUSE B%d",
	"J%d %.8s %s -",
	"J%d %.8s %s +",
	"J%d Button %d",
	"J%d %s",

	NULL
};

const char *default_key[] =
{
	/* key names */
	"A",
	"B",
	"C",
	"D",
	"E",
	"F",
	"G",
	"H",
	"I",
	"J",
	"K",
	"L",
	"M",
	"N",
	"O",
	"P",
	"Q",
	"R",
	"S",
	"T",
	"U",
	"V",
	"W",
	"X",
	"Y",
	"Z",
	"0",
	"1",
	"2",
	"3",
	"4",
	"5",
	"6",
	"7",
	"8",
	"9",
	"0 PAD",
	"1 PAD",
	"2 PAD",
	"3 PAD",
	"4 PAD",
	"5 PAD",
	"6 PAD",
	"7 PAD",
	"8 PAD",
	"9 PAD",
	"F1",
	"F2",
	"F3",
	"F4",
	"F5",
	"F6",
	"F7",
	"F8",
	"F9",
	"F10",
	"F11",
	"F12",
	"ESC",
	"~",
	"-",
	"=",
	"BKSPACE",
	"TAB",
	"[",
	"]",
	"ENTER",
	";",
	":",
	"\\",
	",",
	".",
	"/",
	"SPACE",
	"INS",
	"DEL",
	"HOME",
	"END",
	"PGUP",
	"PGDN",
	"\x81\xd4",
	"\x81\xd6",
	"\x81\xd8",
	"\x81\xd2",
	"/ PAD",
	"* PAD",
	"- PAD",
	"+ PAD",
	". PAD",
	"ENTER PAD",
	"PRTSCR",
	"PAUSE",
	"LSHIFT",
	"RSHIFT",
	"LCTRL",
	"RCTRL",
	"LALT",
	"RALT",
	"SHIFT",
	"CTRL",
	"ALT",
	"LWIN",
	"RWIN",
	"MENU",
	"SCRLOCK",
	"NUMLOCK",
	"CAPSLOCK",
	NULL
};

const char *default_port[] =
{
	"Config Menu",
	"On Screen Display",
	"Single Step",
	"Pause",
	"Reset Game",
	"Show Gfx",
	"Save State",
	"Load State",
	"Save Combo",
	"Load Combo",
	"Frameskip Dec",
	"Frameskip Inc",
	"Throttle",
	"Show FPS",
	"Show Profiler",
	"Save Snapshot",
	"Toggle Cheat",
	"Change Screen Mode",
	"Show Input Log",
	"Toggle Wave Recording",
	"Juke Box",
	"Object Operation",
	"Show Command List",
	"Show Cheat List",
	"Show the Time",
#ifdef KAILLERA
	"Kaillera Chat Open",
	"Kaillera Chat Move",
#endif
	"UI Direction\x81\xd8",
	"UI Direction\x81\xd2",
	"UI Direction\x81\xd4",
	"UI Direction\x81\xd6",
	"UI Select",
	"UI Cancel",
	"Coin 1",
	"Coin 2",
	"Coin 3 (Service)",
	"P1 Start",
	"P1 Select",
	"P1 Direction\x81\xd8",
	"P1 Direction\x81\xd2",
	"P1 Direction\x81\xd4",
	"P1 Direction\x81\xd6",
	"P1 Button\x81\xc0",
	"P1 Button\x81\xc1",
	"P1 Button\x81\xc2",
	"P1 Button\x81\xc3",
	"P1 Custom Button 1",
	"P1 Custom Button 2",
	"P1 Custom Button 3",
	"P1 Custom Button 4",
	"P1 Track Y - (\x81\xd8)",
	"P1 Track Y + (\x81\xd2)",
	"P1 Track X - (\x81\xd4)",
	"P1 Track X + (\x81\xd6)",
	"P2 Start",
	"P2 Select",
	"P2 Direction\x81\xd8",
	"P2 Direction\x81\xd2",
	"P2 Direction\x81\xd4",
	"P2 Direction\x81\xd6",
	"P2 Button\x81\xc0",
	"P2 Button\x81\xc1",
	"P2 Button\x81\xc2",
	"P2 Button\x81\xc3",
	"P2 Custom Button 1",
	"P2 Custom Button 2",
	"P2 Custom Button 3",
	"P2 Custom Button 4",
	"P2 Track Y - (\x81\xd8)",
	"P2 Track Y + (\x81\xd2)",
	"P2 Track X - (\x81\xd4)",
	"P2 Track X + (\x81\xd6)",
	"UnKnown",

	/* NeoGeo Input Port */
	"Test Switch",
	"Coin Chutes?",
	"Autofire (in some games)",
	"COMM Setting",
	"Free Play",
	"Freeze",
	"Territory",
	"Machine Mode",
	"Off",
	"On",
	"Japan",
	"USA",
	"Europe",
	"Home",
	"Arcade",

	NULL
};
#else
static const char *default_text[] =
{
	"NMAME X U ",

	/* copyright stuff */
	"LĂȂROMC[WG~[^Ŏgp邱Ƃ́A쌠@ɂċւĂ܂",
	"h%sh\n\n@IɗVԎiȂꍇESCĂ",
	"OKƓ͂Ƒ܂",

	/* misc stuff */
	"Cj[ɖ߂",
	"Õj[ɖ߂",
	"L[Ă",
	"L[^{^Ă",
	"I",
	"It",
	"Ȃ",
	"OK",
	"",
	"(Ȃ)",
	"CPU",
	"AhX",
	"l",
	"",
	"TEh",
	"XeI",
	"Vector Game",
	"ʉ𑜓x",
	"eLXg",
	"{[",
	"",
	"S`l",
	"邳",
	"K}l",
	"Vector Intensity",
	"I[o[NbN",
	"SCPU",
	"EBhEߗ",
	"qXg[܂",
	"R}hXg܂",

	/* special characters */
	"\x11",
	"\x10",
	"\x18",
	"\x19",
	"\x1a",
	"\x1b",

	/* main menu */
	"W̓͐ݒ",
	"fBbvXCb`",
	"AiORg[",
	"WCXeBbN̒",
	"CJ",
	"ʂ̓͐ݒ",
	"Q[̏",
	"Q[̗j/",
	"R}hꗗ/",
	"Q[Zbg",
	"Q[ɖ߂",
	"`[g(Q[̉)",
	"J[h",
	"JX^{^",
	"Aˑx",

	/* input */
	"Key/Joyx",
	"]",
	"x",

	/* stats */
	"Tickets dispensed",
	"RC",
	"(bN)",

	/* memory card */
	"J[hǂ݂",
	"J[ho",
	"J[h쐬",
	"J[h}l[WĂяo(Zbg)",
	"J[h̓ǂ݂݂Ɏs܂I",
	"[h܂I",
	"J[ho܂I",
	"J[h쐬܂I",
	"J[h̍쐬Ɏs܂I",
	"(łɍ쐬ς݁H)",
	"G[łI",

	/* cheats */
	"EHb`|Cg",
	"`[g",
	"EHb`|Cg",
	"⑫",
	"⑫F",
	"Zbg",
	"܂Fgps",
	"lI܂",

	/* custom button */
	"JX^{^",
	"R{",
	"gpȂ",
	"{^",
	"Aˑx",
	"\x81\xc0\x81\xc4\x81\xc1",
	"\x81\xc0\x81\xc4\x81\xc2",
	"\x81\xc0\x81\xc4\x81\xc3",
	"\x81\xc1\x81\xc4\x81\xc2",
	"\x81\xc1\x81\xc4\x81\xc3",
	"\x81\xc2\x81\xc4\x81\xc3",
	"\x81\xc0\x81\xc4\x81\xc1\x81\xc4\x81\xc2",
	"\x81\xc0\x81\xc4\x81\xc1\x81\xc4\x81\xc3",
	"\x81\xc1\x81\xc4\x81\xc2\x81\xc4\x81\xc3",
	"\x81\xc0\x81\xc4\x81\xc1\x81\xc4\x81\xc2\x81\xc4\x81\xc3",
	"\x81\xc0{^̘A",
	"\x81\xc1{^̘A",
	"\x81\xc2{^̘A",
	"\x81\xc3{^̘A",

	/* save/load state */
	"vCԂ̕ۑ",
	"vCԂ̕A",
	"t",
	"",
	"",
	"vCԂۑ܂",
	"vCԂ𕜋A܂",

	/* save/load combo stuff */
	"R{̕ۑ    ",
	"R{̓ǂݍ",
	"R{ۑ܂",
	"R{ǂݍ݂܂",
	"R{̋L^Jn...",

	/* Wave recording */
	"^",

	/* Juke Box */
	"- W[N{bNX -",
	"oN",
	"R[h1",
	"R[h2",
	"R}h",
	"tJn",
	"t~",
	"^Jn",
	"t~: S",
	"^Jn: R",
	"R[h\: C",
	"tF[hAEg: F",
	"ĐI: S",
	"̋Ȃ: N",
	"~: S",
	"^: R",
	"R[h: C",
	"̪ޱ: F",
	"tF[hAEg̐ݒ",
	"b",
	"_Đ",
	"TEhXgʂ֍s",
	"R[h͉ʂ֍s",
	"obNTEh",
	"Eʉ 1",
	"Eʉ 2",
	"Eʉ 3",
	"Eʉ 4",
	"Eʉ 5",
	"Eʉ 6",
	"Eʉ 7",

	/* object operation */
	"- IuWFNg -",
	"IuWFNgԍ",
	"IuWFNgo",
	"tgCA[",
	"wiF̕ύX",
	"So͂NA",
	"ftHgɖ߂",
	"It",
	"I",

	/* audit.c */
	"oCg",
	"t@CTCYႢ܂",
	"ɋzoĂ܂",
	"ēxzoȂKv܂",
	"`FbNTႢ܂",
	"܂",
	"ROM͐ł",
	"ROMǂݍݒɃsɂȂ܂",

	/* inptport.c */
	"gp",
	"R{o^",

	/* input.c */
	"gps",
	"gp",
	"",
	"",

	/* driver/neogeo.c */
	"X^GtFNg",
	"L",
	"",

	/* win32/DIJoystick.c */
	"O",
	"",
	"E",
	"",
	"}EX {^%d",
	"Joy[%d] %s -",
	"Joy[%d] %s +",
	"Joy[%d] {^%d",
	"Joy[%d] POV%i %s",

	/* win32/Joystick.c */
	"X ",
	"Y ",
	"Z ",
	"R ",
	"U ",
	"V ",
	"POV O",
	"POV E",
	"POV ",
	"POV ",
	"}EX {^%d",
	"J%d %.8s %s -",
	"J%d %.8s %s +",
	"J%d {^ %d",
	"J%d %s",

	NULL
};

const char *default_key[] =
{
	/* key names */
	"A",
	"B",
	"C",
	"D",
	"E",
	"F",
	"G",
	"H",
	"I",
	"J",
	"K",
	"L",
	"M",
	"N",
	"O",
	"P",
	"Q",
	"R",
	"S",
	"T",
	"U",
	"V",
	"W",
	"X",
	"Y",
	"Z",
	"0",
	"1",
	"2",
	"3",
	"4",
	"5",
	"6",
	"7",
	"8",
	"9",
	"eL[ 0",
	"eL[ 1",
	"eL[ 2",
	"eL[ 3",
	"eL[ 4",
	"eL[ 5",
	"eL[ 6",
	"eL[ 7",
	"eL[ 8",
	"eL[ 9",
	"F1",
	"F2",
	"F3",
	"F4",
	"F5",
	"F6",
	"F7",
	"F8",
	"F9",
	"F10",
	"F11",
	"F12",
	"Esc",
	"~",
	"-",
	"=",
	"Back Space",
	"Tab",
	"[",
	"]",
	"Enter",
	";",
	":",
	"\\",
	",",
	".",
	"/",
	"Space",
	"Insert",
	"Delete",
	"Home",
	"End",
	"Page Up",
	"Page Down",
	"\x81\xd4",
	"\x81\xd6",
	"\x81\xd8",
	"\x81\xd2",
	"eL[ /",
	"eL[ *",
	"eL[ -",
	"eL[ +",
	"eL[ .",
	"eL[ Enter",
	"Print Screen",
	"Pause",
	"Shift",
	"EShift",
	"Ctrl",
	"ECtrl",
	"Alt",
	"EAlt",
	"Shift",
	"Ctrl",
	"Alt",
	"Windows",
	"EWindows",
	"Menu",
	"Scroll Lock",
	"Num Lock",
	"Caps Lock",
	NULL
};

const char *default_port[] =
{
	"ݒ胁j[",
	"IXN[fBXvC",
	"R}",
	"|[Y",
	"Q[̃Zbg",
	"OtBbNf[^\",
	"vCԂ̕ۑ",
	"vCԂ̕A",
	"R{̕ۑ",
	"R{̓ǂݍ",
	"t[XLbv",
	"t[XLbv",
	"Xbg",
	"FPS\",
	"vt@C\",
	"XibvVbgۑ",
	"`[g؂ւ",
	"fBXvC[h؂ւ",
	"̓R}h̕\",
	"WAV^Jn/I",
	"W[N{bNX",
	"IuWFNg",
	"R}h/@̕\",
	"`[g̕\",
	"̕\",
#ifdef KAILLERA
	"`bgJn",
	"`bg̕\ʒüړ",
#endif
	"UI \x81\xd8ړ",
	"UI \x81\xd2ړ",
	"UI \x81\xd4ړ",
	"UI \x81\xd6ړ",
	"UI I",
	"UI LZ",
	"RC1",
	"RC2",
	"RC3 (T[rX)",
	"P1 X^[g",
	"P1 ZNg",
	"P1 \x81\xd8",
	"P1 \x81\xd2",
	"P1 \x81\xd4",
	"P1 \x81\xd6",
	"P1 {^\x81\xc0",
	"P1 {^\x81\xc1",
	"P1 {^\x81\xc2",
	"P1 {^\x81\xc3",
	"P1 JX^{^1",
	"P1 JX^{^2",
	"P1 JX^{^3",
	"P1 JX^{^4",
	"P1 gbN{[ \x81\xd8ړ",
	"P1 gbN{[ \x81\xd2ړ",
	"P1 gbN{[ \x81\xd4ړ",
	"P1 gbN{[ \x81\xd6ړ",
	"P2 X^[g",
	"P2 ZNg",
	"P2 \x81\xd8",
	"P2 \x81\xd2",
	"P2 \x81\xd4",
	"P2 \x81\xd6",
	"P2 {^\x81\xc0",
	"P2 {^\x81\xc1",
	"P2 {^\x81\xc2",
	"P2 {^\x81\xc3",
	"P2 JX^{^1",
	"P2 JX^{^2",
	"P2 JX^{^3",
	"P2 JX^{^4",
	"P2 gbN{[ \x81\xd8ړ",
	"P2 gbN{[ \x81\xd2ړ",
	"P2 gbN{[ \x81\xd4ړ",
	"P2 gbN{[ \x81\xd6ړ",
	"UnKnown",

	/* NeoGeo Input Port */
	"eXgXCb`",
	"RCH",
	"A (̃Q[ŗL)",
	"ʐMݒ",
	"t[vC",
	"t[Y",
	"n",
	"}V[h",
	"It",
	"I",
	"{",
	"AJ",
	"[bp",
	"ƒp",
	"A[P[h",

	NULL
};
#endif

const char *ui_getstring(int string_num)
{
	return default_text[string_num];
}

void keyname_init(void)
{
	int i;

	for (i = KEY_first_entry + 1; i < KEY_last_entry; i++)
		strcpy(keyname[i], default_key[i]);
}

void portname_init(void)
{
	int i;

	for (i = PORT_first_entry + 1; i < PORT_last_entry; i++)
		strcpy(portname[i], default_port[i]);
}

